#include "ai.h"

// needed to know which colors are how to be recognized
#include "gamepf.h"

// how many gems have to be grouped before they get destroyed? see main.h!
extern int min_friends_for_group ;

void AI::Init ( int new_ai_level, Pax *pax )
{
	ai_level = new_ai_level ;
	
	quit_state = FALSE ;
	
	GEM_NONE  = pax -> GetCharFromStringVar ( "gem_none_char" ) ;
	
	PLAYFIELD_MAX_X = pax -> GetIntVar ( "playfield_max_x" ) ;
	PLAYFIELD_MAX_Y = pax -> GetIntVar ( "playfield_max_y" ) ;
	PLAYFIELD_SIZE  = pax -> GetIntVar ( "playfield_size" ) ;
	
	fprintf ( stdout, "AI initialized with level %d\n", ai_level ) ;
}

int AI::GetPosInPlayfield ( int x, int y )
{
	return ( y * PLAYFIELD_MAX_X ) + x ;
}

void AI::Think ( char *org_playfield, int user_gem_pos )
{
	if ( ai_level == 0 || quit_state ) return ;
		
	// copy original playfield so that we won't hurt anybody :)
	char playfield [ PLAYFIELD_SIZE ] ;
	for ( int i = 0 ; i < PLAYFIELD_SIZE ; i++ )
		playfield [ i ] = org_playfield [ i ] ;
	
	int user_gem_pos_x = user_gem_pos % PLAYFIELD_MAX_X ;
	int user_gem_pos_y = ( user_gem_pos - user_gem_pos_x ) / PLAYFIELD_MAX_X ;

	fprintf ( stdout, "user gem pos x = %d ; y = %d\n",
				user_gem_pos_x, user_gem_pos_y ) ;

	for ( int x = 0 ; x < PLAYFIELD_MAX_X ; x++ )
	{
		bool the_following_stones_are_not_needed = FALSE ;
		for ( int y = PLAYFIELD_MAX_Y - 1 ; y >= 0 ; y-- )
		{
			int pos = GetPosInPlayfield ( x, y ) ;
			if ( pos != user_gem_pos )
			{
				if ( the_following_stones_are_not_needed )
				{
					if ( playfield [ pos ] >= 32 )
						playfield [ pos ] = 0 ;
				}
				else
				{
					if ( playfield [ pos ] != GEM_NONE )
						the_following_stones_are_not_needed = TRUE ;
					
					if ( org_playfield [ pos ] != org_playfield [ user_gem_pos ] )
					{	// everything other than our wanted color has to die ;)
						if ( playfield [ pos ] >= 32 )
							playfield [ pos ] = 0 ;
						
						if ( org_playfield [ pos ] == GEM_NONE )
						{	// only look left and right if this pos is empty!
							int left  = GetPosInPlayfield ( x - 1 , y ) ;
							int right = GetPosInPlayfield ( x + 1 , y ) ;
							
							if ( x > 0 && left != user_gem_pos &&
									org_playfield [ left ] == 
									playfield [ user_gem_pos ] )
							{
								playfield [ left ] = 2 ;
							}
							
							if ( x < PLAYFIELD_MAX_X - 1 && right != user_gem_pos &&
									org_playfield [ right ] == 
									playfield [ user_gem_pos ] )
							{
								playfield [ right ] = 2 ;
							}
						}
					}
					else
						playfield [ pos ] = 1 ;
				}
			}
		}
	}
	
	fprintf ( stdout, "Dumping AI playfield...\n" ) ;
	for ( int y = 0 ; y < PLAYFIELD_MAX_Y ; y++ )
	{
		for ( int x = 0 ; x < PLAYFIELD_MAX_X ; x++ )
		{
			fprintf ( stdout, "'%02d' ", 
						playfield [ GetPosInPlayfield ( x, y ) ] ) ;
		}
		fprintf ( stdout, "\n" ) ;
	}
}

void AI::Quit ()
{
	quit_state = TRUE ;
}
