#include "audio_sound.h"

extern bool soundeffects_enabled ;
extern FileCache file_cache ;

Audio_Sound::Audio_Sound ()
{
	SoundChunk = NULL ; // we don't have a sound yet..
	SoundChannel = -1 ; // this means any free channel
	panning_pos = 128 ; // set panning to center as default
	volume = MIX_MAX_VOLUME ; // should be 128
}

Mix_Chunk *Audio_Sound::ConvertFromOgg ( char *filename )
{
	char cmd [ 2048 ] ;
	Mix_Chunk *sample ;
	
	struct stat buf ;
	char tempfile [ 2048 ] ;
	
	sprintf ( tempfile, "%s", temp_file ( filename ) ) ;
	
	// first look if theres already a cached file
	if ( stat ( tempfile, &buf ) != 0 )
	{	// if not, try to convert one
#ifndef __WIN32__
		//sprintf ( cmd, "oggdec '%s' --output '%s'", filename, 
		//			tempfile ) ;
		sprintf ( cmd, "ogg123 -d wav -f '%s' '%s'", tempfile, 
					filename ) ;
#else
		sprintf ( cmd, "../GemHunters-win32/bin/oggdec.exe '%s' --output '%s'", filename, 
					tempfile ) ;
#endif
		
		// old method was this but it causes windows to open
		// a new shell-window for every command... ARGH!!!!!
		//system ( cmd ) ;

		// this method catches the output and uses the main
		// shell-window for writing the text into it...
		FILE *stream = popen ( cmd, "r" ) ;

		if ( stream == NULL )
		{
			fprintf ( stdout, "converting file from ogg to wav failed:\n%s\n", cmd ) ;
		}
		else
		{
			fprintf ( stdout, "converting file from ogg to wav:\n%s\n", cmd ) ;
			char line[1024] ;
			while ( fscanf ( stream, "%s", line ) > 0 )
			{
				fgets ( line, 1024, stream ) ;
				fprintf ( stdout, "%s", line ) ;
			}
			pclose ( stream ) ;
		}
		fprintf ( stdout, "\n" ) ;
	}

	// if there is still no useful file create a dummy => indicate failed..
	if ( stat ( tempfile, &buf ) != 0 )
	{
		FILE *stream ;
		stream = fopen ( tempfile, "a" ) ;
		fclose ( stream ) ;
	}
	
	sample = Mix_LoadWAV ( tempfile ) ;
	
	return sample ;
}

void Audio_Sound::LoadSound ( char *filename )
{
	if ( soundeffects_enabled == FALSE ) return ;

	/*if ( SoundChunk != NULL )
	{
		fprintf ( stdout, "loading %s - you should remove the old sound first!\n",
					filename ) ;
		file_cache.DumpStatus () ;
	}*/
	
	SoundChunk = NULL ; // we don't have a sound yet..
	SoundChannel = -1 ; // this means any free channel
	panning_pos = 128 ; // set panning to center as default
	volume = MIX_MAX_VOLUME ; // should be 128
	
	SoundChunk = (Mix_Chunk*) file_cache.GetFromCache ( filename ) ;
	
	if ( SoundChunk == NULL )
	{
		SoundChunk = Mix_LoadWAV ( filename ) ;
		
		if ( SoundChunk == NULL )
		{
			char filebase[2048], fileext[2], tmpnam[2048] ;
			int filebase_length = strlen(filename) - 4 ;
			sprintf ( tmpnam, "%s", filename ) ;
			//fprintf ( stdout, "this doesn't seem to be a wav file.\n" ) ;
			//fprintf ( stdout, "gathering filetype...\n" ) ;
			strncpy ( filebase, tmpnam, filebase_length ) ;
			filebase [ filebase_length ] = '\0' ;
			//fprintf ( stdout, "file name is '%s'\n", filebase ) ;
			strncpy ( fileext, tmpnam + filebase_length + 1, 4 ) ;
			//fprintf ( stdout, "file extension is '%s'\n", fileext ) ;
			
			if ( strcmp ( fileext, "ogg" ) == 0 )
			{
				//fprintf ( stdout, "this seems to be an ogg file! trying...\n" ) ;
				SoundChunk = ConvertFromOgg ( filename ) ;
			}
	
			//if ( SoundChunk == NULL )
			//	fprintf ( stdout, "!Error! Couldn't load sound file %s!\n", filename ) ;
		}
		
	}

	if ( SoundChunk != NULL )
	{
		file_cache.PutIntoCache ( filename, SoundChunk ) ;
	}
}

void Audio_Sound::PlaySound ( int channel, int times )
{
	if ( soundeffects_enabled == FALSE ) return ;

	if ( SoundChunk && soundeffects_enabled )
	{
		if ( channel == AUTO ) channel = -1 ;
		SoundChannel = Mix_PlayChannel ( channel, SoundChunk, times - 1 ) ;
		SetVolume ( volume ) ;
		SetPanning ( panning_pos ) ;
	}
}

void Audio_Sound::StopSound ()
{
	if ( soundeffects_enabled == FALSE ) return ;

	if ( SoundChunk && SoundChannel >= 0 && SoundChannel <= AUDIO_CHANNELS )
	{
		Mix_HaltChannel ( SoundChannel ) ;
		SoundChannel = -1 ;
	}
}

void Audio_Sound::FadeInSound ( int channel, int times, int ms )
{
	if ( soundeffects_enabled == FALSE ) return ;

	if ( SoundChunk && soundeffects_enabled )
	{
		SoundChannel = Mix_FadeInChannel ( channel, SoundChunk, times - 1, ms ) ;
		SetVolume ( volume ) ;
		SetPanning ( panning_pos ) ;
	}
}

void Audio_Sound::FadeOutSound ( int ms )
{
	if ( soundeffects_enabled == FALSE ) return ;

	if ( SoundChunk && SoundChannel >= 0 && SoundChannel <= AUDIO_CHANNELS )
	{
		Mix_FadeOutChannel ( SoundChannel, ms ) ;
		SoundChannel = -1 ;
	}
}

void Audio_Sound::SetVolume ( int value )
{
	if ( soundeffects_enabled == FALSE ) return ;

	volume = value ;
	
	if ( SoundChunk && SoundChannel >= 0 && SoundChannel <= AUDIO_CHANNELS )
	{
		Mix_Volume ( SoundChannel, volume ) ;
	}
}

int Audio_Sound::GetVolume ()
{
	if ( soundeffects_enabled == FALSE ) return 0 ;

	if ( SoundChunk )
	{
		return volume ;
	}
	else
	{
		return 0 ;
	}
}

// note that 0 is left, 128 is center and 255 is right!
void Audio_Sound::SetPanning ( int pos )
{
	if ( soundeffects_enabled == FALSE ) return ;

	panning_pos = pos ;
	
	int left = 256 - pos ;
	int right = pos ;
	
	if ( left < 0 ) left = 0 ;
	if ( left > 255 ) left = 255 ;
	if ( right < 0 ) right = 0 ;
	if ( right > 255 ) right = 255 ;
	
	//fprintf ( stdout, "setting panning to left = %d ; right = %d\n",
	//			left, right ) ;
	
#if SDL_VERSION_ATLEAST(1, 2, 4)
	const SDL_version *version = Mix_Linked_Version () ;
	if ( version -> major >= 1 && version -> minor >= 2 && version -> patch >= 4 &&
			SoundChunk && SoundChannel >= 0 && SoundChannel <= AUDIO_CHANNELS )
		Mix_SetPanning ( SoundChannel, left, right ) ;
#endif
}

int Audio_Sound::GetPanning ()
{
	return panning_pos ;
}

int Audio_Sound::IsPlaying ()
{
	return Mix_Playing ( SoundChannel ) ;
}

void Audio_Sound::UnloadSound ()
{
	if ( soundeffects_enabled == FALSE ) return ;

	if ( SoundChunk )
	{
#if SDL_VERSION_ATLEAST(1, 2, 4)
		const SDL_version *version = Mix_Linked_Version () ;
		if ( version -> major >= 1 && version -> minor >= 2 && version -> patch >= 4 &&
				SoundChannel >= 0 && SoundChannel <= AUDIO_CHANNELS )
			Mix_SetPanning ( SoundChannel, 255, 255 ) ;
#endif
		if ( file_cache.FreeDataFromCache ( SoundChunk ) == 0 )
			Mix_FreeChunk ( SoundChunk ) ;
	}

	SoundChunk = NULL ;
}

