#include "network.h"

#ifdef __NETWORK_SUPPORT__

#include <stdarg.h>
#include <unistd.h>

int MAX_CONNECTIONS = 256 ; // max number of connections

#define PASS_PHRASE      "Hello. I want to be a gemhun client!"
#define PASS_ANSWER      "Of Course! Feel yourself at home!"
#define PASS_ESTABLISHED "Thank you."
#define PASS_ALIAS       "Host is now known as: "
#define PASS_PING        "PING"
#define PASS_ALIVE       "ALIVE"
#define PASS_USERDATA    "USERDATA"

void Network::Init ()
{
	peers = new GlDynArray < peer > () ;
	peers -> Clear () ;
	
	seen_users = new GlDynArray < peer > () ;
	seen_users -> Clear () ;
	
	console = NULL ;
	buddy_list = NULL ;
	
	SDL_EnableUNICODE ( 1 ) ;
	if ( SDLNet_Init () )
	{
		printf ( "Failed to initialize SDL_Net error=%s\n", SDL_GetError () ) ;
		exit ( 1 ) ;
	}
	
	// Allocate the socket set
	socketset = SDLNet_AllocSocketSet(MAX_CONNECTIONS+1);
	if ( socketset == NULL )
	{
		fprintf(stderr, "Couldn't create socket set: %s\n",
		SDLNet_GetError());
		exit(2);
	}

	//serverIP = NULL ; how do this? perhaps set host and port to 0?
	servsock = NULL ;
	
	if ( strlen ( getenv ( "USER" ) ) > 0 )
		aliasname = getenv ( "USER" ) ;
	else if ( strlen ( getenv ( "HOST" ) ) > 0 )
		aliasname = getenv ( "HOST" ) ;
	else if ( strlen ( getenv ( "HOSTNAME" ) ) > 0 )
		aliasname = getenv ( "HOSTNAME" ) ;
	else if ( strlen ( rindex ( getenv ( "HOME" ), '/' ) ) > 0 )
		aliasname = rindex ( getenv ( "HOME" ), '/' ) ;
	else
		aliasname = "you" ;
}

void Network::EnableConsolePrinting ( Console *new_console, Console *new_buddy_list )
{
	console = new_console ;
	buddy_list = new_buddy_list ;
	
	if ( buddy_list )
		UpdateBuddyList () ;
}

void Network::DisableConsolePrinting ()
{
	console = NULL ;
	buddy_list = NULL ;
}

void Network::ResolveIP ( IPaddress *ip, char *hostname )
{
	char *ip_name = (char*) SDLNet_ResolveIP ( ip ) ;
	if ( ip_name == NULL )
	{	// get ip address
		int m1 = 0, m2 = 0, m3 = 0, m4 = 0 ;
		char hips [ 16 ] ;
		sprintf ( hips, "%08x", ip -> host ) ;
		m4 = hex_string_to_int ( hips, 0, 2 ) ;
		m3 = hex_string_to_int ( hips, 2, 2 ) ;
		m2 = hex_string_to_int ( hips, 4, 2 ) ;
		m1 = hex_string_to_int ( hips, 6, 2 ) ;
		sprintf ( hostname, "%d.%d.%d.%d", m1, m2, m3, m4 ) ;
	}
	else	// get hostname
		sprintf ( hostname, "%s", ip_name ) ;
}

void Network::Printf ( const char *format, ... )
{	// this one is copied from kyra's console.cpp file :)
	// slightly modified by me to use either a KrConsole or
	// the stdout stream...
	va_list va ;
	char buffer [ 8192 ] ;
	
	//
	//  format and output the message..
	//
	va_start ( va, format ) ;
	vsprintf ( buffer, format, va ) ;
	va_end ( va ) ;
	
	char *start ;
	char *end ;
	char *next ;
	
	start = buffer ;
	
	while ( start && *start )
	{
		end = strchr ( start, '\n' ) ;
		if ( end )
		{
			next  = end + 1 ;
			*end = 0 ;
			if ( console != NULL )
				console -> PushText ( start ) ;
			else
				fprintf ( stdout, "%s\n", start ) ;
			start = next ;
		}
		else
		{
			if ( console != NULL )
				console -> PushText ( start ) ;
			else
				fprintf ( stdout, "%s", start ) ;
			start = 0 ;
		}
	}
}

std::string Network::AddUsernameAtBeginning ( std::string text )
{
	std::string t_text ;
	
	t_text = aliasname ;
	t_text += ": " ;
	t_text += text ;
	
	return t_text ;
}

int Network::SendMessage ( peer *peer, char *message )
{
	int num_chars = 0 ;
	
	char data [ 256 ] ;
	sprintf ( data, "%s", message ) ;
	
	if ( peer )
	{
		//int sockno = FindPeerBySocketI ( peer -> socket ) ;
		//fprintf ( stdout, "send to socket %d: %s (%d chars)\n",
		//			sockno, message, strlen(message) ) ;
		num_chars = SDLNet_TCP_Send ( peer -> socket, data, 256 ) ;
		
		//fprintf ( stdout, "send %d chars to socket %d\n", num_chars, sockno ) ;
	}
	else
	{
		//fprintf ( stdout, "broadcast message: %s\n", message ) ;
		for ( unsigned int i = 0 ; i < peers -> Count () ; i++ )
			//if ( peers -> Item ( i ) . type == NM_CLIENT )
				num_chars += SDLNet_TCP_Send ( peers -> Item ( i ) . socket,
											data, 256 ) ;
		
		//fprintf ( stdout, "send %d chars to all sockets\n", num_chars ) ;
	}
	
	return num_chars ;
}

int Network::FindPeerBySocketI ( TCPsocket socket )
{
	for ( unsigned int i = 0 ; i < peers -> Count () ; i++ )
		if ( peers -> Item ( i ) . socket == socket )
			return i ;
	
	return -1 ;
}

Network::peer *Network::FindPeerBySocket ( TCPsocket socket )
{
	for ( unsigned int i = 0 ; i < peers -> Count () ; i++ )
		if ( peers -> Item ( i ) . socket == socket )
			return peers -> ItemPointer ( i ) ;
	
	return NULL ;
}

void Network::ConnectTo ( char *client, int port )
{
	Printf ( "|| connecting to %s at port %d...\n",
				client, port ) ;
	
	IPaddress ip ;
	SDLNet_ResolveHost(&ip, client, port);
	TCPsocket socket = SDLNet_TCP_Open ( &ip ) ;
	if ( ip.host == INADDR_NONE || socket == NULL )
	{
		Printf ( "|| can't connect to %s at port %d !\n",
					client, port ) ;
	}
	else
	{
		struct peer t_peer ;
		t_peer . type = NM_SERVER ;
		sprintf ( t_peer . name, "%s", client ) ;
		sprintf ( t_peer . alias, "%s", client ) ;
		t_peer . port = port ;
		t_peer . socket = socket ;
		SDLNet_ResolveHost(&(t_peer . peer), client, port);
		t_peer . established = false ;
		t_peer . timeout = SDL_GetTicks () + 30000 ;
		peers -> PushBack ( t_peer ) ;
		
		SDLNet_TCP_AddSocket(socketset, socket);
		
		SendMessage ( FindPeerBySocket ( socket ), PASS_PHRASE ) ;
	}
}

void Network::ListenOnPort ( int port )
{
	// Create the server socket
	SDLNet_ResolveHost(&serverIP, NULL, 5555);
	servsock = SDLNet_TCP_Open(&serverIP);
	if ( servsock == NULL )
	{
		Quit();
		Printf( "|| couldn't create server socket: %s\n",
				SDLNet_GetError());
		//serverIP = NULL ;
	}
	else
	{
		Printf ( "|| serving at port %d...\n", port ) ;
		SDLNet_TCP_AddSocket(socketset, servsock);
	}
}

void Network::Disconnect ( peer *peer )
{
	if ( peer )
	{
		char word [ 256 ] ;
		if ( peer -> timeout < SDL_GetTicks () )
			sprintf ( word, "%s", "lost" ) ;
		else
			sprintf ( word, "%s", "closed" ) ;

		if ( peer -> type == NM_SERVER )
			Printf ( "|| %s connection to server '%s'!\n", word, peer -> name ) ;
		else
			Printf ( "|| %s connection to client '%s'.\n", word, peer -> name ) ;
		
		int n = FindPeerBySocketI ( peer -> socket ) ;

		SDLNet_TCP_DelSocket(socketset, peer -> socket);
		SDLNet_TCP_Close(peer -> socket);
		
		peers -> Remove ( n ) ;
	}
}

/* obsolete */
void Network::StartAsServer ( int port ) {}
void Network::StartAsClient ( char *server, int port ) {}
int  Network::Connect () { return -1 ; }
void Network::Disconnect () {}
/* */

void Network::TCPAcceptEvent ( TCPsocket socket )
{
	IPaddress ip = *SDLNet_TCP_GetPeerAddress ( socket ) ;
	char name [ 256 ] ;

	ResolveIP ( &ip, (char*) name ) ;
	
	Printf ( "|| incoming connection from '%s' ( %08x ) on port %d...\n",
				name, ip . host, ip . port ) ;

	struct peer t_peer ;
	t_peer . type = NM_CLIENT ;
	sprintf ( t_peer . name, "%s", name ) ;
	sprintf ( t_peer . alias, "%s", name ) ;
	t_peer . port = ip . port ;
	t_peer . socket = socket ;
	t_peer . peer = ip ;
	t_peer . established = false ;
	t_peer . timeout = SDL_GetTicks () + 30000 ;
	t_peer . born = SDL_GetTicks () ;
	peers -> PushBack ( t_peer ) ;
	
	SDLNet_TCP_AddSocket(socketset, socket);
	
	Printf ( "|| host '%s' accepted. waiting for authentication...\n",
				name ) ;
}

bool Network::TCPReceiveEvent ( peer *peer )
{
	char string [ 8192 ] ;
	int  string_len = SDLNet_TCP_Recv ( peer -> socket, string, 8192 ) ;
	//printf ( "get %d chars from peer '%s':\n", string_len, peer->name ) ;
	if ( string_len > 0 )
	{
		string [ min ( sizeof ( string ), string_len ) ] = '\0' ;
		char host [ 256 ], name [ 256 ], alias [ 256 ] ;
		bool is_foreign_message = false ;
		
		//fprintf ( stdout, "string = '%s' (%d chars)\n", string, strlen(string) ) ;
		
		sprintf ( name, "%s", peer -> name ) ;
		sprintf ( alias, "%s", peer -> alias ) ;
		
		if ( strlen ( string ) > 5 && string [ 0 ] == 1 &&
				string [ 1 ] == 2 && string [ 2 ] == 3 )
		{
			char t_host [ 8192 ], t_alias [ 8192 ], t_message [ 8192 ] ;
			sprintf ( t_host, "%s", string + 3 ) ;
			*index ( t_host, ':' ) = '\0' ;
			sprintf ( t_alias , "%s", string + 3 + strlen ( t_host ) + 1 ) ;
			*index ( t_alias, ':' ) = '\0' ;
			sprintf ( t_message , "%s", string + 3 + strlen ( t_host ) +
													strlen ( t_alias ) + 2 ) ;
			//printf ( "t_host = '%s' ; t_alias = '%s' ; t_message = '%s'\n",
			//			t_host, t_alias, t_message ) ;
			
			sprintf ( string, "%s", t_message ) ;
			sprintf ( name, "%s", t_host ) ;
			sprintf ( alias, "%s", t_alias ) ;
			
			is_foreign_message = true ;
		}
		
		if ( strcmp ( name, alias ) != 0 )
		{
			sprintf ( host, "%s", alias ) ;
			if ( strlen ( name ) > 0 &&
				strcmp ( alias, aliasname . c_str () ) == 0 )
			{
				strcat ( host, " (" ) ;
				strcat ( host, name ) ;
				strcat ( host, ")" ) ;
			}
		}
		else
			ResolveIP ( &peer -> peer, (char*) host ) ;
		
		int type = NM_USER_MESSAGE ;
		if ( is_foreign_message &&
				( compare_start_of_string ( string, PASS_PHRASE ) ||
				  compare_start_of_string ( string, PASS_ANSWER ) ||
				  compare_start_of_string ( string, PASS_ESTABLISHED ) ||
				  compare_start_of_string ( string, PASS_ALIAS ) ||
				  compare_start_of_string ( string, PASS_PING ) ||
				  compare_start_of_string ( string, PASS_ALIVE ) ) )
		{
			// is a system message of another host! don't process!!!
			// perhaps there could be something like
			// ProcessRemoteMessages ( name, alias, string ) ;
			type = NM_PING_MESSAGE ;
		}
		else // local or direct communication
			type = ProcessMessage ( peer, string ) ;
		
		if ( type == NM_USER_MESSAGE )
			Printf ( "%s: %s\n", host, string ) ;
		else if ( type != NM_PING_MESSAGE )
			printf ( "|| internal message from %s%s: '%s'\n",
				peer -> type == NM_LOCAL ? "local " :
				peer -> type == NM_SERVER ? "server " :
				peer -> type == NM_CLIENT ? "client " :
				"UNKNOWN (BUG?!)",
				host, string ) ;

		// add new buddies
		if ( compare_start_of_string ( string, PASS_PING ) ||
			compare_start_of_string ( string, PASS_ALIVE ) )
		{
			struct peer *user = NULL ;
			// do we know this user?
			for ( unsigned int i = 0 ; i < seen_users -> Count () ; i++ )
				if ( strcmp ( seen_users -> Item ( i ) . name, name ) == 0 &&
						strcmp ( seen_users -> Item ( i ) . alias, alias ) == 0 )
					user = seen_users -> ItemPointer ( i ) ;
			
			if ( user )
			{
				// user is still active / alive
				user -> timeout = SDL_GetTicks () + 10000 ;
				
				// is this user already visible in our buddy-list?
				if ( ! user -> established )
				{
					if ( user -> born + 2000 < SDL_GetTicks () )
					{	// was the user active long enough?
						user -> established = true ;
						Printf ( "|| %s came to us!\n", host ) ;
					}
				}
			}
			else
			{
				struct peer t_user ;
				t_user . type = NM_CLIENT ;
				sprintf ( t_user . name, "%s", name ) ;
				sprintf ( t_user . alias, "%s", alias ) ;
				t_user . timeout = SDL_GetTicks () + 10000 ;
				t_user . established = false ;
				t_user . born = SDL_GetTicks () ;
				seen_users -> PushBack ( t_user ) ;
				SendUserData () ;
			}
		}

		return true ;
	}
	else
	{
		Printf ( "|| connection to '%s' has died.\n", peer -> name ) ;
		TCPCloseEvent ( peer ) ;
		
		return false ;
	}
}

void Network::TCPCloseEvent ( peer *peer )
{
	Disconnect ( peer ) ;
}

void Network::UpdateBuddyList ()
{	// update buddy list!
	unsigned int users_in_buddy_list = seen_users -> Count () ;
	unsigned int max_lines = max ( users_in_buddy_list, buddy_list -> NumLines () ) ;
	unsigned int uid = 0 ;
	for ( unsigned int i = 0 ; i < max_lines ; i++ )
	{
		if ( uid < users_in_buddy_list )
		{
			struct peer *user = seen_users -> ItemPointer ( uid ) ;
			
			// search next user
			while ( uid < users_in_buddy_list && ! user -> established )
				user = seen_users -> ItemPointer ( ++uid ) ;
			
			if ( user -> established )
				buddy_list -> SetLine ( i, user -> alias ) ;
			else
				buddy_list -> SetLine ( i, "" ) ;
			
			uid++ ;
		}
		else
			buddy_list -> SetLine ( i, "" ) ;
	}
}

void Network::ProcessEvents ( SDL_Event *event )
{
	static unsigned int users_in_buddy_list = seen_users -> Count () ;
	unsigned int i = 0 ;
	
	// wait a moment for events
	SDLNet_CheckSockets(socketset, 0);
	
	// Check for new connections
	if ( SDLNet_SocketReady(servsock) )
	{
		TCPsocket socket = SDLNet_TCP_Accept ( servsock ) ;
		if ( socket )
		{
			TCPAcceptEvent ( socket ) ;
		}
	}
	
	bool send_ping = false ;
	static Uint32 last_ping, this_ping = last_ping = SDL_GetTicks () ;
	this_ping = SDL_GetTicks () ;
	if ( this_ping - last_ping > 5000 )
	{
		last_ping = this_ping ;
		send_ping = true ;
	}
	
	for ( i = 0 ; i < peers -> Count () ; i++ )
	{
		bool still_active = true ;
		peer *peer = peers -> ItemPointer ( i ) ;
		
		// Check for events on existing client
		if ( SDLNet_SocketReady(peer -> socket) )
			still_active = TCPReceiveEvent ( peer ) ;
		
		if ( still_active )
		{
			if ( send_ping )
				SendMessage ( NULL, PASS_PING ) ;
			
			// has the connection timed out?
			if ( peer -> timeout < SDL_GetTicks () )
			{
				Disconnect ( peer ) ;
				continue ;
			}
		}
		else // connection died
			break ;
	}
	
	// remove stale buddies
	for ( unsigned int i = 0 ; i < seen_users -> Count () ; i++ )
	{
		if ( seen_users -> Item ( i ) . timeout < SDL_GetTicks () )
		{
			if ( seen_users -> Item ( i ) . established )
				Printf ( "|| %s left us.\n",
						seen_users -> Item ( i ) . alias ) ;
			seen_users -> Remove ( i ) ;
		}
	}
	
	if ( users_in_buddy_list != seen_users -> Count () && buddy_list )
	{	// update buddy list!
		users_in_buddy_list = seen_users -> Count () ;
		UpdateBuddyList () ;
	}
}

char *Network::StripHeadAndTail ( char *text )
{
	int  length = strlen ( text ) ;
	int  start, end, pos ;
	
	for ( pos = 0 ; pos < length ; pos++ )
		if ( text [ pos ] == 9 )
			text [ pos ] = ' ' ;
	
	for ( pos = 0 ; pos < length ; pos++ )
		if ( text [ pos ] != ' ' )
			break ;
	
	start = pos ;
	
	for ( pos = length - 1 ; pos > 0 ; pos-- )
	{
		if ( text [ pos ] != ' ' )
			break ;
	}
	
	end = pos + 1 ;
	
	for ( pos = start ; pos < end ; pos++ )
		text [ pos - start ] = text [ pos ] ;
	for ( pos = pos - start ; pos < length ; pos++ )
		text [ pos ] = '\0' ;
	
	return text ;
}

char *Network::StripMarks ( char *text )
{
	int old_pos = 0 ;
	
	for ( unsigned int pos = 0 ; pos < strlen ( text ) ; pos ++ )
	{
		if ( !( pos == 0 && text [ pos ] == '"' ) &&
				!( pos == strlen ( text ) - 1 && text [ pos ] == '"' ) )
			text [ old_pos++ ] = text [ pos ] ;
	}
	
	text [ old_pos ] = '\0' ;
	
	return text ;
}

GlDynArray < std::string > *Network::GetArgList ( char *string,
													char delimiter )
{
	GlDynArray < std::string > *bucket = new GlDynArray < std::string > () ;
	
	bucket -> Clear () ;
	unsigned int pos = 0 ;
	
	bool marks_found = FALSE ;
	char eop = 255 ; // end of part mark
	
	//fprintf ( stdout, "args =" ) ;
	
	char *copy = (char*) malloc ( sizeof (char) * 8192 ) ;
	sprintf ( copy, "%s", string ) ;
	
	while ( pos < strlen ( copy ) )
	{
		std::string arg ;
		for ( ; pos < strlen ( copy ) ; pos++ )
		{
			if ( copy [ pos ] == '"' ) marks_found = ! marks_found ;
			if ( copy [ pos ] == delimiter && ! marks_found ) break ;
			arg += copy [ pos ] ;
		}
		arg += '\0' ;
		if ( pos < strlen ( copy ) )
			copy [ pos ] = eop ;
		
		if ( (arg . c_str ()) [ 0 ] == eop ) arg = arg . c_str () + 1 ;

		arg = StripMarks ( StripHeadAndTail ( (char*) arg . c_str () ) ) ;
		
		if ( strlen ( arg . c_str () ) > 0 )
			bucket -> PushBack ( arg ) ;
		
		//fprintf ( stdout, " '%s'", (char*) bucket [ bucket . Count () - 1 ] . c_str () ) ;
	}
	
	//fprintf ( stdout, "\n" ) ;
	
	free ( copy ) ;
	
	return bucket ;
}

int Network::ProcessMessage ( peer *peer, char *t_message )
{
	int message_type = NM_USER_MESSAGE ; // default
	
	char message [ 8192 ] ;
	sprintf ( message, "%s", t_message ) ;
	
	if ( peer )
	{
		//fprintf ( stdout, "ProcessMessage from socket %d: %s (%d chars)\n",
		//			FindPeerBySocketI ( peer -> socket ), message, strlen ( message ) ) ;
		
		if ( peer -> established == false )
		{
			if ( strcmp ( message, PASS_PHRASE ) == 0 )
			{
				SendMessage ( peer, PASS_ANSWER ) ;
				message_type = NM_SYSTEM_MESSAGE ;
			}
			
			if ( strcmp ( message, PASS_ANSWER ) == 0 )
			{
				SendMessage ( peer, PASS_ESTABLISHED ) ;
				message_type = NM_SYSTEM_MESSAGE ;
			}
			
			if ( strcmp ( message, PASS_ESTABLISHED ) == 0 &&
					peer -> established == false )
			{
				if ( peer -> type != NM_SERVER )
				{
					SendMessage ( peer, PASS_ESTABLISHED ) ;
					Printf ( "|| %s connected to us.\n", peer -> alias ) ;
				}
				else
					Printf ( "|| successfully connected to %s.\n", peer -> alias ) ;
				SendUserData ( peer ) ;
				peer -> established = true ;
				message_type = NM_SYSTEM_MESSAGE ;
			}
		}
		
		if ( strcmp ( message, PASS_PING ) == 0 )
		{
			//Printf ( "|| got ping from %s! sending alive signal...\n",
			//			peer -> name ) ;
			SendMessage ( peer, PASS_ALIVE ) ;
			message_type = NM_PING_MESSAGE ;
		}
		
		if ( strcmp ( message, PASS_ALIVE ) == 0 )
		{
			//Printf ( "|| got alive signal from %s! resetting timer.\n",
			//			peer -> name ) ;
			peer -> timeout = SDL_GetTicks () + 30000 ;
			message_type = NM_PING_MESSAGE ;
		}
		
		// also send this message to everyone else!
		for ( unsigned int i = 0 ; i < peers -> Count () ; i++ )
		{
			struct peer *t_peer = peers -> ItemPointer ( i ) ;
			if ( t_peer -> socket != peer -> socket )
			{
				char ext_message [ 8192 ] ;
				sprintf ( ext_message, "%c%c%c%s:%s:%s", 1, 2, 3,
							peer -> name, peer -> alias, message ) ;
				SendMessage ( t_peer, ext_message ) ;
			}
		}
	}

	GlDynArray < std::string > *arg_list = GetArgList ( message ) ;
	if ( arg_list -> Count () > 0 )
	{
		// display about header: //about
		if ( strcmp ( arg_list -> Item ( 0 ) . c_str (), "//about" ) == 0 )
		{
			Printf ( "|| welcome to the gemhun network console !\n" ) ;
			Printf ( "||\n" ) ;
		}
		
		// display command-line help: //help
		if ( strcmp ( arg_list -> Item ( 0 ) . c_str (), "//help" ) == 0 )
		{
			if ( arg_list -> Count () == 1 )
			{	// display global help
				Printf ( "\n|| global help follows:\n" ) ;
				Printf ( "||\n" ) ;
				Printf ( "|| connect hostname [port]  -  connect with specified server at port ( defaults to 5555 )\n" ) ;
				Printf ( "|| alias new_username       -  change username to new_username\n" ) ;
				Printf ( "|| sockets                  -  list all connections\n" ) ;
				Printf ( "\n" ) ;
			}
			else if ( arg_list -> Count () == 2 )
			{	// display command-specific help
				char *command = (char*) arg_list -> Item ( 1 ) . c_str () ;
				Printf ( "\n|| help for '%s' command:\n", command ) ;
				Printf ( "\n" ) ;
			}
			else
			{
				Printf ( "\n|| help command can be used in two ways:\n" ) ;
				Printf ( "|| - give a name for the command you want help for\n" ) ;
				Printf ( "|| - don't specify anything and you'll get the full list\n\n" ) ;
			}
		}
		
		// connect to peer: //connect <hostname> [<port>]
		if ( strcmp ( arg_list -> Item ( 0 ) . c_str (), "//connect" ) == 0 )
		{
			message_type = NM_SYSTEM_MESSAGE ;
			if ( arg_list -> Count () >= 2 && arg_list -> Count () <= 3 )
			{
				char host [ 256 ] ;
				sprintf ( host, "%s", (char*) arg_list -> Item ( 1 ) . c_str () ) ;
				int port = 5555 ;
				
				if ( arg_list -> Count () == 3 )
					port = atoi ( arg_list -> Item ( 2 ) . c_str () ) ;
				
				ConnectTo ( host, port ) ;
			}
			else
			{
				Printf ( "\n|| connect command needs ca. 2 parameters:\n" ) ;
				Printf ( "|| - the name or the ip-address of the computer\n" ) ;
				Printf ( "|| - the port on which your partner listens on (optional / defaults to 5555)\n" ) ;
				Printf ( "|| example:\n" ) ;
				Printf ( "||   connect mycomputer\n" ) ;
				Printf ( "||   or\n" ) ;
				Printf ( "||   connect mycomputer 5555\n\n" ) ;
			}
		}
		
		// change username: //alias <new username>
		if ( strcmp ( arg_list -> Item ( 0 ) . c_str (), "//alias" ) == 0 )
		{
			message_type = NM_SYSTEM_MESSAGE ;
			if ( arg_list -> Count () == 2 )
			{
				aliasname = arg_list -> Item ( 1 ) ;
				SendUserData ( NULL, NM_ALIASNAME ) ;
			}
			else
			{
				Printf ( "\n|| alias command needs exactly 1 parameter:\n" ) ;
				Printf ( "|| - the new name (alias) of the computer\n" ) ;
				Printf ( "|| example:\n  alias oktay\n" ) ;
				Printf ( "|| this would display yourself as oktay.\n" ) ;
				Printf ( "|| note: you could also supply this: \"i'm the master\"\n\n" ) ;
			}
		}
		
		// recieve aliasname from peer
		if ( peer && strncmp ( message, PASS_ALIAS, strlen ( PASS_ALIAS ) ) == 0 )
		{
			message_type = NM_SYSTEM_MESSAGE ;
			char alias [ 256 ] ;
			sprintf ( alias, "%s", message + strlen ( PASS_ALIAS ) ) ;
			Printf ( "|| %s%s has renamed itself from '%s' to '%s'\n",
						peer -> type == NM_LOCAL ? "local " :
						peer -> type == NM_SERVER ? "server " :
						peer -> type == NM_CLIENT ? "client " :
						"",
						peer -> name,
						peer -> alias,
						alias ) ;
			sprintf ( peer -> alias, "%s", alias ) ;
		}
		
		// list sockets: //sockets
		if ( strcmp ( arg_list -> Item ( 0 ) . c_str (), "//sockets" ) == 0 )
		{
			message_type = NM_SYSTEM_MESSAGE ;
			if ( arg_list -> Count () == 1 )
			{
				if ( peers -> Count () == 0 )
					Printf ( "\n|| there are no open connections.\n" ) ;
				else
					Printf ( "\n|| found %d open connections. listing...\n",
								peers -> Count () ) ;
				
				for ( unsigned int i = 0 ; i < peers -> Count () ; i++ )
				{
					struct peer *t_peer = peers -> ItemPointer ( i ) ;
					int s = FindPeerBySocketI ( t_peer -> socket ) ;
					char appendix [ 256 ] ;
					if ( strcmp ( t_peer -> name, t_peer -> alias ) != 0 )
						sprintf ( appendix, "%s ( %s:%d )",
									t_peer -> alias,
									t_peer -> name,
									t_peer -> port ) ;
					else
						sprintf ( appendix, "%s:%d",
									t_peer -> name,
									t_peer -> port ) ;
					
					Printf ( "|| socket %d: %s%s\n",
								s,
								t_peer -> type == NM_LOCAL ? "local " :
								t_peer -> type == NM_SERVER ? "server " :
								t_peer -> type == NM_CLIENT ? "client " :
								"",
								appendix ) ;
				}
				Printf ( "\n" ) ;
			}
			else
			{
				Printf ( "\n|| sockets command doesn't need any arguments.\n\n" ) ;
			}
		}
	}

	return message_type ;
}

void Network::SendUserData ( peer *peer, int type )
{
	if ( ( type == NM_ALLDATA || type == NM_ALIASNAME ) &&
			strlen ( aliasname . c_str () ) > 0 &&
			strcmp ( aliasname . c_str (), "you" ) != 0 )
	{
		char alias [ 256 ] ;
		sprintf ( alias, "%s%s", PASS_ALIAS, aliasname . c_str () ) ;
		SendMessage ( peer, alias ) ;
	}
}

void Network::Quit ()
{
	peers -> Clear () ; delete peers ;
	seen_users -> Clear () ; delete seen_users ;
	
	console = NULL ;
	buddy_list = NULL ;
	
	if ( servsock != NULL )
	{
		SDLNet_TCP_Close(servsock);
		servsock = NULL;
	}
	if ( socketset != NULL )
	{
		SDLNet_FreeSocketSet(socketset);
		socketset = NULL;
	}
	SDLNet_Quit();
}

#endif
