#include "console.h"

#ifdef __NETWORK_SUPPORT__

#include <math.h>

#include "video.h"
extern Video video ;

#include "network.h"
extern Network network ;

void Console::Init ( int new_x, int new_y,
						int new_width, int new_height, int new_linespacing,
						KrImNode *new_basenode, KrScheme new_scheme,
						bool has_command_line )
{
	x = new_x ;
	y = new_y ;
	width = new_width ;
	height = new_height ;
	spacing = new_linespacing ;
	//fprintf ( stdout, "width = %d, height = %d, spacing = %d\n",
	//			width, height, spacing ) ;
	KrImNode *cc_base = new_basenode ;
	/*console = new KrConsole ( width, height, spacing, new_scheme ) ;
	video . GetEngine () -> Tree () -> AddNode ( cc_base, console ) ;
	console -> SetPos ( x, y ) ;*/

	int command_line_height = 0 ;
	if ( has_command_line )
		command_line_height = 12 ;
	
	cmdhistory = new KrTextBox ( new_scheme . font, width, height - command_line_height,
									spacing, KrTextBox::LEFT ) ;
	cmdhistory -> SetPos ( x, y ) ;
	video . GetEngine () -> Tree () -> AddNode ( cc_base, cmdhistory ) ;
	cmdhistorybuf = new GlDynArray < std::string > () ;
	cmdhistorybuf -> Clear () ;

	if ( has_command_line )
	{
		cmdline = new KrTextWidget ( width, command_line_height,
										false, false, true, new_scheme ) ;
		cmdline -> SetPos ( x, y + height - command_line_height );
		video . GetEngine () -> Tree () -> AddNode ( cc_base, cmdline ) ;
	}
	else
		cmdline = NULL ;
	
	cmdlinebuf . Clear () ;
	cmdlinebuf_pos = 0 ;
	cmdline_bak = "" ;
}

void Console::ProcessEvents ( SDL_Event event )
{
	switch ( event.type )
	{
		case SDL_KEYDOWN:
		{
			SDLKey sym = event.key.keysym.sym ;
			
			if ( ! cmdline ) break ;
			
			if ( sym == SDLK_RETURN )
			{
				std::string text ;
				cmdline -> GetTextChar ( &text ) ;
				
				uint entries = cmdlinebuf . Count () ;
				if ( ( entries && strlen ( text . c_str () ) &&
					strcmp ( cmdlinebuf [ entries - 1 ] . c_str (), text . c_str () ) != 0 ) ||
					! entries && strlen ( text . c_str () ) )
				{
					cmdlinebuf . PushBack ( text ) ;
				}
				cmdlinebuf_pos = cmdlinebuf . Count () ;
				
				PushText ( network . AddUsernameAtBeginning ( text ) ) ;
				network . ProcessMessage ( NULL, (char*) text . c_str () ) ;
				cmdline -> SetTextChar ( "" ) ;
			}
			
			if ( sym == SDLK_UP && cmdlinebuf_pos > 0 )
			{
				if ( cmdlinebuf_pos == cmdlinebuf . Count () )
					cmdline -> GetTextChar ( &cmdline_bak ) ;
				
				cmdlinebuf_pos-- ;
				cmdline -> SetTextChar ( cmdlinebuf . Item ( cmdlinebuf_pos ) ) ;
				
				//if ( cmdline_pos > strlen ( cmdline ) )
				//	cmdline_pos = strlen ( cmdline ) ;
			}
			
			if ( sym == SDLK_DOWN && cmdlinebuf_pos < cmdlinebuf . Count () )
			{
				cmdlinebuf_pos++ ;
				if ( cmdlinebuf_pos < cmdlinebuf .  Count () )
					cmdline -> SetTextChar ( cmdlinebuf . Item ( cmdlinebuf_pos ) ) ;
				else
					cmdline -> SetTextChar ( cmdline_bak ) ;
				
				//if ( cmdline_pos > strlen ( cmdline ) )
				//	cmdline_pos = strlen ( cmdline ) ;
			}
		}
		break ;
	}
}

void Console::PushText ( std::string text )
{
	//
	KrFontResource *font = (KrFontResource*) cmdhistory -> FontResource () ;
	int length = (int) floor ( width / font -> FontWidth1 ( 'W' ) ) ;
	GlDynArray < std::string > buffer = BreakTextToMaxCols ( text, length ) ;
	
	for ( unsigned int i = 0 ; i < buffer . Count () ; i++ )
		cmdhistorybuf -> PushBack ( buffer [ i ] ) ;
	//
	
	uint buffer_lines = cmdhistorybuf -> Count () ;
	uint screen_lines = cmdhistory -> NumLines () ;
	uint first_line = max ( buffer_lines - screen_lines, 0 ) ;
	
	for ( uint line = first_line ; line < first_line + screen_lines ; line++ )
	{
		if ( line < buffer_lines )
		{
			char *text = (char*) cmdhistorybuf -> Item ( line ) . c_str () ;
			cmdhistory -> SetTextChar ( text, line - first_line ) ;
		}
		else
			cmdhistory -> SetTextChar ( "", line - first_line ) ;
	}
}

std::string Console::GetLine ( int line )
{
	std::string text ;
	cmdhistory -> GetTextChar ( &text, line ) ;
	return text ;
}

void Console::SetLine ( int line, std::string text )
{
	cmdhistory -> SetTextChar ( text . c_str (), line ) ;
}

int Console::NumLines ()
{
	return cmdhistory -> NumLines () ;
}

void Console::Free ()
{
	if ( cmdline )
		video . GetEngine () -> Tree () -> DeleteNode ( cmdline ) ;
	video . GetEngine () -> Tree () -> DeleteNode ( cmdhistory ) ;
	cmdhistorybuf -> Clear () ;
	delete cmdhistorybuf ;
	cmdlinebuf . Clear () ;
}

#endif
