#include "filecache.h"

void FileCache::Init ()
{	// clear cache and initialize it
	//fprintf ( stdout, "FileCache::Init () ;\n" ) ;
	
	for ( int pos = 0 ; pos < FILECACHE_SIZE ; pos++ )
	{
		sprintf ( cache [ pos ] . filename, "%s", "" ) ;
		cache [ pos ] . data = NULL ;
		cache [ pos ] . ref_count = 0 ;
	}
}

int FileCache::SearchForFilename ( char *filename )
{	// search for filename ; returns pos or -1 = nothing found
	for ( int pos = 0 ; pos < FILECACHE_SIZE ; pos++ )
		if ( cache [ pos ] . filename != NULL &&
				strcmp ( cache [ pos ] . filename, filename ) == 0 )
			return pos ;
	
	return -1 ;
}

int FileCache::SearchForData ( void *data )
{	// search for data ; returns pos or -1 = nothing found
	for ( int pos = 0 ; pos < FILECACHE_SIZE ; pos++ )
		if ( cache [ pos ] . data == data )
			return pos ;
	
	return -1 ;
}

void *FileCache::GetFromCache ( char *filename )
{	// search for filename in cache and return *data or NULL
	//fprintf ( stdout, "FileCache::GetFromCache ( \"%s\" ) ;\n", filename ) ;
	
	int pos = SearchForFilename ( filename ) ;
	
	if ( pos != -1 )
	{
		//fprintf ( stdout, "cache [ %d ] : filename = '%s' has %d references.\n",
		//			pos, cache [ pos ] . filename, cache [ pos ] . ref_count ) ;
		return cache [ pos ] . data ;
	}
	else
		return NULL ;
}

int FileCache::PutIntoCache ( char *filename, void *data )
{	// put *data into cache if neseccary, increase and return ref_count
	//fprintf ( stdout, "FileCache::PutIntoCache ( \"%s\", void *data ) ;\n", filename ) ;
	
	int pos = SearchForFilename ( filename ) ;
	
	if ( pos != -1 )
	{
		cache [ pos ] . ref_count++ ;
		//fprintf ( stdout, "cache [ %d ] : filename = '%s' has %d references.\n",
		//			pos, cache [ pos ] . filename, cache [ pos ] . ref_count ) ;
		return pos ;
	}
	else
	{
		pos = SearchForFilename ( "" ) ;
		
		if ( pos != -1 )
		{
			cache [ pos ] . ref_count++ ;
			sprintf ( cache [ pos ] . filename, "%s", filename ) ;
			cache [ pos ] . data = data ;
			//fprintf ( stdout, "cache [ %d ] : filename = '%s' has %d references.\n",
			//			pos, cache [ pos ] . filename, cache [ pos ] . ref_count ) ;
			return pos ;
		}
	}
	
	DumpStatus () ;
	fprintf ( stdout, "!!! file cache is full! increase cache size!!!\n" ) ;
	
	return 0 ;
}

int  FileCache::IncreaseDependency ( int pos )
{
	if ( cache [ pos ] . ref_count == 0 && cache [ pos ] . data == NULL ) return -1 ;
	
	cache [ pos ] . ref_count++ ;
	
	return cache [ pos ] . ref_count ;
}

int FileCache::FreeFileFromCache ( char *filename )
{	// decrease ref_count
	//fprintf ( stdout, "FileCache::FreeFileFromCache ( \"%s\" ) ;\n", filename ) ;
	
	int pos = SearchForFilename ( filename ) ;
	
	if ( pos != -1 )
	{
		cache [ pos ] . ref_count-- ;
		if ( cache [ pos ] . ref_count == 0 &&
				cache [ pos ] . filename != NULL )
		{
			//fprintf ( stdout, "cache [ %d ] : filename = '%s' has %d references.\n",
			//			pos, cache [ pos ] . filename, cache [ pos ] . ref_count ) ;
			sprintf ( cache [ pos ] . filename, "%s", "" ) ;
		}
		return cache [ pos ] . ref_count ;
	}
	
	return 0 ;
}

int FileCache::FreeDataFromCache ( void *data )
{	// decrease ref_count
	//fprintf ( stdout, "FileCache::FreeDataFromCache ( void *data ) ;\n" ) ;
	
	int pos = SearchForData ( data ) ;
	
	if ( pos != -1 )
	{
		cache [ pos ] . ref_count-- ;
		if ( cache [ pos ] . ref_count == 0 &&
			cache [ pos ] . filename != NULL )
		{
			//fprintf ( stdout, "cache [ %d ] : filename = '%s' has %d references.\n",
			//			pos, cache [ pos ] . filename, cache [ pos ] . ref_count ) ;
			sprintf ( cache [ pos ] . filename, "%s", "" ) ;
		}
		return cache [ pos ] . ref_count ;
	}
	
	return 0 ;
}

void FileCache::Reorganize ()
{	// defragment cache
	//fprintf ( stdout, "FileCache::Reorganize () ;\n" ) ;
	
}

void FileCache::DumpStatus ()
{	// dump current cached files to console
	fprintf ( stdout, "dumping status of cached files...\n" ) ;
	
	int num_found = 0 ;
	
	for ( int pos = 0 ; pos < FILECACHE_SIZE ; pos++ )
	{
		if ( strcmp ( cache [ pos ] . filename, "" ) != 0 ||
				cache [ pos ] . ref_count > 0 )
		{
			fprintf ( stdout, "%02d. '%s' has %d references.\n",
						pos, 
						cache [ pos ] . filename, 
						cache [ pos ] . ref_count ) ;
			num_found++ ;
		}
	}
	
	fprintf ( stdout, "%d cached files found. done.\n", num_found ) ;
}

void FileCache::Quit ()
{	// free cache completely
	//fprintf ( stdout, "FileCache::Quit () ;\n" ) ;
	
	for ( int pos = 0 ; pos < FILECACHE_SIZE ; pos++ )
	{
		if ( cache [ pos ] . ref_count > 0 )
		{
			if ( strcmp ( cache [ pos ] . filename, "" ) != 0 )
				sprintf ( cache [ pos ] . filename, "%s", "" ) ;
			if ( cache [ pos ] . data != NULL )
				free ( cache [ pos ] . data ) ;
			cache [ pos ] . ref_count = 0 ;
		}
	}
}

// additional functions for kyra

void *FileCache::CachedLoad ( char *type, char *filename )
{
	if ( strcmp ( type, "vault" ) == 0 )
	{
		KrResourceVault *Vault = NULL ;
		
		Vault = (KrResourceVault*) GetFromCache ( filename ) ;
	
		if ( Vault == NULL )
		{
			Vault = new KrResourceVault () ;
	
			// load our gamebg
			if ( !Vault->LoadDatFile ( filename ) )
			{
				printf( "Error loading the %s file\n", filename ) ;
				exit ( 255 ) ;
			}
		}
	
		PutIntoCache ( filename, Vault ) ;
	
		return Vault ;
	}
	else
		return NULL ;
}

void FileCache::FreeCachedData ( char *type, void *object )
{
	if ( strcmp ( type, "vault" ) == 0 )
	{
		if ( FreeDataFromCache ( object ) == 0 )
		{
			//fprintf ( stdout,"delete object\n" ) ;
			delete (KrResourceVault*) object ;
		}
	}
}
