#include "fps.h"

extern FileCache file_cache ;

#include "video.h"
extern Video video ;

float current_framerate   = 0 ;
float current_lost_frames = 0 ;

void FPS::Init ()
{
	engine = video . GetEngine () ;
	
	start_time = SDL_GetTicks () ;

#if USE_FRAMERATE_VERSION == 0
	LastTick = SDL_GetTicks () ;
	ThisTick = LastTick ;
	LastCheckTick = LastTick ;
#endif
	framerate = (float) FRAMERATE ;
	lost_frames = 0 ;

	current_framerate = framerate ;
	current_lost_frames = lost_frames ;

#if USE_FRAMERATE_VERSION == 0
	ThisTick = LastTick = LastCheckTick = SDL_GetTicks () ;
#endif
	
	fontVault = (KrResourceVault*) file_cache . CachedLoad ( "vault",
													"gfx/fonts_para.dat" ) ;
	font = fontVault->GetFontResource ( "parafont" ) ;
	GLASSERT ( font ) ;
	textBox = new KrTextBox ( font, 200, 50, 0 ) ;
	if ( video . GetDisplayFramerate () == TRUE )
	{	
		char text [ 256 ] ;
		sprintf ( text, "%02d/%02d FPS", (int) current_framerate, FRAMERATE ) ;
		textBox -> SetTextChar ( text, 0 ) ;
	}
	else
	{
		textBox -> SetTextChar ( "", 0 ) ;
	}
	engine->Tree()->AddNode ( 0, textBox ) ;
	textBox -> SetZDepth ( 1000 ) ;
	textBox -> SetPos ( 5, 5 ) ;
	
	if ( video . GetUsingOpenGL () )
	{	// haven't tested this much..
		fps_throttle_normal = (int) ( (float) FRAMERATE / 2.50 ) ; // was 5.00
		fps_throttle_paused = 0 ;
	}
	else
	{	// changed 1.75 to 2.50 to get more responsibility...
		fps_throttle_normal = (int) ( (float) FRAMERATE / 2.50 ) ; // was 1.75
		fps_throttle_paused = -5 ;
	}
	
	SetSpeedToNormal () ;
}

void FPS::Update ()
{
#if USE_FRAMERATE_VERSION == 0
	ThisTick = SDL_GetTicks () ;
	lost_frames = ( (float) ( ThisTick - LastTick ) / FRAMEDELAY ) ;
	framerate = (float) FRAMERATE / lost_frames ;
	lost_frames -= 1.0 ;
	LastTick = ThisTick ;
#elif USE_FRAMERATE_VERSION == 1
	// These are to calculate our fps
	static Uint32 T0     = 0 ;
	static Uint32 Frames = 0 ;
	
	// Gather our frames per second
	Frames++ ;
	{
		Uint32 t = SDL_GetTicks () ;
		if ( t - T0 > 0 ) 
		{
			float seconds = ( t - T0 ) / 1000.0 ;
			float fps = Frames / seconds ;
			current_framerate = fps ;
			current_lost_frames = FRAMERATE - fps ;
			T0 = t ;
			Frames = 0 ;
		}
	}
#else
	static long tick = 0 ;
	static float fps = 0 ;
	static int fps_count = 0, fps_start = 0 ;
	tick = SDL_GetTicks () ;
	
	if ( tick > fps_start )
	{
		fps = (float) fps_count * 1000.0 / ( tick - fps_start ) ;
		fps_count = 0 ;
		fps_start = tick ;
		current_framerate = fps ;
		current_lost_frames = FRAMERATE - fps ;
	}
	// print fps here if you want :)
	fps_count++ ;
#endif
}

void FPS::DumpStatus ()
{
	fprintf ( stdout, "current framerate = %f of %d ",
				current_framerate, FRAMERATE ) ;
	fprintf ( stdout, "current lost frames = %f\n",
				current_lost_frames ) ;
}

void FPS::PrintStatistics ()
{
	Uint32 i = SDL_GetTicks () - start_time;
	Uint32 ms = i % 1000 ;
	Uint32 s = ( i / 1000 ) % 60 ;
	Uint32 m = ( i / 1000 / 60 ) % 60 ;
	Uint32 h = ( i / 1000 / 60 / 60 ) % 60 ;
	Uint32 d = i / 1000 / 60 / 60 / 24 ;
	printf("\n----------------------------------------------------------------------\n");
	printf("          Total time running: %d days %dh %dm %ds %dms (%d ticks)\n",
			d, h, m, s, ms, i);
	if(!i) i = 1;
	printf("Average rendering frame rate: %.2f fps\n",
			video . GetRenderedFrames () * 1000.0 / i);
	printf("    Average logic frame rate: %.2f fps\n",
			video . GetLogicFrames () * 1000.0 / i);
	printf("----------------------------------------------------------------------\n\n");
}

void FPS::DrawFPSText ( int fps )
{
	if ( video . GetDisplayFramerate () == TRUE )
	{	
		char text [ 256 ] ;
		clamp ( &fps, 0, FRAMERATE ) ;
		sprintf ( text, "%02d/%02d FPS", fps, FRAMERATE ) ;
		textBox -> SetTextChar ( text, 0 ) ;
	}
	else
	{
		textBox -> SetTextChar ( "", 0 ) ;
	}
}

float FPS::GetFramerate ()
{
	return current_framerate ;
}

float FPS::GetFramesLost ()
{
	return current_lost_frames ;
}

void FPS::Draw ()
{
#if USE_FRAMERATE_VERSION == 0
	if ( ThisTick - LastCheckTick >= 1000 )
	{
		current_framerate = framerate ;
		current_lost_frames = lost_frames ;
		DrawFPSText ( (int) current_framerate ) ;
		LastCheckTick = ThisTick ;
	}
#elseif USE_FRAMERATE_VERSION == 1
	// These are to calculate our fps
	/*static Uint32 T0     = 0 ;
	static Uint32 Frames = 0 ;
	
	// Gather our frames per second
	Frames++ ;
	{
		Uint32 t = SDL_GetTicks () ;
		if ( t - T0 >= 500 ) 
		{
			float seconds = ( t - T0 ) / 1000.0 ;
			float fps = Frames / seconds ;
			DrawFPSText ( fps ) ;
			//printf ( "%d frames in %g seconds = %g FPS\n", Frames, seconds, fps ) ;
			current_framerate = fps ;
			current_lost_frames = FRAMERATE - current_framerate ;
			T0 = t ;
			Frames = 0 ;
		}
	}*/
	DrawFPSText ( current_framerate ) ;
#else
	static Uint32 last_tick = 0, this_tick = 0 ;
	this_tick = SDL_GetTicks () ;
	if ( this_tick - last_tick >= 1000 )
	{
		DrawFPSText ( (int) current_framerate ) ;
		last_tick = this_tick ;
	}
#endif
}

bool FPS::FpsReached ()
{
#if USE_FRAMERATE_VERSION == 0
	return ( (int) framerate >= FRAMERATE - 10 ? TRUE : FALSE ) ;
#else //if USE_FRAMERATE_VERSION == 1
	bool frate_reached = ((int) current_framerate) >= FRAMERATE - fps_throttle ;
	return ( frate_reached ) ;
//#else
//	return ( (int) framerate >= FRAMERATE ? TRUE : FALSE ) ;
#endif
}

void FPS::SetSpeedToSlow ()
{	// this one tries to slow down the framerate to save cpu-power
	fps_throttle = fps_throttle_paused ;
	fprintf ( stdout, "slow down framerate a bit...\n" ) ;
}

void FPS::SetSpeedToNormal ()
{	// this one tries to get quiet an accurate framerate
	fps_throttle = fps_throttle_normal ;
	fprintf ( stdout, "set framerate to normal speed...\n" ) ;
}

void FPS::Quit ()
{
	engine->Tree()->DeleteNode ( textBox ) ;
	file_cache . FreeCachedData ( "vault", fontVault ) ;
}
