#include "game.h"

#include "hid.h"
extern HID hid ;

extern char rc_dir[] ;

extern FileCache file_cache ;

#include "video.h"
extern Video video ;

// routines for displaying a splash screen
#include "splash.h"
extern Splash *splash ;

void Game::Init ( char *pkg_filename, char *left_level, char *right_level,
					int left_ai, int right_ai )
{
	if ( splash ) splash -> SetPercent ( 0 ) ;
	
	if ( strncmp ( pkg_filename, "savestate:", 10 ) == 0 )
	{
		{
			char string [ 256 ] ;
			sprintf ( string, "%s", pkg_filename + 10 ) ;
			sprintf ( pkg_filename, "%s", string ) ;
		}
		
		fprintf ( stdout, "Loading saved game '%s'...\n", pkg_filename ) ;
		video . DisplayOSDMessage ( "loading saved game..." ) ;
		
		FILE *stream ;
		char fname [ 256 ] ;
		sprintf ( fname, "%s/%s.state", rc_dir, pkg_filename ) ;
		stream = fopen ( fname, "r" ) ;
		
		if ( stream )
		{
			char name [ 4096 ] ;
			fgets ( name, 4096, stream ) ;
			
			sprintf ( left_level,  pkg_filename ) ;
			sprintf ( right_level, pkg_filename ) ;
			
			sprintf ( pkg_filename, "%s", name ) ;
			fprintf ( stdout, "'%s'\n", pkg_filename ) ;
			
			fclose ( stream ) ;
		}
	}
	else
	{
		char text [ 256 ], file [ 256 ] ;
		if ( rindex ( pkg_filename, '/' ) )
			sprintf ( file, "%s", rindex ( pkg_filename, '/' ) + 1 ) ;
		else
			sprintf ( file, "%s", pkg_filename ) ;
		sprintf ( text, "loading %s...\n", file ) ;
		video . DisplayOSDMessage ( text ) ;
	}
	
	if ( splash ) splash -> SetPercent ( 5 ) ;
	
	///
	
	pax = new Pax () ;
	pax -> Init ( pkg_filename ) ;
	
	///
	
	if ( splash ) splash -> SetPercent ( 70 ) ;
	
	background.Init ( pax ) ;
	
	if ( splash ) splash -> SetPercent ( 80 ) ;
	
	// now load our gems
	pax -> ReadPax ( pax -> GetStringVar ( "gemsfile" ) ) ;
	
	playfield_left.Init ( "left", pax, left_level, left_ai ) ;
	playfield_right.Init ( "right", pax, right_level, right_ai ) ;
	
	//playfield_left.DisableFallingGems () ;
	//playfield_right.DisableFallingGems () ;
	
	if ( splash ) splash -> SetPercent ( 90 ) ;
	
	GEM_COUNT       = pax -> GetIntVar ( "gem_count" ) ;
	GEM_FIRST_COLOR = pax -> GetCharFromStringVar ( "gem_first_color" ) ;
	GEM_BLOCK       = pax -> GetCharFromStringVar ( "gem_block_char" ) ;
	
	Vault = (KrResourceVault*) file_cache . CachedLoad ( "vault", "gfx/misc.dat" ) ;
	
	SetupGetReadySetGoText ( "background" ) ;
	SetupGetReadySetGoText ( "playfield_osd_left" ) ;
	SetupGetReadySetGoText ( "playfield_osd_right" ) ;

	ResetGetReadySetGoText ( "background" ) ;
	ResetGetReadySetGoText ( "playfield_osd_left" ) ;
	ResetGetReadySetGoText ( "playfield_osd_right" ) ;
	
	if ( splash ) splash -> SetPercent ( 95 ) ;
	
	go_sound . LoadSound ( "sfx/fanfare.wav" ) ;
	go_sound . SetVolume ( 128 ) ;
	
	new_game = TRUE ;
	playfield_left_new = FALSE ;
	playfield_right_new = FALSE ;
	
	paused = FALSE ;
	
	quit_state = false ;
	
	if ( splash ) splash -> SetPercent ( 100 ) ;
	
	video . DisplayOSDMessage ( "loading done." ) ;
}

void Game::ProcessEvents ( SDL_Event event )
{
	switch ( event.type )
	{
		case SDL_KEYDOWN:
		{
			SDLKey sym = event.key.keysym.sym ;

			// toggle music on/off
			if ( sym == SDLK_F2 )
				background.ToggleMusic () ;
			
			// dump left playfield and gem status
			if ( sym ==  SDLK_1 && playfield_left . GetIsFreed () == false )
				playfield_left.DumpStatus () ;
			// dump right playfield and gem status
			if ( sym ==  SDLK_2 && playfield_right . GetIsFreed () == false )
				playfield_right.DumpStatus () ;
			
		}
		break ;

		default:
		break ;
	}
	
	if ( hid.EventFound ( event ) )
	{
		// left player
		if ( playfield_left . GetIsFreed () == false )
		{
			if ( hid.Left_up_pressed ( event ) )
				playfield_left.MoveUserGemUpStart () ;
			if ( hid.Left_left_pressed ( event ) )
				playfield_left.MoveUserGemLeftStart () ;
			if ( hid.Left_right_pressed ( event ) )
				playfield_left.MoveUserGemRightStart () ;
	
			if ( hid.Left_fire_pressed ( event ) )
				playfield_left.MoveUserGemToTop () ;
			
			if ( hid.Left_down_pressed ( event ) )
			{
				/*int score = playfield_left.PlayerScoreGet () ;
				if ( score - 25 >= 0 )
				{
					playfield_left.GemQueueAddToOutgoing ( GEM_BLOCK, 5 ) ;
					playfield_left.PlayerScoreSet ( score - 25 ) ;
				}*/
				playfield_left.RotateSlaveGem () ;
			}
			
			if ( hid.Left_up_released ( event ) )
				playfield_left.MoveUserGemUpStop () ;
			if ( hid.Left_left_released ( event ) )
				playfield_left.MoveUserGemLeftStop () ;
			if ( hid.Left_right_released ( event ) )
				playfield_left.MoveUserGemRightStop () ;
			
		}
		
		// right player
		if ( playfield_left . GetIsFreed () == false )
		{
			if ( hid.Right_up_pressed ( event ) )
				playfield_right.MoveUserGemUpStart () ;
			if ( hid.Right_left_pressed ( event ) )
				playfield_right.MoveUserGemLeftStart () ;
			if ( hid.Right_right_pressed ( event ) )
				playfield_right.MoveUserGemRightStart () ;
	
			if ( hid.Right_fire_pressed ( event ) )
				playfield_right.MoveUserGemToTop () ;
	
			if ( hid.Right_down_pressed ( event ) )
			{
				/*int score = playfield_right.PlayerScoreGet () ;
				if ( score - 25 >= 0 )
				{
					playfield_right.GemQueueAddToOutgoing ( GEM_BLOCK, 5 ) ;
					playfield_right.PlayerScoreSet ( score - 25 ) ;
				}*/
				playfield_right.RotateSlaveGem () ;
			}
			
			if ( hid.Right_up_released ( event ) )
				playfield_right.MoveUserGemUpStop () ;
			if ( hid.Right_left_released ( event ) )
				playfield_right.MoveUserGemLeftStop () ;
			if ( hid.Right_right_released ( event ) )
				playfield_right.MoveUserGemRightStop () ;
		}
	}

	if ( playfield_left . GetIsFreed () == false &&
			playfield_right . GetIsFreed () == false )
	{
		for ( int i = 0 ; i < GEM_COUNT ; i++ )
		{
			char gem_color = GEM_FIRST_COLOR + i ;
			
			if ( playfield_left.GemQueueGetFromOutgoing ( gem_color ) )
				playfield_right.GemQueueAddToIncoming ( gem_color, 
						playfield_left.GemQueueGetFromOutgoing ( gem_color, true ) ) ;
	
			if ( playfield_right.GemQueueGetFromOutgoing ( gem_color ) )
				playfield_left.GemQueueAddToIncoming ( gem_color, 
						playfield_right.GemQueueGetFromOutgoing ( gem_color, true ) ) ;
		}
		if ( playfield_left.GemQueueGetFromOutgoing ( GEM_BLOCK ) )
			playfield_right.GemQueueAddToIncoming ( GEM_BLOCK, 
					playfield_left.GemQueueGetFromOutgoing ( GEM_BLOCK, true ) ) ;
	
		if ( playfield_right.GemQueueGetFromOutgoing ( GEM_BLOCK ) )
			playfield_left.GemQueueAddToIncoming ( GEM_BLOCK, 
					playfield_right.GemQueueGetFromOutgoing ( GEM_BLOCK, true ) ) ;
	}
}

void Game::Pause ()
{
	background.Pause () ;
	playfield_left.Pause () ;
	playfield_right.Pause () ;
	paused = TRUE ;
}

void Game::UnPause ()
{
	background.UnPause () ;
	playfield_left.UnPause () ;
	playfield_right.UnPause () ;
	paused = FALSE ;
}

void Game::SetupGetReadySetGoText ( char *surface )
{
	Sprite *gr_text, *s_text, *g_text ;
	int use_cached_scaling = AUTO ;
	GlFixed scale = 1 ;
	
	if ( strcmp ( surface, "background" ) == 0 )
	{
		gr_text = &get_ready_text ;
		s_text  = &set_text ;
		g_text  = &go_text ;
	}
	else if ( strcmp ( surface, "playfield_osd_left" ) == 0 )
	{
		gr_text = &get_ready_text_left ;
		s_text  = &set_text_left ;
		g_text  = &go_text_left ;
		
		use_cached_scaling = TRUE ;
		
		scale = 0.5 ;
	}
	else if ( strcmp ( surface, "playfield_osd_right" ) == 0 )
	{
		gr_text = &get_ready_text_right ;
		s_text  = &set_text_right ;
		g_text  = &go_text_right ;
		
		use_cached_scaling = TRUE ;
		
		scale = 0.5 ;
	}
	else
		return ;
	
	gr_text -> Init ( Vault, "get_ready", 
						pax -> GetNode ( pax -> GetDeclaredObjectFromList ( surface ), surface ),
						use_cached_scaling ) ;
	gr_text -> Glow ( TRUE ) ;
	gr_text -> SetMoveSound ( "sfx/malus.wav", 128 ) ;
	gr_text -> SetStopSound ( NULL, 0 ) ;
	if ( scale != 1 )
		gr_text -> GetSprite () -> Parent () -> SetScale ( scale, scale ) ;
	gr_text -> SetZ ( 100 ) ;
	
	s_text -> Init ( Vault, "set",
						pax -> GetNode ( pax -> GetDeclaredObjectFromList ( surface ), surface ),
						use_cached_scaling ) ;
	s_text -> Glow ( TRUE ) ;
	s_text -> SetMoveSound ( "sfx/malus.wav", 128 ) ;
	s_text -> SetStopSound ( NULL, 0 ) ;
	s_text -> SetZ ( 100 ) ;
	
	g_text -> Init ( Vault, "go",
						pax -> GetNode ( pax -> GetDeclaredObjectFromList ( surface ), surface ),
						use_cached_scaling ) ;
	g_text -> Glow ( TRUE ) ;
	g_text -> SetMoveSound ( "sfx/malus.wav", 128 ) ;
	g_text -> SetStopSound ( NULL, 0 ) ;
	g_text -> SetZ ( 100 ) ;
}

void Game::ResetGetReadySetGoText ( char *surface )
{
	Sprite *gr_text, *s_text, *g_text ;
	int x = -1, y = -1 ;
	int gr_slideinfrom, s_slideinfrom, g_slideinfrom ;
	
	if ( strcmp ( surface, "background" ) == 0 )
	{
		gr_text = &get_ready_text ;
		s_text  = &set_text ;
		g_text  = &go_text ;
		
		text_state = 0 ;
		text_counter = 0 ;
		
		gr_slideinfrom = -video . GetScreenX () ;
		s_slideinfrom  =  video . GetScreenX () ;
		g_slideinfrom  = -video . GetScreenX () ;
	}
	else if ( strcmp ( surface, "playfield_osd_left" ) == 0 )
	{
		gr_text = &get_ready_text_left ;
		s_text  = &set_text_left ;
		g_text  = &go_text_left ;
		
		text_state_left = 0 ;
		text_counter_left = 0 ;
		
		x = -15 ;
		y = 300 ;
		
		gr_slideinfrom = s_slideinfrom = g_slideinfrom = -video . GetScreenX () / 2 ;
	}
	else if ( strcmp ( surface, "playfield_osd_right" ) == 0 )
	{
		gr_text = &get_ready_text_right ;
		s_text  = &set_text_right ;
		g_text  = &go_text_right ;

		text_state_right = 0 ;
		text_counter_right = 0 ;
		
		x = -15 ;
		y = 300 ;
		
		gr_slideinfrom = s_slideinfrom = g_slideinfrom = video . GetScreenX () / 2 ;
	}
	else
		return ;
	
	if ( x == -1 && y == -1 )
		gr_text -> SetXY_Center () ;
	else
		gr_text -> SetXY ( x, y ) ;
	gr_text -> SlideInFrom ( gr_slideinfrom, 0 ) ;
	gr_text -> GetSprite () -> SetVisible ( FALSE ) ;
	
	if ( x == -1 && y == -1 )
		s_text -> SetXY_Center () ;
	else
		s_text -> SetXY ( x, y ) ;
	s_text -> SlideInFrom ( s_slideinfrom, 0 ) ;
	s_text -> GetSprite () -> SetVisible ( FALSE ) ;
	
	if ( x == -1 && y == -1 )
		g_text -> SetXY_Center () ;
	else
		g_text -> SetXY ( x, y ) ;
	//g_text -> SlideInFrom ( g_slideinfrom, 0 ) ;
	g_text -> GetSprite () -> SetVisible ( FALSE ) ;
}

void Game::UpdateGetReadySetGoTextBig ()
{
	if ( playfield_left  . GetIsQuitting () ||
		 playfield_right . GetIsQuitting () )
	{
		get_ready_text . SetAlpha ( get_ready_text . GetAlpha () - 25 ) ;
		set_text       . SetAlpha ( set_text       . GetAlpha () - 25 ) ;
		go_text        . SetAlpha ( go_text        . GetAlpha () - 25 ) ;
	}
	
	if ( text_counter > 100 || ( text_state > 0 && text_state < 2 ) )
	{
		if ( get_ready_text . GetSprite () -> IsVisible () == FALSE )
			get_ready_text . GetSprite () -> SetVisible ( TRUE ) ;
		get_ready_text . Draw () ;
	}
	if ( text_state > 0 && text_state < 3 )
	{
		if ( set_text . GetSprite () -> IsVisible () == FALSE )
			set_text . GetSprite () -> SetVisible ( TRUE ) ;
		get_ready_text . Draw () ;
		set_text . Draw () ;
	}
	if ( text_state > 1 )
	{
		if ( go_text . GetSprite () -> IsVisible () == FALSE )
			go_text . GetSprite () -> SetVisible ( TRUE ) ;
		go_text . Draw () ;
	}
	
	if ( text_state == 0 && text_counter >= 250 &&
		 playfield_left  . PlayfieldLayoutIsSet () &&
		 playfield_right . PlayfieldLayoutIsSet () )
	{
		text_state = 1 ;
		get_ready_text . SlideOutTo ( video . GetScreenX (), 0 ) ;
		
		text_counter = 0 ;
	}
	if ( text_state == 1 &&
		 playfield_left  . PlayfieldIsSet () == TRUE &&
		 playfield_right . PlayfieldIsSet () == TRUE &&
		 get_ready_text . SlideFinished () && text_counter >= 150 )
	{
		text_state = 2 ;
		set_text . SlideOutTo ( -video . GetScreenX (), 0 ) ;
		
		background . SetPlayfieldIsReady ( TRUE ) ;
		playfield_left  . SetPlayfieldIsReady ( TRUE ) ;
		playfield_right . SetPlayfieldIsReady ( TRUE ) ;
		
		go_sound . PlaySound () ;
		
		text_counter = 0 ;
	}
	if ( text_state == 2 && text_counter >= 100 &&
		 set_text . SlideFinished () )
	{
		text_state = 3 ;
		go_text . SlideOutTo ( video . GetScreenX (), 0 ) ;
		
		text_counter = 0 ;
		
		//playfield_left  . DisableFallingGems () ;
		//playfield_right . DisableFallingGems () ;
	}
	if ( text_state < 3 )
		text_counter++ ;
	
	if ( text_state >= 3 && go_text . SlideFinished () )
	{
		new_game = FALSE ;
		ResetGetReadySetGoText ( "background" ) ;
	}
}

void Game::UpdateGetReadySetGoTextSmall ( char *surface )
{
	Sprite *gr_text, *s_text, *g_text ;
	GamePF *playfield ;
	int *t_counter, *t_state ;
	int gr_slideoutto, s_slideoutto, g_slideoutto ;
	bool *playfield_new ;
	
	if ( strcmp ( surface, "playfield_osd_left" ) == 0 )
	{
		gr_text   = &get_ready_text_left ;
		s_text    = &set_text_left ;
		g_text    = &go_text_left ;
		t_counter = &text_counter_left ;
		t_state   = &text_state_left ;
		playfield = &playfield_left ;
		gr_slideoutto = -video . GetScreenX () ;
		s_slideoutto  = -video . GetScreenX () ;
		g_slideoutto  = -video . GetScreenX () ;
		playfield_new = &playfield_left_new ;
	}
	else if ( strcmp ( surface, "playfield_osd_right" ) == 0 )
	{
		gr_text   = &get_ready_text_right ;
		s_text    = &set_text_right ;
		g_text    = &go_text_right ;
		t_counter = &text_counter_right ;
		t_state   = &text_state_right ;
		playfield = &playfield_right ;
		gr_slideoutto = video . GetScreenX () ;
		s_slideoutto  = video . GetScreenX () ;
		g_slideoutto  = video . GetScreenX () ;
		playfield_new = &playfield_right_new ;
	}
	else
	{
		fprintf ( stdout, "Can't handle this playfield: '%s'\n", surface ) ;
		return ;
	}
	
	if ( playfield -> GetIsQuitting () )
	{
		gr_text -> SetAlpha ( gr_text -> GetAlpha () - 25 ) ;
		s_text  -> SetAlpha ( s_text  -> GetAlpha () - 25 ) ;
		g_text  -> SetAlpha ( g_text  -> GetAlpha () - 25 ) ;
	}
	
	if ( *t_counter > 100 || ( *t_state > 0 && *t_state < 2 ) )
	{
		if ( gr_text -> GetSprite () -> IsVisible () == FALSE )
			gr_text -> GetSprite () -> SetVisible ( TRUE ) ;
		gr_text -> Draw () ;
	}
	if ( *t_state > 0 && *t_state < 3 )
	{
		if ( s_text -> GetSprite () -> IsVisible () == FALSE )
			s_text -> GetSprite () -> SetVisible ( TRUE ) ;
		s_text -> Draw () ;
	}
	if ( *t_state > 1 )
	{
		if ( g_text -> GetSprite () -> IsVisible () == FALSE )
			g_text -> GetSprite () -> SetVisible ( TRUE ) ;
		g_text -> Draw () ;
	}
	
	if ( *t_state == 0 && *t_counter >= 250 &&
		 playfield -> PlayfieldLayoutIsSet () )
	{
		*t_state = 1 ;
		gr_text -> SlideOutTo ( gr_slideoutto, 0 ) ;
		
		*t_counter = 0 ;
	}
	if ( *t_state == 1 && *t_counter >= 100 &&
		 playfield -> PlayfieldIsSet () == TRUE &&
		 gr_text -> SlideFinished () &&
		 s_text -> SlideFinished () )
	{
		*t_state = 2 ;
		s_text -> SlideOutTo ( s_slideoutto, 0 ) ;
		
		playfield -> SetPlayfieldIsReady ( TRUE ) ;
		
		go_sound . PlaySound () ;
		
		*t_counter = 0 ;
	}
	if ( *t_state == 2 && *t_counter >= 100 )
	{
		*t_state = 3 ;
		g_text -> SlideOutTo ( g_slideoutto, 0 ) ;
		
		*t_counter = 0 ;
	}
	if ( *t_state < 3 )
		(*t_counter) ++ ;
	
	if ( *t_state >= 3 && g_text -> SlideFinished () )
	{
		*playfield_new = FALSE ;
		ResetGetReadySetGoText ( surface ) ;
	}
}

void Game::FreeGetReadySetGoText ( char *surface )
{
	Sprite *gr_text, *s_text, *g_text ;
	
	if ( strcmp ( surface, "background" ) == 0 )
	{
		gr_text = &get_ready_text ;
		s_text  = &set_text ;
		g_text  = &go_text ;
	}
	else if ( strcmp ( surface, "playfield_osd_left" ) == 0 )
	{
		gr_text = &get_ready_text_left ;
		s_text  = &set_text_left ;
		g_text  = &go_text_left ;
	}
	else if ( strcmp ( surface, "playfield_osd_right" ) == 0 )
	{
		gr_text = &get_ready_text_right ;
		s_text  = &set_text_right ;
		g_text  = &go_text_right ;
	}
	else return ;
	
	gr_text -> Free () ;
	s_text  -> Free () ;
	g_text  -> Free () ;
}

void Game::Draw ()
{
	if ( quit_state == true ) return ;
	
	background.Draw () ;
	if ( background.PlayfieldIsReady () )
	{
		playfield_left.Draw () ;
		playfield_right.Draw () ;
		
		if ( paused == FALSE )
		{
			if ( new_game )
			{
				UpdateGetReadySetGoTextBig () ;
			}
			else
			{
				if ( playfield_left_new == TRUE )
					UpdateGetReadySetGoTextSmall ( "playfield_osd_left" ) ;
				else
				{
					if ( playfield_left.PlayfieldLayoutIsSet () == FALSE &&
							playfield_left.PlayfieldIsSet () == FALSE )
						playfield_left_new = TRUE ;
				}
					
				if ( playfield_right_new == TRUE )
					UpdateGetReadySetGoTextSmall ( "playfield_osd_right" ) ;
				else
				{
					if ( playfield_right.PlayfieldLayoutIsSet () == FALSE &&
							playfield_right.PlayfieldIsSet () == FALSE )
						playfield_right_new = TRUE ;
				}
			}
		}
	}
}

void Game::AfterDraw ()
{
	if ( playfield_left.GetIsDone () && playfield_right.GetIsDone () &&
			background.GetIsQuitting () == FALSE ) 
	{
		FreeGetReadySetGoText ( "background" ) ;
		FreeGetReadySetGoText ( "playfield_osd_left" ) ;
		FreeGetReadySetGoText ( "playfield_osd_right" ) ;
		go_sound . UnloadSound () ;
		if ( file_cache . FreeDataFromCache ( Vault ) == 0 )
			delete Vault ;
		
		background.Quit () ;
	}
}

bool Game::IsQuitting ()
{
	if ( background.GetIsQuitting () || 
			playfield_left.GetIsQuitting () ||
			playfield_right.GetIsQuitting () )
		return TRUE ;
	else
		return FALSE ;
}

bool Game::IsDone ()
{
	if ( quit_state == false &&
			background.GetIsDone() &&
			playfield_left.GetIsDone () &&
			playfield_right.GetIsDone () )
	{
		FILE *stream ;
		char fname [ 256 ] ;
		sprintf ( fname, "%s/autosave.state", rc_dir ) ;
		stream = fopen ( fname, "w" ) ;
		if ( stream )
		{
			fprintf ( stream, "%s", pax -> GetPkgFilename () ) ;
			fclose ( stream ) ;
		}
		
		playfield_left.Free () ;
		playfield_right.Free () ;
		pax -> FreeObjects () ;
		delete pax ;
		quit_state = true ;
	}
	
	return quit_state ;
}

void Game::Quit ()
{
	if ( background.PlayfieldIsReady () == FALSE ) return ;
	if ( playfield_left.GetIsQuitting () == FALSE ) playfield_left.Quit () ;
	if ( playfield_right.GetIsQuitting () == FALSE ) playfield_right.Quit () ;
	background.PrepareQuit () ;
	
	video . DisplayOSDMessage ( "bye bye..." ) ;
}
