#include "gamebg_pkg.h"

extern int level_of_detail ;
extern FileCache file_cache ;

#include "video.h"
extern Video video ;

void GameBG_Pkg::Init ( Pax *new_pax )
{
	///
	
	pax = new_pax ;
	
	///
	
	if ( strstr ( pax -> GetPkgFilename (), "title" ) != NULL ||
			strstr ( pax -> GetPkgFilename (), "networkscreen" ) != NULL )
		is_game_screen = false ;
	else
		is_game_screen = true ;
	
	background = (Sprite*) pax -> GetDeclaredObjectFromList ( "background" ) -> data ;
	background -> GetSprite () -> SetNodeName ( "background" ) ;
	
	bgmusic = (Audio_Music*) pax -> GetDeclaredObjectFromList ( "backgroundmusic" ) -> data ;

	if ( is_game_screen )
	{
		playfield_left  = pax -> GetNode ( pax -> GetDeclaredObjectFromList ( "playfield_left"  ), "playfield_left"  ) ;
		playfield_right = pax -> GetNode ( pax -> GetDeclaredObjectFromList ( "playfield_right" ), "playfield_right" ) ;
	
		preview_field_left  = pax -> GetNode ( pax -> GetDeclaredObjectFromList ( "preview_field_left"  ), "preview_field_left"  ) ;
		preview_field_right = pax -> GetNode ( pax -> GetDeclaredObjectFromList ( "preview_field_right" ), "preview_field_right" ) ;
		
		field_layer_left  = (Sprite*) pax -> GetDeclaredObjectFromList ( "field_layer_left"  ) -> data ;
		field_layer_right = (Sprite*) pax -> GetDeclaredObjectFromList ( "field_layer_right" ) -> data ;
		
		preview_layer_left  = (Sprite*) pax -> GetDeclaredObjectFromList ( "preview_layer_left"  ) -> data ;
		preview_layer_right = (Sprite*) pax -> GetDeclaredObjectFromList ( "preview_layer_right" ) -> data ;
	}
	else
	{
		//fprintf ( stdout, "seems to be a title screen...\n" ) ;
	}
	
	music_off = bgmusic -> GetVolume() == 0 ? TRUE : FALSE ;
	
	// first we have to wait until our playfield is ready
	ready_for_playing = FALSE ;
	
	paused = FALSE ;
	pause_value = 0 ;
	music_paused = FALSE ;
	
	quit = FALSE ;
	quit_state = FALSE ;
	quit_script_started = false ;
	
	pax -> ExecuteScript ( "start" ) ;
}

void GameBG_Pkg::SetLayerAlpha ( KrImNode *layer, int new_alpha )
{
	if ( new_alpha < 0 ) new_alpha = 0 ;
	if ( new_alpha > 255 ) new_alpha = 255 ;

	KrColorTransform color = layer->CTransform() ;
	
	color.SetAlpha ( new_alpha ) ;
	layer->SetColor ( color ) ;
}

void GameBG_Pkg::Draw ()
{
	if ( quit_state == TRUE ) return ;
	
	if ( paused == TRUE && pause_value < 128 )
	{
		pause_value += 20 ;
		if ( video . GetUsingOpenGL () )
			background -> SetDestDarken ( 128 ) ;
		if ( level_of_detail > 1 && is_game_screen )
		{
			field_layer_left -> SetAlpha ( 255 - pause_value ) ;
			field_layer_right -> SetAlpha ( 255 - pause_value ) ;
			SetLayerAlpha ( playfield_left, 255 - pause_value ) ;
			SetLayerAlpha ( playfield_right, 255 - pause_value ) ;
			SetLayerAlpha ( preview_field_left, 255 - pause_value ) ;
			SetLayerAlpha ( preview_field_right, 255 - pause_value ) ;
		}
	}
	else if ( paused == FALSE && pause_value > 0 )
	{
		pause_value -= 20 ;
		if ( video . GetUsingOpenGL () )
			background -> SetDestDarken ( 0 ) ;
		if ( level_of_detail > 1 && is_game_screen )
		{
			field_layer_left -> SetAlpha ( 255 - pause_value ) ;
			field_layer_right -> SetAlpha ( 255 - pause_value ) ;
			SetLayerAlpha ( playfield_left, 255 - pause_value ) ;
			SetLayerAlpha ( playfield_right, 255 - pause_value ) ;
			SetLayerAlpha ( preview_field_left, 255 - pause_value ) ;
			SetLayerAlpha ( preview_field_right, 255 - pause_value ) ;
		}
	}
	
	// background
	if ( background -> SlideFinished() || quit == TRUE ) 
	{	// background is in-place - setup our playfield :)
		if ( ! is_game_screen || field_layer_right -> SlideFinished () )
		{
			if ( is_game_screen )
			{
				field_layer_left -> Draw () ;
				preview_layer_left -> Draw () ;
			}
			if ( quit == FALSE && ( ! is_game_screen || field_layer_left -> SlideFinished () ) )
			{	// this is done while playing the game
				if ( ready_for_playing == FALSE && paused == FALSE )
				{
					ready_for_playing = TRUE ;
					pax -> ExecuteScript ( "ready" ) ;
					//fprintf ( stdout, "background is now ready for playing\n" ) ;
				}
			} 
			else if ( quit == TRUE && background -> SlideFinished () == TRUE /*&&
						bgmusic -> MusicPlaying () == FALSE*/ ) 
			{
				quit_state = TRUE ;
				return ;
			} 
			/*else if ( quit == TRUE )
				background -> SetAlpha ( background -> GetAlpha() - 6 ) ;*/
		}
		if ( is_game_screen )
		{
			field_layer_right -> Draw () ;
			preview_layer_right -> Draw () ;
		}
	}
	
	background -> Draw () ;
	
	pax -> DrawUserDefinedObjects () ;
}

bool GameBG_Pkg::PlayfieldIsReady ()
{
	return ready_for_playing ;
}

void GameBG_Pkg::SetPlayfieldIsReady ( bool value )
{
	if ( value )
		pax -> ExecuteScript ( "go" ) ;

}

void GameBG_Pkg::ToggleMusic ()
{
	if ( paused == TRUE ) return ;
	if ( music_off == TRUE )
	{
		bgmusic -> SetVolume ( 255 ) ;
		music_off = FALSE ;
	}
	else
	{
		bgmusic -> SetVolume ( 0 ) ;
		music_off = TRUE ;
	}
}

void GameBG_Pkg::Pause ()
{
	if ( music_paused == FALSE )
		bgmusic -> PauseMusic () ;
	music_paused = TRUE ;
	paused = TRUE ;
	
	pax -> ExecuteScript ( "pause" ) ;
}

void GameBG_Pkg::UnPause ()
{
	if ( music_paused == TRUE )
		bgmusic -> UnPauseMusic () ;
	music_paused = FALSE ;
	paused = FALSE ;
	
	pax -> ExecuteScript ( "unpause" ) ;
}

bool GameBG_Pkg::GetIsQuitting ()
{
	return quit ;
}

bool GameBG_Pkg::GetIsDone ()
{
	return quit_state ;
}

void GameBG_Pkg::PrepareQuit ()
{
	if ( quit_script_started == false )
	{
		pax -> ExecuteScript ( "quitting" ) ;
		quit_script_started = true ;
	}
}

void GameBG_Pkg::Quit ()
{
	if ( quit_state == TRUE ) return ;
	
	//ready_for_playing = FALSE ;
	quit = TRUE ;
	
	pax -> ExecuteScript ( "quit" ) ;
}
