#include "gamepf.h"

extern FileCache file_cache ;

extern char rc_dir[] ;

// how many gems have to be grouped before they get destroyed? see main.cpp!
extern int min_friends_for_group ;

extern float current_framerate ;

#include "video.h"
extern Video video ;

#ifdef __NETWORK_SUPPORT__
#include "network.h"
extern Network network ;
#endif

GamePF::GamePF ()
{
	quit = FALSE ;
	quit_state = TRUE ;
	
	already_freed = true ;
}

void GamePF::Init ( char *side, Pax *new_pax, char *new_level, int ai_level )
{
	pax = new_pax ;
	
	GEM_COUNT = pax  -> GetIntVar ( "gem_count"  ) ;
	GEM_WIDTH = pax  -> GetIntVar ( "gem_width"  ) ;
	GEM_HEIGHT = pax -> GetIntVar ( "gem_height" ) ;
	GEM_FIRST_COLOR = pax -> GetCharFromStringVar ( "gem_first_color" ) ;
	GEM_LAST_COLOR  = pax -> GetCharFromStringVar ( "gem_last_color"  ) ;
	GEM_BLOCK = pax -> GetCharFromStringVar ( "gem_block_char" ) ;
	GEM_HOLD  = pax -> GetCharFromStringVar ( "gem_hold_char"  ) ;
	GEM_NONE  = pax -> GetCharFromStringVar ( "gem_none_char"  ) ;
	
	//fprintf ( stdout, "GEM_COUNT = %d\n", GEM_COUNT ) ;
	//fprintf ( stdout, "GEM_WIDTH = %d\n", GEM_WIDTH ) ;
	//fprintf ( stdout, "GEM_HEIGHT = %d\n", GEM_HEIGHT ) ;
	//fprintf ( stdout, "GEM_FIRST_COLOR = %c\n", GEM_FIRST_COLOR ) ;
	//fprintf ( stdout, "GEM_BLOCK = %c\n", GEM_BLOCK ) ;
	//fprintf ( stdout, "GEM_HOLD = %c\n", GEM_HOLD ) ;
	//fprintf ( stdout, "GEM_NONE = %c\n", GEM_NONE ) ;
	
	preview_gems = pax -> GetIntVar ( "preview_gems" ) ;
	
	preview_gem_id = (int*) malloc ( sizeof (int) * preview_gems ) ;
	preview_gem_color = (char*) malloc ( sizeof (char) * ( preview_gems + 1 ) ) ;
	
	gems_incoming = (int*) malloc ( sizeof (int) * ( GEM_COUNT + 1 ) ) ;
	gems_outgoing = (int*) malloc ( sizeof (int) * ( GEM_COUNT + 1 ) ) ;
	gems_incoming_timer = (int*) malloc ( sizeof (int) * ( GEM_COUNT + 1 ) ) ;
	gems_incoming_preview = (Sprite**) malloc ( sizeof (Sprite*) * ( GEM_COUNT + 1 ) ) ;
	gems_incoming_preview_text = (KrTextBox**) malloc ( sizeof (KrTextBox*) * ( GEM_COUNT + 1 ) ) ;
	
	PLAYFIELD_MAX_X = pax -> GetIntVar ( "playfield_max_x" ) ;
	PLAYFIELD_MAX_Y = pax -> GetIntVar ( "playfield_max_y" ) ;
	PLAYFIELD_SIZE  = pax -> GetIntVar ( "playfield_size" ) ;
	
	playfield = (char*) malloc ( sizeof (char) * ( PLAYFIELD_SIZE + 1 ) ) ;
	gem = (Sprite**) malloc ( sizeof (Sprite*) * PLAYFIELD_SIZE ) ;
	gem_active = (bool*) malloc ( sizeof (bool) * PLAYFIELD_SIZE ) ;
	gem_pos = (int*) malloc ( sizeof (int) * PLAYFIELD_SIZE ) ;
	gem_frozen = (bool*) malloc ( sizeof (bool) * PLAYFIELD_SIZE ) ;
	gem_frozen_layer = (Sprite**) malloc ( sizeof (Sprite*) * PLAYFIELD_SIZE ) ;
	
	playfield [ PLAYFIELD_SIZE ] = '\0' ;
	
	engine = video . GetEngine () ;
	using_opengl = video . GetUsingOpenGL () ;
	
	playfield_side = side ;
	
	int item = 0 ;
	
	gem_colors_activated = GEM_COUNT ;
	
	char playfield_name [ 256 ], preview_field_name [ 256 ] ;
	char incoming_field_name [ 256 ] ;
	sprintf ( playfield_name, "playfield_%s", side ) ;
	sprintf ( preview_field_name, "preview_field_%s", side ) ;
	sprintf ( incoming_field_name, "incoming_field_%s", side ) ;
	basefield = pax -> GetNode ( pax -> GetDeclaredObjectFromList ( playfield_name ), playfield_name ) ;
	previewfield = pax -> GetNode ( pax -> GetDeclaredObjectFromList ( preview_field_name ), preview_field_name ) ;
	incomingfield = pax -> GetNode ( pax -> GetDeclaredObjectFromList ( incoming_field_name ), incoming_field_name ) ;
	
	Vault = (KrResourceVault*) pax -> GetDeclaredObjectFromList ( "gems" ) -> data ;

	// initialize all gems
	for ( item = 0 ; item < PLAYFIELD_SIZE ; item++ )
	{
		gem [ item ] = new Sprite () ;
		gem_pos [ item ] = -1 ;
		gem_active [ item ] = FALSE ;
		gem_frozen [ item ] = FALSE ;
		
		if ( pax -> GetObjectFromList ( "gem_frozen_layer" ) )
		{
			// now setup frozen layer on top of playfield and gems
			gem_frozen_layer [ item ] = new Sprite () ;
			Sprite *t_gem_fl = gem_frozen_layer [ item ] ;
			int pos = item ; //gem_pos [ id ] ;
			int x = pos % PLAYFIELD_MAX_X ;
			int y = ( pos - x ) / PLAYFIELD_MAX_X ;
			t_gem_fl -> Init ( Vault, pax -> GetStringVar ( "gem_frozen_layer" ), basefield ) ;
			t_gem_fl -> SetXY ( x * GEM_WIDTH, y * GEM_HEIGHT ) ;
			t_gem_fl -> SetAlpha ( 0 ) ;
			t_gem_fl -> SetMoveSound ( NULL, 0 ) ;
			t_gem_fl -> SetStopSound ( NULL, 0 ) ;
			t_gem_fl -> SetMoveSound ( NULL, 0 ) ;
			t_gem_fl -> SetAccelSound ( pax -> GetStringVar ( "gem_freeze_sound" ), 64 ) ;
			//t_gem_fl -> SetBounceSound ( pax -> GetStringVar ( "gem_unfreeze_sound" ), 64 ) ;
			t_gem_fl -> SetZ ( 1 ) ;
		}
		else
		{
			gem_frozen_layer [ item ] = NULL ;
		}
	}
	user_gem_cursor = NULL ;
	slave_gem_cursor = NULL ;

	// setup earthquake :)
	incoming_earthquake = FALSE ;
	earthquake_sound.LoadSound ( "sfx/bang.wav" ) ;
	pre_earthquake_sound.LoadSound ( "sfx/metal.wav" ) ;
	
	// set to first row to pull in
	pull_row = 0 ;
	pull_row_counter = 0 ;
	
	// init bucket for aligned gems ( pair search )
	friend_bucket = new GlDynArray < int > () ;
	friend_bucket -> Clear () ;
	
	action_indicator = 0 ;
	action_step_one = 0 ;
	action_step_two = 7 ;
	action_max = 14 ;
	
	auto_play = TRUE ;
	user_gem_id = -1 ;
	slave_gem_id = -1 ;
	slave_gem_thrown = true ;
	for ( int i = 0 ; i < preview_gems ; i++ )
	{
		preview_gem_id [ i ] = -1 ;
		preview_gem_color [ i ] = GEM_NONE ;
	}
	preview_gem_color [ preview_gems ] = '\0' ;
	move_on_counter = 0 ;
	move_on_delay = 10 ;
	move_up = FALSE ;
	move_left = FALSE ;
	move_right = FALSE ;
	
	idle_rows_enabled = TRUE ;
	idle_row_counter = 0 ;
	idle_row_delay = 50 ; // 5 is about 1 second

	destroyable_gems = TRUE ;
	
	falling_gems_enabled = TRUE ;
	freeze_gems_at_start = TRUE ;
	
	Ai . Init ( ai_level, pax ) ;
	
	char player_name [ 256 ], player_score [ 256 ], player_combo [ 256 ] ;
	sprintf ( player_name, "playername_%s", side ) ;
	sprintf ( player_score, "playerscore_%s", side ) ;
	sprintf ( player_combo, "playercombo_%s", side ) ;
	playernamebox  = (KrTextBox*) pax -> GetNode ( pax -> GetDeclaredObjectFromList ( player_name  ), player_name  ) ;
	playerscorebox = (KrTextBox*) pax -> GetNode ( pax -> GetDeclaredObjectFromList ( player_score ), player_score ) ;
	playercombobox = (KrTextBox*) pax -> GetNode ( pax -> GetDeclaredObjectFromList ( player_combo ), player_combo ) ;
	//playernamebox -> SetTextChar ( player_name, 0 ) ;
	playerscore_cur = playerscore_dest = 0 ;
	playerscore_combo = 1 ;
	playerscore_combo_rest_seconds = 5 ;
	playerscore_combo_max_rest_seconds = 5 ;
	playerscore_inc_sound.LoadSound ( "sfx/score.wav" ) ;
	playerscore_inc_sound.SetVolume ( 32 ) ;
	playerscore_dec_sound.LoadSound ( "sfx/disable.wav" ) ;
	playerscore_dec_sound.SetVolume ( 32 ) ;
	playerscore_bonus_sound.LoadSound ( "sfx/malus.wav" ) ;
	playerscore_bonus_sound.SetVolume ( 128 ) ;
	playerlose_sound.LoadSound ( "sfx/lose.wav" ) ;
	//playercombobox -> SetTextChar ( "please wait...", 0 ) ;
	playerscoreupdatetimer = playerscoreupdatecounter = 0 ;
	
	GemQueueInit () ;
	
	// setup level layout
	level = strdup ( new_level ) ;
	AutoLoadState () ;

	playfield_set = FALSE ;
	playfield_ready = FALSE ;
	
	paused = FALSE ;
	
	quit = FALSE ;
	quit_state = FALSE ;
	
	already_freed = false ;
}

// network stuff
int GamePF::GetData ( char *message )
{
	return 0 ;
}

int GamePF::SendData ( char *message )
{
//#ifdef __NETWORK_SUPPORT__
//	if ( netdata . mode == NM_SERVER )
//		return network . SendMessage ( netdata . address, message ) ;
//	else
//#endif
		return network . SendMessage ( "192.168.0.234", message ) ;
}
//

int GamePF::PlayerScoreGet ()
{
	return playerscore_dest ;
}

void GamePF::PlayerScoreSet ( int value )
{
	playerscore_dest = value ;
}

void GamePF::PlayerScoreReset ()
{
	PlayerScoreSet ( 0 ) ;
	playerlose_sound.SetVolume ( 128 ) ;
	playerlose_sound.PlaySound () ;
}

void GamePF::PlayerScoreUpdate ()
{
	if ( playerscoreupdatetimer > FRAMERATE && playerscore_combo_rest_seconds > 0 &&
			playerscoreupdatecounter > 5 )
	{
		playerscore_combo_rest_seconds-- ;
		//fprintf ( stdout, "combo lasts %d seconds...\n",
		//			playerscore_combo_rest_seconds ) ;

		if ( playerscore_combo_rest_seconds <= 0 )
		{
			playerscore_combo = 1 ;
			playercombobox -> SetTextChar ( " ", 0 ) ;
		}
		
		playerscoreupdatetimer = 0 ;
	}
	else if ( playerscore_combo_rest_seconds )
		playerscoreupdatetimer++ ;


	if ( playerscore_cur != playerscore_dest && playerscoreupdatecounter > 5 )
	{	
		char text [ 256 ] ;
		int diff = max ( playerscore_cur, playerscore_dest ) -
					min ( playerscore_cur, playerscore_dest ) ;
		int step = diff / 5 ;
		step = ( step > 1 ? step : 1 ) ;
		
		if ( playerscore_cur < playerscore_dest )
		{
			playerscore_cur += step ;
			playerscore_inc_sound.PlaySound () ;
		}
		if ( playerscore_cur > playerscore_dest )
		{
			playerscore_cur -= step ;
			playerscore_dec_sound.PlaySound () ;
		}
		
		sprintf ( text, "%d", playerscore_cur ) ;
		playerscorebox -> SetTextChar ( text, 0 ) ;
		
		playerscoreupdatecounter = 0 ;
	}
	else if ( playerscore_cur != playerscore_dest || playerscore_combo_rest_seconds )
		playerscoreupdatecounter++ ;
}

bool GamePF::PlayerScoreIsInComboTime ()
{
	return ( playerscore_combo_rest_seconds ) ;
}

void GamePF::GemQueueUpdate ()
{
	if ( updategemqueuetimer > FRAMERATE && ! PlayerScoreIsInComboTime () &&
			action_indicator > action_step_two )
	{
		int overall_counted_gems = 0, max_gems = PLAYFIELD_MAX_X * 2 ;
		updategemqueuetimer = 0 ;
		
		//fprintf ( stdout, "%s: UpdateGemQueue ();\n", playfield_side ) ;
		
		for ( int i = 0 ; i < GEM_COUNT + 1 && overall_counted_gems < max_gems ; i++ )
		{
			//fprintf ( stdout, "%s: color %d has %d gems to come...\n",
			//			playfield_side, GEM_FIRST_COLOR + i, gems_incoming [ i ] ) ;
			
			// consistency tests...
			if ( gems_incoming [ i ] < 0 )
				gems_incoming [ i ] = 0 ;
			if ( gems_incoming [ i ] > 99 )	// 99 gems in queue
				gems_incoming [ i ] = 99 ;	// are enough!!
			
			if ( gems_incoming [ i ] > 0 && playfield_ready && ! quit )
			{
				if ( gems_incoming_timer [ i ] > 0 )
				{
					gems_incoming_timer [ i ] -- ;
					
					//fprintf ( stdout, "%s: %d seconds left till dropping %d gems with color %d...\n",
					//			playfield_side,
					//			gems_incoming_timer [ i ], gems_incoming [ i ],
					//			GEM_FIRST_COLOR + i ) ;
				}
				else
				{
					//fprintf ( stdout, "%s: dropping %d gems with color %d !!!\n",
					//			playfield_side,
					//			gems_incoming [ i ], GEM_FIRST_COLOR + i ) ;
					
					int color = i < GEM_COUNT ? GEM_FIRST_COLOR + i : GEM_BLOCK ;
					int count = AddFallingGems ( color, min ( gems_incoming [ i ], max_gems ) ) ;
					
					gems_incoming [ i ] -= count ;
					gems_incoming_timer [ i ] = 1 ;
					
					overall_counted_gems += count ;
				}
			}
		}
	}
	else
		updategemqueuetimer++ ;
	
	if ( updategemqueuetimer % 4 == 0 )
	{
		for ( int i = 0 ; i < GEM_COUNT + 1 ; i++ )
		{
			int cur_value = 0, old_value = 0 ;
			std::string value_string ;
			value_string = "0" ;
			gems_incoming_preview_text [ i ] -> GetTextChar ( &value_string, 0 ) ;
			old_value = cur_value = atoi ( (const char*) value_string . c_str () ) ;
			if ( cur_value < gems_incoming [ i ] )
			{
				cur_value++ ;
				playerscore_inc_sound.PlaySound () ;
			}
			if ( cur_value > gems_incoming [ i ] )
			{
				cur_value-- ;
				playerscore_dec_sound.PlaySound () ;
			}
			if ( cur_value != old_value )
			{
				char value [ 255 ] ;
				if ( cur_value > 0 )
					sprintf ( value, "%d", cur_value ) ;
				else
					sprintf ( value, "%s", " " ) ;
				gems_incoming_preview_text [ i ] -> SetTextChar ( value, 0 ) ;
			}
				
			if ( gems_incoming [ i ] > 0 && 
					gems_incoming_preview [ i ] -> GetAlpha () < 255 )
			{
				int brightness = gems_incoming [ i ] * 5 ;
				brightness = brightness > 200 ? 200 : brightness ;
				gems_incoming_preview [ i ] -> SetDestAlpha ( 255 ) ;
				//gems_incoming_preview [ i ] -> Brighten ( brightness ) ;
			}
			else if ( gems_incoming [ i ] < 1 &&
						gems_incoming_preview [ i ] -> GetAlpha () > 0 )
			{
				//gems_incoming_preview [ i ] -> Brighten ( 0 ) ;
				gems_incoming_preview [ i ] -> SetAlpha ( 0 ) ;
			}

		}
	}

	for ( int i = 0 ; i < GEM_COUNT + 1 ; i++ )
	{		
		if ( gems_incoming_preview [ i ] != NULL )
			gems_incoming_preview [ i ] -> Draw () ;
	}
}

void GamePF::GemQueueAddToIncoming ( char gem_color, int count )
{
	// is it a color?
	if ( gem_color >= GEM_FIRST_COLOR && gem_color < GEM_FIRST_COLOR + GEM_COUNT )
	{
		//fprintf ( stdout, "%s: Adding %d gem(s) with color %d to incoming queue...\n",
		//			playfield_side, count, gem_color ) ;
		
		gems_incoming [ gem_color - GEM_FIRST_COLOR ] += count ;
		gems_incoming_timer [ gem_color - GEM_FIRST_COLOR ] += 1 ;
	}
	else if ( gem_color == GEM_BLOCK ) // or is it a block?
	{
		//fprintf ( stdout, "%s: Adding %d smilies to incoming queue...\n", playfield_side, count ) ;
		
		gems_incoming [ GEM_COUNT ] += count ;
		gems_incoming_timer [ GEM_COUNT ] += 1 ;
	}
	else
		fprintf ( stdout, "%s: Unknown color %d can't be added (%d times) to incoming gem queue... sorry.\n",
					playfield_side, gem_color, count ) ;
}

void GamePF::GemQueueAddToOutgoing ( char gem_color, int count )
{
	// is it a color?
	if ( gem_color >= GEM_FIRST_COLOR && gem_color < GEM_FIRST_COLOR + GEM_COUNT )
	{
		//fprintf ( stdout, "%s: Adding %d gem(s) with color %d to outgoing queue...\n",
		//			playfield_side, count, gem_color ) ;
		
		gems_outgoing [ gem_color - GEM_FIRST_COLOR ] += count ;
		gems_incoming [ gem_color - GEM_FIRST_COLOR ] -= count ;
		if ( gems_incoming [ gem_color - GEM_FIRST_COLOR ] < 0 )
			gems_incoming [ gem_color - GEM_FIRST_COLOR ] = 0 ;
	}
	else if ( gem_color == GEM_BLOCK ) // or is it a block?
	{
		//fprintf ( stdout, "%s: Adding %d smilies to outgoing queue...\n", playfield_side, count ) ;
		
		gems_outgoing [ GEM_COUNT ] += count ;
		gems_incoming [ GEM_COUNT ] -= count ;
		if ( gems_incoming [ GEM_COUNT ] < 0 )
			gems_incoming [ GEM_COUNT ] = 0 ;
	}
	else
		fprintf ( stdout, "%s: Unknown color %d can't be added (%d times) to outgoing gem queue... sorry.\n",
					playfield_side, gem_color, count ) ;
}

int GamePF::GemQueueGetFromOutgoing ( char gem_color, bool clear_after_query )
{
	int count = 0 ;
	// is it a color?
	if ( gem_color >= GEM_FIRST_COLOR && gem_color < GEM_FIRST_COLOR + GEM_COUNT )
	{
		count = gems_outgoing [ gem_color - GEM_FIRST_COLOR ] ;
		
		if ( clear_after_query )
		{
			//fprintf ( stdout, "%s: Exporting %d gem(s) with color %d from outgoing queue...\n",
			//			playfield_side, count, gem_color ) ;
			
			gems_outgoing [ gem_color - GEM_FIRST_COLOR ] = 0 ;
		}
	}
	else if ( gem_color == GEM_BLOCK ) // or is it a block?
	{
		count = gems_outgoing [ GEM_COUNT ] ;
		
		if ( clear_after_query )
		{
			//fprintf ( stdout, "%s: Exporting %d smilies from outgoing queue...\n", playfield_side, count ) ;
			
			gems_outgoing [ GEM_COUNT ] = 0 ;
		}
	}
	else
		fprintf ( stdout, "%s: Unknown color %d can't be exported from outgoing gem queue... sorry.\n",
					playfield_side, gem_color ) ;

	return count ;
}

void GamePF::GemQueueInit ()
{
	// get the font vault and the font...
	gems_incoming_preview_text_vault = (KrResourceVault*)
				file_cache . CachedLoad ( "vault", "gfx/fonts_neon.dat" ) ;
	gems_incoming_preview_text_font  = 
				gems_incoming_preview_text_vault ->
					GetFontResource ( "SmallNeonGrey" ) ;
	
	for ( int i = 0 ; i < GEM_COUNT + 1 ; i++ )
	{
		gems_incoming [ i ] = 0 ;
		gems_outgoing [ i ] = 0 ;
		gems_incoming_timer [ i ] = 0 ;

		char color = GEM_FIRST_COLOR + i ;
		if ( i == GEM_COUNT ) color = GEM_BLOCK ;
		gems_incoming_preview [ i ] = new Sprite () ;
		int my_gem = SetupGem ( color, -i - 2, incomingfield, 0, 0, FALSE, 0, 0 ) ;
		if ( my_gem == -1 )
		{	// bad bad bad bad!!! should never happen!!!!
			delete gems_incoming_preview [ i ] ;
			gems_incoming_preview [ i ] = NULL ;
		}
		else
		{	// ah :) our incoming gem indicator :) scare me, scare me =D
			Sprite *gip = gems_incoming_preview [ i ] ;
			gip -> SetXY ( i * GEM_WIDTH, 0 ) ;
			gip -> SetDestAlpha ( 0 ) ;
			gip -> SetCreationSound ( NULL, 0 ) ;
		}
		
		// i'm lazy :)
		KrFontResource *font = gems_incoming_preview_text_font ;
		GLASSERT ( font ) ;
		// now setup our textbox...
		gems_incoming_preview_text [ i ] = new KrTextBox ( font, 30, 30, 0,
												KrTextBox::CENTER ) ;
		// and again :)
		KrTextBox *gipt = gems_incoming_preview_text [ i ] ;
		gipt -> SetTextChar ( " ", 0 ) ;
		video . GetEngine () -> Tree () -> AddNode ( incomingfield, gipt ) ;
		gipt -> SetZDepth ( 50 ) ;
		gipt -> SetPos ( i * GEM_WIDTH, 0 ) ;
	}
	
	updategemqueuetimer = 0 ;
}

void GamePF::GemQueueQuit ()
{
	for ( int i = 0 ; i < GEM_COUNT + 1 ; i++ )
		video . GetEngine () -> Tree () -> DeleteNode ( gems_incoming_preview_text [ i ] ) ;
	file_cache . FreeCachedData ( "vault", gems_incoming_preview_text_vault ) ;
}

char *GamePF::GetPlayfieldLayout ( char *level_name )
{
	char level_filename[2048] ;
	int pos = 0 ;
	char layout [ PLAYFIELD_SIZE + 1 ] ;
	FILE *stream ;
	
	sprintf ( level_filename, "%s", level_name ) ;
	
	stream = fopen ( level_filename, "r" ) ;

	//fprintf ( stdout, "level_name = '%s'\n", level_filename ) ;
	
	if ( ! stream )
	{
		if ( strcmp ( level_filename, "random" ) != 0 )
		{
			fprintf ( stdout, "WARNING: could not open %s !", level_filename ) ;
			fprintf ( stdout, " will use random setup instead...\n" ) ;
		}
		return NULL ;
	}
	else
	{
		char line [ PLAYFIELD_MAX_X + 2 ] ;

		while ( fgets ( line, PLAYFIELD_MAX_X + 2, stream ) )
		{
			char txt [ PLAYFIELD_MAX_X ] ;
			sprintf ( txt, "%s", line ) ;
			
			if ( strlen ( line ) - 1 < (unsigned int) PLAYFIELD_MAX_X )
			{
				int offset = ( PLAYFIELD_MAX_X - ( strlen ( line ) - 1 ) ) / 2 ;
				//fprintf ( stdout, "pos = %d ; line has %d chars ; max_x = %d ; offset = %d\n",
				//			pos, strlen ( line ), PLAYFIELD_MAX_X, offset ) ;
				for ( int i = 0 ; i < offset ; i++ )
					layout [ pos++ ] = GEM_NONE ;
			}
			
			for ( int i = 0 ; i < PLAYFIELD_MAX_X ; i++ )
			{
				if ( (unsigned int) i < strlen ( line ) - 1 )
					layout [ pos ] = txt [ i ] ;
				else
					layout [ pos ] = GEM_NONE ;
				
				pos++ ;
			}
			pos -= pos % PLAYFIELD_MAX_X ;
		}
		
		fclose ( stream ) ;
		
		for ( ; pos < PLAYFIELD_SIZE ; pos++ )
			layout [ pos ] = GEM_NONE ;
		
		layout [ pos ] = '\0' ;
	}
	
	return strdup ( layout ) ;
}

void GamePF::SetupPlayfield ()
{
	char *playfield_setup = GetPlayfieldLayout ( level ) ;
	
	user_gem_id = -1 ;
	slave_gem_id = -1 ;
	
	if ( playfield_setup != NULL )
	{
		//fprintf ( stdout, "sizeof setup = %d ; playfield = %d\n",
		//			strlen ( playfield_setup ), strlen ( playfield ) ) ;
		for ( int pos = 0 ; pos < PLAYFIELD_SIZE ; pos++ )
		{
			if ( playfield_setup [ pos ] >= GEM_FIRST_COLOR &&
					playfield_setup [ pos ] < GEM_FIRST_COLOR + gem_colors_activated )
				playfield [ pos ] = playfield_setup [ pos ] ;
			else if ( playfield_setup [ pos ] == GEM_NONE ||
						playfield_setup [ pos ] == GEM_BLOCK )
				playfield [ pos ] = playfield_setup [ pos ] ;
			else
				playfield [ pos ] = GEM_NONE ;
		}
		//fprintf ( stdout, "sizeof setup = %d ; playfield = %d\n",
		//			strlen ( playfield_setup ), strlen ( playfield ) ) ;
		free ( playfield_setup ) ;
	}
	else
	{
		for ( int ty = 0 ; ty < PLAYFIELD_MAX_Y ; ty++ )
		{
			for ( int tx = 0 ; tx < PLAYFIELD_MAX_X ; tx++ )
			{
				int random = rand() % gem_colors_activated ;
				int item = ( random == 0 ? GEM_NONE : random + GEM_FIRST_COLOR ) ;
				playfield [ ( ty * PLAYFIELD_MAX_X ) + tx ] = item ;
			}
		}
	}

	// now build our playfield from predefined data....
	for ( int pos = 0 ; pos < PLAYFIELD_SIZE ; pos++ )
	{
		if ( playfield [ pos ] != GEM_HOLD && playfield [ pos ] != GEM_NONE )
		{
			ThrowInGem ( pos ) ;
			
			int id = GetMatchingGem ( pos ) ;
			if ( id != -1 && freeze_gems_at_start )
			{
				FreezeGem ( id ) ;
			}
		}
	}
	// clear the preview area
	for ( int i = 0 ; i < preview_gems ; i++ )
	{
		preview_gem_id [ i ] = -1 ;
		preview_gem_color [ i ] = GEM_NONE ;
	}
	// now indicate that we are done here
	playfield_is_set_up = TRUE ;
	
	playfield_ready = FALSE ;
	
	// don't bother us too soon :)
	idle_row_counter = 0 ;
	
	// choose a throw-in effect
	playfield_throw_in_effect = rand () % 4 ;
}

void GamePF::EnableFallingGems ()
{
	falling_gems_enabled = TRUE ;
}

void GamePF::DisableFallingGems ()
{
	falling_gems_enabled = FALSE ;
	
	for ( int ty = 0 ; ty < PLAYFIELD_MAX_Y ; ty++ )
	{
		for ( int tx = 0 ; tx < PLAYFIELD_MAX_X ; tx++ )
		{
			if ( tx == 0 || tx == PLAYFIELD_MAX_X - 1 ||
				   ty == 0 || ty == PLAYFIELD_MAX_Y - 1 )
			{
				int pos = ( ty * PLAYFIELD_MAX_X ) + tx ;
				int matching_gem = GetMatchingGem ( pos ) ;
				if ( playfield [ pos ] != GEM_NONE && 
						matching_gem != user_gem_id && 
						matching_gem != slave_gem_id ) 
				{
					playfield [ pos ] = GEM_HOLD ; 
					UnfreezeGem ( matching_gem ) ;
					gem [ matching_gem ] -> Explode () ;
				}
			}
		}
	}
}

int GamePF::GetMatchingGem ( int dest_pos )
{
	int result_item = -1 ;
	int item = 0 ;
	
	if ( dest_pos < 0 ) return dest_pos ;
	
	//fprintf ( stdout, "search for gem on pos %d... ", dest_pos ) ;
	
	for ( item = 0 ; item < PLAYFIELD_SIZE ; item++ )
	{
		if ( gem_pos [ item ] == dest_pos )
			result_item = item ;
	}
	
	return result_item ;
}

void GamePF::HighlightEqualGems ()
{
	for ( int item = 0 ; item < PLAYFIELD_SIZE ; item++ )
	{
		bool is_preview_gem = FALSE ;
		for ( int i = 0 ; i < preview_gems ; i++  )
			if ( item == preview_gem_id [ i ] )
				is_preview_gem = TRUE ;
		
		if ( gem_active [ item ] == TRUE && 
				//item != user_gem_id &&
				is_preview_gem == FALSE )
			if ( ( gem_active [ user_gem_id ] == TRUE &&
					playfield [ gem_pos [ item ] ] == 
					playfield [ gem_pos [ user_gem_id ] ] ) ||
				 ( gem_active [ slave_gem_id ] == TRUE &&
					playfield [ gem_pos [ item ] ] == 
					playfield [ gem_pos [ slave_gem_id ] ] ) )
				gem [ item ] -> Glow ( TRUE ) ;
			else
				gem [ item ] -> Glow ( FALSE ) ;
	}
}

bool GamePF::SlidingFinishedRow ( int row )
{
	bool finished = TRUE ;
	int item = 0 ;
	int row_start = row * PLAYFIELD_MAX_X ;
	int row_end = row_start + PLAYFIELD_MAX_X ;
	
	for ( item = row_start ; item < row_end ; item++ )
	{
		int match = GetMatchingGem ( item ) ;
		if ( match >= 0 && match <= PLAYFIELD_SIZE && 
				gem [ match ] -> SlideFinished () == FALSE && 
				gem_active [ match ] == TRUE &&
				!( gem_active [ user_gem_id ] == TRUE &&
				  playfield [ item ] == playfield [ gem_pos [ user_gem_id ] ] ) &&
				!( gem_active [ slave_gem_id ] == TRUE &&
				  playfield [ item ] == playfield [ gem_pos [ slave_gem_id ] ] ))
			finished = FALSE ;
	}
	
	return finished ;
}

bool GamePF::SlidingFinished ()
{
	bool finished = TRUE ;
	int row = 0 ;

	for ( row = 0 ; row < PLAYFIELD_MAX_Y ; row++ )
	{
		if ( SlidingFinishedRow ( row ) == FALSE )
			finished = FALSE ;
	}
	
	return finished ;
}

int GamePF::GetRelPos ( int pos_in_playfield, int x, int y )
{
	int col = pos_in_playfield % PLAYFIELD_MAX_X ;
	int row = ( pos_in_playfield - col ) / PLAYFIELD_MAX_X ;
	
	col += x ;
	row += y ;
	
	return ( ( col >= 0 && col < PLAYFIELD_MAX_X &&
				row >= 0 && row < PLAYFIELD_MAX_Y ) ? 
				( row * PLAYFIELD_MAX_X ) + col : -1 ) ;
}

int GamePF::CheckGemForFriends ( int pos_in_playfield )
{
	int friends = 1 ;
	
	if ( pos_in_playfield < 0 || pos_in_playfield >= PLAYFIELD_SIZE )
	{
		fprintf ( stdout, "Error: Access out of playfield not allowed!\n" ) ;
		fprintf ( stdout, "       pos_in_playfield = %d\n", pos_in_playfield ) ;
		DumpStatus () ;
		exit ( 0 ) ;
	}
	
	friend_bucket -> PushBack ( pos_in_playfield ) ;
	if ( friend_bucket -> Find ( pos_in_playfield ) == friend_bucket -> NotFound () )
	{
		fprintf ( stdout, "Warning: Couldn't add friend to list!\n" ) ;
		return 0 ;
	}
	
	int element_above ;
	int element_below ;
	int element_left ;
	int element_right ;
	
	element_above = GetRelPos ( pos_in_playfield,  0, -1 ) ;
	element_below = GetRelPos ( pos_in_playfield,  0,  1 ) ;
	element_left  = GetRelPos ( pos_in_playfield, -1,  0 ) ;
	element_right = GetRelPos ( pos_in_playfield,  1,  0 ) ;
	
	if ( element_above != -1 && 
			playfield [ pos_in_playfield ] == playfield [ element_above ] &&
			friend_bucket -> Find ( element_above ) == friend_bucket -> NotFound () ) 
		friends += CheckGemForFriends ( element_above ) ;
	if ( element_below != -1 && 
			playfield [ pos_in_playfield ] == playfield [ element_below ] &&
			friend_bucket -> Find ( element_below ) == friend_bucket -> NotFound () ) 
		friends += CheckGemForFriends ( element_below ) ;
	if ( element_left != -1 && 
			playfield [ pos_in_playfield ] == playfield [ element_left ] &&
			friend_bucket -> Find ( element_left ) == friend_bucket -> NotFound () ) 
		friends += CheckGemForFriends ( element_left ) ;
	if ( element_right != -1 && 
			playfield [ pos_in_playfield ] == playfield [ element_right ] &&
			friend_bucket -> Find ( element_right ) == friend_bucket -> NotFound () ) 
		friends += CheckGemForFriends ( element_right ) ;
	
	return friends ;
}

int GamePF::GetUnusedGem ()
{
	int item = 0 ;
	for ( item = 0 ; item < PLAYFIELD_SIZE ; item++ )
	{
		if ( gem_active [ item ] == FALSE /*&&
				( 
				  gem [ item ] -> GetIsAlreadyDrawn () == FALSE ||
				  gem [ item ] -> GetIsDone () == TRUE 
				)*/
			)
			return item ;
	}
	
	return -1 ;
}

// give a char ( for ex playfield [ 33 ] ) to get all color data
GamePF::gem_color_data *GamePF::GetGemColorDataFromPlayfieldChar ( char color )
{
	gem_color_data *gem_data = new gem_color_data () ;
	char gem_color[256] ;
	char color_string [ 256 ] ;
	U8 tr = 0, tg = 0, tb = 0 ; // explosion tint-color
	
	// default dummy :)	
	sprintf ( gem_color, "%s", pax -> GetStringVar ( "gem_block" ) ) ;
	
	if ( color >= GEM_FIRST_COLOR && color < GEM_FIRST_COLOR + GEM_COUNT )
	{
		char my_gem [ 256 ] ;
		sprintf ( my_gem, "gem_%d", color - GEM_FIRST_COLOR ) ;
		sprintf ( gem_color, "%s", pax -> GetStringVar ( my_gem ) ) ;
		
		sprintf ( color_string, "gem_%s_color", gem_color ) ;
		
		//fprintf ( stdout, "Creating gem '%c' at %dx%d ; gem_color = '%s' ; my_gem = '%s'\n",
		//			playfield [ pos_in_playfield ], x, y, gem_color, my_gem ) ;
	}
	else if ( color == GEM_BLOCK )
	{
		sprintf ( gem_color, "%s", pax -> GetStringVar ( "gem_block" ) ) ;
		
		sprintf ( color_string, "gem_block_color" ) ;
		
		//fprintf ( stdout, "Creating gem '%c' at %dx%d ; gem_color = '%s'\n",
		//			playfield [ pos_in_playfield ], x, y, gem_color ) ;
	}
	else if ( color < GEM_FIRST_COLOR || color > GEM_LAST_COLOR )
	{
		fprintf ( stdout, "Warning: out-of-bounds gem color '%c' '%s'\n",
					color, gem_color ) ;
		return NULL ;
	}
	
	char color_vars [ 256 ] ;
	sprintf ( color_vars, "%s", pax -> GetStringVar ( color_string ) ) ;
	if ( strlen ( color_vars ) > 0 && strcmp ( color_vars, "random" ) != 0 )
	{	// if setup in pax-file as manually chosen color
		GlDynArray < std::string > *list = pax -> GetArgList ( color_vars ) ;
		
		int value [ 3 ] ;
		
		value [ 0 ] = pax -> CalcValueFromString ( (char*) list -> Item ( 0 ) . c_str () ) ;
		value [ 1 ] = pax -> CalcValueFromString ( (char*) list -> Item ( 1 ) . c_str () ) ;
		value [ 2 ] = pax -> CalcValueFromString ( (char*) list -> Item ( 2 ) . c_str () ) ;
		
		tr = value [ 0 ] ; tg = value [ 1 ] ; tb = value [ 2 ] ;
		list -> Clear () ;
		delete list ;
	}
	else
	{	// if setup in pax-file as randomly chosen color / fallback
		tr = rand () % 255 ; tg = rand () % 255 ; tb = rand () % 255 ;
		//fprintf ( stdout, "gem_color '%s' is beeing chosen randomly...\n",
		//			gem_color ) ;
	}
	
	//fprintf ( stdout, "color_string = %s value is %d, %d, %d\n",
	//			color_string, tr, tg, tb ) ;
	
	sprintf ( gem_data -> name, "%s", gem_color ) ;
	sprintf ( gem_data -> string, "%s", color_string ) ;
	gem_data -> exp_red = tr ;
	gem_data -> exp_green = tg ;
	gem_data -> exp_blue = tb ;
	
	return gem_data ;
}

// pos_in_playfield >=  0 means that this is in the playfield
// pos_in_playfield == -1 means that this is an preview gem
// pos_in_playfield  < -1 means that this is an incoming gem
// when creating an incoming gem we use another sprite array for storage
int GamePF::SetupGem ( char color,
						int pos_in_playfield,
						KrImNode *field,
						int x,
						int y,
						bool do_slide,
						int slide_x,
						int slide_y )
{
	int item = GetUnusedGem () ; // we need to find an unused gem-slot
	
	if ( item == -1 && pos_in_playfield >= -1 ) 
	{
		//fprintf ( stdout, "failed to setup new gem!\n" ) ;
		return -1 ; // no free gems available :(
	}
	
	if ( color == FALSE && pos_in_playfield >= 0 )
	{
		color = playfield [ pos_in_playfield ] ;
	}

	gem_color_data *gem_data = GetGemColorDataFromPlayfieldChar ( color ) ;
	if ( gem_data == NULL )
	{
		if ( pos_in_playfield >= -1 )
		{
			fprintf ( stdout, "Warning: gem in pos %d will be removed...\n",
						pos_in_playfield ) ;
		
			playfield [ pos_in_playfield ] = GEM_NONE ;
		}
		
		return -1 ;
	}
	
	char frame_string     [ 4096 ], delay_string      [ 4096 ] ;
	char createsnd_string [ 4096 ], movesnd_string    [ 4096 ] ;
	char stopsnd_string   [ 4096 ], accelsnd_string   [ 4096 ] ;
	char bumpsnd_string   [ 4096 ] ;
	char explosion_string [ 4096 ], twinkspeed_string [ 4096 ] ;
	
	sprintf ( frame_string,      "gem_%s_frames"      , gem_data -> name ) ;
	sprintf ( delay_string,      "gem_%s_delays"      , gem_data -> name ) ;
	sprintf ( createsnd_string,  "gem_%s_createsound" , gem_data -> name ) ;
	sprintf ( movesnd_string,    "gem_%s_movesound"   , gem_data -> name ) ;
	sprintf ( stopsnd_string,    "gem_%s_stopsound"   , gem_data -> name ) ;
	sprintf ( accelsnd_string,   "gem_%s_accelsound"  , gem_data -> name ) ;
	sprintf ( bumpsnd_string,    "gem_%s_bumpsound"   , gem_data -> name ) ;
	sprintf ( explosion_string,  "gem_%s_explosion"   , gem_data -> name ) ;
	sprintf ( twinkspeed_string, "gem_%s_twinkspeed"  , gem_data -> name ) ;
	
	sprintf ( frame_string,      "%s", pax -> GetStringVar ( frame_string      ) ) ;
	sprintf ( delay_string,      "%s", pax -> GetStringVar ( delay_string      ) ) ;
	sprintf ( createsnd_string,  "%s", pax -> GetStringVar ( createsnd_string  ) ) ;
	sprintf ( movesnd_string,    "%s", pax -> GetStringVar ( movesnd_string    ) ) ;
	sprintf ( stopsnd_string,    "%s", pax -> GetStringVar ( stopsnd_string    ) ) ;
	sprintf ( accelsnd_string,   "%s", pax -> GetStringVar ( accelsnd_string   ) ) ;
	sprintf ( bumpsnd_string,    "%s", pax -> GetStringVar ( bumpsnd_string    ) ) ;
	sprintf ( explosion_string,  "%s", pax -> GetStringVar ( explosion_string  ) ) ;
	sprintf ( twinkspeed_string, "%s", pax -> GetStringVar ( twinkspeed_string ) ) ;
	
	GlDynArray < std::string > *frame_args      = pax -> GetArgList ( frame_string      ) ;
	GlDynArray < std::string > *delay_args      = pax -> GetArgList ( delay_string      ) ;
	GlDynArray < std::string > *createsnd_args  = pax -> GetArgList ( createsnd_string  ) ;
	GlDynArray < std::string > *movesnd_args    = pax -> GetArgList ( movesnd_string    ) ;
	GlDynArray < std::string > *stopsnd_args    = pax -> GetArgList ( stopsnd_string    ) ;
	GlDynArray < std::string > *accelsnd_args   = pax -> GetArgList ( accelsnd_string   ) ;
	GlDynArray < std::string > *bumpsnd_args    = pax -> GetArgList ( bumpsnd_string    ) ;
	GlDynArray < std::string > *explosion_args  = pax -> GetArgList ( explosion_string  ) ;
	GlDynArray < std::string > *twinkspeed_args = pax -> GetArgList ( twinkspeed_string ) ;
	
	Sprite *t_gem = NULL ;
	
	// if this is an incoming gem, use the right one...
	if ( pos_in_playfield < -1 )
	{
		item = pos_in_playfield * -1 -2 ;
		t_gem = gems_incoming_preview [ item ] ;
	}
	else
	{
		t_gem = gem [ item ] ;
	}

	if ( item < 0 || item >= PLAYFIELD_SIZE )
		fprintf ( stdout, "WARNING: SetupGem ( pos_in_playfield = %d ) ; item = %d ;\n",
					pos_in_playfield, item ) ;
	
	// set it up...
	t_gem -> Init ( Vault, gem_data -> name, field ) ;
	t_gem -> SetFrames ( pax -> CalcValueFromString ( (char*) frame_args -> Item ( 0 ) . c_str () ),
							pax -> CalcValueFromString ( (char*) frame_args -> Item ( 1 ) . c_str () ),
							pax -> CalcValueFromString ( (char*) frame_args -> Item ( 2 ) . c_str () ),
							pax -> CalcValueFromString ( (char*) frame_args -> Item ( 3 ) . c_str () ) ) ;
	t_gem -> SetDelays ( pax -> CalcValueFromString ( (char*) delay_args -> Item ( 0 ) . c_str () ),
							pax -> CalcValueFromString ( (char*) delay_args -> Item ( 1 ) . c_str () ),
							pax -> CalcValueFromString ( (char*) delay_args -> Item ( 2 ) . c_str () ),
							pax -> CalcValueFromString ( (char*) delay_args -> Item ( 3 ) . c_str () ),
							pax -> CalcValueFromString ( (char*) delay_args -> Item ( 4 ) . c_str () ),
							pax -> CalcValueFromString ( (char*) delay_args -> Item ( 5 ) . c_str () ) ) ;
	t_gem -> SetXY ( x, y ) ;
	if ( do_slide == TRUE )
		t_gem -> SlideInFrom ( slide_x, slide_y ) ;
	if ( do_slide == FALSE ) 
		t_gem -> SetCreationSound ( (char*) createsnd_args -> Item ( 0 ) . c_str (),
								pax -> CalcValueFromString ( (char*) createsnd_args -> Item ( 1 ) . c_str () ) ) ;
	t_gem -> SetMoveSound ( (char*) movesnd_args -> Item ( 0 ) . c_str (),
								pax -> CalcValueFromString ( (char*) movesnd_args -> Item ( 1 ) . c_str () ) ) ;
	t_gem -> SetStopSound ( (char*) stopsnd_args -> Item ( 0 ) . c_str (),
								pax -> CalcValueFromString ( (char*) stopsnd_args -> Item ( 1 ) . c_str () ) ) ;
	t_gem -> SetAccelSound ( (char*) accelsnd_args -> Item ( 0 ) . c_str (),
								pax -> CalcValueFromString ( (char*) accelsnd_args -> Item ( 1 ) . c_str () ) ) ;
	t_gem -> SetBounceSound ( (char*) bumpsnd_args -> Item ( 0 ) . c_str (),
								pax -> CalcValueFromString ( (char*) bumpsnd_args -> Item ( 1 ) . c_str () ) ) ;
	t_gem -> SetAlpha ( 0 ) ;
	t_gem -> SetExplosionResource ( (char*) explosion_args -> Item ( 0 ) . c_str () ) ;
	t_gem -> SetExplosionColor ( gem_data -> exp_red, gem_data -> exp_green, gem_data -> exp_blue ) ;
	t_gem -> SetTwinklingSpeed ( pax -> CalcValueFromString ( (char*) twinkspeed_args -> Item ( 0 ) . c_str () ),
									pax -> CalcValueFromString ( (char*) twinkspeed_args -> Item ( 1 ) . c_str () ) ) ;
	if ( pos_in_playfield >= -1 )
	{
		gem_pos [ item ] = pos_in_playfield ;
		gem_active [ item ] = TRUE ;
	}
	
	frame_args      -> Clear () ; delete frame_args      ;
	delay_args      -> Clear () ; delete delay_args      ;
	createsnd_args  -> Clear () ; delete createsnd_args  ;
	movesnd_args    -> Clear () ; delete movesnd_args    ;
	stopsnd_args    -> Clear () ; delete stopsnd_args    ;
	accelsnd_args   -> Clear () ; delete accelsnd_args   ;
	bumpsnd_args    -> Clear () ; delete bumpsnd_args    ;
	explosion_args  -> Clear () ; delete explosion_args  ;
	twinkspeed_args -> Clear () ; delete twinkspeed_args ;
	
	delete gem_data ;
	
	return item ;
}

void GamePF::UpdateGemPreview ()
{
	// push gems ahead...
	for ( int i = preview_gems - 2 ; i >= 0 ; i-- )
	{
		if ( preview_gem_id [ i ] != -1 &&  preview_gem_id [ i + 1 ] == -1 )
		{
			gem [ preview_gem_id [ i ] ] -> SetXY ( 0, ( i + 1 ) * GEM_HEIGHT ) ;
			gem [ preview_gem_id [ i ] ] -> SlideInFrom ( 0, -GEM_HEIGHT ) ;
			//gem [ preview_gem_id [ i ] ] -> SetDestXY ( 0, ( i + 1 ) * GEM_HEIGHT ) ;
			gem [ preview_gem_id [ i ] ] -> Jump ( GEM_HEIGHT ) ;
			preview_gem_id [ i + 1 ] = preview_gem_id [ i ] ;
			preview_gem_color [ i + 1 ] = preview_gem_color [ i ] ;
			preview_gem_id [ i ] = -1 ; preview_gem_color [ i ] = GEM_NONE ;
			//fprintf ( stdout, "switch gems %d -> %d\n", i, i+1 ) ;
			if ( preview_gem_id [ i + 1 ] != -1 && using_opengl == TRUE )
				if ( i + 1 == preview_gems - 1 )
					gem [ preview_gem_id [ i + 1 ] ] -> Twinkling ( TRUE ) ;
				else
					gem [ preview_gem_id [ i + 1 ] ] -> Twinkling ( FALSE ) ;
		}
	}
	
	// create new gems if needed
	for ( int i = 0 ; i < preview_gems ; i++ )
	{
		if ( preview_gem_id [ i ] == -1 )
		{
			char color = preview_gem_color [ i ] ;
			if ( preview_gem_color [ i ] == GEM_NONE )
				color = ( rand() % gem_colors_activated ) + GEM_FIRST_COLOR ;
			preview_gem_id [ i ] = SetupGem ( color, -1, previewfield, 
												0, i * GEM_HEIGHT, FALSE, 0, 0 ) ;
			if ( preview_gem_id [ i ] != -1 )
			{
				gem [ preview_gem_id [ i ] ] -> SlideInFrom ( 0, -GEM_HEIGHT ) ;
				//gem [ preview_gem_id [ i ] ] -> SetDestXY ( 0, i * GEM_HEIGHT ) ;
				gem [ preview_gem_id [ i ] ] -> Jump ( GEM_HEIGHT ) ;
				preview_gem_color [ i ] = color ;
				//gem [ preview_gem_id [ i ] ] -> Glow ( TRUE ) ;
				//gem [ preview_gem_id [ i ] ] -> Twinkling ( TRUE ) ;
				//fprintf ( stdout, "%d. new gem created with id %d and color '%c' !!!\n", 
				//			i, preview_gem_id [ i ], preview_gem_color [ i ] ) ;
			}
			//else
				//fprintf ( stdout, "failed to create new gem %d for preview!\n", i ) ;
		}
	}
}

int GamePF::GetEmptyPosAtBottom () // this one gets an randomly chosen pos!!!
{
	int pos = -1 ;
	int row = PLAYFIELD_MAX_Y - 1 ; // last row
	int possible_cols = 0 ;
	int col = 0 ; // current column
	// first calculate how many empty slots we have...
	for ( ; col < PLAYFIELD_MAX_X ; col++ )
	{
		int pos = ( row * PLAYFIELD_MAX_X ) + col ;
		if ( playfield [ pos ] == GEM_NONE ) possible_cols++ ;
	}
	
	//fprintf ( stdout, "we have %d empty slots to throw in...\n", possible_cols ) ;
	
	if ( possible_cols < 1 ) return -1 ;
	
	// decide which one of the empty slots it will be
	int which_col_do_we_want = possible_cols > 0 ? rand () % possible_cols : 0 ;
	
	//fprintf ( stdout, "I'll choose the %d. - that will be ", which_col_do_we_want + 1 ) ;
	
	// now calculate which col this will be...
	int col_counter = 0 ;
	col = 0 ;
	while ( col_counter <= which_col_do_we_want )
	{
		pos = row * PLAYFIELD_MAX_X + col ;
		if ( playfield [ pos ] == GEM_NONE )
			col_counter++ ;
		col++ ;
	}
	col-- ;
	pos = row * PLAYFIELD_MAX_X + col ;
	
	//fprintf ( stdout, "pos %d / col no %d !\n", pos, col ) ;
	
	return pos ;
}

void GamePF::CreateUserGemCursorAtPos ( int pos )
{
	int col = pos % PLAYFIELD_MAX_X ;
	int row = ( pos - col ) / PLAYFIELD_MAX_X ;

	if ( user_gem_cursor != NULL )
	{
		user_gem_cursor -> Free () ;
		delete user_gem_cursor ;
		user_gem_cursor = NULL ;
	}
	
	gem_color_data *gem_data = GetGemColorDataFromPlayfieldChar ( playfield [ pos ] ) ;
	char cursor_string [ 256 ] ;
	sprintf ( cursor_string, "gem_%s_cursor", gem_data -> name ) ;
	delete gem_data ;
	
	if ( pax -> GetObjectFromList ( cursor_string ) )
	{
		user_gem_cursor = new Sprite () ;
		user_gem_cursor -> Init ( Vault, pax -> GetStringVar ( cursor_string ), basefield ) ;
		user_gem_cursor -> SetXY ( col * GEM_WIDTH, row * GEM_HEIGHT ) ;
		user_gem_cursor -> SetFrames ( AUTO, AUTO, AUTO, TRUE ) ;
		user_gem_cursor -> SetDelays ( 3 ) ;
		user_gem_cursor -> SetZDepthPos ( 10 ) ;
		user_gem_cursor -> SetAlpha ( 0 ) ;
		user_gem_cursor -> SetDestAlpha ( 128 ) ;
		user_gem_cursor -> SetXAccelFactor ( 4 ) ;
		user_gem_cursor -> SetYAccelFactor ( 4 ) ;
		user_gem_cursor -> Jump ( GEM_HEIGHT ) ;
	}
}

void GamePF::ThrowInGem ( int pos, bool do_slide )
{
	// when submitted with an invalid playfield position
	// the gem will be a random one comming from the bottom
	// it will also be movable by the user
	if ( ( pos == -1 || pos == -2 ) && playfield_is_set_up == TRUE )
	{
		int gem_mode = pos ; // -1 = user_gem_id ; -2 = slave_gem_id
		// get an empty pos at the bottom of our playfield...
		if ( gem_mode == -1 )
		{
			pos = GetEmptyPosAtBottom () ;
		}
		else
		{
			int left   = GetRelPos ( gem_pos [ user_gem_id ], -1,  0 ) ;
			int right  = GetRelPos ( gem_pos [ user_gem_id ],  1,  0 ) ;
			int top    = GetRelPos ( gem_pos [ user_gem_id ],  0, -1 ) ;
			int bottom = GetRelPos ( gem_pos [ user_gem_id ],  0,  1 ) ;
			
			if ( left != -1 && playfield [ left ] == GEM_NONE )
				pos = left ;
			else if ( right != -1 && playfield [ right ] == GEM_NONE )
				pos = right ;
			else if ( top != -1 && playfield [ top ] == GEM_NONE )
				pos = top ;
			else if ( bottom != -1 && playfield [ bottom ] == GEM_NONE )
				pos = bottom ;
			else
				pos = -1 ;
		}
		// go ahead and create our gem...
		
		UpdateGemPreview () ;
		
		if ( pos == -1 || 
				preview_gem_id [ preview_gems - 1 ] == -1 ||
				SlidingFinished () == FALSE ||
				! playfield_set || ! playfield_ready ) 
		{
			if ( gem_mode == -2 )
				slave_gem_thrown = true ;
			return ;
		}

		//fprintf ( stdout, "throw in new gem!\n" ) ;

		int col = pos % PLAYFIELD_MAX_X ;
		int row = ( pos - col ) / PLAYFIELD_MAX_X ;
		
		int cur_gem_id = preview_gem_id [ preview_gems - 1 ] ;
		preview_gem_id [ preview_gems - 1 ] = -1 ;
		playfield [ pos ] = preview_gem_color [ preview_gems - 1 ] ;
		preview_gem_color [ preview_gems - 1 ] = GEM_NONE ;
		
		if ( gem_mode == -1 )
		{
			user_gem_id  = cur_gem_id ;
			
			CreateUserGemCursorAtPos ( pos ) ;
			
			slave_gem_thrown = false ;
		}
		else
		{
			slave_gem_id = cur_gem_id ;
			slave_gem_thrown = true ;
		}
		
		// wait until gem has finished sliding
		while ( gem [ cur_gem_id ] -> SlideFinished () == false )
			gem [ cur_gem_id ] -> DoSlide () ;
		
		//int x_d1 = col * GEM_WIDTH, y_d1 = row * GEM_HEIGHT ;
		//fprintf ( stdout, "x_d1 = %d ; y_d1 = %d\n", x_d1, y_d1 ) ;
		gem [ cur_gem_id ] -> HandOverToBaseAt ( basefield, col * GEM_WIDTH, row * GEM_HEIGHT ) ;
		
		int x_d2 = gem [ cur_gem_id ] -> ConvertXPos ( previewfield, 0 ) - ( col * GEM_WIDTH ) ;
		int y_d2 = gem [ cur_gem_id ] -> ConvertYPos ( previewfield, preview_gems * GEM_HEIGHT ) - ( ( row - ( preview_gems - 2 ) ) * GEM_HEIGHT ) ;
		//fprintf ( stdout, "x_d2 = %d ; y_d2 = %d\n", x_d2, y_d2 ) ;
		gem [ cur_gem_id ] -> SlideInFrom ( x_d2, y_d2 ) ;
		gem_pos [ cur_gem_id ] = ( row * PLAYFIELD_MAX_X ) + col ;
		gem [ cur_gem_id ] -> Glow ( TRUE ) ;
		if ( using_opengl || ( using_opengl == false && gem_mode == -1 ) )
		{
			gem [ cur_gem_id ] -> Twinkling ( TRUE ) ;
		}
		
		{	// sadly the handovertobaseat function has to re-init the rain
			// so that we have to reload our twinkling-settings :(
			char gem_color[256] ;
			sprintf ( gem_color, "%s", pax -> GetStringVar ( "gem_block" ) ) ;
			char color = playfield [ pos ] ;
			
			if ( color >= GEM_FIRST_COLOR && color < GEM_FIRST_COLOR + GEM_COUNT )
			{
				char my_gem [ 256 ] ;
				sprintf ( my_gem, "gem_%d", color - GEM_FIRST_COLOR ) ;
				sprintf ( gem_color, "%s", pax -> GetStringVar ( my_gem ) ) ;
			}
			else if ( color == GEM_BLOCK )
				sprintf ( gem_color, "%s", pax -> GetStringVar ( "gem_block" ) ) ;
			
			char twinkspeed_string [ 4096 ] ;
			sprintf ( twinkspeed_string, "gem_%s_twinkspeed"  , gem_color ) ;
			sprintf ( twinkspeed_string, "%s", pax -> GetStringVar ( twinkspeed_string ) ) ;
			GlDynArray < std::string > *twinkspeed_args = pax -> GetArgList ( twinkspeed_string ) ;
			gem [ cur_gem_id ] -> SetTwinklingSpeed ( pax -> CalcValueFromString ( (char*) twinkspeed_args -> Item ( 0 ) . c_str () ),
												pax -> CalcValueFromString ( (char*) twinkspeed_args -> Item ( 1 ) . c_str () ) ) ;
			twinkspeed_args -> Clear () ;
			delete twinkspeed_args ;
		}
		
		//int jump_distance = x_d2 ;
		//jump_distance = jump_distance % 2 ? jump_distance + 1 : jump_distance ;
		//jump_distance = jump_distance < 0 ? -jump_distance : jump_distance ;
		
		//int x1 = col * GEM_WIDTH, y1 = row * GEM_HEIGHT ;
		//int x2 = gem [ cur_gem_id ] -> ConvertXPos ( previewfield, 0 ) ;
		//int y2 = gem [ cur_gem_id ] -> ConvertYPos ( previewfield, preview_gems * GEM_HEIGHT ) ;
		//int jump_distance = (x2-x1)^2 + (y2-y1)^2 ;
		//jump_distance = jump_distance < 0 ? -jump_distance : jump_distance ;
		
		int x = x_d2 < 0 ? -x_d2 : x_d2 ;
		int y = y_d2 < 0 ? -y_d2 : y_d2 ;
		int jump_distance = ( x + y ) / 2 + ( ( GEM_WIDTH + GEM_HEIGHT ) / 2 ) * 2 ;
		
		gem [ cur_gem_id ] -> Jump ( jump_distance ) ;

		HighlightEqualGems () ;
		
		UpdateGemPreview () ;
		
		//if ( gem_mode == -1 )
		//	ThrowInGem ( -2 ) ;
	}
	else if ( pos != -1 && pos != -2 )
	{	// when submitted with a valid position gather needed data
		int item ;
		int col = pos % PLAYFIELD_MAX_X ;
		int row = ( pos - col ) / PLAYFIELD_MAX_X ;

		int tmp_playfield_throw_in_effect = playfield_throw_in_effect ;
		// while playing don't throw them from the top
		if ( ( tmp_playfield_throw_in_effect % 2 ) && playfield_is_set_up )
			tmp_playfield_throw_in_effect-- ;
		
		// note: order is: up, down, up, down, etc.
		
		switch ( tmp_playfield_throw_in_effect )
		{
			case 0:	// everything comes up in a plain row
			{
				if ( using_opengl == FALSE && row < PLAYFIELD_MAX_Y / 2 &&
						playfield_is_set_up == FALSE )
				{	// up -> down
					item = SetupGem ( FALSE, pos, basefield, col * GEM_WIDTH, row * GEM_HEIGHT,
										do_slide, 0, ( -2 - row ) * GEM_HEIGHT ) ;
				}
				else
				{	// down -> up
					item = SetupGem ( FALSE, pos, basefield, col * GEM_WIDTH, row * GEM_HEIGHT,
										do_slide, 0, ( PLAYFIELD_MAX_Y + 2 - row ) * GEM_HEIGHT ) ;
				}
			}
			break ;
			case 1:	// everything comes down in a plain row
			{
				if ( using_opengl == FALSE && row < PLAYFIELD_MAX_Y / 2 &&
						playfield_is_set_up == FALSE )
				{	// down -> up
					item = SetupGem ( FALSE, pos, basefield, col * GEM_WIDTH, row * GEM_HEIGHT,
										do_slide, 0, ( PLAYFIELD_MAX_Y + 2 - row ) * GEM_HEIGHT ) ;
				}
				else
				{	// up -> down
					item = SetupGem ( FALSE, pos, basefield, col * GEM_WIDTH, row * GEM_HEIGHT,
										do_slide, 0, ( -2 - row ) * GEM_HEIGHT ) ;
				}
			}
			break ;
			case 2: // everything comes up from the middle and unfolds into a row
			{
				if ( using_opengl == FALSE && row < PLAYFIELD_MAX_Y / 2 &&
						playfield_is_set_up == FALSE )
				{	// up -> down
					item = SetupGem ( FALSE, pos, basefield, col * GEM_WIDTH, row * GEM_HEIGHT, do_slide,
										( PLAYFIELD_MAX_X / 2 - col ) * GEM_WIDTH,
										( -2 - row ) * GEM_HEIGHT ) ;
				}
				else
				{	// down -> up
					item = SetupGem ( FALSE, pos, basefield, col * GEM_WIDTH, row * GEM_HEIGHT, do_slide,
										( PLAYFIELD_MAX_X / 2 - col ) * GEM_WIDTH,
										( PLAYFIELD_MAX_Y + 2 - row ) * GEM_HEIGHT ) ;
				}
			}
			break ;
			case 3: // everything comes down from the middle and unfolds into a row
			{
				if ( using_opengl == FALSE && row < PLAYFIELD_MAX_Y / 2 &&
						playfield_is_set_up == FALSE )
				{	// down -> up
					item = SetupGem ( FALSE, pos, basefield, col * GEM_WIDTH, row * GEM_HEIGHT, do_slide,
										( PLAYFIELD_MAX_X / 2 - col ) * GEM_WIDTH,
										( PLAYFIELD_MAX_Y + 2 - row ) * GEM_HEIGHT ) ;
				}
				else
				{	// up -> down
					item = SetupGem ( FALSE, pos, basefield, col * GEM_WIDTH, row * GEM_HEIGHT, do_slide,
										( PLAYFIELD_MAX_X / 2 - col ) * GEM_WIDTH,
										( -2 - row ) * GEM_HEIGHT ) ;
				}
			}
			break ;
		}
		
		//int color = playfield [ pos ] ;
		//fprintf ( stdout, "new gem id%0d at %dx%d with color '%c'\n", item, col, row, color ) ;
	}
}

void GamePF::DestroyPlayfield ()
{
	int direction = 6 ; // take a look at your numberpad...
	int top, bottom, left, right ;
	top = 0 ;
	bottom = PLAYFIELD_MAX_Y - 1 ;
	left = 0 ;
	right = PLAYFIELD_MAX_X - 1 ;
	
	bool not_finished = true ;
	int current_pos = 0 ;
	int current_time = PLAYFIELD_SIZE ;
	
	while ( not_finished )
	{
		int x = current_pos % PLAYFIELD_MAX_X ;
		int y = ( current_pos - x ) / PLAYFIELD_MAX_X ;
		int current_item = GetMatchingGem ( current_pos ) ;
		
		if ( current_item != -1 )
		{
			playfield [ current_pos ] = GEM_HOLD ;
			UnfreezeGem ( current_item ) ;
			gem [ current_item ] -> Explode ( -(PLAYFIELD_SIZE-current_time)*4, 10 ) ;
		}
		
		current_time-- ;
		
		if ( direction == 6 )
		{
			if ( x < right )
				x++ ;
			else
			{
				if ( y < bottom )
				{
					y++ ;
					top++ ;
					direction = 2 ; // go down
				}
				else
					not_finished = false ;
			}
		}
		else if ( direction == 2 )
		{
			if ( y < bottom )
				y++ ;
			else
			{
				if ( x > left )
				{
					x-- ;
					right-- ;
					direction = 4 ; // go left
				}
				else
					not_finished = false ;
			}
		}
		else if ( direction == 4 )
		{
			if ( x > left )
				x-- ;
			else
			{
				if ( y > top )
				{
					y-- ;
					bottom-- ;
					direction = 8 ; // go up
				}
				else
					not_finished = false ;
			}
		}
		else if ( direction == 8 )
		{
			if ( y > top )
				y-- ;
			else
			{
				if ( x < right )
				{
					x++ ;
					left++ ;
					direction = 6 ; // go right
				}
				else
					not_finished = false ;
			}
		}
			
		current_pos = y * PLAYFIELD_MAX_X + x ;
	}
	
	current_time = PLAYFIELD_SIZE ;
	
	for ( int item = 0 ; item < preview_gems ; item++ )
	{
		if ( preview_gem_id [ item ] != -1 )
		{
			gem [ preview_gem_id [ item ] ] -> Explode ( current_time,
														255 / PLAYFIELD_SIZE ) ;
			preview_gem_id [ item ] = -1 ;
			preview_gem_color [ item ] = -1 ;
		}
		current_time += PLAYFIELD_SIZE / preview_gems ;
	}
	
	playfield_is_set_up = FALSE ;

	user_gem_id = slave_gem_id = -1 ;
	if ( user_gem_cursor )
	{
		user_gem_cursor -> Free () ;
		delete user_gem_cursor ;
		user_gem_cursor = NULL ;
	}
	if ( slave_gem_cursor )
	{
		slave_gem_cursor -> Free () ;
		delete slave_gem_cursor ;
		slave_gem_cursor = NULL ;
	}
}

void GamePF::CheckForFullPlayfield ()
{
	bool is_full = TRUE ;
	
	if ( SlidingFinished () == FALSE || playfield_is_set_up == FALSE ) return ;
	
	for ( int i = 0 ; i < PLAYFIELD_SIZE ; i++ )
	{
		// check for valid entry
		if ( playfield [ i ] != GEM_NONE && playfield [ i ] != GEM_HOLD &&
				playfield [ i ] != GEM_BLOCK && 
				( playfield [ i ] < GEM_FIRST_COLOR ||
				  playfield [ i ] >= GEM_FIRST_COLOR + GEM_COUNT ) )
		{	// AAH!! We've detected some mess in our playfield!!!
			int current_gem = GetMatchingGem  ( i ) ;
			DumpStatus () ;
			if ( current_gem != -1 )
			{
				UnfreezeGem ( current_gem ) ;
				gem [ current_gem ] -> Explode () ;
				playfield [ i ] = GEM_HOLD ;
			}
			else
			{
				playfield [ i ] = GEM_NONE ;
			}
		}
		
		// remove invalid gem from playfield
		if ( playfield [ i ] != GEM_NONE && GetMatchingGem ( i ) == -1 )
			playfield [ i ] = GEM_NONE ;
		
		// check for space
		if ( i >= PLAYFIELD_SIZE - PLAYFIELD_MAX_X - 1 &&
			 (	playfield [ i ] == GEM_NONE ||
				playfield [ i ] == GEM_HOLD ) ) is_full = FALSE ;
	}
	
	if ( is_full == TRUE )
	{	// playfield is full! reset it!
		//DumpStatus () ; // but first dump its contents :)
		
		DestroyPlayfield () ;
		
		playfield_set = FALSE ;
		playfield_is_set_up = FALSE ;
		PlayerScoreReset () ;
	}
}

void GamePF::FreezeGem ( int id )
{
	gem_frozen [ id ] = true ;
	if ( gem_frozen_layer [ gem_pos [ id ] ] != NULL )
		gem_frozen_layer [ gem_pos [ id ] ] -> SetDestAlpha ( 255 ) ;
}

void GamePF::UnfreezeGem ( int id )
{
	gem_frozen [ id ] = false ;
	if ( gem_frozen_layer [ gem_pos [ id ] ] != NULL )
		gem_frozen_layer [ gem_pos [ id ] ] -> SetDestAlpha ( 0 ) ;
}

void GamePF::UnfreezeGemArea ( int id )
{
	int lt = GetMatchingGem ( GetRelPos ( gem_pos [ id ], -1, -1 ) ) ;
	int t  = GetMatchingGem ( GetRelPos ( gem_pos [ id ],  0, -1 ) ) ;
	int rt = GetMatchingGem ( GetRelPos ( gem_pos [ id ],  1, -1 ) ) ;
	int r  = GetMatchingGem ( GetRelPos ( gem_pos [ id ],  1,  0 ) ) ;
	int rb = GetMatchingGem ( GetRelPos ( gem_pos [ id ],  1,  1 ) ) ;
	int b  = GetMatchingGem ( GetRelPos ( gem_pos [ id ],  0,  1 ) ) ;
	int lb = GetMatchingGem ( GetRelPos ( gem_pos [ id ], -1,  1 ) ) ;
	int l  = GetMatchingGem ( GetRelPos ( gem_pos [ id ], -1,  0 ) ) ;
	
	if ( lt != -1 ) UnfreezeGem ( lt ) ;
	if ( t  != -1 ) UnfreezeGem ( t  ) ;
	if ( rt != -1 ) UnfreezeGem ( rt ) ;
	if ( r  != -1 ) UnfreezeGem ( r  ) ;
	if ( rb != -1 ) UnfreezeGem ( rb ) ;
	if ( b  != -1 ) UnfreezeGem ( b  ) ;
	if ( lb != -1 ) UnfreezeGem ( lb ) ;
	if ( l  != -1 ) UnfreezeGem ( l  ) ;
	
	UnfreezeGem ( id ) ;
}

void GamePF::Draw ()
{
	static int counter_for_setup = 0 ;
	
	// Update score field
	PlayerScoreUpdate () ;
	
	if ( quit_state == TRUE || paused == TRUE ) return ;
	
	int item = 0 ; // go to first possible gem
	
	bool have_pair = FALSE ;
	
	int moving_gems = 0 ; // how many gems have to be moved
		
	if ( action_indicator == action_step_one && destroyable_gems == FALSE )
	{
		// before pulling in the first row create a playfield!!!
		if ( playfield_is_set_up == FALSE && quit == FALSE )
		{
			bool someone_left = FALSE ;
			for ( int i = 0 ; i < PLAYFIELD_SIZE ; i++ )
				if ( playfield [ i ] != GEM_NONE ||
						gem_active [ i ] == TRUE /*||
						gem [ i ] -> GetIsDone () == FALSE*/ )
					someone_left = TRUE ;
			if ( someone_left == FALSE )
			{
				pull_row = 0 ;
				// create playfield in memory and indicate correct state
				SetupPlayfield () ;
			}
		}
	
		if ( falling_gems_enabled == TRUE )
		{
			for ( item = 0 ; item < PLAYFIELD_SIZE ; item++ )
			{
				if ( gem_active [ item ] == TRUE && pull_row == PLAYFIELD_MAX_Y &&
						item != user_gem_id && item != slave_gem_id &&
						gem [ item ] -> SlideFinishedX () &&
						gem_pos [ item ] >= PLAYFIELD_MAX_X &&
						gem_frozen [ item ] == false ) // only do unfrozen gems !
				{	// now test for empty spaces and fill them
					int test_item = gem_pos [ item ] ;
					int jump_rows = 0 ;
					bool not_finished = TRUE ;
					
					while ( not_finished )
					{
						bool first = FALSE, second = FALSE ;
						if ( test_item >= PLAYFIELD_MAX_X && 
								playfield [ test_item - PLAYFIELD_MAX_X ] == GEM_NONE )
						{
							test_item -= PLAYFIELD_MAX_X ;
							jump_rows += 1 ;
							first = TRUE ;
						}
						if ( test_item >= PLAYFIELD_MAX_X * 2 && 
									slave_gem_id != -1 && gem_pos [ slave_gem_id ] < gem_pos [ user_gem_id ] &&
									playfield [ test_item - PLAYFIELD_MAX_X ] == playfield [ gem_pos [ slave_gem_id ] ] &&
									playfield [ test_item - ( PLAYFIELD_MAX_X * 2 ) ] == GEM_NONE )
						{
							test_item -= PLAYFIELD_MAX_X * 2 ;
							jump_rows += 2 ;
							second = TRUE ;
						}
						if ( test_item >= PLAYFIELD_MAX_X * 2 && 
									user_gem_id != -1 &&
									playfield [ test_item - PLAYFIELD_MAX_X ] == playfield [ gem_pos [ user_gem_id ] ] &&
									playfield [ test_item - ( PLAYFIELD_MAX_X * 2 ) ] == GEM_NONE )
						{
							test_item -= PLAYFIELD_MAX_X * 2 ;
							jump_rows += 2 ;
							second = TRUE ;
						}
						if ( test_item >= PLAYFIELD_MAX_X * 2 && 
									slave_gem_id != -1 && gem_pos [ slave_gem_id ] >= gem_pos [ user_gem_id ] &&
									playfield [ test_item - PLAYFIELD_MAX_X ] == playfield [ gem_pos [ slave_gem_id ] ] &&
									playfield [ test_item - ( PLAYFIELD_MAX_X * 2 ) ] == GEM_NONE )
						{
							test_item -= PLAYFIELD_MAX_X * 2 ;
							jump_rows += 2 ;
							second = TRUE ;
						}
						if ( first == FALSE && second == FALSE )
							not_finished = FALSE ;
					}
					
					if ( test_item != gem_pos [ item ] ) 
					{
						int rest = gem_pos [ item ] % PLAYFIELD_MAX_X ;
						int row = ( gem_pos [ item ] - rest ) / PLAYFIELD_MAX_X ;
						int a, b ;
						a = playfield [ gem_pos [ item ] ] ;
						b = playfield [ test_item ] ;
						playfield [ gem_pos [ item ] ] = b ;
						playfield [ test_item ] = a ;
						
						//fprintf ( stdout, "switching items %d <-> %d ; '%c' <-> '%c'\n",
						//			gem_pos [ item ], test_item,
						//			a, b ) ;
	
						gem [ item ] -> SetXY ( ( gem_pos [ item ] % PLAYFIELD_MAX_X ) * GEM_WIDTH,
												( row - jump_rows ) * GEM_HEIGHT ) ;
						gem [ item ] -> SlideInFrom ( 0, jump_rows * GEM_HEIGHT ) ;
						gem [ item ] -> Jump ( jump_rows * GEM_HEIGHT ) ;
	
						gem_pos [ item ] = test_item ;
						
						moving_gems++ ;
						
						if ( ! playfield_set )
							counter_for_setup = 0 ;
					}
				}
			}
		}
		
		if ( moving_gems > 5 )
		{
			incoming_earthquake = TRUE ; // :)
			
			pre_earthquake_sound.SetVolume ( 64 ) ;
			if ( moving_gems > 10 ) pre_earthquake_sound.SetVolume ( 128 ) ;
				
			pre_earthquake_sound.PlaySound () ;
		}
	}
	
	if ( incoming_earthquake == TRUE && SlidingFinished () )
	{
		incoming_earthquake = FALSE ;
		earthquake_sound.SetVolume ( 64 ) ;
		earthquake_sound.PlaySound () ;
	}

	if ( action_indicator == action_step_two || quit == TRUE )
	{
		destroyable_gems = FALSE ;
	}
	
	for ( item = 0 ; item < PLAYFIELD_SIZE ; item++ )
	{
		int rest = gem_pos [ item ] % PLAYFIELD_MAX_X ;
		int row = ( gem_pos [ item ] - rest ) / PLAYFIELD_MAX_X ;
		
		if ( gem_active [ item ] == TRUE && row <= pull_row )
		{
			// as soon as all gems have stopped sliding check for pairs or more
			if ( /*destroyable_gems == FALSE &&*/ action_indicator == action_step_two && 
					SlidingFinished () == TRUE && have_pair == FALSE && 
					playfield [ gem_pos [ item ] ] >= GEM_FIRST_COLOR &&
					playfield [ gem_pos [ item ] ] <= GEM_LAST_COLOR &&
					(
					  ( gem_active [ user_gem_id ] == TRUE &&
					    playfield [ gem_pos [ item ] ] == playfield [ gem_pos [ user_gem_id ] ] &&
						gem [ user_gem_id ] -> SlideFinished () == TRUE
					  ) || (
					    gem_active [ slave_gem_id ] == TRUE &&
					    playfield [ gem_pos [ item ] ] == playfield [ gem_pos [ slave_gem_id ] ] &&
						gem [ user_gem_id ] -> SlideFinished () == TRUE
					  ) || (
					    gem_active [ user_gem_id ] == TRUE &&
					    playfield [ gem_pos [ item ] ] != playfield [ gem_pos [ user_gem_id ] ] //&&
						//gem [ user_gem_id ] -> SlideFinished () == TRUE
					  ) || (
					    gem_active [ slave_gem_id ] == TRUE &&
					    playfield [ gem_pos [ item ] ] != playfield [ gem_pos [ slave_gem_id ] ] //&&
						//gem [ slave_gem_id ] -> SlideFinished () == TRUE
					  ) || (
						gem_active [ user_gem_id ] != TRUE
					  ) || (
						gem_active [ slave_gem_id ] != TRUE
					  )
					)
				)
			{
				int friends = 0 ;
				
				//fprintf ( stdout, "checking for group of gems...\n" ) ;
				
				friend_bucket -> Clear () ;
				friends = CheckGemForFriends ( gem_pos [ item ] ) ;
				
				if ( friends >= min_friends_for_group )
				{
					int above, below, left, right ;
					
					//fprintf ( stdout, "found a group of %d gems: \n", friends ) ;
					
					/*for ( int py = 0 ; py < PLAYFIELD_MAX_Y ; py++ )
					{
						for ( int px = 0 ; px < PLAYFIELD_MAX_X ; px++ )
							fprintf ( stdout, "%c ", playfield [ ( py * PLAYFIELD_MAX_X ) + px ] ) ;
						fprintf ( stdout, "\n" ) ;
					}*/
	
					int score_gathered = 0 ;
					char gem_color = playfield [ gem_pos [ item ] ] ;
					
					if ( PlayerScoreIsInComboTime () )
					{
						playerscore_combo *= 2 ;
						//fprintf ( stdout, "combo = %d\n", playerscore_combo ) ;
						char player_combo [ 256 ] ;
						sprintf ( player_combo, "x%d", playerscore_combo ) ;
						playercombobox -> SetTextChar ( player_combo, 0 ) ;
					}
					else
					{
						playerscore_combo = 1 ;
						playercombobox -> SetTextChar ( " ", 0 ) ;
					}
					
					for ( unsigned int i = 0 ; i < friend_bucket -> Count () ; i++ )
					{
						int current = friend_bucket -> Item ( i ) ;
						int cur_item = GetMatchingGem ( current ) ;
						
						//fprintf ( stdout, "%d='%c' ", current, playfield [ current ] ) ;
						
						score_gathered += playfield [ current ] - GEM_FIRST_COLOR + 1 ;
						
						playfield [ current ] = GEM_HOLD ;
						//gem_pos [ cur_item ] = -1 ;
						//gem [ cur_item ] -> Quit () ;
						UnfreezeGemArea ( cur_item ) ;
						gem [ cur_item ] -> Explode () ;
						
						if ( cur_item == user_gem_id )
						{
							gem [ user_gem_id ] -> Glow ( FALSE ) ;
							gem [ user_gem_id ] -> Twinkling ( FALSE ) ;
							// this will remove glow-effects from all gems...
							HighlightEqualGems () ;
							// and now release the gem
							user_gem_id = -1 ;
							if ( slave_gem_id != -1 )
							{
								user_gem_id = slave_gem_id ;
								slave_gem_id = -1 ;
								gem [ user_gem_id ] -> Twinkling ( TRUE ) ;
								int t_x = gem_pos [ user_gem_id ] % PLAYFIELD_MAX_X ;
								int t_y = ( gem_pos [ user_gem_id ] - t_x ) / PLAYFIELD_MAX_X ;
								if ( user_gem_cursor )
									user_gem_cursor -> SetDestXY ( t_x * GEM_WIDTH, t_y * GEM_HEIGHT ) ;
							}
						}
						if ( cur_item == slave_gem_id )
						{
							gem [ slave_gem_id ] -> Glow ( FALSE ) ;
							gem [ slave_gem_id ] -> Twinkling ( FALSE ) ;
							// this will remove glow-effects from all gems...
							HighlightEqualGems () ;
							// and now release the gem
							slave_gem_id = -1 ;
						}
							
						// kill blocks too :)
						above = GetRelPos ( current, 0, -1 ) ;
						below = GetRelPos ( current, 0, 1 ) ;
						left = GetRelPos ( current, -1, 0 ) ;
						right = GetRelPos ( current, 1, 0 ) ;
						if ( playfield [ above ] == GEM_BLOCK ) 
						{
							playfield [ above ] = GEM_HOLD ; 
							UnfreezeGemArea ( GetMatchingGem ( above ) ) ;
							gem [ GetMatchingGem ( above ) ] -> Explode () ;
						}
						if ( playfield [ below ] == GEM_BLOCK ) 
						{
							playfield [ below ] = GEM_HOLD ; 
							UnfreezeGemArea ( GetMatchingGem ( below ) ) ;
							gem [ GetMatchingGem ( below ) ] -> Explode () ;
						}
						if ( playfield [ left ] == GEM_BLOCK ) 
						{
							playfield [ left ] = GEM_HOLD ; 
							UnfreezeGemArea ( GetMatchingGem ( left ) ) ;
							gem [ GetMatchingGem ( left ) ] -> Explode () ;
						}
						if ( playfield [ right ] == GEM_BLOCK ) 
						{
							playfield [ right ] = GEM_HOLD ; 
							UnfreezeGemArea ( GetMatchingGem ( right ) ) ;
							gem [ GetMatchingGem ( right ) ] -> Explode () ;
						}
					}

					if ( score_gathered && playfield_ready )
					{
						if ( score_gathered < 0 )
						{
							fprintf ( stdout, "score += %d\n", score_gathered ) ;
							DumpStatus () ;
							exit ( 1 ) ;
						}
						
						PlayerScoreSet ( PlayerScoreGet () + 
										 score_gathered * playerscore_combo ) ;
						playerscore_combo_rest_seconds = playerscore_combo_max_rest_seconds ;
						
						// punish our opponent :)
						if ( playerscore_combo >= 4 & playerscore_combo < 16 )
						{
							GemQueueAddToOutgoing ( gem_color, friends / 2 ) ;
						}
						else if ( playerscore_combo >= 16 )
						{
							int combo = GlClamp ( playerscore_combo, 16, 64 ) ;
							GemQueueAddToOutgoing ( GEM_BLOCK, friends * combo / 16 ) ;
							playerscore_bonus_sound.PlaySound () ;
						}
					}
					
					if ( ! playfield_set )
						counter_for_setup = 0 ;

					
					//fprintf ( stdout, "\n" ) ;
					
					/*for ( int py = 0 ; py < PLAYFIELD_MAX_Y ; py++ )
					{
						for ( int px = 0 ; px < PLAYFIELD_MAX_X ; px++ )
							fprintf ( stdout, "%c ", playfield [ ( py * PLAYFIELD_MAX_X ) + px ] ) ;
						fprintf ( stdout, "\n" ) ;
					}*/
	
					have_pair = TRUE ;
				}
			}
			
			// now draw it
			gem [ item ] -> Draw () ;
			
			// fade in our gem if needed
			if ( gem_pos [ item ] != -1 && playfield [ gem_pos [ item ] ] != GEM_HOLD &&
					playfield [ gem_pos [ item ] ] != GEM_NONE &&
					gem [ item ] -> GetAlpha () < 255 )
				gem [ item ] -> SetAlpha ( gem [ item ] -> GetAlpha () + 10 ) ;

			// clean up
			if ( gem [ item ] -> GetIsQuitting () )
			{
				if ( gem_pos [ item ] > -1 )
				{
					playfield [ gem_pos [ item ] ] = GEM_NONE ;
					gem_pos [ item ] = -1 ;
					//fprintf ( stdout, "remove gem %d from playfield.\n", item ) ;
				}
				//gem_active [ item ] = FALSE ;
				//gem [ item ] -> Free () ;
			}
			if ( gem [ item ] -> GetIsDone () )
			{
				/*if ( gem_pos [ item ] > -1 )
				{
					playfield [ gem_pos [ item ] ] = GEM_NONE ;
					gem_pos [ item ] = -1 ;
				}*/
				gem_active [ item ] = FALSE ;
				gem [ item ] -> Free () ;
				//fprintf ( stdout, "freed gem %d from memory.\n", item ) ;
			}
			if ( playfield [ gem_pos [ item ] ] == GEM_HOLD ||
					playfield [ gem_pos [ item ] ] == GEM_HOLD &&
					gem [ item ] -> GetIsQuitting () == TRUE &&
					gem [ item ] -> GetIsDone () == FALSE )
				destroyable_gems = TRUE ;
		}
		
		// draw our frozen layer thingy :)
		if ( pull_row == PLAYFIELD_MAX_Y )
		{
			int gmgem = GetMatchingGem ( item ) ;
			if ( gmgem != -1 )
			{
				if ( gem_frozen [ gmgem ] && gem [ gmgem ] -> SlideFinished () )
				{
					int channel = AUDIO_CHANNELS - 5 + ( rand() % 5 ) ;
					gem [ gmgem ] -> PauseAnim () ;
					if ( gem_frozen_layer [ item ] != NULL &&
						gem_frozen_layer [ item ] -> GetAlpha () == 0 )
						gem_frozen_layer [ item ] -> PlayAccelSound ( channel ) ;
				}
				else
					gem [ gmgem ] -> ContinueAnim () ;
			}
			if ( gem_frozen_layer [ item ] != NULL )
				gem_frozen_layer [ item ] -> Draw () ;
		}
	}
	
	// Update gem queue ( incoming and outgoing gems... )
	GemQueueUpdate () ;

	// if they exist draw them
	if ( user_gem_cursor != NULL )
	{
	       if ( user_gem_id == -1 )
	       {
		       user_gem_cursor -> Free () ;
		       delete user_gem_cursor ;
		       user_gem_cursor = NULL ;
	       }
	       else
		       user_gem_cursor -> Draw () ;
	}
	if ( slave_gem_cursor != NULL ) slave_gem_cursor -> Draw () ;
	
	// make sure that our playfield is properly made-up before playing!!!
	if ( ( moving_gems || destroyable_gems || pull_row != PLAYFIELD_MAX_Y || ! SlidingFinished () ) &&
		! playfield_set && ! playfield_ready && counter_for_setup > 0 )
		counter_for_setup = 0 ;
	
	if ( ! moving_gems && ! destroyable_gems && pull_row == PLAYFIELD_MAX_Y &&
		 ! playfield_set && ! playfield_ready &&
		 action_indicator >= action_max )
	{
		if ( counter_for_setup > 1 )
		{
			playfield_set = TRUE ;
			//fprintf ( stdout, "playfield is now ready for playing...\n" ) ;
			counter_for_setup = 0 ;
		}
		else
			counter_for_setup ++ ;
	}
	//
	
	if ( quit == TRUE )
	{
		//fprintf ( stdout, "moving_gems = %d ; destroyable_gems = %s\n",
		//			moving_gems, destroyable_gems ? "yes" : "no" ) ;
		if ( moving_gems || destroyable_gems )
		{
			pull_row = PLAYFIELD_MAX_Y ;
			return ;
		}
		else
		{
			quit_state = TRUE ;
			for ( int i = 0 ; i < PLAYFIELD_SIZE ; i++ )
			{
				if ( gem_active [ i ] == TRUE &&
						//gem [ i ] -> GetIsAlreadyDrawn () == TRUE &&
						gem [ i ] -> GetIsQuitting () == FALSE &&
						gem [ i ] -> GetIsDone () == FALSE )
				{
					//fprintf ( stdout, "gem %d is still active.. trying to kill it...\n", i ) ;
					gem [ i ] -> Quit () ;
				}
				if ( gem_active [ i ] == TRUE &&
						//gem [ i ] -> GetIsAlreadyDrawn () == TRUE &&
						gem [ i ] -> GetIsQuitting () == TRUE &&
						gem [ i ] -> GetIsDone () == TRUE )
				{
					fprintf ( stdout, "gem %d is still active.. trying to free it...\n", i ) ;
					gem [ i ] -> Free () ;
					gem_active [ i ] = FALSE ;
				}
				if ( playfield [ i ] != GEM_NONE ||
						gem [ i ] -> GetIsAlreadyDrawn () == TRUE &&
						gem [ i ] -> GetIsQuitting () == TRUE &&
						gem [ i ] -> GetIsDone () == FALSE )
					quit_state = FALSE ;
			}
			
			if ( quit_state == TRUE )
				return ;
		}
	}
	
	// fade in our preview gems if needed
	for ( int i = 0 ; i < preview_gems ; i++ )
	{
		int gem_id = preview_gem_id [ i ] ;
		if ( gem_id != -1 && gem [ gem_id ] -> GetAlpha () < 255 )
			gem [ gem_id ] -> SetAlpha ( gem [ gem_id ] -> GetAlpha () + 10 ) ;
	}
	
	// pull next row in if needed ( only used when setting up the playfield! )
	if ( pull_row < PLAYFIELD_MAX_Y )
	{
		if ( pull_row_counter < GEM_HEIGHT / 2 )
		{
			pull_row_counter++ ;
		}
		else if ( playfield_is_set_up == TRUE )
		{
			pull_row_counter = 0 ;
			pull_row++ ;
			//fprintf ( stdout, "pullrow is now %d\n", pull_row ) ;
		}
	}
	else if ( playfield_ready && quit == FALSE )
	{
		if ( SlidingFinished () == TRUE && have_pair == FALSE &&
				destroyable_gems == FALSE &&
				action_indicator == action_step_two )
		{
			if ( user_gem_id == -1 && slave_gem_id == -1 )
			{
				ThrowInGem () ;
				//Ai.Think ( playfield, gem_pos [ user_gem_id ] ) ;
				CheckForFullPlayfield () ;
				move_on_counter = 0 ;
			}
			else if ( user_gem_id != -1 )
			{
				if ( move_on_counter >= move_on_delay &&
						gem [ user_gem_id ] -> SlideFinishedY () &&
						( slave_gem_id != -1 || slave_gem_thrown == true ) )
				{
					MoveUserGemUp ( true ) ;
					move_on_counter = 0 ;
					//Ai.Think ( playfield, gem_pos [ user_gem_id ] ) ;
				} 
				else if ( slave_gem_id == -1 && slave_gem_thrown == false )
				{
					ThrowInGem ( -2 ) ;
				}
				else //if ( gem [ user_gem_id ] -> SlideFinishedY () )
				{
					move_on_counter++ ;
				}
			}

			if ( idle_row_counter >= idle_row_delay )
			{	// throw in block-stones
				idle_row_counter = 0 ;
				
				AddFallingGems ( GEM_BLOCK ) ;
			} 
			else
				idle_row_counter++ ;
		}
		
		if ( action_indicator % 3 == 0 )
		{
			if ( user_gem_id != -1 )
			{
				if ( move_up && gem [ user_gem_id ] -> SlideFinishedY () &&
					( ( slave_gem_id != -1 && gem [ slave_gem_id ] -> SlideFinishedY () ) ||
					  slave_gem_id == -1 ) ) MoveUserGemUp () ;
				if ( move_left ) MoveUserGemLeft () ;
				if ( move_right ) MoveUserGemRight () ;
			}
			if ( move_up || move_left || move_right || user_gem_id == -1 )
				idle_row_counter = 0 ;
		}
	}
	
	if ( action_indicator >= action_max )
		action_indicator = 0 ;
	else
		action_indicator++ ;
}

void GamePF::MoveUserGemUp ( bool force )
{
	bool user_slide_finished = user_gem_id != -1 ?
			gem [ user_gem_id ] -> SlideFinished () : true ;
	bool user_slide_finishedz = user_gem_id != -1 ?
			gem [ user_gem_id ] -> SlideFinishedZ () : true ;
	bool slave_slide_finished = slave_gem_id != -1 ?
			gem [ slave_gem_id ] -> SlideFinished () : true ;
	bool slave_slide_finishedz = slave_gem_id != -1 ?
			gem [ slave_gem_id ] -> SlideFinishedZ () : true ;
	
	if ( /*SlidingFinished () == FALSE ||*/ user_gem_id == -1 ||
			( ( user_slide_finished == FALSE ||
				slave_slide_finished == FALSE ) &&
				force == FALSE ) ||
			user_slide_finishedz == FALSE ||
			slave_slide_finishedz == FALSE ) return ;
	
	int current_pos = gem_pos [ user_gem_id ] ;
	int above = GetRelPos ( current_pos, 0, -1 ) ;
	int col = current_pos % PLAYFIELD_MAX_X ;
	int row = ( current_pos - col ) / PLAYFIELD_MAX_X ;
	int slave_pos = gem_pos [ slave_gem_id ] ;
	int slave_above = GetRelPos ( slave_pos, 0, -1 ) ;
	int slave_col = slave_pos % PLAYFIELD_MAX_X ;
	int slave_row = ( slave_pos - slave_col ) / PLAYFIELD_MAX_X ;
	
	bool cur_clear = false, slave_clear = false ;
	
	//fprintf ( stdout, "MoverUserGemUp ( %s );\n", force ? "true" : "false" ) ;
	
	if ( above != -1 )
	{
		if ( playfield [ above ] == GEM_NONE )
			cur_clear = true ;
	}
	
	if ( slave_gem_id != -1 ) // if there is a satellite gem...
	{
		if ( slave_above != -1 && slave_row < row &&
				playfield [ slave_above ] == GEM_NONE )
		{
			cur_clear = true ;
			slave_clear = true ;
		}
		if ( slave_above != -1 && slave_col > col &&
				playfield [ slave_above ] == GEM_NONE )
		{
			//if ( cur_clear )
				slave_clear = true ;
		}
		if ( slave_above != -1 && slave_row > row &&
				slave_above == current_pos )
		{
			if ( cur_clear )
				slave_clear = true ;
		}
		if ( slave_above != -1 && slave_col < col &&
				playfield [ slave_above ] == GEM_NONE )
		{
			//if ( cur_clear )
				slave_clear = true ;
		}
	}
	
	if ( slave_clear && slave_row < row )
	{
		playfield [ slave_above ] = playfield [ slave_pos ] ;
		playfield [ slave_pos ] = GEM_NONE ;
		gem_pos [ slave_gem_id ] = slave_above ;
		
		gem [ slave_gem_id ] -> SetDestXY ( slave_col * GEM_WIDTH,
								( slave_row - 1 ) * GEM_HEIGHT ) ;
		gem [ slave_gem_id ] -> Jump ( GEM_HEIGHT ) ;
	}
	if ( cur_clear ) 
	{
		playfield [ above ] = playfield [ current_pos ] ;
		playfield [ current_pos ] = GEM_NONE ;
		gem_pos [ user_gem_id ] = above ;
		
		gem [ user_gem_id ] -> SetDestXY ( col * GEM_WIDTH,
								( row - 1 ) * GEM_HEIGHT ) ;
		gem [ user_gem_id ] -> Jump ( GEM_HEIGHT ) ;
		if ( user_gem_cursor )
			user_gem_cursor -> SetDestXY ( col * GEM_WIDTH,
							( row - 1 ) * GEM_HEIGHT ) ;
	}
	if ( slave_clear && slave_row >= row )
	{
		playfield [ slave_above ] = playfield [ slave_pos ] ;
		playfield [ slave_pos ] = GEM_NONE ;
		gem_pos [ slave_gem_id ] = slave_above ;
		
		gem [ slave_gem_id ] -> SetDestXY ( slave_col * GEM_WIDTH,
								( slave_row - 1 ) * GEM_HEIGHT ) ;
		gem [ slave_gem_id ] -> Jump ( GEM_HEIGHT ) ;
	}
	
	if ( slave_clear == false && slave_gem_id != -1 && slave_row < row )
	{
		gem [ slave_gem_id ] -> PlayStopSound () ;
		gem [ slave_gem_id ] -> Glow ( FALSE ) ;
		gem [ slave_gem_id ] -> Twinkling ( FALSE ) ;
		//fprintf ( stdout, "lost gem %d ; above = %d, playfield[above] = '%c'\n",
		//			slave_gem_id, slave_above, playfield [ slave_above ] ) ;
		slave_gem_id = -1 ;
	}
	if ( cur_clear == false )
	{
		gem [ user_gem_id ] -> PlayStopSound () ;
		gem [ user_gem_id ] -> Glow ( FALSE ) ;
		gem [ user_gem_id ] -> Twinkling ( FALSE ) ;
		//fprintf ( stdout, "lost gem %d ; above = %d, playfield[above] = '%c'\n",
		//			user_gem_id, above, playfield [ above ] ) ;
		user_gem_id = -1 ;
	}
	if ( slave_clear == false && slave_gem_id != -1 && slave_row >= row )
	{
		gem [ slave_gem_id ] -> PlayStopSound () ;
		gem [ slave_gem_id ] -> Glow ( FALSE ) ;
		gem [ slave_gem_id ] -> Twinkling ( FALSE ) ;
		//fprintf ( stdout, "lost gem %d ; above = %d, playfield[above] = '%c'\n",
		//			slave_gem_id, slave_above, playfield [ slave_above ] ) ;
		slave_gem_id = -1 ;
	}
	
	if ( user_gem_id == -1 && slave_gem_id != -1 )
	{
		user_gem_id = slave_gem_id ;
		slave_gem_id = -1 ;
		gem [ user_gem_id ] -> Twinkling ( TRUE ) ;
		if ( user_gem_cursor )
			user_gem_cursor -> SetDestXY ( slave_col * GEM_WIDTH,
											( slave_row - 1 ) * GEM_HEIGHT ) ;
	}
}

void GamePF::MoveUserGemLeft ()
{
	bool user_slide_finishedx = user_gem_id != -1 ?
			gem [ user_gem_id ] -> SlideFinishedX () : true ;
	bool user_slide_finishedz = user_gem_id != -1 ?
			gem [ user_gem_id ] -> SlideFinishedZ () : true ;
	bool slave_slide_finishedx = slave_gem_id != -1 ?
			gem [ slave_gem_id ] -> SlideFinishedX () : true ;
	bool slave_slide_finishedz = slave_gem_id != -1 ?
			gem [ slave_gem_id ] -> SlideFinishedZ () : true ;
	
	if ( /*SlidingFinished () == FALSE ||*/ user_gem_id == -1 ||
			user_slide_finishedx == FALSE ||
			user_slide_finishedz == FALSE ||
			slave_slide_finishedx == FALSE ||
			slave_slide_finishedz == FALSE ) return ;
	
	int current_pos = gem_pos [ user_gem_id ] ;
	int left = GetRelPos ( current_pos, -1, 0 ) ;
	int col = current_pos % PLAYFIELD_MAX_X ;
	int row = ( current_pos - col ) / PLAYFIELD_MAX_X ;
	int slave_pos = gem_pos [ slave_gem_id ] ;
	int slave_left = GetRelPos ( slave_pos, -1, 0 ) ;
	int slave_col = slave_pos % PLAYFIELD_MAX_X ;
	int slave_row = ( slave_pos - slave_col ) / PLAYFIELD_MAX_X ;
	
	bool cur_clear = false, slave_clear = false ;
	
	if ( left != -1 && playfield [ left ] == GEM_NONE )
		cur_clear = true ;
	
	if ( slave_gem_id != -1 )
	{
		if ( slave_row < row && playfield [ slave_left ] == GEM_NONE &&
				playfield [ left ] == GEM_NONE )
			slave_clear = true ;
		if ( slave_col > col && playfield [ left ] == GEM_NONE )
			slave_clear = true ;
		if ( slave_row > row && playfield [ slave_left ] == GEM_NONE &&
				playfield [ left ] == GEM_NONE )
			slave_clear = true ;
		if ( slave_col < col && playfield [ slave_left ] == GEM_NONE )
			cur_clear = slave_clear = true ;
	}
	
	int jumps = 0 ;
	while ( cur_clear == false && left != -1 &&
			playfield [ left ] != GEM_NONE && col >= 0 )
	{
		col-- ;
		jumps++ ;
		left = GetRelPos ( ( row * PLAYFIELD_MAX_X ) + col, -1, 0 ) ;
	}
	if ( ( jumps && slave_gem_id != -1 && col >= 0 &&
			left != -1 && playfield [ left ] == GEM_NONE ) ||
			( slave_gem_id != -1 && slave_left != -1 && left != -1 &&
				playfield [ slave_left ] != GEM_NONE &&
				playfield [ left ] == GEM_NONE && slave_row != row ) )
	{	// free satellite
		gem [ slave_gem_id ] -> PlayStopSound () ;
		gem [ slave_gem_id ] -> Glow ( FALSE ) ;
		gem [ slave_gem_id ] -> Twinkling ( FALSE ) ;
		slave_gem_id = -1 ;
		
		slave_clear = false ;
	}

	if ( slave_clear && slave_col < col )
	{
		playfield [ slave_left ] = playfield [ slave_pos ] ;
		playfield [ slave_pos ] = GEM_NONE ;
		gem_pos [ slave_gem_id ] = slave_left ;
		
		gem [ slave_gem_id ] -> SetDestXY ( ( slave_col - 1 ) * GEM_WIDTH,
											slave_row * GEM_HEIGHT ) ;
		gem [ slave_gem_id ] -> Jump ( ( ( slave_pos % PLAYFIELD_MAX_X ) - slave_col + 1 ) * GEM_WIDTH ) ;
	}
	if ( ( cur_clear || ( left != -1 && playfield [ left ] == GEM_NONE ) ) &&
			col >= 0 )
	{
		playfield [ left ] = playfield [ current_pos ] ;
		playfield [ current_pos ] = GEM_NONE ;
		gem_pos [ user_gem_id ] = left ;
		
		gem [ user_gem_id ] -> SetDestXY ( ( col - 1 ) * GEM_WIDTH,
								row * GEM_HEIGHT ) ;
		gem [ user_gem_id ] -> Jump ( ( ( current_pos % PLAYFIELD_MAX_X ) - col + 1 ) * GEM_WIDTH ) ;
		if ( user_gem_cursor )
			user_gem_cursor -> SetDestXY ( ( col - 1 ) * GEM_WIDTH,
							row * GEM_HEIGHT ) ;
	}
	if ( slave_clear && slave_col >= col )
	{
		playfield [ slave_left ] = playfield [ slave_pos ] ;
		playfield [ slave_pos ] = GEM_NONE ;
		gem_pos [ slave_gem_id ] = slave_left ;
		
		gem [ slave_gem_id ] -> SetDestXY ( ( slave_col - 1 ) * GEM_WIDTH,
											slave_row * GEM_HEIGHT ) ;
		gem [ slave_gem_id ] -> Jump ( ( ( slave_pos % PLAYFIELD_MAX_X ) - slave_col + 1 ) * GEM_WIDTH ) ;
	}
}

void GamePF::MoveUserGemRight ()
{
	bool user_slide_finishedx = user_gem_id != -1 ?
			gem [ user_gem_id ] -> SlideFinishedX () : true ;
	bool user_slide_finishedz = user_gem_id != -1 ?
			gem [ user_gem_id ] -> SlideFinishedZ () : true ;
	bool slave_slide_finishedx = slave_gem_id != -1 ?
			gem [ slave_gem_id ] -> SlideFinishedX () : true ;
	bool slave_slide_finishedz = slave_gem_id != -1 ?
			gem [ slave_gem_id ] -> SlideFinishedZ () : true ;
	
	if ( /*SlidingFinished () == FALSE ||*/ user_gem_id == -1 ||
			user_slide_finishedx == FALSE ||
			user_slide_finishedz == FALSE ||
			slave_slide_finishedx == FALSE ||
			slave_slide_finishedz == FALSE ) return ;
	
	int current_pos = gem_pos [ user_gem_id ] ;
	int right = GetRelPos ( current_pos, 1, 0 ) ;
	int col = current_pos % PLAYFIELD_MAX_X ;
	int row = ( current_pos - col ) / PLAYFIELD_MAX_X ;
	int slave_pos = gem_pos [ slave_gem_id ] ;
	int slave_right = GetRelPos ( slave_pos, 1, 0 ) ;
	int slave_col = slave_pos % PLAYFIELD_MAX_X ;
	int slave_row = ( slave_pos - slave_col ) / PLAYFIELD_MAX_X ;
	
	bool cur_clear = false, slave_clear = false ;
	
	if ( right != -1 && playfield [ right ] == GEM_NONE )
		cur_clear = true ;
	
	if ( slave_gem_id != -1 )
	{
		if ( slave_row < row && playfield [ slave_right ] == GEM_NONE &&
				playfield [ right ] == GEM_NONE )
			slave_clear = true ;
		if ( slave_col > col && playfield [ slave_right ] == GEM_NONE )
			cur_clear = slave_clear = true ;
		if ( slave_row > row && playfield [ slave_right ] == GEM_NONE &&
				playfield [ right ] == GEM_NONE )
			slave_clear = true ;
		if ( slave_col < col && playfield [ right ] == GEM_NONE )
			slave_clear = true ;
	}
	
	int jumps = 0 ;
	while ( cur_clear == false && right != -1 &&
			playfield [ right ] != GEM_NONE && col < PLAYFIELD_MAX_X )
	{
		col++ ;
		jumps++ ;
		right = GetRelPos ( ( row * PLAYFIELD_MAX_X ) + col, 1, 0 ) ;
	}
	if ( ( jumps && slave_gem_id != -1 && col < PLAYFIELD_MAX_X &&
			right != -1 && playfield [ right ] == GEM_NONE ) ||
			( slave_gem_id != -1 && slave_right != -1 && right != -1 &&
				playfield [ slave_right ] != GEM_NONE &&
				playfield [ right ] == GEM_NONE && slave_row != row ) )
	{	// free satellite
		gem [ slave_gem_id ] -> PlayStopSound () ;
		gem [ slave_gem_id ] -> Glow ( FALSE ) ;
		gem [ slave_gem_id ] -> Twinkling ( FALSE ) ;
		slave_gem_id = -1 ;
		
		slave_clear = false ;
	}

	if ( slave_clear && slave_col > col )
	{
		playfield [ slave_right ] = playfield [ slave_pos ] ;
		playfield [ slave_pos ] = GEM_NONE ;
		gem_pos [ slave_gem_id ] = slave_right ;
		
		gem [ slave_gem_id ] -> SetDestXY ( ( slave_col + 1 ) * GEM_WIDTH,
											slave_row * GEM_HEIGHT ) ;
		gem [ slave_gem_id ] -> Jump ( ( slave_col + 1 - ( slave_pos % PLAYFIELD_MAX_X ) ) * GEM_WIDTH ) ;
	}
	if ( ( cur_clear || ( right != -1 && playfield [ right ] == GEM_NONE ) ) &&
			col < PLAYFIELD_MAX_X )
	{
		playfield [ right ] = playfield [ current_pos ] ;
		playfield [ current_pos ] = GEM_NONE ;
		gem_pos [ user_gem_id ] = right ;
		
		gem [ user_gem_id ] -> SetDestXY ( ( col + 1 ) * GEM_WIDTH,
								row * GEM_HEIGHT ) ;
		gem [ user_gem_id ] -> Jump ( (  col + 1 - ( current_pos % PLAYFIELD_MAX_X ) ) * GEM_WIDTH ) ;
		if ( user_gem_cursor )
			user_gem_cursor -> SetDestXY ( ( col + 1 ) * GEM_WIDTH,
							row * GEM_HEIGHT ) ;
	}
	if ( slave_clear && slave_col <= col )
	{
		playfield [ slave_right ] = playfield [ slave_pos ] ;
		playfield [ slave_pos ] = GEM_NONE ;
		gem_pos [ slave_gem_id ] = slave_right ;
		
		gem [ slave_gem_id ] -> SetDestXY ( ( slave_col + 1 ) * GEM_WIDTH,
											slave_row * GEM_HEIGHT ) ;
		gem [ slave_gem_id ] -> Jump ( ( slave_col + 1 - ( slave_pos % PLAYFIELD_MAX_X ) ) * GEM_WIDTH ) ;
	}
}

void GamePF::MoveUserGemUpStart ()
{
	move_up = TRUE ;
}

void GamePF::MoveUserGemLeftStart ()
{
	move_left = TRUE ;
}

void GamePF::MoveUserGemRightStart ()
{
	move_right = TRUE ;
}

void GamePF::MoveUserGemUpStop ()
{
	move_up = FALSE ;
}

void GamePF::MoveUserGemLeftStop ()
{
	move_left = FALSE ;
}

void GamePF::MoveUserGemRightStop ()
{
	move_right = FALSE ;
}

void GamePF::MoveUserGemToTop ( bool move_satellite )
{
	int cur_gem_id = user_gem_id ; // default
	
	if ( user_gem_id == -1 || ( move_satellite && slave_gem_id == -1 ) )
		return ;
	
	if ( move_satellite == false )
	{
		cur_gem_id = user_gem_id ;
		if ( user_gem_id == -1 || 
				gem [ user_gem_id ] -> SlideFinished () == FALSE ||
				( slave_gem_id != -1 && 
				  gem [ slave_gem_id ] -> SlideFinished () == FALSE ) )
			return ;
	}
	else
		cur_gem_id = slave_gem_id ;
	
	if ( cur_gem_id == -1 || 
			gem [ cur_gem_id ] -> SlideFinished () == FALSE ) return ;
	
	int current_pos = gem_pos [ cur_gem_id ] ;
	int above = GetRelPos ( current_pos, 0, -1 ) ;
	int col = current_pos % PLAYFIELD_MAX_X ;
	int row = ( current_pos - col ) / PLAYFIELD_MAX_X ;
	
	int cur_pos = gem_pos [ user_gem_id ] ;
	int cur_col = cur_pos % PLAYFIELD_MAX_X ;
	int cur_row = ( cur_pos - cur_col ) / PLAYFIELD_MAX_X ;
	
	int slave_pos = slave_gem_id != -1 ? gem_pos [ slave_gem_id ] : cur_pos ;
	int slave_col = slave_pos % PLAYFIELD_MAX_X ;
	int slave_row = ( slave_pos - slave_col ) / PLAYFIELD_MAX_X ;
	
	if ( move_satellite == false && slave_row < cur_row )
		MoveUserGemToTop ( true ) ;
	
	if ( row > 0 && above != -1 && playfield [ above ] == GEM_NONE ) 
	{
		bool not_finished = true ;
		while ( not_finished )
		{
			if ( GetRelPos ( above, 0, -1 ) != -1 &&
					playfield [ GetRelPos ( above, 0, -1 ) ] == GEM_NONE )
				above = GetRelPos ( above, 0, -1 ) ;
			else
				not_finished = false ;
		}
		
		ShootGemTo ( cur_gem_id, above ) ;
	}
	else if ( playfield [ above ] != GEM_HOLD )
	{
		gem [ cur_gem_id ] -> PlayStopSound () ;
	}
	gem [ cur_gem_id ] -> Glow ( FALSE ) ;
	gem [ cur_gem_id ] -> Twinkling ( FALSE ) ;
	//fprintf ( stdout, "throw gem %d ; above = %d, playfield[above] = '%c'\n",
	//			cur_gem_id, above, playfield [ above ] ) ;
	
	if ( move_satellite == false && slave_row >= cur_row )
		MoveUserGemToTop ( true ) ;
	
	if ( move_satellite == false )
		user_gem_id = -1 ;
	else
		slave_gem_id = -1 ;
}

void GamePF::RotateSlaveGem ()
{
	if ( user_gem_id == -1 || slave_gem_id == -1 ) return ;
	
	if ( gem [ user_gem_id ] -> SlideFinished () == FALSE ||
			gem [ slave_gem_id ] -> SlideFinished () == FALSE )
		return ;
	
	int current_pos = gem_pos [ user_gem_id ] ;
	int col = current_pos % PLAYFIELD_MAX_X ;
	int row = ( current_pos - col ) / PLAYFIELD_MAX_X ;
	int slave_pos = gem_pos [ slave_gem_id ] ;
	int slave_col = slave_pos % PLAYFIELD_MAX_X ;
	int slave_row = ( slave_pos - slave_col ) / PLAYFIELD_MAX_X ;
	
	int left   = GetRelPos ( current_pos, -1,  0 ) ;
	int right  = GetRelPos ( current_pos,  1,  0 ) ;
	int top    = GetRelPos ( current_pos,  0, -1 ) ;
	int bottom = GetRelPos ( current_pos,  0,  1 ) ;
	
	int jump_height = GEM_HEIGHT ;
	
	if ( left != -1 )
		left   = playfield [ left   ] != GEM_NONE ? -1 : left   ;
	if ( right != -1 )
		right  = playfield [ right  ] != GEM_NONE ? -1 : right  ;
	if ( top != -1 )
		top    = playfield [ top    ] != GEM_NONE ? -1 : top    ;
	if ( bottom != -1 )
		bottom = playfield [ bottom ] != GEM_NONE ? -1 : bottom ;
	
	int new_slave_pos = slave_pos ;
	
	/*fprintf ( stdout, "col = %d; row = %d; s_col = %d; s_row = %d; l=%d; r=%d; t=%d; b=%d\n",
				col, row, slave_col, slave_row,
				playfield [ left ], playfield [ right  ], 
				playfield [ top  ], playfield [ bottom ] ) ;*/

	if ( slave_row > row )
	{	// rotate slave from bottom to left
		if ( left != -1 )
			new_slave_pos = left ;
		else if ( top != -1 )
		{
			new_slave_pos = top ;
			jump_height *= 2 ;
		}
		else if ( right != -1 )
			new_slave_pos = right ;
	}
	else if ( slave_col < col )
	{	// rotate slave from left to top
		if ( top != -1 )
			new_slave_pos = top ;
		else if ( right != -1 )
		{
			new_slave_pos = right ;
			jump_height *= 2 ;
		}
		else if ( bottom != -1 )
			new_slave_pos = bottom ;
	}
	else if ( slave_row < row )
	{	// rotate from top to right
		if ( right != -1 )
			new_slave_pos = right ;
		else if ( bottom != -1 )
		{
			new_slave_pos = bottom ;
			jump_height *= 2 ;
		}
		else if ( left != -1 )
			new_slave_pos = left ;
	}
	else if ( slave_col > col )
	{	// rotate from right to bottom
		if ( bottom != -1 )
			new_slave_pos = bottom ;
		else if ( left != -1 )
		{
			new_slave_pos = left ;
			jump_height *= 2 ;
		}
		else if ( top != -1 )
			new_slave_pos = top ;
	}
	
	if ( slave_pos != new_slave_pos )
	{
		playfield [ new_slave_pos ] = playfield [ slave_pos ] ;
		playfield [ slave_pos ] = GEM_NONE ;
		
		int new_slave_col = new_slave_pos % PLAYFIELD_MAX_X ;
		int new_slave_row = ( new_slave_pos - new_slave_col ) / PLAYFIELD_MAX_X ;
		
		gem [ slave_gem_id ] -> SetDestXY ( new_slave_col * GEM_WIDTH,
											new_slave_row * GEM_HEIGHT ) ;
		gem [ slave_gem_id ] -> Jump ( jump_height ) ;
		
		gem_pos [ slave_gem_id ] = new_slave_pos ;
	}
	else
	{
		// SWITCH GEMS!!!
		char t = playfield [ current_pos ] ;
		playfield [ current_pos ] = playfield [ slave_pos ] ;
		playfield [ slave_pos ] = t ;
		gem [ user_gem_id ] -> SetDestXY ( slave_col * GEM_WIDTH,
							slave_row * GEM_HEIGHT ) ;
		gem [ slave_gem_id ] -> SetDestXY ( col * GEM_WIDTH,
							row * GEM_HEIGHT ) ;
		gem [ slave_gem_id ] -> Jump ( GEM_HEIGHT ) ;
		// switch gem pos
		int t_pos = gem_pos [ user_gem_id ] ;
		gem_pos [ user_gem_id ] = gem_pos [ slave_gem_id ] ;
		gem_pos [ slave_gem_id ] = t_pos ;
		// switch ids
		int t_id = user_gem_id ;
		user_gem_id = slave_gem_id ;
		slave_gem_id = t_id ;
	}
	
	//fprintf ( stdout, "old slave pos = %d ; new = %d\n", slave_pos, new_slave_pos ) ;
}

void GamePF::ShootGemTo ( int cur_gem_id, int pos, int velocity, int speed )
{
	int current_pos = gem_pos [ cur_gem_id ] ;
	int col = current_pos % PLAYFIELD_MAX_X ;
	int row = ( current_pos - col ) / PLAYFIELD_MAX_X ;

	int cur_y = ( current_pos - ( current_pos % PLAYFIELD_MAX_X ) ) / PLAYFIELD_MAX_X ;
	int new_y = ( pos - ( pos % PLAYFIELD_MAX_X ) ) / PLAYFIELD_MAX_X ;
	int y_diff = cur_y - new_y ;
	
	if ( velocity == AUTO )
		velocity = -y_diff ;
	
	if ( pos != current_pos )
	{
		playfield [ pos ] = playfield [ current_pos ] ;
		playfield [ current_pos ] = GEM_NONE ;
		gem_pos [ cur_gem_id ] = pos ;
	}
	
	GlDynArray < Sprite* > *bounce_list ;
	bounce_list = new GlDynArray < Sprite* > ;
	bounce_list -> Clear () ;
	pos = GetRelPos ( pos, 0, -1 ) ;
	//fprintf ( stdout, "found gem " ) ;
	int counter = 0 ;
	while ( pos != -1 && counter < PLAYFIELD_MAX_Y )
	{
		if ( pos != -1 && 
				playfield [ pos ] != GEM_NONE &&
				playfield [ pos ] != GEM_HOLD )
		{
			if ( gem_frozen [ GetMatchingGem ( pos ) ] == false )
			{
				int next_gem = GetMatchingGem ( pos ) ;
				bounce_list -> PushBack ( gem [ next_gem ] ) ;
			}
			
			//fprintf ( stdout, "%d ", next_gem ) ;
		}
		else
			counter = PLAYFIELD_MAX_Y ; // back out...
		
		pos = GetRelPos ( pos, 0, -1 ) ;
		
		counter++ ;
	}
	
	// this is too much without hardware acceleration
	// -> the whole playfield is moving up and down...
	if ( using_opengl || true ) // seems to work with new FPS :) (TODO)
	{
		// shake our playfield too!
		char basefield_name [ 256 ] ;
		sprintf ( basefield_name, "basefield_%s", playfield_side ) ;
		Pax::object_item *object = pax -> GetObjectFromList ( basefield_name ) ;
		if ( object )
		{
			Sprite *basefield_sprite = (Sprite*) object -> data ;
			if ( basefield_sprite ) bounce_list -> PushBack ( basefield_sprite ) ;
		}
	}
	
	//fprintf ( stdout, "\n" ) ;
	
	//gem [ cur_gem_id ] -> SetXAccelFactor ( speed ) ;
	gem [ cur_gem_id ] -> SetYAccelFactor ( speed ) ;
	gem [ cur_gem_id ] -> SetDestXY ( col * GEM_WIDTH,
							( row - y_diff ) * GEM_HEIGHT ) ;
	gem [ cur_gem_id ] -> SetBounce ( 0, velocity ) ;
	gem [ cur_gem_id ] -> SetBounceList ( bounce_list ) ;
	gem [ cur_gem_id ] -> PlayAccelSound () ;
}

int GamePF::AddFallingGems ( char gem_color, int count )
{
	if ( /*playfield_ready != true ||*/ playfield_set != true ) return 0 ;
	
	int gem_count = 0 ; // count gems which are done
	
	if ( count == -1 )
	{
		for ( int y = PLAYFIELD_MAX_Y - 2 ; y < PLAYFIELD_MAX_Y ; y++ )
		{
			for ( int x = 0 ; x < PLAYFIELD_MAX_X ; x++ )
			{
				int x_copy = x ;
				int row = y * PLAYFIELD_MAX_X ;
				if ( falling_gems_enabled == FALSE )
				{
					x_copy = rand () % ( PLAYFIELD_MAX_X - 1 ) ;
					row = ( rand () % ( PLAYFIELD_MAX_Y - 1 ) ) * PLAYFIELD_MAX_X ;
				}
				if ( playfield [ row + x_copy ] == GEM_NONE )
				{
					int color = gem_color ; //( rand() % gem_colors_activated ) + 'A' ;
					playfield [ row + x_copy ] = color ;
					ThrowInGem ( row + x_copy ) ;
					
					// now check if it worked
					int new_gem = GetMatchingGem ( row + x_copy ) ;
					if ( new_gem == -1 || gem_active [ new_gem ] == FALSE )
					{	// failed
						playfield [ row + x_copy ] = GEM_NONE ;
						//fprintf ( stdout, "remove block-stone from %dx%d ( %d )\n",
						//			x_copy, y, row +x_copy ) ;
					}
					else
					{	// success
						gem_count++ ;
					}
				}
			}
		}
	}
	else
	{
		//fprintf ( stdout, "Adding %d gem(s) with color %d from bottom...\n", count, gem_color ) ;
		
		int act_count = 0 ;
		
		while ( act_count < count )
		{
			act_count++ ;
			
			int pos = GetEmptyPosAtBottom () ;
			if ( pos != -1 )
			{
//#define OLD_SHOOTIN
#ifdef  OLD_SHOOTIN
				if ( gem_color == GEM_BLOCK )
				{
					while ( GetRelPos ( pos, 0, -1 ) != -1 &&
							playfield [ GetRelPos ( pos, 0, -1 ) ] == GEM_NONE )
						pos = GetRelPos ( pos, 0, -1 ) ;
				}
#endif
				
				if ( playfield [ pos ] != GEM_NONE )
					fprintf ( stdout, "WARNING: playfield [ %d ] = '%c'\n",
								pos, playfield [ pos ] ) ;
				
				playfield [ pos ] = gem_color ;
				ThrowInGem ( pos, false ) ;
				
				// now check if it worked
				int new_gem = GetMatchingGem ( pos ) ;
				if ( new_gem == -1 || gem_active [ new_gem ] == FALSE )
				{	// failed
					playfield [ pos ] = GEM_NONE ;
				}
				else
				{	// success
					gem_count++ ;
					
					int item = gem_color - GEM_FIRST_COLOR ;
					if ( gem_color == GEM_BLOCK )
						item = GEM_COUNT ;
					
					if ( item < 0 || item > GEM_COUNT )
						fprintf ( stdout, "gem_color = %d ; GEM_FIRST_COLOR = %d ; item = %d\n",
									gem_color, GEM_FIRST_COLOR, item ) ;
					
#ifdef OLD_SHOOTIN
					Sprite *gip = gems_incoming_preview [ item ] ;
					int gip_x = gem [ new_gem ] -> ConvertXPos ( incomingfield, gip -> GetXPos () ) ;
					int gip_y = gem [ new_gem ] -> ConvertYPos ( incomingfield, gip -> GetYPos () ) ;
#endif
					int dst_x = gem [ new_gem ] -> GetXPos () ;
					int dst_y = gem [ new_gem ] -> GetYPos () ;
#ifdef OLD_SHOOTIN
					int x = gip_x - dst_x ;
					int y = gip_y - dst_y ;
					
					int t_x = x > 0 ? x : -x ;
					int t_y = y > 0 ? y : -y ;
					
					gem [ new_gem ] -> SlideInFrom ( x, y ) ;
					gem [ new_gem ] -> Jump ( max ( t_x, t_y ) ) ;
#endif
					gem [ new_gem ] -> SetAlpha ( 255 ) ;
					gem [ new_gem ] -> Draw () ;
					
#ifndef OLD_SHOOTIN
					pos = gem_pos [ new_gem ] ;
					
					while ( GetRelPos ( pos, 0, -1 ) != -1 &&
							playfield [ GetRelPos ( pos, 0, -1 ) ] == GEM_NONE )
						pos = GetRelPos ( pos, 0, -1 ) ;
					
					gem [ new_gem ] -> SetXY ( dst_x, PLAYFIELD_MAX_Y * GEM_HEIGHT + dst_y ) ;
					int y_pos = ( pos - ( pos % PLAYFIELD_MAX_X ) ) / PLAYFIELD_MAX_X ;
					ShootGemTo ( new_gem, pos, y_pos, 4 ) ;
#endif
				}
			}
		}
	}
	
	action_indicator = action_step_two ;
	
	return gem_count ;
}

void GamePF::Pause ()
{
	paused = TRUE ;
}

void GamePF::UnPause ()
{
	paused = FALSE ;
}

bool GamePF::PlayfieldLayoutIsSet ()
{
	return pull_row == PLAYFIELD_MAX_Y ;
}

bool GamePF::PlayfieldIsSet ()
{
	return playfield_set ;
}

void GamePF::SetPlayfieldIsReady ( bool choice )
{
	playfield_ready = choice ;
}

void GamePF::DumpStatus ()
{
	fprintf ( stdout, "Dumping playfield %s...\n", level ) ;
	for ( int y = 0 ; y < PLAYFIELD_MAX_Y ; y++ )
	{
		for ( int x = 0 ; x < PLAYFIELD_MAX_X ; x++ )
		{
			int current = ( y * PLAYFIELD_MAX_X ) + x ;
			fprintf ( stdout, "'%c' ", playfield [ current ] ) ;
		}
		fprintf ( stdout, "        " ) ;
		for ( int x = 0 ; x < PLAYFIELD_MAX_X ; x++ )
		{
			int current = ( y * PLAYFIELD_MAX_X ) + x ;
			fprintf ( stdout, "'%03d' ", GetMatchingGem ( current ) ) ;
		}
		fprintf ( stdout, "\n" ) ;
	}
	fprintf ( stdout, "Dumping gem status... (active/pos)\n" ) ;
	for ( int i = 0 ; i < PLAYFIELD_SIZE ; i++ )
	{
		int is_preview_gem = false ;
		fprintf ( stdout, "('%d'/", gem_active [ i ] ) ;
		for ( int j = 0 ; j < preview_gems ; j++ )
			if ( i == preview_gem_id [ j ] )
			{
				fprintf ( stdout, "'P%d') ", j ) ;
				is_preview_gem = true ;
			}
		if ( is_preview_gem == false )
			fprintf ( stdout, "'%03d') ", gem_pos [ i ] ) ;
		if ( (i+1) % PLAYFIELD_MAX_X == 0 )
			fprintf ( stdout, "\n" ) ;
	}
	fprintf ( stdout, "Dumping freeze status of playfield...\n" ) ;
	for ( int i = 0 ; i < PLAYFIELD_SIZE ; i++ )
	{
		if ( i > 0 && ( i % PLAYFIELD_MAX_X == 0 ) )
			fprintf ( stdout, "\n" ) ;
		fprintf ( stdout, "%d ", gem_frozen [ GetMatchingGem ( i ) ] == true ? 1 : 0 ) ;
	}
	fprintf ( stdout, "\n" ) ;
	fprintf ( stdout, "Dumping preview area...\n" ) ;
	for ( int i = 0 ; i < preview_gems ; i++ )
	{
		fprintf ( stdout, "%d. preview gem id %d has colour '%c' (%d)\n",
					i, preview_gem_id [ i ], preview_gem_color [ i ],
					preview_gem_color [ i ] ) ;
	}
	fprintf ( stdout, "user_gem_id is %d ; slave_gem_id is %d\n",
				user_gem_id, slave_gem_id ) ;
	fprintf ( stdout, "done.\n" ) ;
}

bool GamePF::GetIsQuitting ()
{
	return quit ;
}

bool GamePF::GetIsDone ()
{
	return quit_state ;
}

bool GamePF::GetIsFreed ()
{
	return already_freed ;
}

void GamePF::AutoLoadState ()
{
	if ( strcmp ( level, "autosave" ) == 0 )
	{
		FILE *stream ;
		char  autosave_fname [ 256 ] ;
		int t_user_gem_id = -1, t_slave_gem_id ;
		char t_preview_gem_color [ preview_gems + 1 ] ;
		bool is_frozen [ PLAYFIELD_SIZE ] ;
		for ( int i = 0 ; i < preview_gems ; i++ )
			t_preview_gem_color [ i ] = GEM_NONE ;
		
		sprintf ( autosave_fname, "%s/autosave.state.%s", rc_dir, playfield_side ) ;
		stream = fopen ( autosave_fname, "r" ) ;
		if ( stream )
		{
			int i = 0 ;
			fscanf ( stream, "%d", &i ) ;
			if ( i != preview_gems )
				fprintf ( stdout, "Warning! Preview gems amount in savestate doesn't match actual amount!\n" ) ;
			
			fscanf ( stream, "%s", t_preview_gem_color ) ;
			for ( int j = 0 ; j < i ; j++ )
				if ( t_preview_gem_color [ j ] == '#' ||
						t_preview_gem_color [ j ] < GEM_FIRST_COLOR ||
						t_preview_gem_color [ j ] > GEM_LAST_COLOR )
					t_preview_gem_color [ j ] = GEM_NONE ;
			
			fscanf ( stream, "%d", &t_user_gem_id ) ;
			fscanf ( stream, "%d", &t_slave_gem_id ) ;
			fscanf ( stream, "%d", &i ) ; falling_gems_enabled = i ;
			fscanf ( stream, "%d", &gem_colors_activated ) ;
			fscanf ( stream, "%d", &auto_play ) ;
			fscanf ( stream, "%d", &playerscore_combo ) ;
			fscanf ( stream, "%d", &i ) ;
			fscanf ( stream, "%d", &i ) ;
			fscanf ( stream, "%d", &playerscore_dest ) ;
			for ( int i = 0 ; i < GEM_COUNT + 1 ; i++ )
				fscanf ( stream, "%d", &gems_incoming [ i ] ) ;
			for ( int i = 0 ; i < GEM_COUNT + 1 ; i++ )
				fscanf ( stream, "%d", &gems_outgoing [ i ] ) ;
			for ( int i = 0 ; i < GEM_COUNT + 1 ; i++ )
				fscanf ( stream, "%d", &gems_incoming_timer [ i ] ) ;
			for ( int i = 0 ; i < PLAYFIELD_SIZE ; i++ )
			{
				int result = 0 ;
				fscanf ( stream, "%d", &result ) ;
				is_frozen [ i ] = result == 1 ? true : false ;
			}
			
			/*for ( int i = 0 ; i < GEM_COUNT + 1 ; i++ )
				fprintf ( stdout, "gi [ %d ] = %d ; go [ %d ] = %d ; git [ %d ] = %d\n",
							i, gems_incoming [ i ], i, gems_outgoing [ i ],
							i, gems_incoming_timer [ i ] ) ;
			*/
			fclose ( stream ) ;
		}
		
		freeze_gems_at_start = false ;
		
		free ( level ) ;
		level = (char*) malloc ( sizeof ( char ) * 4096 ) ;
		sprintf ( level, "%s/autosave.level.%s", rc_dir, playfield_side ) ;
		
		SetupPlayfield () ;

		for ( int i = 0 ; i < PLAYFIELD_SIZE ; i++ )
			if ( is_frozen [ i ] && GetMatchingGem ( i ) != -1 )
				FreezeGem ( GetMatchingGem ( i ) ) ;
		
		sprintf ( preview_gem_color, "%s", t_preview_gem_color ) ;
		
		user_gem_id = t_user_gem_id ;
		slave_gem_id = t_slave_gem_id ;
		
		//fprintf ( stdout, "gem pos %d has now", user_gem_id ) ;
		user_gem_id = GetMatchingGem ( user_gem_id ) ;
		if ( user_gem_id != -1 )
		{
			gem [ user_gem_id ] -> Glow ( TRUE ) ;
			gem [ user_gem_id ] -> Twinkling ( TRUE ) ;
			CreateUserGemCursorAtPos ( gem_pos [ user_gem_id ] ) ;
		}
		//fprintf ( stdout, " gem id %d\n", user_gem_id ) ;
		
		slave_gem_id = GetMatchingGem ( slave_gem_id ) ;
		if ( slave_gem_id != -1 )
		{
			gem [ slave_gem_id ] -> Glow ( TRUE ) ;
			gem [ slave_gem_id ] -> Twinkling ( TRUE ) ;
		}

		HighlightEqualGems () ;
		
		UpdateGemPreview () ;
		
		sprintf ( level, "random" ) ;
		playercombobox -> SetTextChar ( " ", 0 ) ;
	}
	else
		SetupPlayfield () ;
}

void GamePF::AutoSaveState ()
{
	FILE *stream ;
	char  autosave_fname [ 256 ] ;
	
	sprintf ( autosave_fname, "%s/autosave.state.%s", rc_dir, playfield_side ) ;
	stream = fopen ( autosave_fname, "w" ) ;
	if ( stream )
	{
		fprintf ( stream, "%d\n", preview_gems ) ;
		
		char tmp_color [ preview_gems + 1 ] ;
		sprintf ( tmp_color, "%s", preview_gem_color ) ;
		for ( int i = 0 ; i < preview_gems ; i++ )
			if ( tmp_color [ i ] == GEM_NONE )
				tmp_color [ i ] = '#' ;
		fprintf ( stream, "%s\n", tmp_color ) ;
		
		int t_user_gem_id = user_gem_id != -1 ? gem_pos [ user_gem_id ] : -1 ;
		fprintf ( stream, "%d\n", t_user_gem_id ) ;
		int t_slave_gem_id = slave_gem_id != -1 ? gem_pos [ slave_gem_id ] : -1 ;
		fprintf ( stream, "%d\n", t_slave_gem_id ) ;
		fprintf ( stream, "%d\n", falling_gems_enabled ) ;
		fprintf ( stream, "%d\n", gem_colors_activated ) ;
		fprintf ( stream, "%d\n", auto_play ) ;
		fprintf ( stream, "%d\n", playerscore_combo ) ;
		fprintf ( stream, "%d\n", playerscore_combo_rest_seconds ) ;
		fprintf ( stream, "%d\n", playerscore_combo_max_rest_seconds ) ;
		fprintf ( stream, "%d\n", playerscore_dest ) ;
		for ( int i = 0 ; i < GEM_COUNT + 1 ; i++ )
			fprintf ( stream, "%d ", gems_incoming [ i ] ) ;
		fprintf ( stream, "\n" ) ;
		for ( int i = 0 ; i < GEM_COUNT + 1 ; i++ )
			fprintf ( stream, "%d ", gems_outgoing [ i ] ) ;
		fprintf ( stream, "\n" ) ;
		for ( int i = 0 ; i < GEM_COUNT + 1 ; i++ )
			fprintf ( stream, "%d ", gems_incoming_timer [ i ] ) ;
		fprintf ( stream, "\n" ) ;
		for ( int i = 0 ; i < PLAYFIELD_SIZE ; i++ )
		{
			bool is_frozen = false ;
			// translate gem id into matrix pos
			int  match = GetMatchingGem ( i ) ;
			if ( match != -1 )
				is_frozen = gem_frozen [ match ] ;
				
			if ( i > 0 && ( i % PLAYFIELD_MAX_X == 0 ) )
				fprintf ( stream, "\n" ) ;
			fprintf ( stream, "%d ", is_frozen ? 1 : 0 ) ;
		}
		fprintf ( stream, "\n" ) ;
		
		fclose ( stream ) ;
	}
	
	sprintf ( autosave_fname, "%s/autosave.level.%s", rc_dir, playfield_side ) ;
	stream = fopen ( autosave_fname, "w" ) ;
	if ( stream )
	{
		for ( int y = 0 ; y < PLAYFIELD_MAX_Y ; y++ )
		{
			for ( int x = 0 ; x < PLAYFIELD_MAX_X ; x++ )
				fprintf ( stream, "%c", playfield [ y * PLAYFIELD_MAX_X + x ] ) ;
			fprintf ( stream, "\n" ) ;
		}
		
		fclose ( stream ) ;
	}
}

void GamePF::Quit ()
{
	Ai.Quit () ;
	
	AutoSaveState () ;
	
	PlayerScoreReset () ;

	DestroyPlayfield () ;

	quit = TRUE ;
	
	//fprintf ( stdout, "quitting playfield...\n" ) ;
}

void GamePF::Free ()
{
	if ( quit == false )
	{
		Quit () ;
		return ;
	}
	if ( quit_state == false )
		return ;
	
	if ( already_freed )
	{
		fprintf ( stdout, "%s playfield already freed.\n", playfield_side ) ;
		return ;
	}
	
	fprintf ( stdout, "freeing %s playfield data...\n", playfield_side	) ;
	
	GemQueueQuit () ;
	
	if ( user_gem_cursor )
	{
		user_gem_cursor -> Free () ;
		delete user_gem_cursor ;
		user_gem_cursor = NULL ;
	}
	
	if ( slave_gem_cursor )
	{
		slave_gem_cursor -> Free () ;
		delete slave_gem_cursor ;
		slave_gem_cursor = NULL ;
	}
	
	for ( int item = 0 ; item < PLAYFIELD_SIZE ; item++ )
	{
		if ( gem_active [ item ] == true )
		{
			fprintf ( stdout, "gem %d is still active! trying to free it...\n", item ) ;
			gem [ item ] -> Free () ;
			delete gem [ item ] ;
		}
		if ( gem_frozen_layer [ item ] != NULL )
		{
			gem_frozen_layer [ item ] -> Free () ;
			delete gem_frozen_layer [ item ] ;
		}
	}
	free ( gem ) ;
	free ( gem_active ) ;
	free ( gem_pos ) ;
	free ( gem_frozen ) ;
	free ( gem_frozen_layer ) ;
	
	earthquake_sound.UnloadSound () ;
	pre_earthquake_sound.UnloadSound () ;
	playerscore_inc_sound.UnloadSound () ;
	playerscore_dec_sound.UnloadSound () ;
	playerscore_bonus_sound.UnloadSound () ;
	playerlose_sound.UnloadSound () ;
	
	free ( preview_gem_id ) ;
	free ( preview_gem_color ) ;
	
	free ( gems_incoming ) ;
	free ( gems_outgoing ) ;
	free ( gems_incoming_timer ) ;
	for ( int item = 0 ; item < GEM_COUNT ; item++ )
	{
		if ( gems_incoming_preview [ item ] != NULL )
		{
			gems_incoming_preview [ item ] -> Free () ;
			delete gems_incoming_preview [ item ] ;
		}
	}
	free ( gems_incoming_preview ) ;
	
	friend_bucket -> Clear () ; delete friend_bucket ;
	
	free ( playfield ) ;
	fprintf ( stdout, "done.\n" ) ;
	
	already_freed = true ;
}
