#include "gamescreen.h"

extern char pkg_filename[] ;

int last_game_screen = GS_NONE ;
int current_game_screen = GS_NONE ;
int next_game_screen = GS_NONE ;

#include "video.h"
extern Video video ;

PaxList *pax_list = new PaxList () ;

#include "splash.h"
Splash *splash = NULL ;

void GameScreen::Init ()
{
	// say which screen do we want next...
	if ( strcmp ( pkg_filename, "" ) == 0 || strcmp ( pkg_filename, "title" ) == 0 )
		next_game_screen = GS_TITLE ;
#ifdef __NETWORK_SUPPORT__
	else if ( strcmp ( pkg_filename, "network" ) == 0 )
		next_game_screen = GS_NETWORK ;
#endif
	else
		next_game_screen = GS_GAME ;
	
	quit = FALSE ;
	quit_state = FALSE ;
	
	pax_list -> Init () ;
	
	is_done = FALSE ;

	initialized = false ;
}

void GameScreen::ProcessEvents ( SDL_Event event )
{
	if ( is_done || ! initialized ) return ;
	
	switch ( current_game_screen )
	{
		case GS_TITLE: titleloop.ProcessEvents ( event ) ; break ;
		case GS_GAME: gameloop.ProcessEvents ( event ) ; break ;
#ifdef __NETWORK_SUPPORT__
		case GS_NETWORK: netloop.ProcessEvents ( event ) ; break ;
#endif
		default: break ;
	}
}

void GameScreen::Pause ()
{
	if ( is_done || IsQuitting () || ! initialized ) return ;
	
	switch ( current_game_screen )
	{
		case GS_TITLE: titleloop.Pause () ; break ;
		case GS_GAME: gameloop.Pause () ; break ;
#ifdef __NETWORK_SUPPORT__
		case GS_NETWORK: netloop.Pause () ; break ;
#endif
		default: break ;
	}
}

void GameScreen::UnPause ()
{
	if ( is_done || ! initialized ) return ;
	
	switch ( current_game_screen )
	{
		case GS_TITLE: titleloop.UnPause () ; break ;
		case GS_GAME: gameloop.UnPause () ; break ;
#ifdef __NETWORK_SUPPORT__
		case GS_NETWORK: netloop.UnPause () ; break ;
#endif
		default: break ;
	}
}

void GameScreen::Draw ()
{
	if ( is_done || ! initialized ) return ;
	
	switch ( current_game_screen )
	{
		case GS_TITLE: titleloop.Draw () ; break ;
		case GS_GAME: gameloop.Draw () ; break ;
#ifdef __NETWORK_SUPPORT__
		case GS_NETWORK: netloop.Draw () ; break ;
#endif
		default: break ;
	}
}

void GameScreen::AfterDraw ()
{
	if ( ! is_done && initialized )
	{
		switch ( current_game_screen )
		{
			case GS_TITLE: titleloop.AfterDraw () ; break ;
			case GS_GAME: gameloop.AfterDraw () ; break ;
#ifdef __NETWORK_SUPPORT__
			case GS_NETWORK: netloop.AfterDraw () ; break ;
#endif
			default: break ;
		}
	}

	if ( ( current_game_screen == GS_NONE || is_done == TRUE ) &&
		quit_state == FALSE && ! initialized )
	{
		last_game_screen = current_game_screen ;
		current_game_screen = next_game_screen ;
		
		//sprintf ( left_levelfname , "levels/level%d.dat", rand() % 5 ) ;
		//sprintf ( right_levelfname, "levels/level%d.dat", rand() % 5 ) ;
		sprintf ( left_levelfname , "levels/level%d.dat", 1 ) ;
		sprintf ( right_levelfname, "levels/level%d.dat", 1 ) ;
		
		video . SetClearBackground ( true ) ;
		
		switch ( next_game_screen )
		{
			case GS_TITLE:
			{
				titleloop.Init ( left_levelfname, right_levelfname ) ;
				// if the titleloop won't decide use this... (quit)
				next_game_screen = GS_NONE ;
			}
			break ;
			
			case GS_GAME:
			{
				splash = new Splash () ;
				splash -> Init ( pkg_filename ) ;
				
				gameloop.Init ( pkg_filename, left_levelfname,
								right_levelfname, 0, 0 ) ;
				next_game_screen = GS_TITLE ;
				
				splash -> Free () ;
				delete splash ;
				splash = NULL ;
			}
			break ;
			
#ifdef __NETWORK_SUPPORT__
			case GS_NETWORK:
			{
				netloop.Init () ;
				next_game_screen = GS_TITLE ;
			}
			break ;
#endif
			
			// no following screens left? seems like we are done...
			case GS_NONE:
			break ;

			default:
			break ;
		}
		video . SetClearBackground ( false ) ;
		initialized = true ;
	}
	
	switch ( current_game_screen )
	{
		case GS_TITLE: is_done = titleloop.IsDone () ; break ;
		case GS_GAME: is_done = gameloop.IsDone () ; break ;
#ifdef __NETWORK_SUPPORT__
		case GS_NETWORK: is_done = netloop.IsDone () ; break ;
#endif
		default: break ;
	}

	if ( is_done )
		initialized = false ;
	
	// indicate that we are done if this is the case...
	if ( current_game_screen == GS_NONE && next_game_screen == GS_NONE )
	{
		if ( pax_list )
		{
			pax_list -> Free () ;
			delete pax_list ;
			pax_list = NULL ;
		}
		
		quit_state = TRUE ;
	}
}

bool GameScreen::IsQuitting ()
{
	switch ( current_game_screen )
	{
		case GS_TITLE: return titleloop.IsQuitting () ; break ;
		case GS_GAME: return gameloop.IsQuitting () ; break ;
#ifdef __NETWORK_SUPPORT__
		case GS_NETWORK: return netloop.IsQuitting () ; break ;
#endif
		default: return true ; break ;
	}
}

bool GameScreen::IsDone ()
{
	return quit_state ;
}

void GameScreen::Quit ()
{
	video . SetClearBackground ( true ) ;
	
	switch ( current_game_screen )
	{
		case GS_TITLE: titleloop.Quit () ; break ;
		case GS_GAME: gameloop.Quit () ; break ;
#ifdef __NETWORK_SUPPORT__
		case GS_NETWORK: netloop.Quit () ; break ;
#endif
		default: break ;
	}
}
