#include "joystick.h"

void Joystick::Init ()
{
	int first = 0 ;
	int second = 1 ;
	
	left = NULL ; right = NULL ;

	SDL_InitSubSystem ( SDL_INIT_JOYSTICK ) ;
	
	SetupDefaultMapping () ;
	
	ResetToDefaultsLeft () ;
	ResetToDefaultsRight () ;
	
	fprintf ( stdout, "found %d joysticks.\n", GetNumJoysticks () ) ;
	
	if ( GetNumJoysticks () == 0 )
	{
		fprintf ( stdout, "\n" ) ;
		return ;
	}

	left = SDL_JoystickOpen ( first ) ;
	if ( left == NULL )
		fprintf ( stdout,"couldn't open %d. joystick.\n", first + 1) ;
	else
		Inspect ( left ) ;

	right = SDL_JoystickOpen ( second ) ;
	if ( right == NULL )
		fprintf ( stdout,"couldn't open %d. joystick.\n", second + 1 ) ;
	else
		Inspect ( right ) ;

	fprintf ( stdout, "\n" ) ;
}

// our ultimate default mapping - preset for gamepads

void Joystick::SetupDefaultMapping ()
{
	default_mapping [ JOY_UP    ] . type  = JOY_AXIS    ;
	default_mapping [ JOY_UP    ] . num   = 1           ;
	default_mapping [ JOY_UP    ] . value = -32767      ;
	default_mapping [ JOY_DOWN  ] . type  = JOY_AXIS    ;
	default_mapping [ JOY_DOWN  ] . num   = 1           ;
	default_mapping [ JOY_DOWN  ] . value = 32767       ;
	default_mapping [ JOY_LEFT  ] . type  = JOY_AXIS    ;
	default_mapping [ JOY_LEFT  ] . num   = 0           ;
	default_mapping [ JOY_LEFT  ] . value = -32767      ;
	default_mapping [ JOY_RIGHT ] . type  = JOY_AXIS    ;
	default_mapping [ JOY_RIGHT ] . num   = 0           ;
	default_mapping [ JOY_RIGHT ] . value = 32767       ;
	default_mapping [ JOY_FIRE  ] . type  = JOY_BUTTON  ;
	default_mapping [ JOY_FIRE  ] . num   = 0           ;
	default_mapping [ JOY_FIRE  ] . value = SDL_PRESSED ;
}

void Joystick::ResetToDefaultsLeft ()
{
	left_mapping [ JOY_UP    ] . type  = default_mapping [ JOY_UP    ] . type  ;
	left_mapping [ JOY_UP    ] . num   = default_mapping [ JOY_UP    ] . num   ;
	left_mapping [ JOY_UP    ] . value = default_mapping [ JOY_UP    ] . value ;
	left_mapping [ JOY_DOWN  ] . type  = default_mapping [ JOY_DOWN  ] . type  ;
	left_mapping [ JOY_DOWN  ] . num   = default_mapping [ JOY_DOWN  ] . num   ;
	left_mapping [ JOY_DOWN  ] . value = default_mapping [ JOY_DOWN  ] . value ;
	left_mapping [ JOY_LEFT  ] . type  = default_mapping [ JOY_LEFT  ] . type  ;
	left_mapping [ JOY_LEFT  ] . num   = default_mapping [ JOY_LEFT  ] . num   ;
	left_mapping [ JOY_LEFT  ] . value = default_mapping [ JOY_LEFT  ] . value ;
	left_mapping [ JOY_RIGHT ] . type  = default_mapping [ JOY_RIGHT ] . type  ;
	left_mapping [ JOY_RIGHT ] . num   = default_mapping [ JOY_RIGHT ] . num   ;
	left_mapping [ JOY_RIGHT ] . value = default_mapping [ JOY_RIGHT ] . value ;
	left_mapping [ JOY_FIRE  ] . type  = default_mapping [ JOY_FIRE  ] . type  ;
	left_mapping [ JOY_FIRE  ] . num   = default_mapping [ JOY_FIRE  ] . num   ;
	left_mapping [ JOY_FIRE  ] . value = default_mapping [ JOY_FIRE  ] . value ;
}

void Joystick::ResetToDefaultsRight ()
{
	right_mapping [ JOY_UP    ] . type  = default_mapping [ JOY_UP    ] . type  ;
	right_mapping [ JOY_UP    ] . num   = default_mapping [ JOY_UP    ] . num   ;
	right_mapping [ JOY_UP    ] . value = default_mapping [ JOY_UP    ] . value ;
	right_mapping [ JOY_DOWN  ] . type  = default_mapping [ JOY_DOWN  ] . type  ;
	right_mapping [ JOY_DOWN  ] . num   = default_mapping [ JOY_DOWN  ] . num   ;
	right_mapping [ JOY_DOWN  ] . value = default_mapping [ JOY_DOWN  ] . value ;
	right_mapping [ JOY_LEFT  ] . type  = default_mapping [ JOY_LEFT  ] . type  ;
	right_mapping [ JOY_LEFT  ] . num   = default_mapping [ JOY_LEFT  ] . num   ;
	right_mapping [ JOY_LEFT  ] . value = default_mapping [ JOY_LEFT  ] . value ;
	right_mapping [ JOY_RIGHT ] . type  = default_mapping [ JOY_RIGHT ] . type  ;
	right_mapping [ JOY_RIGHT ] . num   = default_mapping [ JOY_RIGHT ] . num   ;
	right_mapping [ JOY_RIGHT ] . value = default_mapping [ JOY_RIGHT ] . value ;
	right_mapping [ JOY_FIRE  ] . type  = default_mapping [ JOY_FIRE  ] . type  ;
	right_mapping [ JOY_FIRE  ] . num   = default_mapping [ JOY_FIRE  ] . num   ;
	right_mapping [ JOY_FIRE  ] . value = default_mapping [ JOY_FIRE  ] . value ;
}

// don't ask what the following methods do, ok? =)

int Joystick::GetNumJoysticks ()
{
	return SDL_NumJoysticks () ;
}

int Joystick::GetIndex ( SDL_Joystick *joystick )
{
	return SDL_JoystickIndex ( joystick ) ;
}

char *Joystick::GetName ( SDL_Joystick *joystick )
{
	return strdup ( SDL_JoystickName ( GetIndex ( joystick ) ) ) ;
}

int Joystick::GetNumAxes ( SDL_Joystick *joystick )
{
	return SDL_JoystickNumAxes ( joystick ) ;
}

int Joystick::GetNumBalls ( SDL_Joystick *joystick )
{
	return SDL_JoystickNumBalls ( joystick ) ;
}

int Joystick::GetNumHats ( SDL_Joystick *joystick )
{
	return SDL_JoystickNumHats ( joystick ) ;
}

int Joystick::GetNumButtons ( SDL_Joystick *joystick )
{
	return SDL_JoystickNumButtons ( joystick ) ;
}

// what do we have here?! :)

void Joystick::Inspect ( SDL_Joystick *joystick )
{
	fprintf ( stdout, "%d. joystick called '%s' has\n   ", 
				GetIndex ( joystick ) + 1, GetName ( joystick ) ) ;
	fprintf ( stdout, "%d axes, ", GetNumAxes ( joystick ) ) ;
	fprintf ( stdout, "%d balls, ", GetNumBalls ( joystick ) ) ;
	fprintf ( stdout, "%d hats and ", GetNumHats ( joystick ) ) ;
	fprintf ( stdout, "%d buttons\n", GetNumButtons ( joystick ) ) ;
}

// test if a player presses something

int Joystick::Test ( SDL_Joystick *joystick, JoyMap mapping_struct [ JOY_NUM ],
						Uint8 key, SDL_Event event )
{
	// get current mapping
	JoyMap mapping = mapping_struct [ key ] ;
	// this will be used to convert negative vars to positive if necessary...
	int reversal = mapping . value < 0 ? -1 : 1 ;
	
	//fprintf ( stdout, "joy %d: axis = %d, value = %d\n",
	//			GetIndex ( joystick ), event.jaxis.axis, event.jaxis.value ) ;
	
	// SDL_JOYAXISMOTION
	if ( mapping . type == JOY_AXIS &&
			event . jaxis . which == GetIndex ( joystick ) && 
			event . jaxis . axis == mapping . num )
		return ( event . jaxis . value * reversal >= mapping . value * reversal ) ;
	
	// SDL_JOYBALLMOTION
	if ( mapping . type == JOY_BALL &&
			event . jball . which == GetIndex ( joystick) && 
			event . jball . ball == mapping . num )
	{
		if ( key == JOY_UP || key == JOY_DOWN )
			return ( event . jball . yrel * reversal >= mapping . value * reversal ) ;
		if ( key == JOY_LEFT || key == JOY_RIGHT )
			return ( event . jball . xrel * reversal >= mapping . value * reversal ) ;
	}
	
	// SDL_JOYHATMOTION
	if ( mapping . type == JOY_HAT &&
			event . jhat . which == GetIndex ( joystick ) && 
			event . jhat . hat == mapping . num )
		return ( event . jhat . value == mapping . value ) ;
	
	// SDL_JOYBUTTONEVENT
	if ( mapping . type == JOY_BUTTON &&
			event . jbutton . which  == GetIndex ( joystick ) && 
			event . jbutton . button == mapping . num )
		return ( event . jbutton . state  == mapping . value ) ;
	
	// this event has nothing to do with our setup!
	return NOTHING ;
}

// left player pressed something...

bool Joystick::Left_up_pressed ( SDL_Event event )
{
	return ( Test ( left, left_mapping, JOY_UP, event ) == TRUE ) ;
}

bool Joystick::Left_down_pressed ( SDL_Event event )
{
	return ( Test ( left, left_mapping, JOY_DOWN, event ) == TRUE ) ;
}

bool Joystick::Left_left_pressed ( SDL_Event event )
{
	return ( Test ( left, left_mapping, JOY_LEFT, event ) == TRUE ) ;
}

bool Joystick::Left_right_pressed ( SDL_Event event )
{
	return ( Test ( left, left_mapping, JOY_RIGHT, event ) == TRUE ) ;
}

bool Joystick::Left_fire_pressed ( SDL_Event event )
{
	return ( Test ( left, left_mapping, JOY_FIRE, event ) == TRUE ) ;
}

// left player released something

bool Joystick::Left_up_released ( SDL_Event event )
{
	return ( Test ( left, left_mapping, JOY_UP, event ) == FALSE ) ;
}

bool Joystick::Left_down_released ( SDL_Event event )
{
	return ( Test ( left, left_mapping, JOY_DOWN, event ) == FALSE ) ;
}

bool Joystick::Left_left_released ( SDL_Event event )
{
	return ( Test ( left, left_mapping, JOY_LEFT, event ) == FALSE ) ;
}

bool Joystick::Left_right_released ( SDL_Event event )
{
	return ( Test ( left, left_mapping, JOY_RIGHT, event ) == FALSE ) ;
}

bool Joystick::Left_fire_released ( SDL_Event event )
{
	return ( Test ( left, left_mapping, JOY_FIRE, event ) == FALSE ) ;
}

// right player pressed something

bool Joystick::Right_up_pressed ( SDL_Event event )
{
	return ( Test ( right, right_mapping, JOY_UP, event ) == TRUE ) ;
}

bool Joystick::Right_down_pressed ( SDL_Event event )
{
	return ( Test ( right, right_mapping, JOY_DOWN, event ) == TRUE ) ;
}

bool Joystick::Right_left_pressed ( SDL_Event event )
{
	return ( Test ( right, right_mapping, JOY_LEFT, event ) == TRUE ) ;
}

bool Joystick::Right_right_pressed ( SDL_Event event )
{
	return ( Test ( right, right_mapping, JOY_RIGHT, event ) == TRUE ) ;
}

bool Joystick::Right_fire_pressed ( SDL_Event event )
{
	return ( Test ( right, right_mapping, JOY_FIRE, event ) == TRUE ) ;
}

// right player released something

bool Joystick::Right_up_released ( SDL_Event event )
{
	return ( Test ( right, right_mapping, JOY_UP, event ) == FALSE ) ;
}

bool Joystick::Right_down_released ( SDL_Event event )
{
	return ( Test ( right, right_mapping, JOY_DOWN, event ) == FALSE ) ;
}

bool Joystick::Right_left_released ( SDL_Event event )
{
	return ( Test ( right, right_mapping, JOY_LEFT, event ) == FALSE ) ;
}

bool Joystick::Right_right_released ( SDL_Event event )
{
	return ( Test ( right, right_mapping, JOY_RIGHT, event ) == FALSE ) ;
}

bool Joystick::Right_fire_released ( SDL_Event event )
{
	return ( Test ( right, right_mapping, JOY_FIRE, event ) == FALSE ) ;
}

// :(

void Joystick::Quit ()
{
	if ( left != NULL )
		SDL_JoystickClose ( left ) ;
	if ( right != NULL )
		SDL_JoystickClose ( right ) ;

	SDL_QuitSubSystem ( SDL_INIT_JOYSTICK ) ;
}
