// include sdl, kyra and some difinitions...
#include "definitions.h"
#include "math.h"

// will be filled with the correct path gained from RC_DIR
char rc_dir[4096] ;
// where our packages (pax) reside - optional to the data-dir
char pax_dir[4096] ;

bool soundeffects_enabled = TRUE ;
bool music_enabled = TRUE ;
bool disable_audio_output = FALSE ;

bool screen_saver_mode = FALSE ;

// how many gems have to be grouped before they get destroyed?
int min_friends_for_group = 4 ;

// level of detail can be:
// 1 = no alpha blending, tinting or sliding
// 2 = alpha blending, tinting but no sliding
// 3 = alpha blending, tinting and sliding
int level_of_detail = 3 ;

// our theme package
char pkg_filename[4096] ;

// human interface design :) - my api for user input
#include "hid.h"
HID hid ;

// our file cache
FileCache file_cache ;

#include "audio.h"
Audio audio ;

#include "video.h"
Video video ;

#include "utils.h"

#if defined __GLIBC__ && __GLIBC_MINOR__
	#include <gnu/libc-version.h>
#endif

// some network stuff
#ifdef __NETWORK_SUPPORT__
	#include "network.h"
	Network network ;
	
	// the default port on which a server socket will be available
	Uint16 default_server_port = 5555 ;
#endif
//

bool dedicated_server = false ;

SDL_TimerID eventTimer = NULL ;

Uint32 TimerCallback ( Uint32 interval, void *param )
{
	SDL_Event event;
	event.type = SDL_TIMER_EVENT;

	SDL_PeepEvents ( &event, 1, SDL_ADDEVENT, 0 ) ;
	return TIMER_INTERVAL;
}

void quit_main ()
{
	network . Quit () ;
	
	if ( ! dedicated_server )
	{
		hid . Quit () ;
		
		audio . Quit () ;
		
		video . Quit () ;
	}
	
	SDL_Quit () ;	
	
	cleanup_temp_dir () ;
	
	fprintf ( stdout, "Let's see what's left to be done...\n" ) ;
	file_cache.DumpStatus () ;
	
	file_cache . Quit () ;
	
	fprintf ( stdout, "Cleanup finished. Bye!\n" ) ;
	
	exit ( 0 ) ;
}

int main ( int argc, char *argv[] )
{
	fprintf ( stdout, "\nWelcome to %s by TheBlackLion'2003/2004 !!! =)\n", GAME_TITLE ) ;

#if defined __GLIBC__ && defined __GLIBC_MINOR__
	fprintf ( stdout, "Compiled with glibc %d.%d ",
			__GLIBC__,
			__GLIBC_MINOR__ ) ;
#if defined __GNUC__ && defined __GNUC_MINOR__
	fprintf ( stdout, "( gcc %d.%d ) ",
			__GNUC__, __GNUC_MINOR__ ) ;
#endif

	fprintf ( stdout, "\nRunning with glibc %s (%s)\n\n",
			gnu_get_libc_version (),
			gnu_get_libc_release () ) ;
#else
	fprintf ( stdout, "\n" ) ;
#endif

	// theme file to start with...
	//pkg_filename = (char*) malloc ( sizeof(char) * 4096 ) ;
	sprintf ( pkg_filename, "%s", "" ) ;

	// setup config dir and subdirs...
	create_config_dir () ;
	
	const SDL_version *sdlVersion = SDL_Linked_Version () ;
	if ( sdlVersion->minor < 2 )
	{
		fprintf ( stdout, "SDL version must be at least 1.2.0" ) ;
		exit ( 254 ) ;
	}
	
	/* Initialize the SDL library */
	int failure = SDL_Init ( SDL_INIT_TIMER /* | SDL_INIT_NOPARACHUTE */ ) ;
	if ( failure )
	{
		printf ( "Failed to initialize SDL error=%s\n", SDL_GetError () ) ;
		exit ( 1 ) ;
	}
	
#ifdef __NETWORK_SUPPORT__
	// init network stuff
	network . Init () ;
	// may be modified by commandline args
	Uint16 server_port = default_server_port ;
	// disable ncurses gui?
	bool no_curs_gui = false ;
#endif
	
	{	// make sure to search for files where we've been executed !
		char *bname, *dname ;
		char *path = argv [ 0 ] ;
		
		dname = dirname ( path ) ;
		bname = basename ( path ) ;
		//fprintf ( stdout, "dirname = '%s' ; basename = '%s' ;\n", dname, bname ) ;
		chdir ( dname ) ;
		free ( dname ) ;
		free ( bname ) ;
	}
	
	// now go into our data directory...
	if ( chdir ( "data" ) != 0 ) // this works in source-tree
		if ( chdir ( "../../data" ) != 0 ) // this works in source-tree
			if ( chdir ( "../data" ) != 0 ) // this works in release
#ifndef PACKAGE_DATA_DIR
				chdir ( "../../GemHunters-data" ) ; // last resort :)
#else
				if ( chdir ( "../../GemHunters-data" ) != 0 ) // nearly last resort :)
					chdir ( PACKAGE_DATA_DIR ) ;
#endif

	file_cache . Init () ;

	video . Init () ;

	// look if user wants our help-page
	for ( int i = 1 ; i < argc ; i++ )
	{
		if ( *argv[i] == '-' && *(argv[i]+1) )
		{
			switch ( *(argv[i]+1) )
			{
				case 'h':
				{
					int depth = SDL_VideoModeOK ( video . GetScreenX (),
									video . GetScreenY (),
									video . GetDefaultDepth (), 0 ) ;
					char prog_name [ 2048 ] ;
					sprintf ( prog_name, "%s", argv [ 0 ] ) ;
					
					fprintf ( stdout, "Usage: %s [ options ]\n\n", prog_name ) ;
					fprintf ( stdout, "Options are...\n\n" ) ;
					fprintf ( stdout, "-f  = play fullscreen\n" ) ;
					fprintf ( stdout, "-o  = force the use of OpenGL\n" ) ;
					fprintf ( stdout, "-s  = force the use of software rendering\n" ) ;
					fprintf ( stdout, "-c# = force color-depth for renderer\n" ) ;
					fprintf ( stdout, "      1 = 16 bit ( highcolor %s)\n",
								depth == 16 ? " / suggested " : "" ) ;
					fprintf ( stdout, "      2 = 32 bit ( truecolor %s)\n",
								depth == 32 ? " / suggested " : "" ) ;
					fprintf ( stdout, "-cs = cache scaled bitmaps ( needs much memory! )\n" ) ;
					fprintf ( stdout, "-d# = set level of detail which can be one of\n" ) ;
					fprintf ( stdout, "      1 = no alpha blending, tinting or sliding%s\n",
								level_of_detail == 1 ? " ( default )" : "" ) ;
					fprintf ( stdout, "      2 = alpha blending, tinting but no sliding%s\n",
								level_of_detail == 2 ? " ( default )" : "" ) ;
					fprintf ( stdout, "      3 = alpha blending, tinting and sliding%s\n",
								level_of_detail == 3 ? " ( default )" : "" ) ;
					fprintf ( stdout, "-r  = create resizable main window\n" ) ;
					fprintf ( stdout, "      ( YUV overlay resizing / forces -s and disables -c# !!! )\n" ) ;
					fprintf ( stdout, "-m# = specifies YUV mode\n" ) ;
					fprintf ( stdout, "      1 = YV12 ( planar mode %s)\n",
								video . GetOverlayMode () == SDL_YV12_OVERLAY ? "/ default " : "" ) ;
					fprintf ( stdout, "      2 = YUY2 ( packed mode %s)\n",
								video . GetOverlayMode () == SDL_YUY2_OVERLAY ? "/ default " : "" ) ;
					fprintf ( stdout, "-x# = resize width of window to # ( default is %d )\n", video . GetScreenX () ) ;
					fprintf ( stdout, "-y# = resize height of window to # ( default is %d )\n", video . GetScreenY () ) ;
					fprintf ( stdout, "-ns = initially disable sound effects\n" ) ;
					fprintf ( stdout, "-nm = initially disable music\n" ) ;
					fprintf ( stdout, "-na = don't initialize audio at all\n" ) ;
					fprintf ( stdout, "-sm = startup in screen-saver-mode\n" ) ;
					fprintf ( stdout, "-t# = theme to start with\n" ) ;
					fprintf ( stdout, "      # = filename ( default = '%s' )\n", pkg_filename ) ;
#ifdef __NETWORK_SUPPORT__
					fprintf ( stdout, "-network    = start with networkscreen\n" ) ;
					fprintf ( stdout, "-connectto computer.example.net[:5555]\n" ) ;
					fprintf ( stdout, "-nogui [#]  = start as server and listen on port # ( default is %d )\n", default_server_port ) ;
#ifdef __NCURSES_SUPPORT__
					fprintf ( stdout, "              ( will provide a text-interface using ncurses )\n" ) ;
					fprintf ( stdout, "-nocursgui  = disable even ncurses interface :(\n" ) ;
#else
					fprintf ( stdout, "              ( compile with ncurses to get a beautiful text-interface :) )\n" ) ;
#endif
					fprintf ( stdout, "-netdebug   = prints many messages to the screen for debugging\n" ) ;
#endif
					fprintf ( stdout, "\n" ) ;
					fprintf ( stdout, "Examples\n" ) ;
					fprintf ( stdout, "========\n" ) ;
					fprintf ( stdout, "For playing fullscreen with medium level of detail for\n" ) ;
					fprintf ( stdout, "slower computers use something like this:\n" ) ;
					fprintf ( stdout, "%s -f -d2\n\n", prog_name ) ;
					fprintf ( stdout, "Or for faster computers and people who want to freely\n" ) ;
					fprintf ( stdout, "scale the main window to their desires:\n" ) ;
					fprintf ( stdout, "%s -r\n\n", prog_name ) ;
					fprintf ( stdout, "Or as a simple screensaver without sliding:\n" ) ;
					fprintf ( stdout, "%s -sm -d2\n", prog_name ) ;
					fprintf ( stdout, "\n" ) ;
					exit ( 0 ) ;
				} // case 'h'
			}
		}
	}

	fprintf ( stdout, "Try -h as an option for command-line help...\n\n" ) ;

	int resize_window_width_to = -1, resize_window_height_to = -1 ;
	
	// look for arguments from the command-line...
	for ( int i = 1 ; i < argc ; i++ )
	{
		if ( *argv[i] == '-' && *(argv[i]+1) )
		{
			const char* param = argv[i] + 2 ;
			switch ( *(argv[i]+1) )
			{
				case 'f':
				{
					video . SetFullscreen ( TRUE ) ;
				}
				break ;
				case 'o':
				{
					video . SetForceOpenGL ( TRUE ) ;
				}
				break ;
				case 's':
				{
					if ( *param == 'm' )
					{
						screen_saver_mode = TRUE ;
						disable_audio_output = TRUE ;
						video . SetFullscreen ( TRUE ) ;
					}
					else
						video . SetForceSoftware ( TRUE ) ;
				}
				break ;
				case 'c':
				{
					if ( atoi ( param ) >= 1 && atoi ( param ) <= 2 )
					{
						video . SetDefaultDepth ( atoi ( param ) == 1 ? 16 : 32 ) ;
						video . SetForceDepth ( TRUE ) ;
					}
					else if ( *param == 's' )
						video . SetUseCachedScaling ( TRUE ) ;
#ifdef __NETWORK_SUPPORT__
					else if ( strcmp ( param - 1, "connectto" ) == 0 )
					{
						char server [ 256 ] ;
						Network::simple_host host ;
						host . port = default_server_port ;
						sprintf ( server, "%s", argv[++i] ) ;
						if ( rindex ( server, ':' ) )
						{
							host . port = atoi ( rindex ( server, ':' ) + 1 ) ;
							*rindex ( server, ':' ) = '\0' ;
						}
						sprintf ( host . address, "%s", server ) ;
						network . AddHostToConnectionList ( &host ) ;
						
						char string [] = "network" ;
						sprintf ( pkg_filename, "%s", string ) ;
					}
#endif
					else
					{
						fprintf ( stdout, "color-depth can only be 1 ( highcolor ) or 2 ( truecolor / default ) !\n" ) ;
						exit ( 1 ) ;
					}
				}
				break ;
				case 'd':
				{
					if ( atoi ( param ) < 1 || atoi ( param ) > 3 )
					{
						fprintf ( stdout, "level of detail can only be one of 1 (min), 2 (medium), 3 (max/default) !\n" ) ;
						exit ( 1 ) ;
					}
					level_of_detail = atoi ( param ) ;
				}
				break ;
				case 'r':
				{
					//if ( *param == 'y' )
					{
						video . SetWindowResizeMode ( video . RESIZE_MODE_OVERLAY ) ;
						video . SetForceSoftware ( TRUE ) ;
					}
					//else
						//video . SetWindowResizeMode ( video . RESIZE_MODE_AUTO ) ;
				}
				break ;
				case 'm':
				{
					if ( atoi ( param ) == 1 )
						video . SetOverlayMode ( SDL_YV12_OVERLAY ) ;
					else
						video . SetOverlayMode ( SDL_YUY2_OVERLAY ) ;
				}
				break ;
				case 'x':
				{
					if ( atoi ( param ) >= video . GetScreenX () )
						resize_window_width_to = atoi ( param ) ;
					else
						fprintf ( stdout, "the minimum width is %d pixels. %d is too small...\n",
									video . GetScreenX (), atoi ( param ) ) ;
				}
				break ;
				case 'y':
				{
					if ( atoi ( param ) >= video . GetScreenY () )
						resize_window_height_to = atoi ( param ) ;
					else
						fprintf ( stdout, "the minimum height is %d pixels. %d is too small...\n",
									video . GetScreenY (), atoi ( param ) ) ;
				}
				break ;
				case 'n':
				{
					if ( *param == 's' ) soundeffects_enabled = FALSE ;
					if ( *param == 'm' ) music_enabled = FALSE ;
					if ( *param == 'a' ) disable_audio_output = TRUE ;
#ifdef __NETWORK_SUPPORT__
					if ( strcmp ( param - 1, "network" ) == 0 )
					{	// this could be done in a better way :)
						sprintf ( pkg_filename, "%s", param - 1 ) ;
					}
					if ( strcmp ( param - 1, "netdebug" ) == 0 )
					{
						network . EnableDebugging () ;
					}
					if ( strcmp ( param - 1, "nogui" ) == 0 )
					{
						if ( argc < i + 1 && argv[i+1][0] != '-' )
							server_port =  atoi ( argv[++i] ) ;
						
						char string [] = "network" ;
						sprintf ( pkg_filename, "%s", string ) ;
						dedicated_server = true ;
						printf ( "- starting as dedicated server, gui disabled!\n" ) ;
					}
					if ( strcmp ( param - 1, "nocursgui" ) == 0 )
					{
						no_curs_gui = true ;
					}
#endif
				}
				break ;
				case 't':
				{
					sprintf ( pkg_filename, "%s", param ) ;
				}
			}
		}
	}

	if ( ! dedicated_server )
	{
		// Setup audio
		audio . Init () ;
		
		// Setup video
		video . Setup () ;
		
		// Check if the user wanted to resize the main window...
		if ( resize_window_width_to != -1 || resize_window_height_to != -1 )
		{
			if ( resize_window_width_to < video . GetScreenX () )
				resize_window_width_to = video . GetScreenWidth () ;
			if ( resize_window_height_to < video . GetScreenY () )
				resize_window_height_to = video . GetScreenHeight () ;
			
			if ( resize_window_width_to != video . GetScreenX () ||
				resize_window_height_to != video . GetScreenY () )
				video . ResizeVideo ( resize_window_width_to, resize_window_height_to ) ;
		}
	}
	
	// Start timing!
	SDL_Event event ;
	eventTimer = SDL_AddTimer ( TIMER_INTERVAL, TimerCallback, NULL ) ;

	// whether the game is done
	bool done = false ;
	// whether the game is paused ( user / focus )
	bool pause = false ;
	// whether a user wants to un-/pause the game
	bool toggle_pause_by_user = false ;
	// whether we lost our focus ( gamewindow )
	bool focus_lost = false ;
	
	//video . SetReactOnFocusLost ( true ) ;
	
	//
	//// Game starts here
	//

	if ( ! dedicated_server )
		hid . Init () ;
	
	atexit ( quit_main ) ;

	// code for framecounter
	Uint32 last_tick = SDL_GetTicks () ;
	float frames, dt ;
	float time = 0 ;

#ifdef __NETWORK_SUPPORT__
	if ( dedicated_server && ! no_curs_gui )
		network . EnableConsolePrinting ( NULL, NULL ) ;
	network . ListenOnPort ( server_port ) ;
	network . UpdateNConsole () ;
	if ( strcmp ( pkg_filename, "network" ) != 0 ||
		( dedicated_server && no_curs_gui ) )
		network . ConnectToServersInList () ;
#endif

	while ( !done && 
		( ( ! dedicated_server && SDL_WaitEvent ( &event ) ) ||
	          dedicated_server ) )
	{
		if ( event.type == SDL_QUIT )
			break ;

		// Calculate time since last update ( framecounter )
		Uint32 tick = SDL_GetTicks () ;
		dt = ( tick - last_tick ) * 0.001 ;
		frames = dt * (float) FRAMERATE ;
		//printf ( "frames = %f\r", frames ) ;
		
		if ( ! dedicated_server )
			video . Everytime ( event ) ;
		
		switch ( event.type )
		{
			// someone resized our window!
			case SDL_VIDEORESIZE:
			{	// user resized main window..
				video . ResizeVideo ( event.resize.w, event.resize.h ) ;
			}
			break ;
			
			// press a key
			case SDL_KEYDOWN:
			{
				switch ( event.key.keysym.sym )
				{
					// quit current module?
					case SDLK_ESCAPE:
					{
						if ( pause == FALSE &&
								video . GetIsQuitting () == FALSE )
							video . QuitGameScreen () ;
					}
					break ;
					
					// toggle full-screen
					case SDLK_F1:
					{
						video . ToggleFullscreen () ;
					}
					break ;
					
					// toggle sound-effects on/off
					case SDLK_F3:
							soundeffects_enabled = ! soundeffects_enabled ;
					break ;
					
					// toggle framerate display
					case SDLK_F9:
						video . SetDisplayFramerate ( ! video . GetDisplayFramerate () ) ;
					break ;
					
					// dump current framerate
					case SDLK_F10:
						video . DumpFPS () ;
					break ;
					
					// dump cached files as list
					case SDLK_F11:
						file_cache . DumpStatus () ;
					break ;

					// pause game
					case SDLK_p:
						toggle_pause_by_user = true ;
					break ;
					
					default:
					break ;
				}
			}
			break ;
			
			// release a key
			case SDL_KEYUP:
			{
				switch ( event.key.keysym.sym )
				{
					default:
					break ;
				}
			}
			break ;

			// it is time for a visual update!
			case SDL_TIMER_EVENT:
			{
				// nothing here
			}
			break ;

			// nothing to do..
			default:
			break ;
		} // switch ( event.type )
		
#ifdef __NETWORK_SUPPORT__
		// network stuff
		network . ProcessEvents () ;
#endif
		
		// we now use a different update scheme where we update
		// everything often enough internally and finally draw
		// something on the screen too - when our framerate matches...
		// our internal updates aren't very cpu intensive...
		int i = (int) ( floor ( time + frames ) - floor ( time ) ) ;
		if ( i && ! dedicated_server )
		{
			while ( i-- )
			{
				video . SetTime ( time ) ;
				//video . Update ( pause, toggle_pause_by_user, focus_lost, false ) ;
				video . Update ( pause, toggle_pause_by_user, focus_lost, i > 0 ? false : AUTO, i ) ;
			}
		}
		time += frames ;
		video . SetTime ( time ) ;
		
		if ( ! dedicated_server )
		{
			done = video . GetIsDone () ;
			audio . Update () ;
		}
		else
			SDL_Delay ( (Uint32) FRAMEDELAY ) ;
		
		last_tick = tick ;
		
	} // while ( !done && SDL_WaitEvent ( &event ) )

	// Print some statistics
	video . PrintFPSStatistics () ;
	
	// kill our timer...
	if ( eventTimer != NULL )
		SDL_RemoveTimer ( eventTimer ) ;
	
	// cleanup will be done by the atexit() function which calls quit_main() !
	fprintf ( stdout, "Clean up started...\n" ) ;
	
	return 0 ;
}
