#include "nconsole.h"

#ifdef __NETWORK_SUPPORT__

#ifdef __NCURSES_SUPPORT__

#include "video.h"
extern Video video ;

#include "network.h"
extern Network network ;

#include <ncurses.h>

NConsole::NConsole ()
{
	initscr () ;
	clear () ;
	noecho () ;
	cbreak () ; // disable line buffering
	nodelay ( stdscr, true ) ; // make getch non-blocking :)
	keypad ( stdscr, true ) ; // enable cursor-keys etc.
	meta ( stdscr, true ) ;
	
	start_color () ;
	
	init_pair ( 1, COLOR_WHITE , COLOR_BLACK ) ; // normal text
	init_pair ( 2, COLOR_YELLOW, COLOR_BLUE  ) ; // borders and titles
	init_pair ( 3, COLOR_RED   , COLOR_BLACK ) ; // system messages
	init_pair ( 4, COLOR_BLUE  , COLOR_BLACK ) ; // debug messages
	
	textbuffer . Clear () ;
	buddies . Clear () ;
	
	Network::packet packet ;
	cmdline_length = ARRAY_SIZE ( packet.data ) ;
	cmdline = (char*) calloc ( cmdline_length, sizeof(char) ) ;
	cmdline_bak = (char*) calloc ( cmdline_length, sizeof(char) ) ;
	cmdline_pos = 0 ; // cursor pos ( horizontal )
	cmdhistory . Clear () ;
	cmdhistory_pos = 0 ;
	
	getmaxyx ( stdscr, screenh, screenw ) ;
}

void NConsole::PushText ( std::string text )
{
	GlDynArray < std::string > buffer = BreakTextToMaxCols ( text, screenw - 21 ) ;
	
	for ( unsigned int i = 0 ; i < buffer . Count () ; i++ )
		textbuffer . PushBack ( buffer [ i ] ) ;
}

void NConsole::Printf ( const char *format, ... )
{	// this one is copied from kyra's console.cpp file :)
	// slightly modified by me to use either a KrConsole or
	// the stdout stream...
	va_list va ;
	char buffer [ 8192 ] ;
	
	//
	//  format and output the message..
	//
	va_start ( va, format ) ;
	vsprintf ( buffer, format, va ) ;
	va_end ( va ) ;
	
	char *start ;
	char *end ;
	char *next ;
	
	start = buffer ;
	
	while ( start && *start )
	{
		end = strchr ( start, '\n' ) ;
		if ( end )
		{
			next  = end + 1 ;
			*end = 0 ;
			PushText ( start ) ;
			start = next ;
		}
		else
		{
			PushText ( start ) ;
			start = 0 ;
		}
	}
}

void NConsole::PrintText ( uint x, uint y, char *text,
							int num_chars, int color_pair )
{
	char t_text [ strlen ( text ) ] ;
	sprintf ( t_text, "%s", text ) ;
	
	if ( num_chars == AUTO )
		num_chars = strlen ( text ) ;
	else if ( num_chars < (int) strlen ( t_text ) )
		t_text [ num_chars ] = '\0' ;
	
	if ( color_pair != NOTHING )
		attron ( COLOR_PAIR ( color_pair ) ) ;
	
	mvwprintw ( stdscr, y, x, "%s", t_text ) ;
	for ( uint i = strlen ( t_text ) ; i < (uint) num_chars ; i++ )
		mvwprintw ( stdscr, y, x + i, "%s", " " ) ;
	
	if ( color_pair != NOTHING )
		attroff ( COLOR_PAIR ( color_pair ) ) ;
}

void NConsole::PrintHLine ( uint x, uint y, uint num_chars, uint color_pair )
{
	attron ( COLOR_PAIR ( color_pair ) ) ;
	for ( uint i = 0 ; i < num_chars ; i++ )
		mvwprintw ( stdscr, y, x + i, "%s", " " ) ;
	attroff ( COLOR_PAIR ( color_pair ) ) ;
}

void NConsole::PrintVLine ( uint x, uint y, uint num_chars, uint color_pair )
{
	attron ( COLOR_PAIR ( color_pair ) ) ;
	for ( uint i = 0 ; i < num_chars ; i++ )
		mvwprintw ( stdscr, y + i, x, "%s", " " ) ;
	attroff ( COLOR_PAIR ( color_pair ) ) ;
}

void NConsole::UpdateBuddyList ( GlDynArray < std::string > *buddy_list )
{
	if ( ! buddy_list ) return ;
	
	buddies . Clear () ;
	
	for ( uint i = 0 ; i < buddy_list -> Count () ; i++ )
		buddies . PushBack ( buddy_list -> Item ( i ) ) ;
}

void NConsole::Update ()
{
	PrintText ( 0, 0, "Gem Hunters - Server Console", screenw - 20, 2 ) ;
	PrintText ( screenw - 19, 0, "Buddy list", 20, 2 ) ;
	PrintVLine ( screenw - 21, 0, screenh - 2, 2 ) ;
	PrintVLine ( screenw - 20, 0, screenh - 2, 2 ) ;
	PrintText ( 0, screenh - 2, "Command line", screenw, 2 ) ;
	
	uint first_line = max ( textbuffer . Count () - ( screenh - 3 ), 0 ) ;
	
	for ( uint line = first_line ; line < first_line + ( screenh - 3 ) ; line++ )
	{
		if ( line < textbuffer . Count () )
		{
			int color_pair = 1 ;
			char *text = (char*) textbuffer [ line ] . c_str () ;
			
			if ( strlen ( text ) > 1 && text [ 0 ] == '|' && text [ 1 ] == '|' )
				color_pair = 3 ;
			else if ( strlen ( text ) > 0 && text [ 0 ] == '|' )
				color_pair = 4 ;
			
			PrintText ( 0, 1 + line - first_line, text, screenw - 21, color_pair ) ;
		}
		else
			PrintText ( 0, 1 + line - first_line, "", screenw - 21, 1 ) ;
	}
	
	first_line = 0 ;
	for ( uint line = first_line ; line < first_line + ( screenh - 3 ) ; line++ )
	{
		if ( line < buddies . Count () )
			PrintText ( screenw - 19, 1 + line - first_line,
						(char*) buddies [ line ] . c_str (), 20, 1 ) ;
		else
			PrintText ( screenw - 19, 1 + line - first_line, "", 20, 1 ) ;
	}
	
	int c = getch () ;
	
	if ( c != ERR )
	{
		//char text [ 256 ] ;
		//sprintf ( text, "char was '%c' in int %d.\n", c, c ) ;
		//PrintText ( screenw + 1 - strlen ( text ), screenh - 2, text,
		//			AUTO, NOTHING ) ;
		
		// my gnome-terminal returns 10 here
		if ( ( c == KEY_ENTER || c == 10 ) /*&& strlen ( cmdline )*/ )
		{
			std::string text ;
			text = cmdline ;
			uint entries = cmdhistory . Count () ;
			if ( ( entries && strlen ( text . c_str () ) &&
				strcmp ( cmdhistory [ entries - 1 ] . c_str (), text . c_str () ) != 0 ) ||
				! entries && strlen ( text . c_str () ) )
			{
				cmdhistory . PushBack ( text ) ;
			}
			cmdhistory_pos = cmdhistory . Count () ;
			PushText ( network . AddUsernameAtBeginning ( text ) ) ;
			network . ProcessMessage ( NULL, (char*) text . c_str () ) ;
			sprintf ( cmdline, "%s", "" ) ;
			cmdline_pos = 0 ;
			
			PrintText ( 0, screenh-1, cmdline, screenw, 1 ) ;
		}
		else if ( c == KEY_LEFT && cmdline_pos > 0 )
			cmdline_pos-- ;
		else if ( c == KEY_RIGHT && cmdline_pos < strlen ( cmdline ) )
			cmdline_pos++ ;
		else if ( c == KEY_HOME )
			cmdline_pos = 0 ;
		else if ( c == KEY_END )
			cmdline_pos = strlen ( cmdline ) ;
		else if ( c == KEY_UP && cmdhistory_pos > 0 )
		{
			if ( cmdhistory_pos == cmdhistory . Count () )
				sprintf ( cmdline_bak, "%s", cmdline ) ;
			
			cmdhistory_pos-- ;
			sprintf ( cmdline, "%s", (char*) cmdhistory . Item ( cmdhistory_pos ) . c_str () ) ;
			
			//if ( cmdline_pos > strlen ( cmdline ) )
				cmdline_pos = strlen ( cmdline ) ;
			
			PrintText ( 0, screenh-1, cmdline, screenw, 1 ) ;
		}
		else if ( c == KEY_DOWN && cmdhistory_pos < cmdhistory . Count () )
		{
			cmdhistory_pos++ ;
			if ( cmdhistory_pos < cmdhistory . Count () )
				sprintf ( cmdline, "%s", (char*) cmdhistory . Item ( cmdhistory_pos ) . c_str () ) ;
			else
				sprintf ( cmdline, "%s", cmdline_bak ) ;
			
			//if ( cmdline_pos > strlen ( cmdline ) )
				cmdline_pos = strlen ( cmdline ) ;
			
			PrintText ( 0, screenh-1, cmdline, screenw, 1 ) ;
		}
		// my gnome-terminal returns 127 here
		else if ( ( c == KEY_BACKSPACE || c == 127 ) &&
					cmdline_pos > 0 && strlen ( cmdline ) )
		{
			for ( uint p = cmdline_pos - 1 ; p < strlen ( cmdline ) ; p++ )
				cmdline [ p ] = cmdline [ p + 1 ] ;
			cmdline_pos-- ;
			
			PrintText ( 0, screenh-1, cmdline, screenw, 1 ) ;
		}
		// my gnome-terminal returns 330 here
		else if ( ( c == KEY_DC || c == 330 ) &&
					cmdline_pos < strlen ( cmdline )  )
		{
			for ( uint p = cmdline_pos ; p < strlen ( cmdline ) ; p++ )
				cmdline [ p ] = cmdline [ p + 1 ] ;
			
			PrintText ( 0, screenh-1, cmdline, screenw, 1 ) ;
		}
		else if ( ( c >= '!' && c <= '~' ) || c == ' ' )
		{
			if ( cmdline_pos < strlen ( cmdline ) )
			{
				for ( uint p = strlen ( cmdline ) ; p > cmdline_pos ; p-- )
				{
					cmdline [ p ] = cmdline [ p - 1 ] ;
				}
			}
			else
				cmdline [ cmdline_pos + 1 ] = '\0' ;
			cmdline [ cmdline_pos ] = c ;
			
			cmdline_pos++ ;
			
			PrintText ( 0, screenh-1, cmdline, screenw, 1 ) ;
		}
	}
	
	move ( screenh-1, cmdline_pos ) ;
	
	redrawwin ( stdscr ) ;
	refresh () ;
}

NConsole::~NConsole ()
{
	endwin () ;
	
	textbuffer . Clear () ;
	buddies . Clear () ;
	
	free ( cmdline ) ;
	cmdhistory . Clear () ;
}

#endif

#endif
