#include "networkscreen.h"

#ifdef __NETWORK_SUPPORT__

extern char rc_dir[], pax_dir[] ;

extern char pkg_filename[] ;

extern FileCache file_cache ;

#include "video.h"
extern Video video ;

#include "gamescreen.h"
extern PaxList *pax_list ;
extern int next_game_screen ;

#include "hid.h"
extern HID hid ;

#include "network.h"
extern Network network ;

void NetworkScreen::Init ()
{
	using_opengl = video . GetUsingOpenGL () ;

	///
	
	pax = new Pax () ;
	pax -> Init ( "networkscreen" ) ;
	
	///

	background.Init ( pax ) ;
	
	//pax_list -> AttachListToNode () ;
	
	vault = (KrResourceVault*) pax -> GetDeclaredObjectFromList ( "vault" ) -> data ;
	bg_sprite = (Sprite*) pax -> GetDeclaredObjectFromList ( "background" ) -> data ;

	paused = FALSE ;
	
	quit = FALSE ;
	quit_state = FALSE ;
	
	activate_sound . LoadSound ( "sfx/disable.wav" ) ;

	video . DisplayOSDMessage ( "welcome to your network :)" ) ;
	video . SetPausePossible ( false ) ;
	
	mouse = new MouseCursor () ;
	mouse -> Init () ;
	
	KrFontResource *font = (KrFontResource*) pax -> GetObjectFromList ( "console_font" ) -> data ;
	KrScheme scheme ( font ) ;
	
	/*{
		KrRGBA *c = &scheme . primary ;
		fprintf ( stdout, "primary color = %d, %d, %d\n", c->c.red, c->c.green, c->c.blue ) ;
		c = &scheme . cursor ;
		fprintf ( stdout, "cursor color = %d, %d, %d\n", c->c.red, c->c.green, c->c.blue ) ;
	}*/

	console = new Console () ;
	console -> Init ( 	pax -> GetIntVar ( "cc_offset_x" ),
						pax -> GetIntVar ( "cc_offset_y" ),
						pax -> GetIntVar ( "cc_width" ),
						pax -> GetIntVar ( "cc_height" ),
						pax -> GetIntVar ( "cc_spacing" ),
						pax -> GetNode ( pax -> GetObjectFromList ( "cc_base" ), "cc_base" ),
						scheme ) ;

	buddy_list = new Console () ;
	buddy_list -> Init ( 	pax -> GetIntVar ( "bl_offset_x" ),
							pax -> GetIntVar ( "bl_offset_y" ),
							pax -> GetIntVar ( "bl_width" ),
							pax -> GetIntVar ( "bl_height" ),
							pax -> GetIntVar ( "bl_spacing" ),
							pax -> GetNode ( pax -> GetObjectFromList ( "bl_base" ), "bl_base" ),
							scheme,
							false ) ;
	
	KrSpriteResource *play_res = vault -> GetSpriteResource ( "play_button" ) ;
	GLASSERT ( play_res ) ;
	KrSprite *play_sprite = new KrSprite ( play_res ) ;
	GLASSERT ( play_sprite ) ;
	play_button = new KrPushButton ( play_sprite, scheme ) ;
	video . GetEngine () -> Tree () -> AddNode ( bg_sprite -> GetSprite (), play_button ) ;
	play_button -> SetPos ( 547 + ( pax -> GetIntVar ( "using_c1mode" ) ? 160 : 0 ), 396 ) ;
	
	network . EnableConsolePrinting ( console, buddy_list ) ;
}

void NetworkScreen::ProcessEvents ( SDL_Event event )
{
	console -> ProcessEvents ( event ) ;
	buddy_list -> ProcessEvents ( event ) ;
	
	switch ( event.type )
	{
		case SDL_KEYDOWN:
		{
			SDLKey sym = event.key.keysym.sym ;
			if ( sym == SDLK_F2 )
			{
				// toggle music on/off
				background . ToggleMusic () ;
			}
		}
		break ;
		
		case SDL_MOUSEMOTION:
		{
			int x = (int) ( event.motion.x / video . GetZoomX () ) ;
			int y = (int) ( event.motion.y / video . GetZoomY () ) ;
			
			mouse -> SetXY ( x, y ) ;
		}
		break ;

		default:
		break ;
	}
	
	if ( hid.EventFound ( event ) && background.PlayfieldIsReady () &&
			paused == false && quit == false )
	{
		if ( hid.Left_fire_pressed ( event ) ||
				hid.Right_fire_pressed ( event ) )
		{
			
		}
		else if ( hid.Left_left_pressed ( event ) ||
				hid.Right_left_pressed ( event ) )
		{
			
		}
		else if ( hid.Left_right_pressed ( event ) ||
				hid.Right_right_pressed ( event ) )
		{
			
		}
		else if ( hid.Left_up_pressed ( event ) ||
				  hid.Right_up_pressed ( event ) )
		{
			
		}
		else if ( hid.Left_down_pressed ( event ) ||
				  hid.Right_down_pressed ( event ) )
		{
			
		}
	}
	
	// argh :( cause we're using our own mouse cursor (KrSprite)
	// we have to hide it before kyra can test button clicks etc...
	// if we wouldn't do this, our mouse cursor stands between
	// the button and the original mouse cursor - no hits :(
	mouse -> Hide () ;
	
	// event handler for our widgets...
	KrEventManager::Instance () -> HandleEvent ( event, video . GetEngine () ) ;
	
	// now show it again....
	mouse -> Show () ;
}

void NetworkScreen::Pause ()
{
	/*background . Pause () ;
	paused = TRUE ;*/
}

void NetworkScreen::UnPause ()
{
	/*background . UnPause () ;
	paused = FALSE ;*/
}

void NetworkScreen::Draw ()
{
	background . Draw () ;
	
	mouse -> Draw () ;
	
	if ( background . PlayfieldIsReady () )
	{
		//pax_list -> Draw () ;
	}
}

void NetworkScreen::AfterDraw ()
{
	if ( quit_state == TRUE ) return ;

	if ( quit == TRUE && 
		background . GetIsQuitting () == FALSE ) 
	{
		background . Quit () ;
		mouse -> Quit () ;
	}

	if ( quit == TRUE && background . GetIsDone () == TRUE )
	{
		network . DisableConsolePrinting () ;
		
		video . GetEngine () -> Tree () -> DeleteNode ( play_button ) ;
		delete KrEventManager::Instance () ;

		console -> Free () ; delete console ;
		buddy_list -> Free () ; delete buddy_list ;
		
		//pax_list -> DetachList () ;
		
		mouse -> Free () ;
		delete mouse ;
		
		activate_sound . UnloadSound () ;
		
		pax -> FreeObjects () ;
		delete pax ;
		
		video . SetPausePossible ( true ) ;
		
		quit_state = TRUE ;
	}
}

bool NetworkScreen::IsQuitting ()
{
	return quit ;
}

bool NetworkScreen::IsDone ()
{
	return quit_state ;
}

void NetworkScreen::Quit ()
{
	quit = TRUE ;
	
	background . PrepareQuit () ;
	
	video . DisplayOSDMessage ( "please wait..." ) ;
}

#endif
