#include "pax.h"

extern int level_of_detail ;
extern FileCache file_cache ;

#include "video.h"
extern Video video ;

extern char rc_dir[], pax_dir[] ;

// routines for displaying a splash screen
#include "splash.h"
extern Splash *splash ;

// theme engine

void Pax::Init ( char *filename, bool new_scan_mode )
{
	// normally we are only reporting errors and warnings...
	be_verbose = false ;
	// exit on errors ( = true ) or try to continue ( = false )
	be_strict = true ;
	// are we just scanning this pax? don't create sprites, sounds etc.
	scan_mode = new_scan_mode ;
	
	sprintf ( pkg_dir, "%s", "" ) ;
	sprintf ( pkg_filename, "%s", "" ) ;
	
	object_list = new GlDynArray < object_item > () ;
	object_list -> Clear () ;
	
	script_start    = new GlDynArray < std::string > () ;
	script_ready    = new GlDynArray < std::string > () ;
	script_go       = new GlDynArray < std::string > () ;
	script_pause    = new GlDynArray < std::string > () ;
	script_unpause  = new GlDynArray < std::string > () ;
	script_quitting = new GlDynArray < std::string > () ;
	script_quit     = new GlDynArray < std::string > () ;
	script_event    = new GlDynArray < std::string > () ;
	
	script_start    -> Clear () ;
	script_ready    -> Clear () ;
	script_go       -> Clear () ;
	script_pause    -> Clear () ;
	script_unpause  -> Clear () ;
	script_quitting -> Clear () ;
	script_quit     -> Clear () ;
	script_event    -> Clear () ;
	
	// setup some variables to work with
	
	//fprintf ( stdout, "\nAdding internal read-only variables...\n" ) ;
	
	// system internals
	ObjectListNewEntryFull ( object_type_var   , "null"      , (int *) NULL      , FALSE, TRUE ) ;
	ObjectListNewEntryFull ( object_type_var   , "true"      , (int *) TRUE      , FALSE, TRUE ) ;
	ObjectListNewEntryFull ( object_type_var   , "false"     , (int *) FALSE     , FALSE, TRUE ) ;
	ObjectListNewEntryFull ( object_type_var   , "auto"      , (int *) AUTO      , FALSE, TRUE ) ;
	ObjectListNewEntryFull ( object_type_var   , "random"    , (int *) RANDOM    , FALSE, TRUE ) ;
	ObjectListNewEntryFull ( object_type_var   , "nothing"   , (int *) NOTHING   , FALSE, TRUE ) ;
	ObjectListNewEntryFull ( object_type_var   , "forever"   , (int *) FOREVER   , FALSE, TRUE ) ;
	ObjectListNewEntryFull ( object_type_var   , "once"      , (int *) ONCE      , FALSE, TRUE ) ;
	ObjectListNewEntryFull ( object_type_string, "game_title", (char*) GAME_TITLE, FALSE, TRUE ) ;
	ObjectListNewEntryFull ( object_type_var   , "framerate" , (int *) FRAMERATE , FALSE, TRUE ) ;
	ObjectListNewEntryFull ( object_type_var   , "screenx"   , (int *) video . GetScreenX (), FALSE, TRUE ) ;
	ObjectListNewEntryFull ( object_type_var   , "screeny"   , (int *) video . GetScreenY (), FALSE, TRUE ) ;
	ObjectListNewEntryFull ( object_type_var   , "level_of_detail", (int*) level_of_detail, FALSE, TRUE ) ;
	ObjectListNewEntryFull ( object_type_var   , "sdl_ticks" , (int *) SDL_GetTicks (), FALSE, TRUE ) ;
	ObjectListNewEntryFull ( object_type_var   , "using_opengl", (int *) ( video . GetUsingOpenGL () ? 1 : 0 ), FALSE, TRUE ) ;
	ObjectListNewEntryFull ( object_type_var   , "using_c1mode", (int *) ( video . GetC1ModeActive () ? 1 : 0 ), FALSE, TRUE ) ;
	
	// will be described in level-sets
	int PLAYFIELD_MAX_X = 7, PLAYFIELD_MAX_Y = 13 ;
	int PLAYFIELD_SIZE = PLAYFIELD_MAX_X * PLAYFIELD_MAX_Y ;
	char *GEM_FIRST_COLOR  = (char*) malloc ( sizeof (char) + 16 ) ;
	sprintf ( GEM_FIRST_COLOR, "A" ) ;
	char *GEM_LAST_COLOR  = (char*) malloc ( sizeof (char) + 16 ) ;
	sprintf ( GEM_LAST_COLOR, "Z" ) ;
	ObjectListNewEntryFull ( object_type_var, "playfield_max_x", (int*) PLAYFIELD_MAX_X, TRUE, TRUE ) ;
	ObjectListNewEntryFull ( object_type_var, "playfield_max_y", (int*) PLAYFIELD_MAX_Y, TRUE, TRUE ) ;
	ObjectListNewEntryFull ( object_type_var, "playfield_size" , (int*) PLAYFIELD_SIZE , TRUE, TRUE ) ;
	ObjectListNewEntryFull ( object_type_string, "gem_first_color", GEM_FIRST_COLOR, TRUE, TRUE ) ;
	ObjectListNewEntryFull ( object_type_string, "gem_last_color" , GEM_LAST_COLOR , TRUE, TRUE ) ;
	
	char *GEM_BLOCK = (char*) malloc ( sizeof (char) + 16 ) ;
	char *GEM_HOLD  = (char*) malloc ( sizeof (char) + 16 ) ;
	char *GEM_NONE  = (char*) malloc ( sizeof (char) + 16 ) ;
	sprintf ( GEM_BLOCK, "*" ) ;
	sprintf ( GEM_HOLD , "-" ) ;
	sprintf ( GEM_NONE , " " ) ;
	ObjectListNewEntryFull ( object_type_string, "gem_block_char", GEM_BLOCK, TRUE, TRUE ) ;
	ObjectListNewEntryFull ( object_type_string, "gem_hold_char" , GEM_HOLD , TRUE, TRUE ) ;
	ObjectListNewEntryFull ( object_type_string, "gem_none_char" , GEM_NONE , TRUE, TRUE ) ;
	
	//fprintf ( stdout, "\nIf anything strange should happen while reading the level-data\n" ) ;
	//fprintf ( stdout, "try to put a line like this at the beginning of its pax-file:\n" ) ;
	//fprintf ( stdout, "  engine:verbose true\n\n" ) ;
	
	if ( scan_mode == false && video . GetC1ModeActive () )
		video . SetDisplayC1Layer ( true ) ;

	if ( filename && ReadPax ( filename, scan_mode ) == false )
	{
		fprintf ( stdout, "Quitting...\n" ) ;
		exit ( 1 ) ;
	}
}

bool Pax::ReadPax ( char *pax_filename, bool be_quiet )
{
	struct stat buf ;
	
	char filename [ strlen ( pax_filename ) + 4096 ] ;
	sprintf ( filename, "%s", pax_filename ) ;
	
	//printf ( "1. pax_filename = %s ; filename = %s ; pkg_dir = %s ; pkg_filename = %s\n",
	//		pax_filename, filename, pkg_dir, pkg_filename ) ;
	
	if ( strlen ( pkg_dir ) == 0 )
	{
		sprintf ( pkg_dir, "%s", filename ) ;
		
		sprintf ( pkg_filename, "%s/index.pax", pkg_dir ) ;
	
		//printf ( "2. pax_filename = %s ; filename = %s ; pkg_dir = %s ; pkg_filename = %s\n",
		//		pax_filename, filename, pkg_dir, pkg_filename ) ;
		
		if ( strcmp ( pkg_dir + strlen ( pkg_dir ) - 4, ".pax" ) == 0 )
		{
			sprintf ( pkg_filename, "%s", pkg_dir ) ;
			if ( rindex ( pkg_dir, '/' ) )
				*rindex ( pkg_dir, '/' ) = '\0' ;
		}
		else if ( ! index ( pkg_dir, '/' ) && stat ( pkg_dir, &buf ) != 0 )
		{
			// pax dir in rc_dir
			char rc_pax_path [ 256 ] ;
			sprintf ( rc_pax_path, "%s/%s", pax_dir, pkg_dir ) ;
			
			// pax dir in data dir
			char data_pax_path [ 256 ] ;
			sprintf ( data_pax_path, "pax/%s", pkg_dir ) ;
			
			if ( be_verbose )
				fprintf ( stdout, "\nPax file/dir not found at first. Searching...\n" ) ;
			
			if ( stat ( rc_pax_path, &buf ) == 0 )
			{
				sprintf ( pkg_dir, "%s", rc_pax_path ) ;
				sprintf ( pkg_filename, "%s/index.pax", pkg_dir ) ;
				if ( be_verbose )
					fprintf ( stdout, "Pax dir found: %s\n", pkg_dir ) ;
			}
			else if ( stat ( data_pax_path, &buf ) == 0 )
			{
				sprintf ( pkg_dir, "%s", data_pax_path ) ;
				sprintf ( pkg_filename, "%s/index.pax", pkg_dir ) ;
				if ( be_verbose )
					fprintf ( stdout, "Pax dir found: %s\n", pkg_dir ) ;
			}
			else
			{
				fprintf ( stdout, "Sorry, couldn't open the pax file/dir:\n%s\n", filename ) ;
				exit ( 1 ) ;
			}
			
		}
		
		sprintf ( filename, "%s", pkg_filename ) ;
		
		if ( splash && ! be_quiet ) splash -> SetPercent ( 10 ) ;
	}
	
	if ( be_verbose )
		fprintf ( stdout, "Trying to read '%s'...\n", filename ) ;
	
	FILE *stream ;
	stream = fopen ( filename, "r" ) ;
	
	if ( stream == NULL )
	{
		if ( be_verbose )
			fprintf ( stdout, "Warning: Couldn't open pax-file '%s'!\n", filename ) ;
		
		if ( rindex ( filename, '/' ) != NULL )
			sprintf ( filename, "%s/%s", pkg_dir, rindex ( filename, '/' ) + 1 ) ;
		else
		{
			char tmp [ strlen ( filename ) + 256 ] ;
			sprintf ( tmp, "%s/%s", pkg_dir, filename ) ;
			sprintf ( filename, "%s", tmp ) ;
		}
		
		if ( be_verbose )
			fprintf ( stdout, "         Trying to open '%s' as a replacement...\n", filename ) ;
		
		stream = fopen ( filename, "r" ) ;
		
		if ( stream == NULL )
		{
			if ( be_verbose )
				fprintf ( stdout, "Warning: Couldn't open pax-file '%s'!\n", filename ) ;
			
			if ( rindex ( filename, '/' ) != NULL )
				sprintf ( filename, "gfx/%s", rindex ( filename, '/' ) + 1 ) ;
			else
			{
				char tmp [ strlen ( filename ) + 256 ] ;
				sprintf ( tmp, "gfx/%s", filename ) ;
				sprintf ( filename, "%s", tmp ) ;
			}
			
			if ( be_verbose )
				fprintf ( stdout, "         Trying to open '%s' as a replacement...\n", filename ) ;
			
			stream = fopen ( filename, "r" ) ;
		}
	}
	
	if ( stream != NULL )
	{
		int current_line_number = 0 ;
		char *current_line ;
		
		current_line = (char*) malloc ( sizeof (char) * 4096 ) ;
		
		bool already_done = false ;
		
		int lines_counted = 0 ;
		while ( fgets ( current_line, 4096, stream ) )
			lines_counted++ ;
		rewind ( stream ) ;
		
		while ( fgets ( current_line, 4096, stream ) && !already_done )
		{
			if ( index ( current_line, '#' ) )
				*( index ( current_line, '#' ) ) = '\0' ;
			
			current_line = StripHeadAndTail ( current_line ) ;
			
			current_line_number++ ;
			
			if ( current_line [ 0 ] == '\0' ||
					strlen ( current_line ) <= 0 ||
					current_line [ 0 ] == 10 ||
					strncmp ( current_line, "header", 6 ) == 0 )
				continue ;
			
			current_line [ strlen ( current_line ) -1 ] = '\0' ;
			
			if ( be_verbose )
				fprintf ( stdout, "%03d: %s\n     ", current_line_number, current_line ) ;
			
			char *object_name = GetObjectName ( current_line ) ;
			char *object_action = GetObjectAction ( current_line ) ;
			GlDynArray < std::string > *object_arglist = GetObjectArgList ( current_line ) ;
			
			//fprintf ( stdout, "'%s' : '%s'\n", object_name, object_action )  ;
			
			InterpretLine ( object_name, object_action, object_arglist ) ;
			
			float percent = 60.00 / (float) lines_counted * (float) current_line_number ;
			if ( splash && ! be_quiet ) splash -> SetPercent ( 10.00 + percent ) ;
			
			free ( object_name ) ;
			free ( object_action ) ;
			object_arglist -> Clear () ;
			delete object_arglist ;
			
			if ( scan_mode )
			{
				object_item *object = GetObjectFromList ( "screenshot" ) ;
				
				if ( object != NULL )
				{
					if ( object -> data != NULL )
					{
						char string_var [ 4096 ] ;
						sprintf ( string_var, "%s", (char*) object -> data ) ;
						
						if ( strcmp ( string_var, "hidden" ) != 0 )
						{
							GlDynArray < std::string > *shot_args ;
							shot_args = GetArgList ( string_var ) ;
							
							char *gfx_pack = (char*) shot_args -> Item ( 0 ) . c_str () ;
							
							//fprintf ( stdout, "  Screenshot will be loaded from %s...\n", gfx_pack ) ;
							
							object = GetObjectFromList ( gfx_pack ) ;
							
							shot_args -> Clear () ;
							delete shot_args ;
						}
						
						if ( object )
							already_done = true ;
					}
				}
			}
		}
		
		if ( be_verbose )
			fprintf ( stdout, "Done reading '%s'.\n", filename ) ;
		
		free ( current_line ) ;
		
		fclose ( stream ) ;
	}
	else
	{
		fprintf ( stdout, "Error opening pax-file '%s' !\n", filename ) ;
		return false ;
	}
	
	return true ;
}

char *Pax::GetPkgDirname ()
{
	return pkg_dir ;
}

char *Pax::GetPkgFilename ()
{
	return pkg_filename ;
}

bool Pax::ObjectListNewEntryFull ( unsigned int type, char *name,
									void *data, bool writeable,
									bool auto_declare )
{
	object_item object ;
	
	if ( GetObjectFromList ( name ) )
	{
		fprintf ( stdout, "Warning: Can't add item called '%s' twice!\n", name ) ;
		return false ;
	}
	
	object . type = type ;
	sprintf ( object . name, "%s", name ) ;
	object . data = data ;
	object . writeable = writeable ;
	sprintf ( object . parent, "%s", "" ) ;
	object . child_list = new GlDynArray < std::string > () ;
	object . child_list -> Clear () ;
	if ( auto_declare )
		sprintf ( object . declared_as, "%s", name ) ;
	else
		sprintf ( object . declared_as, "%s", "" ) ;
	
	// fill the rest with '\0' for cleaner debugging
	memset ( object . name + strlen ( object . name ),
				'\0', 256 - strlen ( object . name ) ) ;
	memset ( object . parent + strlen ( object . parent ),
				'\0', 256 - strlen ( object . parent ) ) ;
	memset ( object . declared_as + strlen ( object . declared_as ),
				'\0', 256 - strlen ( object . declared_as ) ) ;

	//fprintf ( stdout, "Adding item %d to objectlist called %s\n",
	//			object_list -> Count (), name ) ;
	
	object_list -> PushBack ( object ) ;
	
	return true ;
}

void Pax::ObjectListAddChildToParent ( char *child, char *parent,
										bool set_parent_var )
{
	object_item *child_object = GetObjectFromList ( child ) ;
	object_item *parent_object = GetObjectFromList ( parent ) ;
	
	if ( child_object == NULL )
	{
		fprintf ( stdout, "Can't find child '%s'. Quitting...\n", child ) ;
		exit ( 1 ) ;
	}
	if ( GetObjectFromList ( child_object -> parent ) )
	{
		fprintf ( stdout, "Can't add child '%s' to parent '%s' when child already has parent '%s'! Quitting...\n",
					child, parent, child_object -> parent ) ;
		exit ( 1 ) ;
	}
	if ( parent_object == NULL )
	{
		fprintf ( stdout, "Can't add child '%s' to non-existent parent '%s'! Quitting...\n",
					child, parent ) ;
		exit ( 1 ) ;
	}
	
	if ( set_parent_var )
		sprintf ( child_object -> parent, "%s", parent ) ;
	
	parent_object -> child_list -> PushBack ( child ) ;
}

char *Pax::StripHeadAndTail ( char *text )
{
	int  length = strlen ( text ) ;
	int  start, end, pos ;
	
	for ( pos = 0 ; pos < length ; pos++ )
		if ( text [ pos ] == 9 )
			text [ pos ] = ' ' ;
	
	for ( pos = 0 ; pos < length ; pos++ )
		if ( text [ pos ] != ' ' )
			break ;
	
	start = pos ;
	
	for ( pos = length - 1 ; pos > 0 ; pos-- )
	{
		if ( text [ pos ] != ' ' )
			break ;
	}
	
	end = pos + 1 ;
	
	for ( pos = start ; pos < end ; pos++ )
		text [ pos - start ] = text [ pos ] ;
	for ( pos = pos - start ; pos < length ; pos++ )
		text [ pos ] = '\0' ;
	
	return text ;
}

char *Pax::StripMarks ( char *text )
{
	int old_pos = 0 ;
	
	for ( unsigned int pos = 0 ; pos < strlen ( text ) ; pos ++ )
	{
		if ( !( pos == 0 && text [ pos ] == '"' ) &&
				!( pos == strlen ( text ) - 1 && text [ pos ] == '"' ) )
			text [ old_pos++ ] = text [ pos ] ;
	}
	
	text [ old_pos ] = '\0' ;
	
	return text ;
}

char *Pax::GetObjectName ( char *line )
{
	char *copy ;
	
	copy = (char*) malloc ( sizeof (char) * 4096 ) ;
	
	//memset ( copy, '\0', 4096 ) ;
	sprintf ( copy, "%s", "" ) ;
	
	if ( index ( line, ':' ) )
	{
		strcpy ( copy, line ) ;
		*( index ( copy, ':' ) ) = '\0' ;
	}
	else
	{
		fprintf ( stdout, "ERROR: Expected object name - what's this?? '%s'\n", line ) ;
		exit ( 1 ) ;
	}
	
	return StripHeadAndTail ( copy ) ;
}

char *Pax::GetObjectAction ( char *line )
{
	char *copy ;
	
	copy = (char*) malloc ( sizeof (char) * strlen ( line ) ) ;
	
	//memset ( copy, '\0', strlen ( line ) ) ;
	sprintf ( copy, "%s", "" ) ;
	
	if ( index ( line, ':' ) )
	{
		strcpy ( copy, StripHeadAndTail ( index ( line, ':' ) + 1 ) ) ;
		if ( index ( copy, ' ' ) )
			*( index ( copy, ' ' ) ) = '\0' ;
	}
	else
	{
		fprintf ( stdout, "ERROR: Expected object action - what's this?? '%s'\n", line ) ;
		exit ( 1 ) ;
	}
	
	return StripHeadAndTail ( copy ) ;
}

GlDynArray < std::string > *Pax::GetObjectArgList ( char *line )
{
	GlDynArray < std::string > *bucket = new GlDynArray < std::string > () ;
	char *copy ;
	
	bucket -> Clear () ;

	copy = (char*) malloc ( sizeof (char) * strlen ( line ) ) ;
	
	//memset ( copy, '\0', strlen ( line ) ) ;
	sprintf ( copy, "%s", "" ) ;
	
	if ( index ( line, ':' ) )
	{
		strcpy ( copy, StripHeadAndTail ( index ( line, ':' ) + 1 ) ) ;
		if ( index ( copy, ' ' ) )
		{
			strcpy ( copy, StripHeadAndTail ( index ( copy, ' ' ) + 1 ) ) ;
			
			bucket = GetArgList ( copy ) ;
		}
	}
	else
	{
		fprintf ( stdout, "ERROR: Expected object arglist - what's this?? '%s'\n", line ) ;
		exit ( 1 ) ;
	}
	
	free ( copy ) ;
	
	return bucket ;
}

Pax::object_item *Pax::GetDeclaredObjectFromList ( char *object_name )
{
	for ( unsigned int pos = 0 ; pos < object_list -> Count () ; pos ++ )
	{
		object_item *object = object_list -> ItemPointer ( pos ) ;
		
		if ( strcmp ( object -> declared_as, object_name ) == 0 )
			return object ;
	}
	
	fprintf ( stdout, "ERROR: Couldn't find declared object '%s'.\n", object_name ) ;
	fprintf ( stdout, "       Perhaps it's not declared yet?\n" ) ;
	
	object_item *object = GetObjectFromList ( object_name ) ;
	
	if ( object )
		fprintf ( stdout, "       Found undeclared object '%s'.\n", object -> name ) ;
	
	fprintf ( stdout, "       Check the '%s' file and declare one!\n", pkg_filename ) ;
	if ( object && ! be_strict )
	{
		fprintf ( stdout, "       Trying to continue anyway...\n" ) ;
		return object ;
	}
	fprintf ( stdout, "       Quitting...\n" ) ;
	exit ( 1 ) ;
	
	return NULL ;
}

Pax::object_item *Pax::GetObjectFromList ( char *object_name )
{
	for ( unsigned int pos = 0 ; pos < object_list -> Count () ; pos ++ )
	{
		object_item *object = object_list -> ItemPointer ( pos ) ;
		
		if ( strcmp ( object -> name, object_name ) == 0 )
			return object ;
	}
	
	return NULL ;
}

KrImNode *Pax::GetNode ( Pax::object_item *item, char *object_name )
{
	KrImNode *node = NULL ;
	
	if ( item == NULL )
	{
		fprintf ( stdout, "Can't get node from non-existent object '%s'! Quitting...\n",
					object_name ) ;
		exit ( 1 ) ;
	}
	
	char type [ 256 ] ;
	sprintf ( type, "%s", item -> type == object_type_sprite ? "sprite" :
				item -> type == object_type_node ? "node" :
				item -> type == object_type_textbox ? "textbox" :
				"unsuitable" ) ;
	
	if ( strcmp ( type, "unsuitable" ) == 0 )
	{
		fprintf ( stdout, "Can't get node from unsuitable object '%s'!\n", object_name ) ;
		fprintf ( stdout, "Has to be one of node, sprite or textbox. Quitting...\n" ) ;
		exit ( 1 ) ;
	}
	
	if ( item -> type == object_type_node )
		node = (KrImNode*) item -> data ;
	else if ( item -> type == object_type_sprite )
		node = ( (Sprite*) item -> data ) -> GetSprite () ;
	else if ( item -> type == object_type_textbox )
		node = (KrTextBox*) item -> data ;
	else
		fprintf ( stdout, "Unkown type at line %d in %s !! Contact author!!\n",
					__LINE__, __FILE__ ) ;
	
	return node ;
}

int   Pax::GetIntVar ( char *object_name )
{
	//fprintf ( stdout, "GetIntVar %s\n", object_name ) ;
	return (int) GetDeclaredObjectFromList ( object_name ) -> data ;
}

char *Pax::GetStringVar ( char *object_name )
{
	//fprintf ( stdout, "GetStringVar %s\n", object_name ) ;
	return (char*) GetDeclaredObjectFromList ( object_name ) -> data ;
}

char  Pax::GetCharFromStringVar ( char *object_name, int pos )
{
	//fprintf ( stdout, "GetCharFromStringVar %s\n", object_name ) ;
	return ( (char*) GetDeclaredObjectFromList ( object_name ) -> data ) [ pos ] ;
}

GlDynArray < std::string > *Pax::GetArgList ( char *string )
{
	GlDynArray < std::string > *bucket = new GlDynArray < std::string > () ;
	
	bucket -> Clear () ;
	unsigned int pos = 0 ;
	
	bool marks_found = FALSE ;
	
	//fprintf ( stdout, "args =" ) ;
	
	char *copy = (char*) malloc ( sizeof (char) * 4096 ) ;
	sprintf ( copy, "%s", string ) ;
	
	while ( pos < strlen ( copy ) )
	{
		std::string arg ;
		for ( ; pos < strlen ( copy ) ; pos++ )
		{
			if ( copy [ pos ] == '"' ) marks_found = ! marks_found ;
			if ( copy [ pos ] == ',' && ! marks_found ) break ;
			arg += copy [ pos ] ;
		}
		arg += '\0' ;
		if ( pos < strlen ( copy ) )
			copy [ pos ] = ' ' ;
		
		arg = StripMarks ( StripHeadAndTail ( (char*) arg . c_str () ) ) ;
		
		bucket -> PushBack ( arg ) ;
		
		//fprintf ( stdout, " '%s'", (char*) bucket [ bucket . Count () - 1 ] . c_str () ) ;
	}
	
	//fprintf ( stdout, "\n" ) ;
	
	free ( copy ) ;
	
	return bucket ;
}

void Pax::DrawUserDefinedObjects ()
{	// now draw objects which are not declared to be managed by the engine...
	
	ExecuteScript ( "event" ) ;
	
	for ( unsigned int pos = 0 ; pos < object_list -> Count () ; pos ++ )
	{
		object_item *object = object_list -> ItemPointer ( pos ) ;
		
		// update sprites
		if ( object -> type == object_type_sprite &&
				strcmp ( object -> declared_as, "" ) == 0 )
			( (Sprite*) object -> data ) -> Draw () ;
		// or rain
		if ( object -> type == object_type_rain &&
				strcmp ( object -> declared_as, "" ) == 0 )
			( (Rain*) object -> data ) -> Draw () ;
	}
}

void Pax::ExecuteScript ( char *scriptname )
{
	GlDynArray < std::string > *script = NULL ;
	
	if      ( strcmp ( scriptname, "start"    ) == 0 ) script = script_start    ;
	else if ( strcmp ( scriptname, "ready"    ) == 0 ) script = script_ready    ;
	else if ( strcmp ( scriptname, "go"       ) == 0 ) script = script_go       ;
	else if ( strcmp ( scriptname, "pause"    ) == 0 ) script = script_pause    ;
	else if ( strcmp ( scriptname, "unpause"  ) == 0 ) script = script_unpause  ;
	else if ( strcmp ( scriptname, "quitting" ) == 0 ) script = script_quitting ;
	else if ( strcmp ( scriptname, "quit"     ) == 0 ) script = script_quit     ;
	else if ( strcmp ( scriptname, "event"    ) == 0 ) script = script_event    ;
	else
	{
		fprintf ( stdout, "Warning: Script '%s' does not exist!\n", scriptname ) ;
		return ;
	}
	
	if ( script -> Count () == 0 ) return ;
	
	if ( be_verbose )
		fprintf ( stdout, "Executing %s script...\n", scriptname ) ;
	
	for ( unsigned int line = 0 ; line < script -> Count () ; line++ )
	{
		char text [ 4096 ] ;
		
		sprintf ( text, "%s", (char*) script -> Item ( line ) . c_str () ) ;
		
		if ( be_verbose )
			fprintf ( stdout, "%d: %s\n", line, text ) ;
		
		// get actual ticks from sdl...
		GetObjectFromList ( "sdl_ticks" ) -> data = (int*) SDL_GetTicks () ;
		
		GlDynArray < std::string > *args = new GlDynArray < std::string > () ;
		
		args -> Clear () ;
		
		// get arg list if this should be the eventscript
		if ( strcmp ( scriptname, "event" ) == 0 )
			args = GetArgList ( text ) ;
		
		bool do_action = true ;
		
		if ( args -> Count () > 1 )
		{	// first check all conditions
			for ( unsigned int arg_num = 0 ; arg_num < args -> Count () - 1 ; arg_num++ )
			{
				char *line = (char*) args -> Item ( arg_num ) . c_str () ;
				char *object_name = GetObjectName ( line ) ;
				char *object_action = GetObjectAction ( line ) ;
				
				if ( strcmp ( object_action, "musicplaying" ) == 0 )
				{
					Audio_Music *music = (Audio_Music*) GetObjectFromList ( object_name ) -> data ;
					
					if ( do_action != false )
						do_action = ( music == NULL ) || ( music -> MusicPlaying () ) ;
					
					if ( be_verbose )
						fprintf ( stdout, "musicplaying = %s\n", do_action ? "true" : "false" ) ;
				}
				else if ( strcmp ( object_action, "musicstopped" ) == 0 )
				{
					Audio_Music *music = (Audio_Music*) GetObjectFromList ( object_name ) -> data ;
					
					if ( do_action != false )
						do_action = ( music != NULL ) && ( music -> MusicFinished () ) ;
					
					if ( be_verbose )
						fprintf ( stdout, "musicstopped = %s\n", do_action ? "true" : "false" ) ;
				}
				else if ( strcmp ( object_action, "hasvar" ) == 0 )
				{
					object_item *tmp_object = GetObjectFromList ( object_name ) ;
					
					// get value from current condition
					GlDynArray < std::string > *object_arglist = GetObjectArgList ( line ) ;
					int value = object_arglist -> Count () > 0 ?
								CalcValueFromString ( (char*) object_arglist -> Item ( 0 ) . c_str () ) : 0 ;
					
					int tmp_var = tmp_object ? (int) tmp_object -> data : 0 ;
					
					if ( do_action != false )
						do_action = ( tmp_object != NULL ) &&
									( object_arglist -> Count () > 0 ) &&
									( tmp_var == value ) ;
					
					if ( be_verbose )
						fprintf ( stdout, "hasvar ( %d == %d ) = %s ; action = %s\n",
									tmp_var, value,
									tmp_var == value ? "true" : "false",
									do_action ? "true" : "false" ) ;
					
					object_arglist -> Clear () ;
					delete object_arglist ;
				}
				else if ( strcmp ( object_action, "isgreaterthan" ) == 0 )
				{
					object_item *tmp_object = GetObjectFromList ( object_name ) ;
					
					// get value from current condition
					GlDynArray < std::string > *object_arglist = GetObjectArgList ( line ) ;
					int value = object_arglist -> Count () > 0 ?
								CalcValueFromString ( (char*) object_arglist -> Item ( 0 ) . c_str () ) : 0 ;
					
					if ( do_action != false )
						do_action = ( tmp_object != NULL ) &&
									( object_arglist -> Count () > 0 ) &&
									( (int) tmp_object -> data > value ) ;
					
					if ( be_verbose )
						fprintf ( stdout, "isgreaterthan = %s\n", do_action ? "true" : "false" ) ;

					object_arglist -> Clear () ;
					delete object_arglist ;
				}
				else if ( strcmp ( object_action, "issmallerthan" ) == 0 )
				{
					object_item *tmp_object = GetObjectFromList ( object_name ) ;
					
					// get value from current condition
					GlDynArray < std::string > *object_arglist = GetObjectArgList ( line ) ;
					int value = object_arglist -> Count () > 0 ?
								CalcValueFromString ( (char*) object_arglist -> Item ( 0 ) . c_str () ) : 0 ;
					
					if ( do_action != false )
						do_action = ( tmp_object != NULL ) &&
									( object_arglist -> Count () > 0 ) &&
									( (int) tmp_object -> data < value ) ;
					
					if ( be_verbose )
						fprintf ( stdout, "issmallerthan = %s\n", do_action ? "true" : "false" ) ;
					
					object_arglist -> Clear () ;
					delete object_arglist ;
				}
				else
				{
					fprintf ( stdout, "Error: Unknown condition '%s' in event-script.\n", object_action ) ;
					fprintf ( stdout, "       Line looks like this:\n       %s\n", text ) ;
					free ( object_name ) ;
					free ( object_action ) ;
					//free ( line ) ;
					exit ( 1 ) ;
				}
				
				free ( object_name ) ;
				free ( object_action ) ;
				//free ( line ) ;
			}
		}
		
		if ( strcmp ( scriptname, "event" ) == 0 )
			sprintf ( text, "%s", (char*) args -> Item ( args -> Count () - 1 ) . c_str () ) ;
		
		if ( do_action )
		{	// if not eventscript... or just an action-line without condition...
			char *object_name = GetObjectName ( text ) ;
			char *object_action = GetObjectAction ( text ) ;
			GlDynArray < std::string > *object_arglist = GetObjectArgList ( text ) ;
			
			//fprintf ( stdout, "'%s' : '%s'\n", object_name, object_action )  ;
			
			InterpretLine ( object_name, object_action, object_arglist ) ;
			
			free ( object_name ) ;
			free ( object_action ) ;
			object_arglist -> Clear () ;
			delete object_arglist ;
		}
		
		if ( strcmp ( scriptname, "event" ) == 0 && args -> Count () < 1 )
			exit ( 1 ) ;
		
		args -> Clear () ;
		delete args ;
	}
	
	if ( be_verbose )
		fprintf ( stdout, "Done.\n" ) ;
}

int Pax::FindObjectInList ( char *object_name )
{
	for ( unsigned int pos = 0 ; pos < object_list -> Count () ; pos++ )
		if ( strcmp ( object_list -> ItemPointer ( pos ) -> name, object_name ) == 0 )
			return pos ;
	
	return -1 ;
}

void Pax::FreeObject ( object_item *object )
{
	int childs = object -> child_list -> Count () ;
	
	if ( object == NULL || childs ) return ;
	
	if ( be_verbose )
		fprintf ( stdout, "Removing object '%s'... ", object -> name ) ;
	
	switch ( object -> type )
	{
		case object_type_none:
		{
			// nothing to do here...
		}
		break ;
		case object_type_var:
		{
			if ( object -> writeable && object -> data != NULL )
			{
				if ( be_verbose )
					fprintf ( stdout, "freeing var... " ) ;
				object -> data = NULL ;
			}
		}
		break ;
		case object_type_string:
		{
			if ( object -> writeable && object -> data != NULL )
			{
				if ( be_verbose )
					fprintf ( stdout, "freeing string... " ) ;
				free ( object -> data ) ;
			}
		}
		break ;
		case object_type_vault:
		{
			KrResourceVault *Vault = (KrResourceVault*) object -> data ;
			if ( be_verbose )
				fprintf ( stdout, "unref vault... " ) ;
			if ( file_cache . FreeDataFromCache ( Vault ) == 0 )
			{
				if ( be_verbose )
					fprintf ( stdout, "deleting... " ) ;
				delete Vault ;
			}
		}
		break ;
		case object_type_node:
		{
			if ( be_verbose )
				fprintf ( stdout, "deleting node... " ) ;
			if ( ! object -> data )
				fprintf ( stdout, "non-existent node ! BAD BAD BAD !!!\n" ) ;
			else
			{
				KrImNode *node = (KrImNode*) object -> data ;
				if ( ! node )
					fprintf ( stdout, "couldn't get node ! BAD BAD BAD !!!\n" ) ;
				else
					video . GetEngine () -> Tree () -> DeleteNode ( node ) ;
			}
		}
		break ;
		case object_type_sprite:
		{
			if ( be_verbose )
				fprintf ( stdout, "freeing sprite... " ) ;
			if ( object -> data )
			{
				( (Sprite*) object -> data ) -> Free () ;
				delete (Sprite*) object -> data ;
			}
			else
				fprintf ( stdout, "found sprite without data ! BAD BAD BAD !!! " ) ;
		}
		break ;
		case object_type_font:
		{
			if ( be_verbose )
				fprintf ( stdout, "deleting font... " ) ;
			// don't do this! this would harm the vault!!!
			//delete (KrFontResource*) object -> data ;
		}
		break ;
		case object_type_textbox:
		{
			if ( be_verbose )
				fprintf ( stdout, "deleting textbox... " ) ;
			if ( ! object -> data )
				fprintf ( stdout, "non-existent node ! BAD BAD BAD !!!\n" ) ;
			else
			{
				KrImNode *node = (KrImNode*) object -> data ;
				if ( ! node )
					fprintf ( stdout, "couldn't get node ! BAD BAD BAD !!!\n" ) ;
				else
					video . GetEngine () -> Tree () -> DeleteNode ( node ) ;
			}
		}
		break ;
		case object_type_music:
		{
			if ( be_verbose )
				fprintf ( stdout, "deleting music... " ) ;
			( (Audio_Music*) object -> data ) -> UnloadMusic () ;
			delete (Audio_Music*) object -> data ;
		}
		break ;
		case object_type_rain:
		{
			if ( be_verbose )
				fprintf ( stdout, "deleting rain... " ) ;
			( (Rain*) object -> data ) -> Quit () ;
			delete (Rain*) object -> data ;
		}
		break ;
	}
	
	for ( unsigned int pp = 0 ; pp < object_list -> Count () ; pp++ )
	{
		object_item *parent = object_list -> ItemPointer ( pp ) ;

		if ( parent -> child_list -> Count () == 0 )
			continue ;
		
		for ( unsigned int pos = 0 ; pos < parent -> child_list -> Count () ; pos++ )
		{
			char child [ 256 ] ;
			sprintf ( child, "%s", parent -> child_list -> ItemPointer ( pos ) -> c_str () ) ;
			
			if ( strcmp ( child, object -> name ) == 0 )
				parent -> child_list -> Remove ( pos ) ;
		}
	}
	
	object -> child_list -> Clear () ;
	delete object -> child_list ;
	object -> child_list = NULL ;
	
	if ( be_verbose )
		fprintf ( stdout, "done.\n" ) ;
}

void Pax::FreeObjects ()
{
	//be_verbose = true ;
	if ( be_verbose )
		fprintf ( stdout, "Freeing objects...\n" ) ;
	
	int wait_time = 0 ;
	unsigned int objects_left = object_list -> Count () ;
	while ( object_list -> Count () && wait_time < 100 )
	{
		if ( be_verbose && objects_left != object_list -> Count () )
			fprintf ( stdout, "Objects left: %d\n", objects_left ) ;
		
		bool object_killed = false ;
		for ( unsigned int pos = 0 ; pos < objects_left && object_killed == false ; pos ++ )
		{
			object_item *object = object_list -> ItemPointer ( pos ) ;
			
			if ( object != NULL && object -> child_list -> Count () == 0 )
			{
				FreeObject ( object ) ;
				object_list -> Remove ( pos ) ;
				object_killed = true ;
			}
		}
		
		if ( object_list -> Count () == objects_left )
			wait_time++ ;
		else
			wait_time-- ;
		
		objects_left = object_list -> Count () ;
	}
	
	if ( wait_time >= 100 )
	{
		fprintf ( stdout, "Warning: Backed out with %d objects left to kill!\n",
					object_list -> Count () ) ;
		
		for ( unsigned int pos = 0 ; pos < object_list -> Count () ; pos++ )
		{
			object_item *object = object_list -> ItemPointer ( pos ) ;
			fprintf ( stdout, "object name = '%s'\n", object -> name ) ;
			fprintf ( stdout, "     parent = '%s'\n", object -> parent ) ;
			/*if ( object -> child_list -> Count () )
				for ( unsigned int pos2 = 0 ; pos2 < object -> child_list -> Count () ; pos2++ )
				{
					fprintf ( stdout, "      child = '%s'\n",
								object -> child_list -> ItemPointer ( pos2 ) -> c_str () ) ;
				}*/
			fprintf ( stdout, "\n" ) ;
		}
	}
	
	object_list -> Clear () ; delete object_list ;
	
	script_start    -> Clear () ; delete script_start    ;
	script_ready    -> Clear () ; delete script_ready    ;
	script_go       -> Clear () ; delete script_go       ;
	script_pause    -> Clear () ; delete script_pause    ;
	script_unpause  -> Clear () ; delete script_unpause  ;
	script_quitting -> Clear () ; delete script_quitting ;
	script_quit     -> Clear () ; delete script_quit     ;
	script_event    -> Clear () ; delete script_event    ;
	
	if ( scan_mode == false && video . GetC1ModeActive () )
		video . SetDisplayC1Layer ( true ) ;
}

char *Pax::CalcString ( char *string )
{
	int open = 0, close = 0 ;
	int length = strlen ( string ) ;
	
	{	// test if there is more than one number...
		int test_var = atoi ( string ) ;
		char test_var_string [ strlen ( string ) ] ;
		sprintf ( test_var_string, "%d", test_var ) ;
		if ( strlen ( test_var_string ) == strlen ( string ) )
			return string ;
	}
	
	//fprintf ( stdout, "got string: '%s'\n", string ) ;
	
	if ( index ( string, '(' ) )
	{
		unsigned int start = 0 ;
		unsigned int end = 0 ;
		for ( unsigned int pos = 0 ; pos < strlen ( string ) ; pos++ )
		{
			if ( string [ pos ] == '(' )
			{
				start = pos ;
				break ;
			}
		}
				
		char *line = StripHeadAndTail ( index ( string, '(' ) ) ;
		for ( unsigned int pos = 0 ; pos < strlen ( line ) ; pos++ )
		{
			if ( line [ pos ] == '(' ) open++ ;
			if ( line [ pos ] == ')' ) close++ ;
			
			if ( open == close )
			{
				end = start + pos ;
				
				//fprintf ( stdout, "     pos = %d\n", pos ) ;
				//fprintf ( stdout, "     start = %d ; end = %d\n", start, end ) ;
				
				char *tmp = (char*) malloc ( sizeof (char) * strlen (line) ) ;
				strcpy ( tmp, line + 1 ) ;
				tmp [ pos - 2 ] = '\0' ;
				
				sprintf ( tmp, "%s", StripHeadAndTail ( tmp ) ) ;
				
				//fprintf ( stdout, "     part looks like this: '%s'\n", tmp ) ;
				
				tmp = CalcString ( tmp ) ;
				
				//fprintf ( stdout, "     replacing brackets...\n" ) ;
				
				strcpy ( line, tmp ) ;
				
				//fprintf ( stdout, "     line looks like this: '%s'\n", line ) ;
				
				for ( int c = end ; c < length ; c++ )
				{
					string [ c - pos + strlen ( tmp ) ] = string [ c + 1 ] ;
				}
				string [ length - pos + strlen ( tmp ) ] = '\0' ;
				
				//fprintf ( stdout, "     string looks like this: '%s'\n", string ) ;
				
				//exit ( 0 ) ;
			}
		}
	}
	//else
	{
		char line [ strlen ( string ) ] ;
		int  lpos = 0 ;
		
		for ( unsigned int pos = 0 ; pos < strlen ( string ) ; pos++ )
		{
			if ( string [ pos ] != ' ' )
				line [ lpos++ ] = string [ pos ] ;
		}
		line [ lpos ] = '\0' ;
		
		//fprintf ( stdout, "     got plain calculation: '%s'\n", line ) ;
		
		int  last_char_pos = 0 ;
		char last_operation = 0 ;
		int a = -1, b = -1, c = -1 ;
		
		for ( unsigned int pos = 0 ; pos <= strlen ( line ) ; pos++ )
		{
			if ( a == -1 && pos > 0 &&
					( line [ pos ] == '*' || line [ pos ] == '/' ||
					  line [ pos ] == '+' || line [ pos ] == '-' ) &&
					( line [ pos+1 ] != '*' || line [ pos+1 ] != '/' ||
					  line [ pos+1 ] != '+' || line [ pos+1 ] != '-' ) )
			{
				char var [ 4096 ] ;
				sprintf ( var, "%s", line + last_char_pos ) ;
				var [ pos ] = '\0' ;
				a = atoi ( var ) ;
				//fprintf ( stdout, "     var_a = %d\n", a ) ;
				last_char_pos = pos + 1 ;
				last_operation = line [ pos ] ;
			}
			
			if ( a != -1 && b == -1 &&
/*					( ( line [ pos ] == '*' && line [ pos + 1 ] != '*' ) ||
					  ( line [ pos ] == '/' && line [ pos + 1 ] != '/' ) ||
					  ( line [ pos ] == '+' && line [ pos + 1 ] != '+' ) ||
					  ( line [ pos ] == '-' && line [ pos + 1 ] != '-' ) ) )*/
					( line [ pos ] == '*' || line [ pos ] == '/' ||
					  line [ pos ] == '+' || line [ pos ] == '-' ) &&
					( line [ pos+1 ] != '*' && line [ pos+1 ] != '/' &&
					  line [ pos+1 ] != '+' && line [ pos+1 ] != '-' ) )
			{
				//fprintf ( stdout, "     1 last char pos = %d\n", last_char_pos ) ;
				char var [ 4096 ] ;
				sprintf ( var, "%s", line + last_char_pos ) ;
				b = atoi ( var ) ;
				sprintf ( var, "%d", b ) ;
				//fprintf ( stdout, "     var_b = %d\n", b ) ;
				
				int modifier = ( b < 0 ? 1 : 0 ) ;
				
				last_char_pos = pos + strlen ( var ) + 1 + !modifier ;
				//fprintf ( stdout, "     2 last char pos = %d\n", last_char_pos ) ;
				
				if ( last_operation == 0 || a == c )
					last_operation = line [ pos - modifier ] ;
				
				if ( last_operation == '*' )
					c = a * b ;
				if ( last_operation == '/' )
					c = a / b ;
				if ( last_operation == '+' )
					c = a + b ;
				if ( last_operation == '-' )
					c = a - b ;
				
				//fprintf ( stdout, "     calc %d %c %d = %d\n", a, last_operation, b, c ) ;
				
				a = c ;
				b = -1 ;
			}
		}
		
		if ( c == -1 ) c = a ;
		
		//fprintf ( stdout, "     summary: %s = %d\n", line, c ) ;
		
		sprintf ( string, "%d", c ) ;
	}
	
	return string ;
}

int Pax::CalcValueFromString ( char *string )
{
	int value = atoi ( string ) ;
	unsigned int string_length = strlen ( string ) ;
	
	if ( string == NULL || strlen ( string ) < 1 )
	{
		if ( be_verbose )
		{
			fprintf ( stdout, "Warning: Returning a value of 0 because of empty string in\n" ) ;
			fprintf ( stdout, "         CalcValueFromString ( char *string ) method.\n" ) ;
		}
		return 0 ;
	}
	
	// check if we have a plain value or what...
	bool has_chars = false ;	
	for ( unsigned int pos = 0 ; pos < string_length ; pos++ )
	{
		if ( ( string [ pos ] >= 'a' && string [ pos ] <= 'z' ) ||
				( string [ pos ] >= 'A' && string [ pos ] <= 'Z' ) ||
				string [ pos ] == '_' )
		{
			has_chars = true ;
		}
	}
	if ( has_chars == false )
	{
		value = atoi ( CalcString ( string ) ) ;
		//fprintf ( stdout, "     Returning a value of %d\n     ", value ) ;
		return value ;
	}
	
	
	//fprintf ( stdout, "CalcValueFromString got string '%s'\n", string ) ;
	
	// replace variables in string first !
	
	//fprintf ( stdout, "     Searching for variables to replace... " ) ;
	
	bool word_started = false ;
	
	std::string word ;
	
	for ( unsigned int pos = 0 ; pos <= string_length ; pos++ )
	{
		if ( pos < strlen ( string ) &&
				( ( string [ pos ] >= 'a' && string [ pos ] <= 'z' ) ||
				( string [ pos ] >= 'A' && string [ pos ] <= 'Z' ) ||
				string [ pos ] == '_' ) )
		{
			if ( word_started == false )
				word_started = true ;
			
			word += string [ pos ] ;
		}
		else
		{
			if ( word_started == true )
			{
				int var_value = 0 ;
				word_started = false ;
				
				if ( be_verbose )
					fprintf ( stdout, "Got var called '%s'\n",
								(char*) word . c_str () ) ;
				
				// get the value of the var
				//fprintf ( stdout, "     Searching for var in list... " ) ;
				
				object_item *item = GetObjectFromList ( (char*) word. c_str () ) ;
				
				if ( item == NULL )
				{
					fprintf ( stdout, "Var '%s' not found in list. Value set to 0.\n",
								(char*) word. c_str () ) ;
					var_value = 0 ;
				}
				else if ( item -> type != object_type_var )
				{
					fprintf ( stdout, "Var '%s' found but seems not to be a value. Value set to 0.\n",
								(char*) word. c_str () ) ;
				}
				else
				{
					var_value = (int) item -> data ;
					if ( be_verbose )
						fprintf ( stdout, "Var '%s' found in list. Value is %d\n",
									(char*) word. c_str (), var_value ) ;
				}
				
				// now replace the var in our string
				char var_string [ 4096 ] ;
				sprintf ( var_string, "%d", var_value ) ;
				
				// go to the beginning of our var in string
				pos -= strlen ( (char*) word . c_str () ) ;
				
				// now put the var_string in our string
				strcpy ( string + pos, var_string ) ;
				
				// now get the rest to the right position
				int var_string_length = strlen ( var_string ) ;
				int word_length = strlen ( (char*) word . c_str () ) ;
				for ( unsigned int c = pos + strlen ( var_string ) ; c < string_length ; c++ )
				{
					string [ c ] = string [ c - var_string_length + word_length ] ;
				}
				
				// now correct the length and be happy
				string_length -= word_length ;
				string_length += var_string_length ;
				
				word = "" ;
			}
		}
	}
	
	//fprintf ( stdout, "Done.\n" ) ;
	
	//fprintf ( stdout, "     Now string looks like this '%s'\n", string ) ;
	
	// now calculate our string
	
	// but first do some basic consistency check...
	int open = 0, close = 0 ;
	for ( unsigned int pos = 0 ; pos < string_length ; pos++ )
	{
		if ( string [ pos ] == '(' ) open++ ;
		if ( string [ pos ] == ')' ) close++ ;
		
		if ( close > open )
		{
			fprintf ( stdout, "     Error: Closed more brackets than opened. Quitting...\n" ) ;
			exit ( 1 ) ;
		}
	}
	if ( open != close )
	{
		fprintf ( stdout, "     Error: Opened %d but closed %d brackets. Quitting...\n",
					open, close ) ;
		exit ( 1 ) ;
	}
	
	value = atoi ( CalcString ( string ) ) ;
	
	//fprintf ( stdout, "     Returning a value of %d\n     ", value ) ;
	
	// return the summary
	return value ;
}

void Pax::InterpretLine ( char *object_name, char *object_action,
							GlDynArray < std::string > *object_arglist )
{
	object_item *object ;
	
	object = GetObjectFromList ( object_name ) ;
	
	int   arg_count = object_arglist -> Count () ;
	char *arg [ arg_count ] ;
	for ( int i = 0 ; i < arg_count ; i++ )
		arg [ i ] = (char*) object_arglist -> Item ( i ) . c_str () ;
	
	if ( strcmp ( object_name, "engine" ) == 0 )
	{	// engine-internals
		if ( strcmp ( object_action, "verbose" ) == 0 )
		{
			if ( strcmp ( arg [ 0 ], "true" ) == 0 )
				be_verbose = true ;
			else if ( strcmp ( arg [ 0 ], "false" ) == 0 )
				be_verbose = false ;
			else
				fprintf ( stdout, "Warning: engine verbosity can only be one of true or false.\n" ) ;
		}
		else if ( strcmp ( object_action, "strict" ) == 0 )
		{
			if ( strcmp ( arg [ 0 ], "true" ) == 0 )
				be_strict = true ;
			else if ( strcmp ( arg [ 0 ], "false" ) == 0 )
				be_strict = false ;
			else
				fprintf ( stdout, "Warning: engine strictness can only be one of true or false.\n" ) ;
		}
		else if ( strcmp ( object_action, "c1layer" ) == 0 )
		{
			if ( strcmp ( arg [ 0 ], "true" ) == 0 )
				video . SetDisplayC1Layer ( true ) ;
			else if ( strcmp ( arg [ 0 ], "false" ) == 0 )
				video . SetDisplayC1Layer ( false ) ;
			else
				fprintf ( stdout, "Warning: value if c1-layer should be shown can only be one of true or false.\n" ) ;
		}
		else if ( strcmp ( object_action, "startscript" ) == 0 )
		{
			std::string line = arg [ 0 ] ;
			if ( be_verbose )
				fprintf ( stdout, "Adding line to start script: '%s'\n", line . c_str () ) ;
			script_start -> PushBack ( line ) ;
		}
		else if ( strcmp ( object_action, "readyscript" ) == 0 )
		{
			std::string line = arg [ 0 ] ;
			if ( be_verbose )
				fprintf ( stdout, "Adding line to get_ready script: '%s'\n", line . c_str () ) ;
			script_ready -> PushBack ( line ) ;
		}
		else if ( strcmp ( object_action, "goscript" ) == 0 )
		{
			std::string line = arg [ 0 ] ;
			if ( be_verbose )
				fprintf ( stdout, "Adding line to go script: '%s'\n", line . c_str () ) ;
			script_go -> PushBack ( line ) ;
		}
		else if ( strcmp ( object_action, "pausescript" ) == 0 )
		{
			std::string line = arg [ 0 ] ;
			if ( be_verbose )
				fprintf ( stdout, "Adding line to pause script: '%s'\n", line . c_str () ) ;
			script_pause -> PushBack ( line ) ;
		}
		else if ( strcmp ( object_action, "unpausescript" ) == 0 )
		{
			std::string line = arg [ 0 ] ;
			if ( be_verbose )
				fprintf ( stdout, "Adding line to unpause script: '%s'\n", line . c_str () ) ;
			script_unpause -> PushBack ( line ) ;
		}
		else if ( strcmp ( object_action, "quittingscript" ) == 0 )
		{
			std::string line = arg [ 0 ] ;
			if ( be_verbose )
				fprintf ( stdout, "Adding line to quitting script: '%s'\n", line . c_str () ) ;
			script_quitting -> PushBack ( line ) ;
		}
		else if ( strcmp ( object_action, "quitscript" ) == 0 )
		{
			std::string line = arg [ 0 ] ;
			if ( be_verbose )
				fprintf ( stdout, "Adding line to quit script: '%s'\n", line . c_str () ) ;
			script_quit -> PushBack ( line ) ;
		}
		else if ( strcmp ( object_action, "eventscript" ) == 0 )
		{
			std::string line ;
			line = "\"" ;
			line += arg [ 0 ] ;
			line += "\"" ;
			for ( int i = 1 ; i < arg_count ; i++ )
			{
				line += ", \"" ;
				line += arg [ i ] ;
				line += "\"" ;
			}
			if ( be_verbose )
				fprintf ( stdout, "Adding line to event script: '%s'\n", line . c_str () ) ;
			script_event -> PushBack ( line ) ;
		}
		else if ( strcmp ( object_action, "include" ) == 0 )
		{
			char line [ 1024 ] ;
			sprintf ( line, "%s/%s", pkg_dir, arg [ 0 ] ) ;
			if ( be_verbose )
				fprintf ( stdout, "Including script '%s'...\n", line ) ;
			ReadPax ( line ) ;
		}
		else
			fprintf ( stdout, "Warning: Action '%s' is unknown to engine.\n", object_action ) ;
	}
	else if ( strcmp ( object_action, "setvar" ) == 0 )
	{
		int *value ;
		value = (int*) CalcValueFromString ( arg [ 0 ] ) ;
		
		if ( object )
		{
			if ( be_verbose )
				fprintf ( stdout, "Changing value of var '%s' to '%d'\n", object_name, (int) value ) ;
			if ( object -> writeable )
			{
				object -> data = value ;
			}
			else
			{
				fprintf ( stdout, "Warning: Can't change value of read-only var '%s' to '%d'!\n", object_name, (int) value ) ;
			}
		}
		else
		{
			if ( be_verbose )
				fprintf ( stdout, "Creating new var '%s' with value '%d'\n", object_name, (int) value ) ;
			if ( ObjectListNewEntryFull ( object_type_var, object_name, value, TRUE ) )
				sprintf ( GetObjectFromList ( object_name ) -> declared_as, "%s", object_name ) ;
			else
			{
				fprintf ( stdout, "Warning: Couldn't create new var '%s' with value '%d'!\n",
							object_name, (int) value ) ;
			}
		}
	}
	else if ( strcmp ( object_action, "setstring" ) == 0 )
	{
		char *value = (char*) malloc ( sizeof (char) * 4096 ) ;
		value [ 0 ] = '\0' ;
		
		if ( arg_count > 1 )
		{
			for ( int i = 0 ; i < arg_count ; i++ )
				sprintf ( value, "%s%s\"%s\"", value,
							( i > 0 && i < arg_count ) ? "," : "",
							arg [ i ] ) ;
		}
		else
			sprintf ( value, "%s", arg [ 0 ] ) ;
		
		if ( object )
		{
			if ( be_verbose )
				fprintf ( stdout, "Changing value of string '%s' to '%s'\n", object_name, value ) ;
			if ( object -> writeable )
			{
				if ( object -> data )
				{
					if ( be_verbose )
						fprintf ( stdout, "     Old value was '%s' ; new will be '%s'\n",
								(char*) object -> data, value ) ;
					free ( object -> data ) ;
				}
				object -> data = value ;
			}
			else
			{
				fprintf ( stdout, "Warning: Can't change value of read-only string '%s' to '%s'!\n", object_name, value ) ;
				free ( value ) ;
			}
		}
		else
		{
			if ( be_verbose )
				fprintf ( stdout, "Creating new string '%s' with value '%s'\n", object_name, value ) ;
			if ( ! ObjectListNewEntryFull ( object_type_string, object_name, value, TRUE, TRUE ) )
			{
				fprintf ( stdout, "Warning: Couldn't create new string '%s' with value '%s'!\n",
							object_name, value ) ;
				free ( value ) ;
			}
		}
	}
	else if ( strcmp ( object_action, "newvaultfromfile" ) == 0 )
	{	// add a KrResourceVault
		if ( object )
		{
			fprintf ( stdout, "Vault '%s' already created! Quitting...\n", object_name ) ;
			exit ( 1 ) ;
		}
		
		if ( be_verbose )
			fprintf ( stdout, "Creating vault called '%s'\n", object_name ) ;
		
		char filename [ 4096 ] ;
		sprintf ( filename , "%s/%s", pkg_dir, arg [ 0 ] ) ;
		
		KrResourceVault *Vault = (KrResourceVault*) file_cache . GetFromCache ( filename ) ;
		
		if ( Vault == NULL )
		{
			Vault = new KrResourceVault () ;
			
			// load our gfx pack
			if ( !Vault->LoadDatFile ( filename ) )
			{
				char altfname [ 4096 ] ;
				sprintf ( altfname , "gfx/%s", arg [ 0 ] ) ;
				
				if ( be_verbose )
				{
					fprintf ( stdout, "Warning: Couldn't load vault '%s' in '%s'!\n",
								arg [ 0 ], pkg_dir ) ;
					fprintf ( stdout, "         Trying '%s' as a replacement...\n", altfname ) ;
				}
				if ( !Vault->LoadDatFile ( altfname ) )
				{
					fprintf ( stdout, "Failed! Error loading the '%s' file\n", filename ) ;
					exit ( 255 ) ;
				}
				else
				{
					if ( be_verbose )
						fprintf ( stdout, "         Successfully loaded '%s' as a replacement!\n", altfname ) ;
					sprintf ( filename, "%s", altfname ) ;
				}
			}
		}
		
		file_cache . PutIntoCache ( filename, Vault ) ;
		
		ObjectListNewEntryFull ( object_type_vault, object_name, Vault, FALSE ) ;
	}
	else if ( strcmp ( object_action, "newsprite" ) == 0 )
	{	// add a Sprite
		if ( scan_mode ) return ;
		if ( object )
		{
			fprintf ( stdout, "Sprite '%s' already created! Quitting...\n", object_name ) ;
			exit ( 1 ) ;
		}
		
		if ( be_verbose )
			fprintf ( stdout, "Creating a sprite called '%s'\n", object_name ) ;
		
		KrImNode *node = NULL ;
		
		char vault [ 256 ] ;
		sprintf ( vault, "%s", arg [ 0 ] ) ;
		char name [ 256 ] ;
		sprintf ( name, "%s", arg [ 1 ] ) ;
		char parent [ 256 ] ;
		if ( arg_count > 2 )
		{
			sprintf ( parent, "%s", arg [ 2 ] ) ;
			object_item *item = GetObjectFromList ( parent ) ;
			
			node = GetNode ( item, object_name ) ;
		}
		else
			sprintf ( parent , "%s", "" ) ;
		
		if ( GetObjectFromList ( vault ) == NULL )
		{
			fprintf ( stdout, "Vault '%s' not loaded! Quitting...\n", vault ) ;
			exit ( 1 ) ;
		}
		
		Sprite *tmp_obj ;
		tmp_obj = new Sprite () ;
		tmp_obj -> Init ( (KrResourceVault*) GetObjectFromList ( vault ) -> data, name, node ) ;
		
		ObjectListNewEntryFull ( object_type_sprite, object_name, tmp_obj, TRUE ) ;
		if ( strcmp ( parent, "" ) != 0 )
			ObjectListAddChildToParent ( object_name, parent ) ;
		else
			ObjectListAddChildToParent ( object_name, vault ) ;
		
		char width_string [ 1024 ], height_string [ 1024 ] ;
		sprintf ( width_string , "%s_width",  object_name ) ;
		sprintf ( height_string, "%s_height", object_name ) ;
		ObjectListNewEntryFull ( object_type_var, width_string,
									(int*) tmp_obj -> GetWidth (), true ) ;
		ObjectListNewEntryFull ( object_type_var, height_string,
									(int*) tmp_obj -> GetHeight (), true ) ;
	}
	else if ( strcmp ( object_action, "newnode" ) == 0 )
	{
		if ( scan_mode ) return ;
		if ( object )
		{
			fprintf ( stdout, "Node '%s' already created! Quitting...\n", object_name ) ;
			exit ( 1 ) ;
		}
		
		if ( be_verbose )
			fprintf ( stdout, "Creating a node called '%s'\n", object_name ) ;
		
		KrImNode *node = new KrImNode () ;
		
		ObjectListNewEntryFull ( object_type_node, object_name, node, TRUE ) ;
	}
	else if ( strcmp ( object_action, "newrain" ) == 0 )
	{
		if ( scan_mode ) return ;
		if ( object )
		{
			fprintf ( stdout, "Rain '%s' already created! Quitting...\n", object_name ) ;
			exit ( 1 ) ;
		}
		
		if ( be_verbose )
			fprintf ( stdout, "Creating a rain-object called '%s'\n", object_name ) ;
		
		char parent [ 256 ] ;
		KrImNode *node = NULL ;
		
		if ( arg_count > 0 )
		{
			sprintf ( parent, "%s", arg [ 0 ] ) ;
			object_item *item = GetObjectFromList ( parent ) ;
			
			node = GetNode ( item, object_name ) ;
		}
		
		Rain *rain = new Rain () ;
		
		rain -> Init ( node ) ;
		
		ObjectListNewEntryFull ( object_type_rain, object_name, rain, TRUE ) ;
		
		if ( node != NULL )
			ObjectListAddChildToParent ( object_name, parent ) ;
	}
	else if ( strcmp ( object_action, "newfont" ) == 0 )
	{
		if ( scan_mode ) return ;
		if ( object )
		{
			fprintf ( stdout, "Font '%s' already created! Quitting...\n", object_name ) ;
			exit ( 1 ) ;
		}
		
		if ( be_verbose )
			fprintf ( stdout, "Creating a font called '%s'\n", object_name ) ;
		
		char vault [ 256 ] ;
		sprintf ( vault, "%s", arg [ 0 ] ) ;
		char name [ 256 ] ;
		sprintf ( name, "%s", arg [ 1 ] ) ;
		
		if ( GetObjectFromList ( vault ) == NULL )
		{
			fprintf ( stdout, "Vault '%s' not loaded! Quitting...\n", vault ) ;
			exit ( 1 ) ;
		}
		
		KrFontResource *tmp_obj ;
		tmp_obj = ( (KrResourceVault*) GetObjectFromList ( vault ) -> data ) -> GetFontResource ( name ) ;
		if ( tmp_obj == NULL )
		{
			fprintf ( stdout, "Error loading font '%s' from vault '%s'! Quitting...\n", name, vault ) ;
			exit ( 1 ) ;
		}
		
		ObjectListNewEntryFull ( object_type_font, object_name, tmp_obj, FALSE ) ;
		ObjectListAddChildToParent ( object_name, vault ) ;
	}
	else if ( strcmp ( object_action, "newtextbox" ) == 0 )
	{
		if ( scan_mode ) return ;
		if ( object )
		{
			fprintf ( stdout, "TextBox '%s' already created! Quitting...\n", object_name ) ;
			exit ( 1 ) ;
		}
		
		if ( be_verbose )
			fprintf ( stdout, "Creating a textbox called '%s'\n", object_name ) ;
		
		char fontname [ 256 ] ;
		sprintf ( fontname, "%s", arg [ 0 ] ) ;
		KrFontResource *font = (KrFontResource*) GetObjectFromList ( fontname ) -> data ;
		int width   = CalcValueFromString ( arg [ 1 ] ) ;
		int height  = CalcValueFromString ( arg [ 2 ] ) ;
		int spacing = CalcValueFromString ( arg [ 3 ] ) ;
		KrTextBox::Alignment alignment = KrTextBox::LEFT ;
		if ( arg_count > 4 )
		{
			if ( strcmp ( arg [ 4 ], "left" ) == 0 )
				alignment = KrTextBox::LEFT ;
			else if ( strcmp ( arg [ 4 ], "right" ) == 0 )
				alignment = KrTextBox::RIGHT ;
			else if ( strcmp ( arg [ 4 ], "center" ) == 0 )
				alignment = KrTextBox::CENTER ;
			else
				fprintf ( stdout, "Unknown alignment '%s' - should be one of left, right or center.\n",
							arg [ 3 ] ) ;
		}
		else
			alignment = KrTextBox::LEFT ;
		
		if ( GetObjectFromList ( fontname ) == NULL )
		{
			fprintf ( stdout, "Font '%s' not loaded! Quitting...\n", fontname ) ;
			exit ( 1 ) ;
		}
		
		KrTextBox *tmp_obj = new KrTextBox ( font, width, height, spacing, alignment ) ;
		
		ObjectListNewEntryFull ( object_type_textbox, object_name, tmp_obj, TRUE ) ;
		ObjectListAddChildToParent ( object_name, fontname, false ) ;
	}
	else if ( strcmp ( object_action, "newmusicfromfile" ) == 0 )
	{
		if ( scan_mode ) return ;
		if ( object )
		{
			fprintf ( stdout, "Music '%s' already created! Quitting...\n", object_name ) ;
			exit ( 1 ) ;
		}
		
		if ( be_verbose )
			fprintf ( stdout, "Creating a music object called '%s' from file\n", object_name ) ;
		
		char filename [ 4096 ] ;
		sprintf ( filename , "%s/%s", pkg_dir, arg [ 0 ] ) ;
		
		Audio_Music *music = new Audio_Music () ;
		if ( ! music -> LoadMusic ( filename ) )
		{
			char altfname [ 4096 ] ;
			sprintf ( altfname , "music/%s", arg [ 0 ] ) ;
			
			if ( be_verbose )
			{
				fprintf ( stdout, "Warning: Couln't load music '%s' in '%s'!\n",
							arg [ 0 ], pkg_dir ) ;
				fprintf ( stdout, "         Trying '%s' as a replacement...\n", altfname ) ;
			}
			if ( ! music -> LoadMusic ( altfname ) )
			{
				fprintf ( stdout, "         Failed! Error loading music '%s'!\n", filename ) ;
			}
			else if ( be_verbose )
				fprintf ( stdout, "         Successfully loaded '%s' as a replacement!\n", altfname ) ;
		}
		
		ObjectListNewEntryFull ( object_type_music, object_name, music, FALSE ) ;
	}
	else if ( strcmp ( object_action, "playmusic" ) == 0 )
	{
		if ( scan_mode ) return ;
		if ( !object )
		{
			fprintf ( stdout, "Can't play non-existing music '%s'! Quitting...\n", object_name ) ;
			exit ( 1 ) ;
		}
		
		if ( be_verbose )
			fprintf ( stdout, "Play a music called '%s'\n", object_name ) ;
		
		int loops = CalcValueFromString ( arg [ 0 ] ) ;
		
		( (Audio_Music*) object -> data ) -> PlayMusic ( loops ) ;
	}
	else if ( strcmp ( object_action, "stopmusic" ) == 0 )
	{
		if ( scan_mode ) return ;
		if ( !object )
		{
			fprintf ( stdout, "Can't stop non-existing music '%s'! Quitting...\n", object_name ) ;
			exit ( 1 ) ;
		}
		
		if ( be_verbose )
			fprintf ( stdout, "Stop a music called '%s'\n", object_name ) ;
		
		( (Audio_Music*) object -> data ) -> StopMusic () ;
	}
	else if ( strcmp ( object_action, "unloadmusic" ) == 0 )
	{
		if ( scan_mode ) return ;
		if ( !object )
		{
			fprintf ( stdout, "Can't unload non-existing music '%s'! Quitting...\n", object_name ) ;
			exit ( 1 ) ;
		}
		
		if ( be_verbose )
			fprintf ( stdout, "Unload a music called '%s'\n", object_name ) ;
		
		( (Audio_Music*) object -> data ) -> UnloadMusic () ;
	}
	else if ( strcmp ( object_action, "loadmusic" ) == 0 )
	{
		if ( scan_mode ) return ;
		if ( !object )
		{
			fprintf ( stdout, "Can't load into non-existing music-object '%s'! Quitting...\n", object_name ) ;
			exit ( 1 ) ;
		}
		
		char filename [ 4096 ] ;
		sprintf ( filename , "%s/%s", pkg_dir, arg [ 0 ] ) ;
		
		if ( be_verbose )
			fprintf ( stdout, "Load a music called '%s'\n", filename ) ;

		Audio_Music *music = (Audio_Music*) object -> data ;
		if ( ! music -> LoadMusic ( filename ) )
		{
			char altfname [ 4096 ] ;
			sprintf ( altfname , "music/%s", arg [ 0 ] ) ;
			
			if ( be_verbose )
			{
				fprintf ( stdout, "Warning: Couln't load music '%s' in '%s'!\n",
							arg [ 0 ], pkg_dir ) ;
				fprintf ( stdout, "         Trying '%s' as a replacement...\n", altfname ) ;
			}
			if ( ! music -> LoadMusic ( altfname ) )
			{
				fprintf ( stdout, "         Failed! Error loading music '%s'!\n", filename ) ;
			}
			else if ( be_verbose )
				fprintf ( stdout, "         Successfully loaded '%s' as a replacement!\n", altfname ) ;
		}
	}
	else if ( strcmp ( object_action, "declareas" ) == 0 )
	{
		if ( scan_mode ) return ;
		char name [ 256 ] ;
		sprintf ( name, "%s", arg [ 0 ] ) ;
		
		if ( be_verbose )
			fprintf ( stdout, "Declaring object '%s' as '%s'\n", object_name, name ) ;
		
		if ( !object )
		{
			fprintf ( stdout, "Can't declare non-existing object '%s'! Quitting...\n", object_name ) ;
			exit ( 1 ) ;
		}
		
		if ( strcmp ( object -> declared_as, "" ) != 0 )
		{
			fprintf ( stdout, "Warning: Changing declaration of already declared object '%s'!\n", object_name ) ;
			fprintf ( stdout, "         Was declared as '%s'. Is now declared as '%s'.\n", 
						object -> declared_as, name ) ;
		}
		
		sprintf ( object -> declared_as, "%s", name ) ;
	}
	else if ( strcmp ( object_action, "addnodeto" ) == 0 )
	{
		if ( scan_mode ) return ;
		char dest_node [ 256 ] ;
		sprintf ( dest_node, "%s", arg [ 0 ] ) ;
		
		if ( be_verbose )
			fprintf ( stdout, "Adding a node called '%s' to node '%s'\n", object_name, dest_node ) ;
		
		if ( !object )
		{
			fprintf ( stdout, "Node '%s' does not exist! Quitting...\n", object_name ) ;
			exit ( 1 ) ;
		}
		
		if ( GetObjectFromList ( dest_node ) == NULL )
		{
			fprintf ( stdout, "Error adding node '%s' to non-existent node '%s'!\n",
						object_name, dest_node ) ;
			exit ( 1 ) ;
		}
		
		KrImNode *src = GetNode ( object, object_name ) ;
		
		KrImNode *dest = GetNode ( GetObjectFromList ( dest_node ), dest_node ) ;
		
		video . GetEngine () -> Tree () -> AddNode ( dest, src ) ;
		
		ObjectListAddChildToParent ( object_name, dest_node ) ;
	}
	else if ( strcmp ( object_action, "setnodename" ) == 0 )
	{
		if ( scan_mode ) return ;
		char name [ 256 ] ;
		sprintf ( name, "%s", arg [ 0 ] ) ;
		
		char type [ 256 ] ;
		sprintf ( type, "%s", object -> type == object_type_sprite ? "sprite" :
					object -> type == object_type_node ? "node" :
					object -> type == object_type_textbox ? "textbox" :
					"unsuitable" ) ;
		
		if ( strcmp ( type, "unsuitable" ) == 0 )
		{
			fprintf ( stdout, "Can't change pos of unsuitable object '%s'!\n", object_name ) ;
			fprintf ( stdout, "Has to be one of sprite, node or textbox. Quitting...\n" ) ;
			exit ( 1 ) ;
		}
		
		if ( be_verbose )
			fprintf ( stdout, "Setting external %s name of '%s' to '%s'\n", type, object_name, name ) ;
		
		if ( !object )
		{
			fprintf ( stdout, "Can't set external name of non-existing %s '%s'! Quitting...\n", type, object_name ) ;
			exit ( 1 ) ;
		}
		
		KrImNode *node = GetNode ( object, object_name ) ;
		
		node -> SetNodeName ( name ) ;
	}
	else if ( strcmp ( object_action, "setxy" ) == 0 )
	{
		if ( scan_mode ) return ;
		int x = CalcValueFromString ( arg [ 0 ] ) ;
		int y = CalcValueFromString ( arg [ 1 ] ) ;
		
		char type [ 256 ] ;
		sprintf ( type, "%s", object -> type == object_type_sprite ? "sprite" :
					object -> type == object_type_node ? "node" :
					object -> type == object_type_textbox ? "textbox" :
					"unsuitable" ) ;
		
		if ( strcmp ( type, "unsuitable" ) == 0 )
		{
			fprintf ( stdout, "Can't change pos of unsuitable object '%s'!\n", object_name ) ;
			fprintf ( stdout, "Has to be one of sprite, node or textbox. Quitting...\n" ) ;
			exit ( 1 ) ;
		}
		
		if ( be_verbose )
			fprintf ( stdout, "Setting x and y pos of %s '%s' to %dx%d\n", type, object_name, x, y ) ;
		
		if ( !object )
		{
			fprintf ( stdout, "Can't set pos of non-existing %s '%s'! Quitting...\n", type, object_name ) ;
			exit ( 1 ) ;
		}
		
		if ( object -> type == object_type_sprite )
			( (Sprite*) ( object -> data ) ) -> SetXY ( x, y ) ;
		else
			GetNode ( GetObjectFromList ( object_name ), object_name ) -> SetPos ( x, y ) ;
	}
	else if ( strcmp ( object_action, "setz" ) == 0 )
	{
		if ( scan_mode ) return ;
		int z = CalcValueFromString ( arg [ 0 ] ) ;
		
		char type [ 256 ] ;
		sprintf ( type, "%s", object -> type == object_type_sprite ? "sprite" :
					object -> type == object_type_node ? "node" :
					object -> type == object_type_textbox ? "textbox" :
					"unsuitable" ) ;
		
		if ( strcmp ( type, "unsuitable" ) == 0 )
		{
			fprintf ( stdout, "Can't change z pos of unsuitable object '%s'!\n", object_name ) ;
			fprintf ( stdout, "Has to be one of sprite, node or textbox. Quitting...\n" ) ;
			exit ( 1 ) ;
		}
		
		if ( be_verbose )
			fprintf ( stdout, "Setting z pos of %s '%s' to %d\n", type, object_name, z ) ;
		
		if ( !object )
		{
			fprintf ( stdout, "Can't set z pos of non-existing %s '%s'! Quitting...\n", type, object_name ) ;
			exit ( 1 ) ;
		}
		
		if ( object -> type == object_type_sprite )
			( (Sprite*) ( object -> data ) ) -> SetZ ( z ) ;
		else
			GetNode ( GetObjectFromList ( object_name ), object_name ) -> SetZDepth ( z ) ;
	}
	else if ( strcmp ( object_action, "setscaleinpercent" ) == 0 )
	{
		if ( scan_mode ) return ;
		int scaleinpercent = CalcValueFromString ( arg [ 0 ] ), scaleinpercent2 ;
		if ( arg_count > 1 )
			scaleinpercent2 = CalcValueFromString ( arg [ 1 ] ) ;
		else
			scaleinpercent2 = scaleinpercent ;
		
		char type [ 256 ] ;
		sprintf ( type, "%s", object -> type == object_type_sprite ? "sprite" :
					object -> type == object_type_node ? "node" :
					object -> type == object_type_textbox ? "textbox" :
					"unsuitable" ) ;
		
		if ( strcmp ( type, "unsuitable" ) == 0 )
		{
			fprintf ( stdout, "Can't change scaling of unsuitable object '%s'!\n", object_name ) ;
			fprintf ( stdout, "Has to be one of sprite, node or textbox. Quitting...\n" ) ;
			exit ( 1 ) ;
		}
		
		if ( be_verbose )
			fprintf ( stdout, "Setting scale of %s '%s' to %d%%x%d%%\n", type, object_name, scaleinpercent, scaleinpercent2 ) ;
		
		if ( !object )
		{
			fprintf ( stdout, "Can't set scaling of non-existing %s '%s'! Quitting...\n", type, object_name ) ;
			exit ( 1 ) ;
		}
		
		double scale_x = (double) scaleinpercent / (double) 100 ;
		GlFixed dscale_x = scale_x ;
		double scale_y = (double) scaleinpercent2 / (double) 100 ;
		GlFixed dscale_y = scale_y ;
		
		GetNode ( GetObjectFromList ( object_name ), object_name ) -> SetScale ( dscale_x, dscale_y ) ;
	}
	else if ( strcmp ( object_action, "setsize" ) == 0 )
	{
		if ( scan_mode ) return ;
		int new_width = CalcValueFromString ( arg [ 0 ] ) ;
		int new_height = CalcValueFromString ( arg [ 1 ] ) ;
		
		char type [ 256 ] ;
		sprintf ( type, "%s", object -> type == object_type_sprite ? "sprite" :
					"unsuitable" ) ;
		
		if ( strcmp ( type, "unsuitable" ) == 0 )
		{
			fprintf ( stdout, "Can't change scaling of unsuitable object '%s'!\n", object_name ) ;
			fprintf ( stdout, "Has to be a sprite. Quitting...\n" ) ;
			exit ( 1 ) ;
		}
		
		if ( be_verbose )
			fprintf ( stdout, "Scaling %s '%s' to %dx%d pixels\n", type, object_name, new_width, new_height ) ;
		
		if ( !object )
		{
			fprintf ( stdout, "Can't set scaling of non-existing %s '%s'! Quitting...\n", type, object_name ) ;
			exit ( 1 ) ;
		}
		
		double scale_x = ((double) new_width)  / ((double) ((Sprite*) object -> data) -> GetWidth  ()) ;
		double scale_y = ((double) new_height) / ((double) ((Sprite*) object -> data) -> GetHeight ()) ;
		GlFixed dscale_x = scale_x ;
		GlFixed dscale_y = scale_y ;
		
		GetNode ( GetObjectFromList ( object_name ), object_name ) -> SetScale ( dscale_x, dscale_y ) ;
	}
	else if ( strcmp ( object_action, "setalpha" ) == 0 )
	{
		if ( scan_mode ) return ;
		int alpha = CalcValueFromString ( arg [ 0 ] ) ;
		
		char type [ 256 ] ;
		sprintf ( type, "%s", object -> type == object_type_sprite ? "sprite" :
					object -> type == object_type_node ? "node" :
					object -> type == object_type_textbox ? "textbox" :
					"unsuitable" ) ;
		
		if ( strcmp ( type, "unsuitable" ) == 0 )
		{
			fprintf ( stdout, "Can't change transparency of unsuitable object '%s'!\n", object_name ) ;
			fprintf ( stdout, "Has to be one of sprite, node or textbox. Quitting...\n" ) ;
			exit ( 1 ) ;
		}
		
		if ( be_verbose )
			fprintf ( stdout, "Setting transparency of %s '%s' to %d\n", type, object_name, alpha ) ;
		
		if ( !object )
		{
			fprintf ( stdout, "Can't set transparency of non-existing %s '%s'! Quitting...\n", type, object_name ) ;
			exit ( 1 ) ;
		}
		
		if ( object -> type == object_type_sprite )
			( (Sprite*) ( object -> data ) ) -> SetAlpha ( alpha ) ;
		else
		{
			fprintf ( stdout, "Setting transparency of non-sprite object not implemented yet.\n" ) ;
			fprintf ( stdout, "As for now use a sprite as a parent. You could do something like this:\n" ) ;
			fprintf ( stdout, "  dummyvault:newvaultfromfile \"misc.dat\"\n" ) ;
			fprintf ( stdout, "  dummysprite:newsprite dummyvault, \"dummy\" %s\n",
						strcmp ( object -> parent, "" ) != 0 ? object -> parent : "" ) ;
			fprintf ( stdout, "  dummysprite:setalpha %d\n", alpha ) ;
			fprintf ( stdout, "  %s:addnodeto dummysprite\n", object_name ) ;
			fprintf ( stdout, "Note: Child nodes will be affected by their parents.\n" ) ;
			//GetNode ( GetObjectFromList ( object_name ), object_name ) -> SetAlpha ( alpha ) ;
		}
	}
	else if ( strcmp ( object_action, "setdestalpha" ) == 0 )
	{
		if ( scan_mode ) return ;
		int alpha = CalcValueFromString ( arg [ 0 ] ) ;
		
		char type [ 256 ] ;
		sprintf ( type, "%s", object -> type == object_type_sprite ? "sprite" :
					object -> type == object_type_node ? "node" :
					object -> type == object_type_textbox ? "textbox" :
					"unsuitable" ) ;
		
		if ( strcmp ( type, "unsuitable" ) == 0 )
		{
			fprintf ( stdout, "Can't change destination transparency of unsuitable object '%s'!\n", object_name ) ;
			fprintf ( stdout, "Has to be one of sprite, node or textbox. Quitting...\n" ) ;
			exit ( 1 ) ;
		}
		
		if ( be_verbose )
			fprintf ( stdout, "Setting destination transparency of %s '%s' to %d\n", type, object_name, alpha ) ;
		
		if ( !object )
		{
			fprintf ( stdout, "Can't set destination transparency of non-existing %s '%s'! Quitting...\n", type, object_name ) ;
			exit ( 1 ) ;
		}
		
		if ( object -> type == object_type_sprite )
			( (Sprite*) ( object -> data ) ) -> SetDestAlpha ( alpha ) ;
		else
		{
			fprintf ( stdout, "Setting destination transparency of non-sprite object not implemented yet.\n" ) ;
			fprintf ( stdout, "As for now use a sprite as a parent. You could do something like this:\n" ) ;
			fprintf ( stdout, "  dummyvault:newvaultfromfile \"misc.dat\"\n" ) ;
			fprintf ( stdout, "  dummysprite:newsprite dummyvault, \"dummy\" %s\n",
						strcmp ( object -> parent, "" ) != 0 ? object -> parent : "" ) ;
			fprintf ( stdout, "  dummysprite:setdestalpha %d\n", alpha ) ;
			fprintf ( stdout, "  %s:addnodeto dummysprite\n", object_name ) ;
			fprintf ( stdout, "Note: Child nodes will be affected by their parents.\n" ) ;
			//GetNode ( GetObjectFromList ( object_name ), object_name ) -> SetDestAlpha ( alpha ) ;
		}
	}
	else if ( strcmp ( object_action, "settextchar" ) == 0 )
	{
		if ( scan_mode ) return ;
		char text [ 256 ] ;
		sprintf ( text, "%s", arg [ 0 ] ) ;
		int line = CalcValueFromString ( arg [ 1 ] ) ;
		
		char type [ 256 ] ;
		sprintf ( type, "%s", object -> type == object_type_textbox ? "textbox" :
					"unsuitable" ) ;
		
		if ( strcmp ( type, "unsuitable" ) == 0 )
		{
			fprintf ( stdout, "Can't change text of unsuitable object '%s'!\n", object_name ) ;
			fprintf ( stdout, "Has to be a textbox... Quitting...\n" ) ;
			exit ( 1 ) ;
		}
		
		if ( be_verbose )
			fprintf ( stdout, "Setting text of %s '%s' to '%s'\n", type, object_name, text ) ;
		
		if ( !object )
		{
			fprintf ( stdout, "Can't set text of non-existing %s '%s'! Quitting...\n", type, object_name ) ;
			exit ( 1 ) ;
		}
		
		// set text of textbox
		( (KrTextBox*) ( object -> data ) ) -> SetTextChar ( text, line ) ;
	}
	else if ( strcmp ( object_action, "setstopsound" ) == 0 )
	{
		if ( scan_mode ) return ;
		char name [ 256 ] ;
		sprintf ( name, "%s", arg [ 0 ] ) ;
		int volume = CalcValueFromString ( arg [ 1 ] ) ;
		
		char type [ 256 ] ;
		sprintf ( type, "%s", object -> type == object_type_sprite ? "sprite" :
					"unsuitable" ) ;
		
		if ( strcmp ( type, "unsuitable" ) == 0 )
		{
			fprintf ( stdout, "Can't change stop-sound of unsuitable object '%s'!\n", object_name ) ;
			fprintf ( stdout, "Has to be a sprite... Quitting...\n" ) ;
			exit ( 1 ) ;
		}
		
		if ( be_verbose )
			fprintf ( stdout, "Setting stop-sound of %s '%s' to %s with a volume of %d\n",
						type, object_name, name, volume ) ;
		
		if ( !object )
		{
			fprintf ( stdout, "Can't set stop-sound of non-existing %s '%s'! Quitting...\n", type, object_name ) ;
			exit ( 1 ) ;
		}
		
		( (Sprite*) (object -> data) ) -> SetStopSound ( name, volume ) ;
	}
	else if ( strcmp ( object_action, "playsoundatframe" ) == 0 )
	{
		if ( scan_mode ) return ;
		char name [ 256 ] ;
		sprintf ( name, "%s", arg [ 0 ] ) ;
		int frame = CalcValueFromString ( arg [ 1 ] ) ;
		
		char type [ 256 ] ;
		sprintf ( type, "%s", object -> type == object_type_sprite ? "sprite" :
					"unsuitable" ) ;
		
		if ( strcmp ( type, "unsuitable" ) == 0 )
		{
			fprintf ( stdout, "Can't setup frame-sound of unsuitable object '%s'!\n", object_name ) ;
			fprintf ( stdout, "Has to be a sprite... Quitting...\n" ) ;
			exit ( 1 ) ;
		}
		
		if ( be_verbose )
			fprintf ( stdout, "Setting frame-sound of %s '%s' to %s at frame %d\n",
						type, object_name, name, frame ) ;
		
		if ( !object )
		{
			fprintf ( stdout, "Can't setup frame-sound of non-existing %s '%s'! Quitting...\n", type, object_name ) ;
			exit ( 1 ) ;
		}
		
		( (Sprite*) (object -> data) ) -> PlaySoundAtFrame ( name, frame ) ;
	}
	else if ( strcmp ( object_action, "setdelays" ) == 0 )
	{
		if ( scan_mode ) return ;
		int delay = CalcValueFromString ( arg [ 0 ] ) ;
		int delay_counter = CalcValueFromString ( arg [ 1 ] ) ;
		int pause_interval = CalcValueFromString ( arg [ 2 ] ) ;
		int pause_interval_counter = CalcValueFromString ( arg [ 3 ] ) ;
		bool pause_interval_randomness = ( strcmp ( arg [ 4 ], "true" ) == 0 ) ;
		int max_pause_interval = CalcValueFromString ( arg [ 5 ] ) ;
		
		char type [ 256 ] ;
		sprintf ( type, "%s", object -> type == object_type_sprite ? "sprite" :
					"unsuitable" ) ;
		
		if ( strcmp ( type, "unsuitable" ) == 0 )
		{
			fprintf ( stdout, "Can't change delays of unsuitable object '%s'!\n", object_name ) ;
			fprintf ( stdout, "Has to be a sprite... Quitting...\n" ) ;
			exit ( 1 ) ;
		}
		
		if ( be_verbose )
			fprintf ( stdout, "Setting delays of %s '%s'\n", type, object_name ) ;
		
		if ( !object )
		{
			fprintf ( stdout, "Can't set delays of non-existing %s '%s'! Quitting...\n", type, object_name ) ;
			exit ( 1 ) ;
		}
		
		( (Sprite*) (object -> data) ) -> SetDelays ( delay, delay_counter,
														pause_interval,
														pause_interval_counter,
														pause_interval_randomness,
														max_pause_interval ) ;
	}
	else if ( strcmp ( object_action, "setframes" ) == 0 )
	{
		if ( scan_mode ) return ;
		int pos = CalcValueFromString ( arg [ 0 ] ) ;
		int frames = CalcValueFromString ( arg [ 1 ] ) ;
		int direction = CalcValueFromString ( arg [ 2 ] ) ;
		bool bounce = ( strcmp ( arg [ 3 ], "true" ) == 0 ) ;
		bool once = arg_count > 4 ? 
					( strcmp ( arg [ 4 ], "true" ) == 0 ) : false ;
		
		char type [ 256 ] ;
		sprintf ( type, "%s", object -> type == object_type_sprite ? "sprite" :
					"unsuitable" ) ;
		
		if ( strcmp ( type, "unsuitable" ) == 0 )
		{
			fprintf ( stdout, "Can't change frames of unsuitable object '%s'!\n", object_name ) ;
			fprintf ( stdout, "Has to be a sprite... Quitting...\n" ) ;
			exit ( 1 ) ;
		}
		
		if ( be_verbose )
			fprintf ( stdout, "Setting frames of %s '%s'\n", type, object_name ) ;
		
		if ( !object )
		{
			fprintf ( stdout, "Can't set frames of non-existing %s '%s'! Quitting...\n", type, object_name ) ;
			exit ( 1 ) ;
		}
		
		( (Sprite*) (object -> data) ) -> SetFrames ( pos, frames, direction,
														bounce, once ) ;
	}
	else if ( strcmp ( object_action, "setbrighten" ) == 0 )
	{
		if ( scan_mode ) return ;
		U8 value = CalcValueFromString ( arg [ 0 ] ) ;
		
		char type [ 256 ] ;
		sprintf ( type, "%s", object -> type == object_type_sprite ? "sprite" :
					"unsuitable" ) ;
		
		if ( strcmp ( type, "unsuitable" ) == 0 )
		{
			fprintf ( stdout, "Can't change brightness of unsuitable object '%s'!\n", object_name ) ;
			fprintf ( stdout, "Has to be a sprite... Quitting...\n" ) ;
			exit ( 1 ) ;
		}
		
		if ( be_verbose )
			fprintf ( stdout, "Setting brightness of %s '%s'\n", type, object_name ) ;
		
		if ( !object )
		{
			fprintf ( stdout, "Can't set brightness of non-existing %s '%s'! Quitting...\n", type, object_name ) ;
			exit ( 1 ) ;
		}
		
		( (Sprite*) (object -> data) ) -> Brighten ( value ) ;
	}
	else if ( strcmp ( object_action, "setdestbrighten" ) == 0 )
	{
		if ( scan_mode ) return ;
		U8 value = CalcValueFromString ( arg [ 0 ] ) ;
		
		char type [ 256 ] ;
		sprintf ( type, "%s", object -> type == object_type_sprite ? "sprite" :
					"unsuitable" ) ;
		
		if ( strcmp ( type, "unsuitable" ) == 0 )
		{
			fprintf ( stdout, "Can't change destination brightness of unsuitable object '%s'!\n", object_name ) ;
			fprintf ( stdout, "Has to be a sprite... Quitting...\n" ) ;
			exit ( 1 ) ;
		}
		
		if ( be_verbose )
			fprintf ( stdout, "Setting destination brightness of %s '%s'\n", type, object_name ) ;
		
		if ( !object )
		{
			fprintf ( stdout, "Can't set destination brightness of non-existing %s '%s'! Quitting...\n", type, object_name ) ;
			exit ( 1 ) ;
		}
		
		( (Sprite*) (object -> data) ) -> SetDestBrighten ( value ) ;
	}
	else if ( strcmp ( object_action, "setdarken" ) == 0 )
	{
		if ( scan_mode ) return ;
		U8 value = CalcValueFromString ( arg [ 0 ] ) ;
		
		char type [ 256 ] ;
		sprintf ( type, "%s", object -> type == object_type_sprite ? "sprite" :
					"unsuitable" ) ;
		
		if ( strcmp ( type, "unsuitable" ) == 0 )
		{
			fprintf ( stdout, "Can't change darkness of unsuitable object '%s'!\n", object_name ) ;
			fprintf ( stdout, "Has to be a sprite... Quitting...\n" ) ;
			exit ( 1 ) ;
		}
		
		if ( be_verbose )
			fprintf ( stdout, "Setting darkness of %s '%s'\n", type, object_name ) ;
		
		if ( !object )
		{
			fprintf ( stdout, "Can't set darkness of non-existing %s '%s'! Quitting...\n", type, object_name ) ;
			exit ( 1 ) ;
		}
		
		( (Sprite*) (object -> data) ) -> Darken ( value ) ;
	}
	else if ( strcmp ( object_action, "setdestdarken" ) == 0 )
	{
		if ( scan_mode ) return ;
		U8 value = CalcValueFromString ( arg [ 0 ] ) ;
		
		char type [ 256 ] ;
		sprintf ( type, "%s", object -> type == object_type_sprite ? "sprite" :
					"unsuitable" ) ;
		
		if ( strcmp ( type, "unsuitable" ) == 0 )
		{
			fprintf ( stdout, "Can't change destination darkness of unsuitable object '%s'!\n", object_name ) ;
			fprintf ( stdout, "Has to be a sprite... Quitting...\n" ) ;
			exit ( 1 ) ;
		}
		
		if ( be_verbose )
			fprintf ( stdout, "Setting destination darkness of %s '%s'\n", type, object_name ) ;
		
		if ( !object )
		{
			fprintf ( stdout, "Can't set destination darkness of non-existing %s '%s'! Quitting...\n", type, object_name ) ;
			exit ( 1 ) ;
		}
		
		( (Sprite*) (object -> data) ) -> SetDestDarken ( value ) ;
	}
	else if ( strcmp ( object_action, "slideinfrom" ) == 0 )
	{
		if ( scan_mode ) return ;
		int x = CalcValueFromString ( arg [ 0 ] ) ;
		int y = CalcValueFromString ( arg [ 1 ] ) ;
		
		char type [ 256 ] ;
		sprintf ( type, "%s", object -> type == object_type_sprite ? "sprite" :
					"unsuitable" ) ;
		
		if ( strcmp ( type, "unsuitable" ) == 0 )
		{
			fprintf ( stdout, "Can't set slide-in params of unsuitable object '%s'!\n", object_name ) ;
			fprintf ( stdout, "Has to be a sprite... Quitting...\n" ) ;
			exit ( 1 ) ;
		}
		
		if ( be_verbose )
			fprintf ( stdout, "Setting up slide-in params of %s '%s' to %dx%d\n", type, object_name, x, y ) ;
		
		if ( !object )
		{
			fprintf ( stdout, "Can't set slide-in params of non-existing %s '%s'! Quitting...\n", type, object_name ) ;
			exit ( 1 ) ;
		}
		
		( (Sprite*) object -> data ) -> SlideInFrom ( x, y ) ;
	}
	else if ( strcmp ( object_action, "slideoutto" ) == 0 )
	{
		if ( scan_mode ) return ;
		int x = CalcValueFromString ( arg [ 0 ] ) ;
		int y = CalcValueFromString ( arg [ 1 ] ) ;
		
		char type [ 256 ] ;
		sprintf ( type, "%s", object -> type == object_type_sprite ? "sprite" :
					"unsuitable" ) ;
		
		if ( strcmp ( type, "unsuitable" ) == 0 )
		{
			fprintf ( stdout, "Can't set slide-out params of unsuitable object '%s'!\n", object_name ) ;
			fprintf ( stdout, "Has to be a sprite... Quitting...\n" ) ;
			exit ( 1 ) ;
		}
		
		if ( be_verbose )
			fprintf ( stdout, "Setting up slide-out params of %s '%s' to %dx%d\n", type, object_name, x, y ) ;
		
		if ( !object )
		{
			fprintf ( stdout, "Can't set slide-out params of non-existing %s '%s'! Quitting...\n", type, object_name ) ;
			exit ( 1 ) ;
		}
		
		( (Sprite*) object -> data ) -> SlideOutTo ( x, y ) ;
	}
	else if ( strcmp ( object_action, "settwinkling" ) == 0 )
	{
		if ( scan_mode ) return ;
		bool activate = ( strcmp ( arg [ 0 ], "true" ) == 0 ) ;
		int new_max_drops = -1 ;
		int new_lifetime = -1 ;
		if ( arg_count > 1 )
			new_max_drops = CalcValueFromString ( arg [ 1 ] ) ;
		if ( arg_count > 2 )
			new_lifetime = CalcValueFromString ( arg [ 2 ] ) ;
		
		char type [ 256 ] ;
		sprintf ( type, "%s", object -> type == object_type_sprite ? "sprite" :
					"unsuitable" ) ;
		
		if ( strcmp ( type, "unsuitable" ) == 0 )
		{
			fprintf ( stdout, "Can't setup twinkling of unsuitable object '%s'!\n", object_name ) ;
			fprintf ( stdout, "Has to be a sprite... Quitting...\n" ) ;
			exit ( 1 ) ;
		}
		
		if ( be_verbose )
			fprintf ( stdout, "Settingup twinkling of %s '%s'\n",
						type, object_name ) ;
		
		if ( !object )
		{
			fprintf ( stdout, "Can't setup twinkling of non-existing %s '%s'! Quitting...\n", type, object_name ) ;
			exit ( 1 ) ;
		}
		
		( (Sprite*) (object -> data) ) -> Twinkling ( activate,
														new_max_drops,
														new_lifetime ) ;
	}
	else if ( strcmp ( object_action, "settwinklingspeed" ) == 0 )
	{
		if ( scan_mode ) return ;
		int speed_x = CalcValueFromString ( arg [ 0 ] ) ;
		int speed_y = CalcValueFromString ( arg [ 1 ] ) ;
		
		char type [ 256 ] ;
		sprintf ( type, "%s", object -> type == object_type_sprite ? "sprite" :
					"unsuitable" ) ;
		
		if ( strcmp ( type, "unsuitable" ) == 0 )
		{
			fprintf ( stdout, "Can't set twinkling speed of unsuitable object '%s'!\n", object_name ) ;
			fprintf ( stdout, "Has to be a sprite... Quitting...\n" ) ;
			exit ( 1 ) ;
		}
		
		if ( be_verbose )
			fprintf ( stdout, "Setting twinkling speed of %s '%s' to %dx%d\n",
						type, object_name, speed_x, speed_y ) ;
		
		if ( !object )
		{
			fprintf ( stdout, "Can't set twinkling speed of non-existing %s '%s'! Quitting...\n", type, object_name ) ;
			exit ( 1 ) ;
		}
		
		( (Sprite*) (object -> data) ) -> SetTwinklingSpeed ( speed_x, speed_y ) ;
	}
	else if ( strcmp ( object_action, "setrain" ) == 0 )
	{
		if ( scan_mode ) return ;
		bool activate = ( strcmp ( arg [ 0 ], "true" ) == 0 ) ;
		int new_max_drops = -1 ;
		int new_lifetime = -1 ;
		if ( arg_count > 1 )
			new_max_drops = CalcValueFromString ( arg [ 1 ] ) ;
		if ( arg_count > 2 )
			new_lifetime = CalcValueFromString ( arg [ 2 ] ) ;
		
		char type [ 256 ] ;
		sprintf ( type, "%s", object -> type == object_type_rain ? "rain" :
					"unsuitable" ) ;
		
		if ( strcmp ( type, "unsuitable" ) == 0 )
		{
			fprintf ( stdout, "Can't setup rain of unsuitable object '%s'!\n", object_name ) ;
			fprintf ( stdout, "Has to be a rain... Quitting...\n" ) ;
			exit ( 1 ) ;
		}
		
		if ( be_verbose )
			fprintf ( stdout, "Settingup rain of %s '%s'\n",
						type, object_name ) ;
		
		if ( !object )
		{
			fprintf ( stdout, "Can't setup rain of non-existing %s '%s'! Quitting...\n", type, object_name ) ;
			exit ( 1 ) ;
		}
		
		if ( activate )
			( (Rain*) (object -> data) ) -> Start () ;
		if ( arg_count > 1 )
			( (Rain*) (object -> data) ) -> SetMaxDrops ( new_max_drops ) ;
		if ( arg_count > 2 )
			( (Rain*) (object -> data) ) -> SetLifeTime ( new_lifetime  ) ;
		if ( ! activate )
			( (Rain*) (object -> data) ) -> Stop  () ;
	}
	else if ( strcmp ( object_action, "setrainspeed" ) == 0 )
	{
		if ( scan_mode ) return ;
		int speed_x = CalcValueFromString ( arg [ 0 ] ) ;
		int speed_y = CalcValueFromString ( arg [ 1 ] ) ;
		
		char type [ 256 ] ;
		sprintf ( type, "%s", object -> type == object_type_rain ? "rain" :
					"unsuitable" ) ;
		
		if ( strcmp ( type, "unsuitable" ) == 0 )
		{
			fprintf ( stdout, "Can't set rain speed of unsuitable object '%s'!\n", object_name ) ;
			fprintf ( stdout, "Has to be rain... Quitting...\n" ) ;
			exit ( 1 ) ;
		}
		
		if ( be_verbose )
			fprintf ( stdout, "Setting rain speed of %s '%s' to %dx%d\n",
						type, object_name, speed_x, speed_y ) ;
		
		if ( !object )
		{
			fprintf ( stdout, "Can't set rain speed of non-existing %s '%s'! Quitting...\n", type, object_name ) ;
			exit ( 1 ) ;
		}
		
		( (Rain*) (object -> data) ) -> SetSpeed ( speed_x, speed_y ) ;
	}
	else if ( strcmp ( object_action, "setfadespeed" ) == 0 )
	{
		if ( scan_mode ) return ;
		int speed = CalcValueFromString ( arg [ 0 ] ) ;
		
		char type [ 256 ] ;
		sprintf ( type, "%s", object -> type == object_type_sprite ? "sprite" :
					"unsuitable" ) ;
		
		if ( strcmp ( type, "unsuitable" ) == 0 )
		{
			fprintf ( stdout, "Can't set fade speed of unsuitable object '%s'!\n", object_name ) ;
			fprintf ( stdout, "Has to be sprite... Quitting...\n" ) ;
			exit ( 1 ) ;
		}
		
		if ( be_verbose )
			fprintf ( stdout, "Setting fade speed of %s '%s' to %d\n",
						type, object_name, speed ) ;
		
		if ( !object )
		{
			fprintf ( stdout, "Can't set fade speed of non-existing %s '%s'! Quitting...\n", type, object_name ) ;
			exit ( 1 ) ;
		}
		
		( (Sprite*) (object -> data) ) -> SetFadeSpeed ( speed ) ;
	}
	else if ( strcmp ( object_action, "setglow" ) == 0 )
	{
		if ( scan_mode ) return ;
		bool activate = ( strcmp ( arg [ 0 ], "true" ) == 0 ) ;
		
		char type [ 256 ] ;
		sprintf ( type, "%s", object -> type == object_type_sprite ? "sprite" :
					"unsuitable" ) ;
		
		if ( strcmp ( type, "unsuitable" ) == 0 )
		{
			fprintf ( stdout, "Can't set glowing of unsuitable object '%s'!\n", object_name ) ;
			fprintf ( stdout, "Has to be a sprite... Quitting...\n" ) ;
			exit ( 1 ) ;
		}
		
		if ( be_verbose )
			fprintf ( stdout, "Setting glowing of %s to %s\n",
						type, activate ? "true" : "false" ) ;
		
		if ( !object )
		{
			fprintf ( stdout, "Can't set glowing of non-existing %s '%s'! Quitting...\n", type, object_name ) ;
			exit ( 1 ) ;
		}
		
		( (Sprite*) (object -> data) ) -> Glow ( activate ) ;
	}
	else if ( strcmp ( object_action, "settint" ) == 0 )
	{
		if ( scan_mode ) return ;
		U8 r = (U8) CalcValueFromString ( arg [ 0 ] ) ;
		U8 g = (U8) CalcValueFromString ( arg [ 1 ] ) ;
		U8 b = (U8) CalcValueFromString ( arg [ 2 ] ) ;
		
		char type [ 256 ] ;
		sprintf ( type, "%s", object -> type == object_type_sprite ? "sprite" :
					"unsuitable" ) ;
		
		if ( strcmp ( type, "unsuitable" ) == 0 )
		{
			fprintf ( stdout, "Can't set tinting of unsuitable object '%s'!\n", object_name ) ;
			fprintf ( stdout, "Has to be a sprite... Quitting...\n" ) ;
			exit ( 1 ) ;
		}
		
		if ( be_verbose )
			fprintf ( stdout, "Tinting %s to r = %d, g = %d, b = %d\n",
						type, r, g, b ) ;
		
		if ( !object )
		{
			fprintf ( stdout, "Can't set tinting of non-existing %s '%s'! Quitting...\n", type, object_name ) ;
			exit ( 1 ) ;
		}
		
		( (Sprite*) (object -> data) ) -> SetTint ( r, g, b ) ;
	}
	else
	{
		fprintf ( stdout, "%s : Unkown action '%s'\n", object_name, object_action ) ;
		//exit ( 1 ) ;
	}
	
	object_arglist -> Clear () ;
}
