#include "pax_list.h"

extern char rc_dir[], pax_dir[] ;

extern FileCache file_cache ;

#include "video.h"
extern Video video ;

void PaxList::Init ( void *user_data )
{
	char pax_path_in_data_dir [ 4096 ] ;
	char pax_path_in_home_dir [ 4096 ] ;
	
	fprintf ( stdout, "initialize pax list...\n" ) ;
	
	video . DisplayOSDMessage ( "searching for paxs..." ) ;
	video . UpdateDisplayOnly () ;
	
	border = PLB_NONE ;
	pax_list_attached = false ;
	pax_list_created = false ;
	
	MiscVault = (KrResourceVault*) file_cache . CachedLoad ( "vault", "gfx/misc.dat" ) ;
	
	sprintf ( pax_path_in_data_dir, "pax" ) ;
	sprintf ( pax_path_in_home_dir, "%s", pax_dir ) ;
	
	list = new GlDynArray < pax_item* > () ;
	list -> Clear () ;
	
	ScanDirForPax ( pax_path_in_data_dir, user_data ) ;
	ScanDirForPax ( pax_path_in_home_dir, user_data ) ;
	
	fprintf ( stdout, "found %d paxs which are suiteable for playing.\n\n", list -> Count () ) ;

	if ( list -> Count () == 0 )
	{
		video . DisplayOSDMessage ( "no paxs found!" ) ;
		video . UpdateDisplayOnly () ;
	}
}

void PaxList::ScanDirForPax ( char *current_path, void *user_data )
{
	struct stat buf ;
	GlDynArray < dirent > namelist ;
	int namelist_count = 0, namelist_pos = 0 ;
	
	fprintf ( stdout, "searching for pax in %s ...\n", current_path ) ;
	
	namelist_count = myscandir ( current_path, &namelist ) ;
	
	if ( namelist_count < 1 )
	{
		namelist . Clear () ;
		return ;
	}
	
	for ( namelist_pos = 0 ; namelist_pos < namelist_count ; namelist_pos++ )
	{
		char current_file [ 4096 ], current_index [ 4096 ] ;
		
		sprintf ( current_file, "%s/%s", current_path, namelist[namelist_pos].d_name ) ;
		sprintf ( current_index, "%s/index.pax", current_file ) ;
		
		if ( isdir ( current_file ) && stat ( current_index, &buf ) == 0 )
		{
			fprintf ( stdout, "  found: %s ...", current_file ) ;
			
			pax_item *item = new pax_item () ;
			
			char *file = (char*) malloc ( sizeof (char) * ( strlen ( current_file ) + 256 ) ) ;
			sprintf ( file, "%s", current_file ) ;
			
			item -> pathname = file ;
			
			item -> shortname = NULL ;
			
			item -> vault = NULL ;
			item -> gfx_name = NULL ;
			item -> node = NULL ;
			item -> screenshot = NULL ;
			item -> border = NULL ;
			
			item -> name = NULL ;
			item -> author = NULL ;
			item -> description = NULL ;
			
			item -> user_data = NULL ;
			
			fprintf ( stdout, "inspecting...\n" ) ;
			
			Pax *tmp_pax = new Pax () ;
			
			tmp_pax -> Init ( current_index, true ) ;
			
			char string_var [ 4096 ] ;
			sprintf ( string_var, "%s", tmp_pax -> GetStringVar ( "screenshot" ) ) ;

			if ( strcmp ( string_var, "hidden" ) == 0 )
			{
				tmp_pax -> FreeObjects () ;
				free ( item -> pathname ) ;
				item -> pathname = NULL ;
				continue ;
			}
			
			GlDynArray < std::string > *shot_args ;
			shot_args = tmp_pax -> GetArgList ( string_var ) ;
			
			char *gfx_pack = (char*) shot_args -> Item ( 0 ) . c_str () ;
			if ( shot_args -> Count () > 1 )
				item -> gfx_name = strdup ( (char*) shot_args -> Item ( 1 ) . c_str () ) ;
			
			shot_args -> Clear () ;
			delete shot_args ;
			
			//fprintf ( stdout, "  screenshot will be loaded from %s...\n", gfx_pack ) ;
			
			Pax::object_item *object = tmp_pax -> GetObjectFromList ( gfx_pack ) ;
			
			if ( ! object )
			{
				fprintf ( stdout, "  gfx pack not created in pax file.\n" ) ;
				fprintf ( stdout, "  this pax will be ignored.\n" ) ;
				tmp_pax -> FreeObjects () ;
				free ( item -> pathname ) ;
				item -> pathname = NULL ;
				continue ;
			}
			
			// get shortname for easier searching....
			char shortname [ 4096 ] ;
			if ( strlen ( rindex ( tmp_pax -> GetPkgDirname (), '/' ) + 1 ) )
				sprintf ( shortname, "%s", rindex ( tmp_pax -> GetPkgDirname (), '/' ) + 1 ) ;
			else
				sprintf ( shortname, tmp_pax -> GetPkgDirname () ) ;
			item -> shortname = strdup ( shortname ) ;
			
			//item -> vault = new KrResourceVault () ;
			item -> vault = (KrResourceVault*) object -> data ;
			int cache_pos = file_cache . SearchForData ( item -> vault ) ;
			if ( cache_pos == -1 ) exit ( 1 ) ;
			file_cache . IncreaseDependency ( cache_pos ) ;
			
			tmp_pax -> FreeObjects () ;
			delete tmp_pax ;
			
			if ( user_data )
				item -> user_data = user_data ;
			
			list -> PushBack ( item ) ;
			
			pax_list_created = true ;
		}
	}
	
	namelist . Clear () ;
}

void PaxList::AttachListToNode ( KrImNode *basenode,
									unsigned int border_type )
{
	if ( pax_list_created == false ||
			list == NULL ) return ;
	
	if ( pax_list_attached )
		DetachList () ;
	
	border = border_type ;	
	
	//fprintf ( stdout, "attaching pax list... ( %d items )\n", list -> Count () ) ;
	
	for ( unsigned int i = 0 ; i < list -> Count () ; i++ )
	{
		pax_item *item = list -> Item ( i ) ;
		
		item -> node = new Sprite () ;
		item -> node -> Init ( MiscVault, "dummy",
								basenode,
								true ) ;
		item -> node -> SetXY ( 0, 0 ) ;
		item -> node -> SetZ ( 10 ) ;
		
		item -> screenshot = new Sprite () ;
		item -> screenshot -> Init ( item -> vault, item -> gfx_name,
									item -> node -> GetSprite (), true ) ;
		item -> screenshot -> SetXY ( 0, 0 ) ;
		
		KrSpriteResource *item_res = item -> screenshot -> GetSprite () -> SpriteResource () ;
		if ( /*video . GetUseCachedScaling () &&*/ !item_res -> IsScaleCached ( 0.25, 0.25 ) )
			item_res -> CacheScale ( 0.25, 0.25 ) ;
		item -> screenshot -> SetXY ( 0, 0 ) ;
		
		if ( border == PLB_WHITE )
		{
			item -> border = new Sprite () ;
			item -> border -> Init ( MiscVault, "shot_border",
									item -> node -> GetSprite (), true ) ;
			item -> border -> SetFrames ( 0, AUTO, 1, false, true ) ;
			item -> border -> SetXY ( 0, 0 ) ;
			item -> border -> SetAlpha ( 255 ) ;
			// adjust screenshot position according to border...
			item -> screenshot -> SetXY ( 40, 30 ) ;
		}
		// scaling has to be set after SetXY which resets scaling...
		item -> screenshot -> GetSprite () -> SetScale ( 0.25, 0.25 ) ;
		
		item -> node -> SetAlpha ( 0 ) ;
		item -> screenshot -> SetAlpha ( 255 ) ;
		
		item -> node -> SetDestAlpha ( 128 ) ;
		item -> node -> SetStopSound ( NULL, 0 ) ;
		item -> node -> SetMoveSound ( "sfx/malus.wav", 64 ) ;
	}
	
	pax_list_attached = true ;
	
	//fprintf ( stdout, "done.\n" ) ;
}

void PaxList::DetachList ()
{
	if ( pax_list_attached == false ||
			pax_list_created == false ||
			list == NULL ) return ;
	
	//fprintf ( stdout, "detaching pax list... ( %d items )\n", list -> Count () ) ;
	
	for ( unsigned int pos = 0 ; pos < list -> Count () ; pos++ )
	{
		if ( list -> Item ( pos ) -> screenshot )
		{
			list -> Item ( pos ) -> screenshot -> Free () ;
			delete list -> Item ( pos ) -> screenshot ;
			list -> Item ( pos ) -> screenshot = NULL ;
		}
		
		if ( list -> Item ( pos ) -> border )
		{
			list -> Item ( pos ) -> border -> Free () ;
			delete list -> Item ( pos ) -> border ;
			list -> Item ( pos ) -> border = NULL ;
		}
		
		if ( list -> Item ( pos ) -> node )
		{
			list -> Item ( pos ) -> node -> Free () ;
			delete list -> Item ( pos ) -> node ;
			list -> Item ( pos ) -> node = NULL ;
		}
		
		if ( list -> Item ( pos ) -> user_data )
		{
			free ( list -> Item ( pos ) -> user_data ) ;
			list -> Item ( pos ) -> user_data = NULL ;
		}
	}
	
	pax_list_attached = false ;
	
	//fprintf ( stdout, "done.\n" ) ;
}

int PaxList::FindPaxElement ( char *pkg_fname )
{
	if ( pax_list_created == false || list == NULL ) return -1 ;
	
	//fprintf ( stdout, "searching for pax element '%s'...\n", pkg_fname ) ;
	
	int pos = -1 ;
	
	for ( unsigned int i = 0 ; i < list -> Count () ; i++ )
	{
		pax_item *item = list -> Item ( i ) ;
		
		// now check if we should activate a specific item...
		char *item_wo_path = rindex ( item -> pathname, '/' ) ;
		char *pkg_wo_path  = rindex ( pkg_fname,    '/' ) ;
		
		if ( item_wo_path == NULL )
			item_wo_path = item -> pathname ;
		else
			item_wo_path++ ;
		if ( pkg_wo_path  == NULL )
			pkg_wo_path = pkg_fname ;
		else
			pkg_wo_path++ ;
		
		if ( strcmp ( item_wo_path, pkg_wo_path ) == 0 )
			pos = i ;
	}
	
	//fprintf ( stdout, "returning pos = %d\n", pos ) ;
	
	return pos ;
}

// emulate GLDynArray *pax_list
unsigned int PaxList::Count ()
{
	if ( pax_list_created == false || list == NULL ) return 0 ;
	
	return list -> Count () ;
}

PaxList::pax_item *PaxList::Item ( unsigned int n )
{
	if ( pax_list_created == false || list == NULL ) return NULL ;
	
	return list -> Item ( n ) ;
}

void PaxList::Draw ( int item )
{
	if ( pax_list_attached == false ||
			pax_list_created == false ||
			list == NULL ) return ;
	
	unsigned int min = 0, max = list -> Count () ;
	if ( item > -1 )
	{
		min = item ;
		max = item + 1 ;
	}
	
	for ( unsigned int t_item = min ; t_item < max ; t_item++ )
	{
		list -> Item ( t_item ) -> node -> Draw () ;
		list -> Item ( t_item ) -> screenshot -> Draw () ;
		if ( list -> Item ( t_item ) -> border )
			list -> Item ( t_item ) -> border -> Draw () ;
	}
}

void PaxList::Free ()
{
	if ( pax_list_created == false || list == NULL )
	{
		if ( list != NULL )
		{
			delete list ;
			list = NULL ;
		}
		file_cache . FreeCachedData ( "vault", MiscVault ) ;
		return ;
	}
	
	DetachList () ;
	
	fprintf ( stdout, "freeing pax list...\n" ) ;
	
	for ( unsigned int pos = 0 ; pos < list -> Count () ; pos++ )
	{
		if ( list -> Item ( pos ) -> pathname    ) free ( list -> Item ( pos ) -> pathname    ) ;
		if ( list -> Item ( pos ) -> shortname   ) free ( list -> Item ( pos ) -> shortname   ) ;
		if ( list -> Item ( pos ) -> name        ) free ( list -> Item ( pos ) -> name        ) ;
		if ( list -> Item ( pos ) -> author      ) free ( list -> Item ( pos ) -> author      ) ;
		if ( list -> Item ( pos ) -> description ) free ( list -> Item ( pos ) -> description ) ;

		if ( list -> Item ( pos ) -> gfx_name    ) free ( list -> Item ( pos ) -> gfx_name    ) ;
		
		if ( list -> Item ( pos ) -> vault )
		{
			if ( file_cache . FreeDataFromCache ( 
									list -> Item ( pos ) -> vault ) == 0 )
				delete list -> Item ( pos ) -> vault ;
		}
		
		delete list -> Item ( pos ) ;
	}
	
	list -> Clear () ;
	delete list ;
	list = NULL ;
	
	file_cache . FreeCachedData ( "vault", MiscVault ) ;
	
	fprintf ( stdout, "done.\n" ) ;
}
