#include "rain.h"

extern float current_lost_frames ;

extern FileCache file_cache ;

#include "video.h"
extern Video video ;

extern KrImNode *video_base_node ;

void Rain::Init ( KrImNode *new_basefield, char *rain_type )
{
	engine = video . GetEngine () ;
	using_opengl = video . GetUsingOpenGL () ;

	MAX_RAIN_DROPS = using_opengl ? 512 : 10 ;

	cur_alpha = (int*) malloc ( sizeof(int) * MAX_RAIN_DROPS ) ;
	dest_alpha = (int*) malloc ( sizeof(int) * MAX_RAIN_DROPS ) ;
	pos_in_life_time = (int*) malloc ( sizeof(int) * MAX_RAIN_DROPS ) ;
	drop = (KrSprite**) malloc ( sizeof(KrSprite*) * MAX_RAIN_DROPS ) ;
	drop_active = (bool*) malloc ( sizeof(bool) * MAX_RAIN_DROPS ) ; ;

	if ( new_basefield )
		basefield = new_basefield ;
	else
		basefield = video_base_node ;
	
	Vault = (KrResourceVault*) file_cache.CachedLoad ( "vault", "gfx/rain.dat" ) ;
	
	//fprintf ( stdout, "getting rainRes\n" ) ;
	rainRes = Vault->GetSpriteResource ( rain_type ) ;
	GLASSERT ( rainRes ) ;
	//fprintf( stdout, "creating raindrops... " ) ;
	for ( int i = 0 ; i < MAX_RAIN_DROPS ; i++ )
	{
		drop [ i ] = NULL ;
		drop_active [ i ] = FALSE ;
		pos_in_life_time [ i ] = 0 ;
	}
	//fprintf ( stdout, "done!\n" ) ;
	
	max_drops = 0 ;
	dest_max_drops = 0 ;
	dest_max_drops_counter = 0 ;
	dest_max_drops_delay = 2 ;
	old_max_drops = 0 ;

	rain_stopped = TRUE ;
	
	ThisTick = LastTick = SDL_GetTicks () ;
	
	/*{	// doesn't work yet :(( - I'm too stupid ;) or maybe too lazy =)
		KrRect basefield_rect ;
		basefield->ToSprite()->QueryBoundingBox ( &basefield_rect ) ;
		SetCreationArea ( basefield->X(), basefield->Y(), 
							basefield_rect.Width(), basefield_rect.Height() ) ;
	}*/
	SetCreationArea ( 0, -video . GetScreenY (), video . GetScreenX (), video . GetScreenY () ) ;
	SetSpeed ( 0, 20 ) ;

	SetLifeTime ( video . GetScreenY () * 2 ) ;
}

void Rain::SetAlpha ( int id, int new_alpha )
{
	if ( new_alpha < 0 ) new_alpha = 0 ;
	if ( new_alpha > 255 ) new_alpha = 255 ;

	dest_alpha [ id ] = new_alpha ;
}

void Rain::UpdateAlpha ( int id )
{
	KrColorTransform color = drop [ id ] -> CTransform () ;
	
	if ( cur_alpha [ id ] < dest_alpha [ id ] )
		cur_alpha [ id ] += 5 ;
	if ( cur_alpha [ id ] > dest_alpha [ id ] )
		cur_alpha [ id ] -= 5 ;
	
	if ( cur_alpha [ id ] < 0 ) cur_alpha [ id ] = 0 ;
	if ( cur_alpha [ id ] > 255 ) cur_alpha [ id ] = 255 ;
	
	color.SetAlpha ( cur_alpha [ id ] ) ;
	drop [ id ] -> SetColor ( color ) ;
}

void Rain::SetMaxDrops ( int new_max_drops )
{
	dest_max_drops = new_max_drops ;
	if ( dest_max_drops < 0 ) dest_max_drops = 0 ;
	if ( dest_max_drops > MAX_RAIN_DROPS ) dest_max_drops = MAX_RAIN_DROPS ;
}

void Rain::SetCreationArea ( int new_x, int new_y, int new_w, int new_h )
{
	area_x = new_x ;
	area_y = new_y ;
	area_w = new_w ;
	area_h = new_h ;
}

void Rain::SetLifeTime ( int new_life_time )
{
	life_time = new_life_time ;
}

void Rain::SetSpeed ( int new_speed_x, int new_speed_y, bool new_scheduled_waits )
{
	speed_x = new_speed_x ;
	speed_y = new_speed_y ;
	scheduled_waits = new_scheduled_waits ;
}

void Rain::AddDrop ( int id )
{
	if ( drop [ id ] == NULL )
	{
		drop [ id ] = new KrSprite ( rainRes ) ;
		//drop [ id ] -> SetPos ( 0, 0 ) ;
		drop [ id ] -> SetFrame ( 0 ) ;
		//drop [ id ] -> SetVisible ( FALSE ) ;
		engine->Tree()->AddNode ( basefield, drop [ id ] ) ;
		drop [ id ] -> SetZDepth ( 5 ) ;
		cur_alpha [ id ] = 128 ;
		dest_alpha [ id ] = 128 ;
	}
	
	drop [ id ] -> SetVisible ( TRUE ) ;
	drop [ id ] -> SetPos ( area_x + rand () % area_w, 
							area_y + rand () % area_h ) ;
	drop_active [ id ] = TRUE ;
	pos_in_life_time [ id ] = 0 ;

	/*{
		U8 red, green, blue ;
		red = blue = 255 ; green = 0 ;
		
		KrColorTransform color = drop [ id ] -> CTransform () ;
		
		color.TintRed ( red ) ;
		color.TintGreen ( green ) ;
		color.TintBlue ( blue ) ;
		
		drop [ id ] -> SetColor ( color ) ;
	}*/
	
	drop_wait_counter = 0 ;
	if ( scheduled_waits == TRUE )
		drop_wait_delay = life_time / 
							( dest_max_drops > 0 ? dest_max_drops : 1 ) ;
	else
		drop_wait_delay = 0 ;
}

void Rain::DelDrop ( int id )
{
	if ( drop [ id ] != NULL )
	{
		engine->Tree()->DeleteNode ( drop [ id ] ) ;
		drop [ id ] = NULL ;
		drop_active [ id ] = FALSE ;
	}
}

void Rain::Start ()
{
	if ( old_max_drops > 0 )
		dest_max_drops = old_max_drops ;
	
	rain_stopped = FALSE ;
}

void Rain::Stop ()
{
	if ( dest_max_drops > 0 )
		old_max_drops = dest_max_drops ;
	dest_max_drops = 0 ;
}

void Rain::Draw ()
{
	int drops_counted = 0 ;
	
	//if ( using_opengl == false ) return ;
	
	if ( rain_stopped == TRUE && dest_max_drops <= 0 )
		return ;

	if ( rain_stopped == FALSE )
	{
		for ( int i = 0 ; i < MAX_RAIN_DROPS ; i++ )
		{
			if ( drop_active [ i ] == TRUE )
			{
				double age_in_life = (double) 128 / (double) life_time * (double) pos_in_life_time [ i ] ;
				double pre_result = (double) 100 / (double) 128 * age_in_life ;
				int result = 128 - (int) ( 1.28 * pre_result ) ;
				
				//if ( using_opengl )
					SetAlpha ( i, result ) ;
				UpdateAlpha ( i ) ;
				
				if ( result > 0 )
				{
					drop [ i ] -> SetPos ( drop [ i ] -> X () + speed_x, 
											drop [ i ] -> Y () + speed_y ) ;
					drop [ i ] -> SetVisible ( ! drop [ i ] -> IsVisible () ) ;
					int average_speed = ( 	( speed_x < 0 ? -speed_x : speed_x ) + 
											( speed_y < 0 ? -speed_y : speed_y )
										) / 2 ;
					average_speed = average_speed < 1 ? 1 : average_speed ;
					pos_in_life_time [ i ] += average_speed ;
				}
				else if ( drops_counted < max_drops &&
							drop_wait_counter >= drop_wait_delay )
					AddDrop ( i ) ;
				else
					DelDrop ( i ) ;

				drops_counted++ ;
			}
			else if ( drops_counted < max_drops )
			{
				if ( drop_wait_counter >= drop_wait_delay )
				{
					AddDrop ( i ) ;
					drops_counted++ ;
				}
			}
		}
		drop_wait_counter++ ;
	}

	if ( dest_max_drops_counter >= dest_max_drops_delay )
	{
		ThisTick = SDL_GetTicks () ;
		/*if ( (int) current_lost_frames > 0 &&
				dest_max_drops > 0 && ThisTick - LastTick >= 500 )
		{
			dest_max_drops -= 2 ;
			if ( dest_max_drops <= 0 )
				dest_max_drops = 1 ;
		}*/
		if ( (int) current_lost_frames > 0 || dest_max_drops < max_drops )
		{
			max_drops -= ( using_opengl ? 4 : 4 ) ;
			int t_counter = 0 ;
			while ( drops_counted > dest_max_drops && ! rain_stopped &&
					t_counter < 2 )
			{
				int id = rand () % MAX_RAIN_DROPS ;
				if ( id >= 0 && id < MAX_RAIN_DROPS && drop_active [ id ] )
				{
					DelDrop ( id ) ;
					drops_counted-- ;
				}
				t_counter++ ;
			}
		}
		else if ( ThisTick - LastTick >= 250 && dest_max_drops > max_drops )
		{
			max_drops += ( using_opengl ? 8 : 4 ) ;
			LastTick = ThisTick ;
		}
		
		if ( max_drops < 0 ) 
			max_drops = 0 ;
		if ( max_drops > MAX_RAIN_DROPS ) 
			max_drops = MAX_RAIN_DROPS ;
		if ( dest_max_drops > 0 && max_drops <= 0 )
			max_drops = 1 ;

		dest_max_drops_counter = 0 ;
	}
	else
		dest_max_drops_counter++ ;

	if ( drops_counted <= 0 && max_drops <= 0 )
		rain_stopped = TRUE ;
	else
		rain_stopped = FALSE ;
}

bool Rain::RainHasStopped ()
{
	/*int drops_counted = 0 ;
	
	for ( int i = 0 ; i < MAX_RAIN_DROPS ; i++ )
		if ( drop [ i ] != NULL )
			drops_counted++ ;
	
	//fprintf ( stdout, "drops counted = %d\n", drops_counted ) ;
	if ( drops_counted <= 0 )
		fprintf ( stdout, "drops counted = %d\n", drops_counted ) ;
		
	return drops_counted > 0 ? FALSE : TRUE ;*/
	
	return rain_stopped ;
}

void Rain::Quit ()
{
	if ( drop == NULL ) return ;

	for ( int i = 0 ; i < MAX_RAIN_DROPS ; i++ )
	{
		if ( drop_active [ i ] == TRUE )
		{
			if ( drop [ i ] != NULL )
				engine->Tree()->DeleteNode ( drop [ i ] ) ;
			drop_active [ i ] = FALSE ;
		}
	}

	free ( cur_alpha ) ; cur_alpha = NULL ;
	free ( dest_alpha ) ; dest_alpha = NULL ;
	free ( pos_in_life_time ) ; pos_in_life_time = NULL ;
	free ( drop ) ; drop = NULL ;
	free ( drop_active ) ; drop_active = NULL ;
	
	// free references from vault - rest is handeled by kyra
	file_cache.FreeCachedData ( "vault", Vault ) ;
}
