#include "splash.h"

extern char rc_dir[] ;

extern FileCache file_cache ;

#include "video.h"
extern Video video ;

#include "gamescreen.h"
extern PaxList *pax_list ;

extern bool music_enabled ;

void Splash::Init ( char *t_pkg_filename )
{
	char pkg_filename [ 4096 ] ;
	sprintf ( pkg_filename, "%s", t_pkg_filename ) ;

	// copied from game.cpp ( could be better -> savegame.cpp ?? )
	if ( strncmp ( pkg_filename, "savestate:", 10 ) == 0 )
	{
		{
			char string [ 256 ] ;
			sprintf ( string, "%s", pkg_filename + 10 ) ;
			sprintf ( pkg_filename, "%s", string ) ;
		}
		
		FILE *stream ;
		char fname [ 256 ] ;
		sprintf ( fname, "%s/%s.state", rc_dir, pkg_filename ) ;
		stream = fopen ( fname, "r" ) ;
		
		if ( stream )
		{
			char name [ 4096 ] ;
			fgets ( name, 4096, stream ) ;
			
			sprintf ( pkg_filename, "%s", name ) ;
			
			fclose ( stream ) ;
			
			// changed against game.cpp!!! game.cpp would return index.pax too!
			*rindex ( pkg_filename, '/' ) = '\0' ;
		}
		else
		{
			fprintf ( stdout, "error: couldn't load saved game!\n" ) ;
			exit ( 1 ) ;
		}
	}
	//
	
	//fprintf ( stdout, "searching for pax: '%s'\n", pkg_filename ) ;
	
	unsigned int pos = 0 ;
	bool found = false ;
	for ( ; pos < pax_list -> Count () ; pos++ )
		if ( strcmp ( pax_list -> Item ( pos ) -> shortname, pkg_filename ) == 0 ||
				strcmp ( pax_list -> Item ( pos ) -> pathname, pkg_filename ) == 0 )
			{
				found = true ;
				break ;
			}
	
	if ( pos >= pax_list -> Count () || found == false )
	{
		fprintf ( stdout, "warning: couldn't create splash screen from pax:\n%s\n",
					pkg_filename ) ;
		item = NULL ;
		return ;
	}
	
	jingle = new Audio_Sound () ;
	jingle -> LoadSound ( "music/pink.ogg" ) ;
	// nasty hack to get volume of previously played music
	if ( Mix_VolumeMusic ( -1 ) && music_enabled ) // music enabled?
		jingle -> SetVolume ( Mix_VolumeMusic ( -1 ) / 2 ) ;
	else
		jingle -> SetVolume ( 0 ) ;

	finish = new Audio_Sound () ;
	finish -> LoadSound ( "sfx/glissup.wav" ) ;
	finish -> SetVolume ( 64 ) ;
	
	item = pax_list -> Item ( pos ) ;
	//fprintf ( stdout, "pax has path: '%s'\n", item -> pathname ) ;
	
	//fprintf ( stdout, "getting screenshot and creating splash screen...\n" ) ;
	
	item -> screenshot = new Sprite () ;
	item -> screenshot -> Init ( item -> vault, item -> gfx_name, NULL, true ) ;
	item -> screenshot -> SetMoveSound ( NULL ) ;
	item -> screenshot -> SetStopSound ( NULL ) ;
	item -> screenshot -> SetZ ( 1000 ) ;
	KrSpriteResource *item_res = item -> screenshot -> GetSprite () -> SpriteResource () ;
	if ( !item_res -> IsScaleCached ( 2.00, 2.00 ) )
		item_res -> CacheScale ( 2.00, 1.50 ) ;
	item -> screenshot -> SetXY_Center ( 320, 320 ) ;
	item -> screenshot -> GetSprite () -> SetScale ( 2.00, 1.50 ) ;
	
	item -> screenshot -> SetAlpha ( 0 ) ;
	item -> screenshot -> SetDestAlpha ( 255 ) ;
	
	item -> screenshot -> Draw () ;
	
	MiscVault = (KrResourceVault*) file_cache . CachedLoad ( "vault", "gfx/misc.dat" ) ;

	pb_border = new Sprite () ;
	pb_border -> Init ( MiscVault, "progressbar_border" ) ;
	pb_border -> SetMoveSound ( NULL ) ;
	pb_border -> SetStopSound ( NULL ) ;
	pb_border -> SetZ ( 1005 ) ;
	pb_border -> SetXY ( 20, 425 ) ;
	pb_border -> SetX_Center () ;
	pb_border -> SlideInFrom ( 0, 100 ) ;
	pb_border -> Draw () ;
	
	pb_bar = new Sprite () ;
	pb_bar -> Init ( MiscVault, "progressbar_bar", pb_border -> GetSprite () ) ;
	pb_bar -> SetMoveSound ( NULL ) ;
	pb_bar -> SetStopSound ( NULL ) ;
	pb_bar -> SetXY ( 4, 4 ) ;
	
	KrSpriteResource *pb_bar_res = pb_bar -> GetSprite () -> SpriteResource () ;
	if ( !pb_bar_res -> IsScaleCached ( 0.00, 1.00 ) )
		pb_bar_res -> CacheScale ( 0.00, 1.00 ) ;
	pb_bar -> GetSprite () -> SetScale ( 0.00, 1.00 ) ;
	pb_bar -> Draw () ;

	jingle -> PlaySound ( -1, FOREVER ) ;
	
	while ( item -> screenshot -> GetAlpha () < 255 ||
			! pb_border -> SlideFinished () )
	{
		item -> screenshot -> Draw () ;
		if ( item -> screenshot -> GetAlpha () >= 255 )
		{
			SDL_Delay ( 10 ) ;
			pb_border -> Draw () ;
		}
		video . UpdateDisplayOnly () ;
	}
	item -> screenshot -> Draw () ;
	pb_border -> Draw () ;
	video . UpdateDisplayOnly () ;
	SDL_Delay ( 500 ) ;
	video . UpdateDisplayOnly () ;
}

void Splash::SetPercent ( float value )
{
	if ( item == NULL || item -> screenshot == NULL ) return ;

	double percent = (double) value / 100.00 ;
	
	double cur_percent = pb_bar -> GetSprite () -> XScale () . ToDouble () ;
	
	while ( cur_percent != percent )
	{
		if ( cur_percent < percent )
		{
			cur_percent += 0.01 ;
			if ( cur_percent > percent )
				cur_percent = percent ;
		}
		
		if ( cur_percent > percent )
		{
			cur_percent -= 0.01 ;
			if ( cur_percent < percent )
				cur_percent = percent ;
		}
		
		if ( pb_border )
		{
			pb_border -> Draw () ;
			
			if ( pb_bar )
			{
				//KrSpriteResource *pb_bar_res = pb_bar -> GetSprite () -> SpriteResource () ;
				//if ( !pb_bar_res -> IsScaleCached ( cur_percent, 1.00 ) )
				//	pb_bar_res -> CacheScale ( cur_percent, 1.00 ) ;
				pb_bar -> GetSprite () -> SetScale ( cur_percent, 1.00 ) ;
				pb_bar -> Draw () ;
			}
		}
		if ( item && item -> screenshot )
			item -> screenshot -> Draw () ;
		
		video . UpdateDisplayOnly () ;
	}
	
	video . UpdateDisplayOnly () ;
	
	//fprintf ( stdout, "%.2f percent done\n", value ) ;
}

void Splash::Free ()
{
	if ( item == NULL || item -> screenshot == NULL ) return ;
	
	item -> screenshot -> SetDestAlpha ( 0 ) ;
	pb_border -> SlideOutTo ( 0, 100 ) ;
	jingle -> FadeOutSound ( 1000 ) ;
	finish -> PlaySound () ;
	while ( item -> screenshot -> GetAlpha () > 0 ||
			! pb_border -> SlideFinished () ||
			// jingle -> IsPlaying () ||
			finish -> IsPlaying () )
	{
		if ( pb_border -> SlideFinished () )
			item -> screenshot -> Draw () ;
		else
			SDL_Delay ( 10 ) ;
		pb_border -> Draw () ;
		video . UpdateDisplayOnly () ;
	}
	
	pb_bar -> Quit () ;
	pb_bar -> Free () ;
	delete pb_bar ;
	pb_bar = NULL ;
	
	pb_border -> Quit () ;
	pb_border -> Free () ;
	delete pb_border ;
	pb_border = NULL ;
	
	item -> screenshot -> Quit () ;
	item -> screenshot -> Free () ;
	delete item -> screenshot ;
	item -> screenshot = NULL ;
	item = NULL ;

	jingle -> UnloadSound () ;
	delete jingle ;
	jingle = NULL ;

	finish -> UnloadSound () ;
	delete finish ;
	finish = NULL ;
	
	file_cache . FreeCachedData ( "vault", MiscVault ) ;
}
