#include "sprite.h"

#include "math.h"

extern int level_of_detail ;

#include "video.h"
extern Video video ;

extern KrImNode *video_base_node ;

Sprite::Sprite ()
{
	quit = FALSE ;
	quit_state = FALSE ;
	already_freed = FALSE ;
}

void Sprite::Init ( KrResourceVault *Vault, char *name, KrImNode *node,
					int cached_scaling, int rain_speed_x, int rain_speed_y )
{
	KrEngine *engine = video . GetEngine () ;
	
	quit = FALSE ;
	quit_state = FALSE ;
	already_freed = FALSE ;
	
	using_opengl = video . GetUsingOpenGL () ;
	use_cached_scaling = cached_scaling == AUTO ? 
							video . GetUseCachedScaling () : cached_scaling ;
	
	anim_ipos . ox = 0 ; anim_ipos . oy = 0 ;
	anim_ipos . gx = 0 ; anim_ipos . gy = 0 ;
	
	anim_x_pos = 0 ;
	anim_y_pos = 0 ;
	anim_z_pos = 0 ;
	anim_zdepth_pos = 0 ;
	anim_dest_x_pos = 0 ;
	anim_dest_y_pos = 0 ;
	anim_dest_z_pos = 0 ;
	anim_x_speed = 5 ; anim_x_accel_factor = 1 ;
	anim_y_speed = 5 ; anim_y_accel_factor = 1 ;
	anim_xy_speed = ( anim_x_speed + anim_y_speed ) / 2 ;
	anim_xy_accel_factor = ( anim_x_accel_factor + anim_y_accel_factor ) / 2 ;
	anim_bounce_x_mod = anim_bounce_x_speed = 0 ;
	anim_bounce_y_mod = anim_bounce_y_speed = 0 ;
	anim_z_speed = anim_xy_speed ;
	anim_z_accel_factor = anim_xy_accel_factor ;
	anim_cur_frame = 0 ;
	anim_direction = 1 ;
	anim_bounce = FALSE ;
	anim_play_once = FALSE ;
	anim_delay = 0 ;
	anim_delay_counter = 0 ;
	anim_pause_interval = 0 ;
	anim_pause_interval_counter = 0 ;
	anim_pause_interval_randomness = FALSE ;
	anim_max_pause_interval = 0 ;

	anim_paused = FALSE ;

	anim_bounce_list = new GlDynArray < Sprite* > () ;
	anim_bounce_list -> Clear () ;

	base_vault = Vault ;
	
	animRes = base_vault->GetSpriteResource ( name ) ;
	//GLASSERT ( animRes ) ;
	if ( animRes == NULL )
	{
		fprintf ( stdout, "ERROR: Sprite not found '%s'\n", name ) ;
		exit ( 1 ) ;
	}
	anim = new KrSprite ( animRes ) ;
	SetIPos ( anim_x_pos, anim_y_pos ) ;
	anim->SetFrame ( anim_cur_frame ) ;
	if ( node )
		engine->Tree()->AddNode ( node, anim ) ;
	else
		engine->Tree()->AddNode ( video_base_node, anim ) ;
	anim_zdepth_pos = anim -> ZDepth () ;
	
	parent_node = node != NULL ? node -> ToSprite () : NULL ;
	
	anim_frames = anim->NumFrames () ;
	
	anim->QueryBoundingBox ( &anim_rect ) ;
	anim_w = anim_rect.Width() ;
	anim_h = anim_rect.Height() ;
	
	anim_x_pos_mod = 0 ;
	anim_y_pos_mod = 0 ;
	
	glow_activated = FALSE ;
	glow_direction = 1 ;
	glow_value = 0 ;
	glow_counter = 0 ;
	
	jump_height = 0 ;
	jump_pos = 0 ;
	jump_factor = 0 ;

	already_drawn = FALSE ;
	anim->SetVisible ( FALSE ) ;

	move_sound.LoadSound ( "sfx/swoosh.wav" ) ;
	move_sound.SetVolume ( 128 ) ;
	move_sound_enabled = true ;
	move_sound_play = false ;
	move_sound_started = false ;
	move_sound_channel = AUDIO_CHANNELS - 40 + ( rand() % 2 ) ;
	
	stop_sound.LoadSound ( "sfx/rumble.wav" ) ;
	stop_sound.SetVolume ( 128 ) ;
	stop_sound_enabled = true ;
	stop_sound_play = false ;
	stop_sound_started = false ;
	stop_sound_channel = AUDIO_CHANNELS - 35 + ( rand() % 10 ) ;

	// will be used by PlaySoundAtFrame ()
	extra_sound_enabled = false ;
	extra_sound_play = false ;
	extra_sound_started = false ;
	extra_sound_channel = AUDIO_CHANNELS - 25 + ( rand() % 10 ) ;
	
	exploding = FALSE ;
	sprintf ( explosion_resource, "%s", "" ) ;
	explode_sound.LoadSound ( "sfx/metal.wav" ) ;
	explode_sound.SetVolume ( 64 ) ;
	explosion_tint_red = 0 ;
	explosion_tint_green = 0 ;
	explosion_tint_blue = 0 ;
	pre_explosion_phase_done = FALSE ;
	pre_explosion_phase_pos = 0 ;
	explosion_sound_channel = AUDIO_CHANNELS - 15 + ( rand() % 10 ) ;

	creation_sound_enabled = false ;
	creation_sound_started = false ;
	creation_sound_channel = -1 ; //AUDIO_CHANNELS - 5 + ( rand() % 5 ) ;
	
	// could be played when accellerating ( set anim_*_accel_factor > 1 )
	accel_sound_enabled = false ;
	accel_sound_started = false ;
	accel_sound_channel = -1 ;
	
	// will be played when stopping after accellerating
	bump_sound_enabled = false ;
	bump_sound_started = false ;
	bump_sound_channel = -1 ;
	
	InitRain ( node, rain_speed_x, rain_speed_y ) ;

	current_alpha = 255 ;
	dest_alpha = 255 ;
	SetAlpha ( 255 ) ;
	Darken ( 0 ) ;
	Brighten ( 0 ) ;
	
	SetFadeSpeed ( using_opengl ? 5 : 17 ) ;
}

void Sprite::SetIPos ( int x, int y )
{
/*	ARGH!! i'm too dumb for this :)
	float time = video . GetTime () ;
	float fframe = time - floor ( time ) ;
	
	anim_ipos . ox = GetXPos () ;
	anim_ipos . oy = GetYPos () ;

	anim_ipos . gx = anim_ipos . ox * (1 - fframe) + GetXPos() * fframe ;
	anim_ipos . gy = anim_ipos . oy * (1 - fframe) + GetYPos() * fframe ;

	anim->SetPos ( (int) anim_ipos . gx, (int) anim_ipos . gy ) ;*/
	
	anim->SetPos ( x, y ) ;
}

void Sprite::SetXY ( int x, int y )
{
	anim_x_pos = anim_dest_x_pos = x ;
	anim_y_pos = anim_dest_y_pos = y ;
	
	if ( quit_state == TRUE ) return ;
	
	double dscale_mod = ( using_opengl ? 1.00 : 0.50 ) ;
	
	double dscale = (double) anim_z_pos * 0.01 * dscale_mod ;
	int sx = (int) ( ( (double) ( anim_w / 2 ) ) * dscale ) ;
	int sy = (int) ( ( (double) ( anim_h / 2 ) ) * dscale ) ;
	
	if ( level_of_detail < 3 ) { sx = 0 ; sy = 0 ; }

	SetIPos ( anim_x_pos + anim_x_pos_mod - sx + anim_bounce_x_mod, 
				anim_y_pos + anim_y_pos_mod - sy + anim_bounce_y_mod ) ;

	dscale = (double) 1 + ( (double) anim_z_pos * 0.01 * dscale_mod ) ;
	GlFixed scale = dscale ;
	if ( level_of_detail > 2 )
	{
		if ( use_cached_scaling &&
				!animRes->IsScaleCached ( scale, scale ) )
			animRes->CacheScale ( scale, scale ) ;
		anim->SetScale ( scale, scale ) ;
		//anim->SetZDepth ( anim_zdepth_pos + scale.v ) ;
	}
}

void Sprite::SetDestXY ( int x, int y )
{
	anim_dest_x_pos = x ;
	anim_dest_y_pos = y ;
	
	if ( quit_state == TRUE ) return ;
	
	if ( level_of_detail < 3 )
		SetXY ( x, y ) ;
}

void Sprite::SetX_Center ( int pseudo_width )
{
	int node_w = pseudo_width == 0 ? video . GetScreenX () : pseudo_width ;
	int node_h = video . GetScreenY () ;
	
	if ( parent_node != NULL )
	{
		KrRect parent_rect ;
		parent_node -> QueryBoundingBox ( &parent_rect ) ;
		node_w = pseudo_width == 0 ? parent_rect.Width() : pseudo_width ;
		node_h = parent_rect.Height() ;
		
		/*fprintf ( stdout, "parent node has %d x %d pixels - ",
					node_w, node_h ) ;
		fprintf ( stdout, "this node has %d x %d pixels and %d frames\n",
					anim_w, anim_h, anim_frames ) ;*/
	}

	SetXY ( ( node_w - ( anim_w / anim_frames ) ) / 2, 
			GetYPos () ) ;
}

void Sprite::SetY_Center ( int pseudo_height )
{
	int node_w = video . GetScreenX () ;
	int node_h = pseudo_height == 0 ? video . GetScreenY () : pseudo_height ;
	
	if ( parent_node != NULL )
	{
		KrRect parent_rect ;
		parent_node -> QueryBoundingBox ( &parent_rect ) ;
		node_w = parent_rect.Width() ;
		node_h = pseudo_height == 0 ? parent_rect.Height() : pseudo_height ;
		
		/*fprintf ( stdout, "parent node has %d x %d pixels - ",
					node_w, node_h ) ;
		fprintf ( stdout, "this node has %d x %d pixels and %d frames\n",
					anim_w, anim_h, anim_frames ) ;*/
	}

	SetXY ( GetXPos (), 
			( node_h - anim_h ) / 2 ) ;
}

void Sprite::SetXY_Center ( int pseudo_width, int pseudo_height )
{
	int node_w = pseudo_width  == 0 ? video . GetScreenX () : pseudo_width  ;
	int node_h = pseudo_height == 0 ? video . GetScreenY () : pseudo_height ;
	
	if ( parent_node != NULL )
	{
		KrRect parent_rect ;
		parent_node -> QueryBoundingBox ( &parent_rect ) ;
		node_w = pseudo_width  == 0 ? parent_rect.Width()  : pseudo_width  ;
		node_h = pseudo_height == 0 ? parent_rect.Height() : pseudo_height ;
		
		/*fprintf ( stdout, "parent node has %d x %d pixels - ",
					node_w, node_h ) ;
		fprintf ( stdout, "this node has %d x %d pixels and %d frames\n",
					anim_w, anim_h, anim_frames ) ;*/
	}

	SetXY ( ( node_w - ( anim_w / anim_frames ) ) / 2, 
			( node_h - anim_h ) / 2 ) ;
}

void Sprite::SetXY_LeftOf ( int x )
{
	SetXY ( x - ( anim_w / anim_frames ), anim_y_pos ) ;
}

void Sprite::SetXY_RightOf ( int x )
{
	SetXY ( x, anim_y_pos ) ;
}

void Sprite::SetXY_AboveOf ( int y )
{
	SetXY ( anim_x_pos, y - anim_h ) ;
}

void Sprite::SetXY_BelowOf ( int y )
{
	SetXY ( anim_x_pos, y ) ;
}

void Sprite::SetZ ( int z )
{
	anim->SetZDepth ( anim_zdepth_pos + z ) ;
}

void Sprite::SetZDepthPos ( int z )
{
	anim_zdepth_pos = z ;
	anim->SetZDepth ( anim_zdepth_pos ) ;
}

void Sprite::SetBounce ( int x_speed, int y_speed )
{
	if ( level_of_detail < 3 ) return ;
	anim_bounce_x_speed = x_speed ;
	anim_bounce_y_speed = y_speed ;
}

void Sprite::SetXSpeed ( int value )
{
	anim_x_speed  = value ;
	anim_xy_speed = ( anim_x_speed + anim_y_speed ) / 2 ;
}

void Sprite::SetBounceList ( GlDynArray < Sprite* > *new_bounce_list )
{
	anim_bounce_list -> Clear () ;
	delete anim_bounce_list ;
	anim_bounce_list = new_bounce_list ;
	if ( level_of_detail < 3 )
		anim_bounce_list -> Clear () ;
}

void Sprite::SetYSpeed ( int value )
{
	anim_y_speed  = value ;
	anim_xy_speed = ( anim_x_speed + anim_y_speed ) / 2 ;
}

void Sprite::SetXYSpeed ( int value )
{
	anim_xy_speed = value ;
}

void Sprite::SetXAccelFactor ( int value )
{
	anim_x_accel_factor  = value ;
	anim_xy_accel_factor = ( anim_x_accel_factor + anim_y_accel_factor ) / 2 ;
}

void Sprite::SetYAccelFactor ( int value )
{
	anim_y_accel_factor  = value ;
	anim_xy_accel_factor = ( anim_x_accel_factor + anim_y_accel_factor ) / 2 ;
}

void Sprite::SetXYAccelFactor ( int value )
{
	anim_xy_accel_factor = value ;
}

void Sprite::SetFrames ( int pos, int frames, int direction, bool bounce, bool once )
{
	if ( pos == RANDOM )
	{
		pos = rand () % anim_frames ;
		//fprintf ( stdout, "pos (RANDOM) = %d\n", pos ) ;
	}
	if ( pos != AUTO ) 
	{
		if ( pos < 0 )
			anim_cur_frame = anim_frames - pos ;
		else
			anim_cur_frame = pos ;
		
		if ( anim_cur_frame > anim_frames )
			anim_cur_frame = anim_frames ;
		if ( anim_cur_frame < 0 )
			anim_cur_frame = 0 ;
	}
	if ( frames != AUTO ) anim_frames = frames ;
	if ( direction != AUTO ) anim_direction = direction ;
	anim_bounce = bounce ;
	anim_play_once = once ;

	if ( quit_state == TRUE ) return ;
	
	anim->SetFrame ( anim_cur_frame ) ;
}

void Sprite::SetDelays ( int delay, int delay_counter,
									int pause_interval,
									int pause_interval_counter,
									bool pause_interval_randomness,
									int max_pause_interval )
{
	anim_delay = delay ;
	anim_delay_counter = delay_counter ;
	anim_pause_interval = pause_interval ;
	anim_pause_interval_counter = pause_interval_counter ;
	anim_pause_interval_randomness = pause_interval_randomness ;
	anim_max_pause_interval = max_pause_interval ;
	
}

void Sprite::SetExplosionResource ( char *name )
{
	sprintf ( explosion_resource, "%s", name ) ;
}

void Sprite::SetExplosionColor ( U8 red, U8 green, U8 blue )
{
	explosion_tint_red = red ;
	explosion_tint_green = green ;
	explosion_tint_blue = blue ;
}

void Sprite::SetTint ( U8 red, U8 green, U8 blue )
{
	if ( quit_state == TRUE ) return ;
	
	if ( level_of_detail < 2 ) return ;
	
	GLASSERT ( anim ) ;
	KrColorTransform color = anim -> CTransform () ;
	GLASSERT ( color ) ;

	color.TintRed ( red ) ;
	color.TintGreen ( green ) ;
	color.TintBlue ( blue ) ;
	anim -> SetColor ( color ) ;
}

void Sprite::Brighten ( U8 value )
{
	if ( quit_state == TRUE ) return ;
	
	if ( level_of_detail < 2 ) return ;
	
	KrColorTransform color = anim -> CTransform () ;
	
	current_brighten = dest_brighten = value ;
	
	color.Brighten ( current_brighten ) ;
	anim -> SetColor ( color ) ;
}

void Sprite::SetDestBrighten ( U8 value )
{
	if ( quit_state == TRUE ) return ;
	
	if ( level_of_detail < 2 ) return ;
	
	if ( value != dest_brighten )
	{
		dest_brighten = value ;
	}
}

void Sprite::Darken ( U8 value )
{
	if ( quit_state == TRUE ) return ;
	
	if ( level_of_detail < 2 ) return ;
	
	KrColorTransform color = anim -> CTransform () ;
	
	current_darken = dest_darken = value ;
	
	color.Darken ( current_darken ) ;
	anim -> SetColor ( color ) ;
}

void Sprite::SetDestDarken ( U8 value )
{
	if ( quit_state == TRUE ) return ;
	
	if ( level_of_detail < 2 ) return ;
	
	if ( value != dest_darken )
	{
		dest_darken = value ;
	}
}

void Sprite::SetFadeSpeed ( U8 value )
{
	fade_speed = value ;
}

U8   Sprite::GetFadeSpeed ()
{
	return fade_speed ;
}

void Sprite::SetAlpha ( int new_alpha )
{
	if ( new_alpha < 0 ) new_alpha = 0 ;
	if ( new_alpha > 255 ) new_alpha = 255 ;

	if ( new_alpha != dest_alpha && quit_state == FALSE )
	{
		current_alpha = dest_alpha = new_alpha ;
	
		if ( level_of_detail > 1 )
		{	
			KrColorTransform color = anim -> CTransform () ;
			
			color.SetAlpha ( current_alpha ) ;
			anim -> SetColor ( color ) ;
		}
		
		if ( anim -> IsVisible () != ( current_alpha > 0 ) )
			anim -> SetVisible ( current_alpha > 0 ) ;
	}
}

void Sprite::SetDestAlpha ( int new_alpha )
{
	if ( new_alpha < 0 ) new_alpha = 0 ;
	if ( new_alpha > 255 ) new_alpha = 255 ;

	if ( new_alpha != dest_alpha )
	{
		dest_alpha = new_alpha ;
	}
}

int Sprite::GetAlpha ()
{
	return current_alpha ;
}

KrSprite *Sprite::GetSprite ()
{
	return ( quit_state == FALSE ? anim : NULL ) ;
}

KrImNode *Sprite::FindNodeByName ( char *name )
{
	KrImNode *node = video . GetEngine () -> Tree () -> FindNodeByName ( name ) ;
	
	if ( node == NULL )
		fprintf ( stdout, "Couldn't find node called '%s'.\n", name ) ;
	
	return node ;
}

int Sprite::GetXPos ()
{
	return anim_x_pos ;
}

int Sprite::GetYPos ()
{
	return anim_y_pos ;
}

int Sprite::GetWidth ()
{
	return anim_w ;
}

int Sprite::GetHeight ()
{
	return anim_h ;
}

KrSprite *Sprite::GetParent ()
{
	return parent_node ;
}

void Sprite::InitRain ( KrImNode *node, int speed_x, int speed_y )
{
	rain.Init ( node, "dot" ) ;
	if ( using_opengl )
		rain.SetMaxDrops ( 64 / 3 * level_of_detail ) ;
	else
		rain.SetMaxDrops ( 32 / 3 * level_of_detail ) ;
	//SetRainCreationArea ( anim->X(), anim->Y(), GetWidth (), GetHeight () ) ;
	SetRainCreationArea ( 0, GetHeight () / 3, GetWidth (), GetHeight () * 2 / 3 ) ;
	rain.SetLifeTime ( GetHeight () * 4 ) ;
	rain.SetSpeed ( speed_x, speed_y, TRUE ) ;
	rain.Stop () ;
	twinkling_activated = FALSE ;
}

void Sprite::SetRainCreationArea ( int x, int y, int w, int h )
{
	if ( x != AUTO ) rain_x = x ;
	if ( y != AUTO ) rain_y = y ;
	if ( w != AUTO ) rain_w = w ;
	if ( h != AUTO ) rain_h = h ;
	
	rain.SetCreationArea ( anim->X() + rain_x, anim->Y() + rain_y, rain_w, rain_h ) ;
}

int Sprite::ConvertXPos ( KrImNode *src_base, int x )
{
	if ( quit_state == TRUE ) return 0 ;
	
	KrVector2T< GlFixed > object ;
	KrImNode *current_node = src_base ;
	int current_x = src_base->X () + x ;
	
	//fprintf ( stdout, "current x = %d\n", current_x ) ;
	
	while ( current_node->Parent () )
	{
		current_node = current_node->Parent () ;
		current_x += current_node->X () ;
		//fprintf ( stdout, "current x = %d\n", current_x ) ;
	}
	
	anim->ScreenToObject ( current_x, 0, &object ) ;
	
	current_x = object.x.ToIntRound () ;
	
	//fprintf ( stdout, "dest x = %d\n", current_x ) ;
	
	return current_x ;
}

int Sprite::ConvertYPos ( KrImNode *src_base, int y )
{
	if ( quit_state == TRUE ) return 0 ;
	
	KrVector2T< GlFixed > object ;
	KrImNode *current_node = src_base ;
	int current_y = src_base->Y () + y ;
	
	//fprintf ( stdout, "current y = %d\n", current_y ) ;
	
	while ( current_node->Parent () )
	{
		current_node = current_node->Parent () ;
		current_y += current_node->Y () ;
		//fprintf ( stdout, "current y = %d\n", current_y ) ;
	}
	
	anim->ScreenToObject ( 0, current_y, &object ) ;
	
	current_y = object.y.ToIntRound () ;
	
	//fprintf ( stdout, "dest y = %d\n", current_y ) ;
	
	return current_y ;
}

void Sprite::HandOverToBaseAt ( KrImNode *new_base, int x, int y )
{
	if ( quit_state == TRUE ) return ;

	rain.Stop () ;
	rain.Quit () ;
	
	KrEngine *engine = video . GetEngine () ;
	
	//fprintf ( stdout, "delete node from tree\n" ) ;
	engine->Tree()->DeleteNode ( anim ) ;
	
	//fprintf ( stdout, "create new node...\n" ) ;
	anim = new KrSprite ( animRes ) ;
	SetIPos ( anim_x_pos, anim_y_pos ) ;
	anim->SetFrame ( anim_cur_frame ) ;

	//fprintf ( stdout, "add node to tree\n" ) ;
	engine->Tree()->AddNode ( new_base, anim ) ;
	
	//fprintf ( stdout, "set position\n" ) ;
	SetXY ( x, y ) ;
	
	InitRain ( new_base ) ;
}

void Sprite::SlideInFrom ( int x, int y )
{
	// slide-in effect setup
	step_x = step_y = 0.5 ; // for initialization
	slide_speed = 1.05 ;

	// precalculate slide-in-effect horizontally...
	if ( x < 0 )
	{
		for ( anim_x_pos_mod = 0 ; anim_x_pos_mod >= x - (int) step_x ; anim_x_pos_mod -= (int) step_x )
		{
				step_x = step_x * slide_speed ;
		}
	
		anim_x_pos_mod += (int) step_x ;
	} 
	else if ( x > 0 ) 
	{
		for ( anim_x_pos_mod = 0 ; anim_x_pos_mod <= x + (int) step_x ; anim_x_pos_mod += (int) step_x )
		{
				step_x = step_x * slide_speed ;
		}
	
		anim_x_pos_mod -= (int) step_x ;
	}

	// precalculate slide-in-effect vertically...
	if ( y < 0 )
	{
		for ( anim_y_pos_mod = 0 ; anim_y_pos_mod >= y - (int) step_y ; anim_y_pos_mod -= (int) step_y )
		{
				step_y = step_y * slide_speed ;
		}
	
		anim_y_pos_mod += (int) step_y ;
	} 
	else if ( y > 0 ) 
	{
		for ( anim_y_pos_mod = 0 ; anim_y_pos_mod <= y + (int) step_y ; anim_y_pos_mod += (int) step_y )
		{
				step_y = step_y * slide_speed ;
		}
	
		anim_y_pos_mod -= (int) step_y ;
	}

	if ( level_of_detail < 3 )
		anim_x_pos_mod = anim_y_pos_mod = 0 ;
	
	SetXY ( anim_x_pos, anim_y_pos ) ;
	
	move_sound_play = true ;
	move_sound_started = false ;
}

void Sprite::SlideOutTo ( int x, int y )
{
	// slide-out effect setup
	step_x = step_y = 0.5 ; // for initialization
	slide_speed = 1.05 ;

	// precalculate slide-out-effect horizontally...
	if ( x < 0 )
	{
		for ( anim_x_pos_mod = 0 ; anim_x_pos_mod >= x - (int) step_x ; anim_x_pos_mod -= (int) step_x )
		{
				step_x = step_x * slide_speed ;
		}
	
		anim_x_pos_mod += (int) step_x ;
	} 
	else if ( x > 0 ) 
	{
		for ( anim_x_pos_mod = 0 ; anim_x_pos_mod <= x + (int) step_x ; anim_x_pos_mod += (int) step_x )
		{
				step_x = step_x * slide_speed ;
		}
		anim_x_pos_mod -= (int) step_x ;
	}

	// precalculate slide-out-effect vertically...
	if ( y < 0 )
	{
		for ( anim_y_pos_mod = 0 ; anim_y_pos_mod >= y - (int) step_y ; anim_y_pos_mod -= (int) step_y )
		{
				step_y = step_y * slide_speed ;
		}
	
		anim_y_pos_mod += (int) step_y ;
	}
	else if ( y > 0 )
	{
		for ( anim_y_pos_mod = 0 ; anim_y_pos_mod <= y + (int) step_y ; anim_y_pos_mod += (int) step_y )
		{
				step_y = step_y * slide_speed ;
		}
	
		anim_y_pos_mod -= (int) step_y ;
	}
	
	anim_x_pos_mod = -anim_x_pos_mod ;
	anim_y_pos_mod = -anim_y_pos_mod ;
	
	if ( level_of_detail < 3 )
	{
		anim_x_pos_mod = anim_y_pos_mod = 0 ;

		SetXY ( anim_x_pos + x, anim_y_pos + y ) ;
	}
	else
		SetXY ( anim_x_pos - anim_x_pos_mod, anim_y_pos - anim_y_pos_mod ) ;
	
	if ( level_of_detail < 3 )
		anim_x_pos_mod = anim_y_pos_mod = 0 ;

	move_sound_play = true ;
	move_sound_started = false ;
}

void Sprite::DoSlide ()
{
	// slide-in horizontally
	if ( anim_x_pos_mod > 0 )
	{
		anim_x_pos_mod -= (int) step_x ;
		
		step_x = step_x / slide_speed ;
		
		if ( anim_x_pos_mod < 0 )
			anim_x_pos_mod = 0 ;
		
		SetXY ( anim_x_pos, anim_y_pos ) ;
	} else if ( anim_x_pos_mod < 0 )
	{
		anim_x_pos_mod += (int) step_x ;
		
		step_x = step_x / slide_speed ;
		
		if ( anim_x_pos_mod > 0 )
			anim_x_pos_mod = 0 ;

		SetXY ( anim_x_pos, anim_y_pos ) ;
	}
	//
	// slide-in vertically
	if ( anim_y_pos_mod > 0 )
	{
		anim_y_pos_mod -= (int) step_y ;
		
		step_y = step_y / slide_speed ;
		
		if ( anim_y_pos_mod < 0 )
			anim_y_pos_mod = 0 ;

		SetXY ( anim_x_pos, anim_y_pos ) ;
	} else if ( anim_y_pos_mod < 0 )
	{
		anim_y_pos_mod += (int) step_y ;
		
		step_y = step_y / slide_speed ;
		
		if ( anim_y_pos_mod > 0 )
			anim_y_pos_mod = 0 ;
		
		SetXY ( anim_x_pos, anim_y_pos ) ;
	}
	//
	
	if ( move_sound_play == true && move_sound_started == false )
	{
		move_sound.PlaySound ( move_sound_channel ) ;
		move_sound_started = true ;
	}
}

bool Sprite::SlideFinishedX ()
{
	if ( anim_x_pos_mod == 0 && anim_x_pos == anim_dest_x_pos )
		return true ;
	else
		return false ;
}

bool Sprite::SlideFinishedY ()
{
	if ( anim_y_pos_mod == 0 && anim_y_pos == anim_dest_y_pos )
		return true ;
	else
		return false ;
}

bool Sprite::SlideFinishedZ ()
{
	if ( anim_z_pos == 0 && anim_z_pos == anim_dest_z_pos )
		return true ;
	else
		return false ;
}

bool Sprite::SlideFinished ()
{
	if ( SlideFinishedX () && SlideFinishedY () && SlideFinishedZ () &&
			current_alpha == dest_alpha &&
			current_brighten == dest_brighten &&
			current_darken == dest_darken )
		return true ;
	else
		return false ;
}

void Sprite::PlaySoundAtFrame ( char *filename, int framenumber )
{
	if ( extra_sound_enabled )
	{
		extra_sound.UnloadSound () ;
		extra_sound_enabled = false ;
	}
	
	if ( filename != NULL )
	{
		if ( framenumber < 0 )
		{
			framenumber = anim_frames - framenumber ;
		}
		
		extra_sound.LoadSound ( filename ) ;
		extra_sound_enabled = true ;
		extra_sound_framenumber = framenumber ;
		extra_sound.SetVolume ( 32 ) ;
	}
}

void Sprite::SetMoveSound ( char *filename, int volume )
{
	if ( move_sound_enabled )
	{
		move_sound.UnloadSound () ;
		move_sound_enabled = false ;
	}
	
	if ( filename != NULL )
	{
		move_sound.LoadSound ( filename ) ;
		move_sound.SetVolume ( volume ) ;
		move_sound_enabled = TRUE ;
	}
}

void Sprite::SetStopSound ( char *filename, int volume )
{
	if ( stop_sound_enabled )
	{
		stop_sound.UnloadSound () ;
		stop_sound_enabled = false ;
	}
	
	if ( filename != NULL )
	{
		stop_sound.LoadSound ( filename ) ;
		stop_sound.SetVolume ( volume ) ;
		stop_sound_enabled = TRUE ;
	}
}

void Sprite::SetCreationSound ( char *filename, int volume )
{
	if ( creation_sound_enabled )
	{
		creation_sound.UnloadSound () ;
		creation_sound_enabled = false ;
	}
	
	if ( filename != NULL )
	{
		creation_sound.LoadSound ( filename ) ;
		creation_sound.SetVolume ( volume ) ;
		creation_sound_enabled = TRUE ;
	}
}

void Sprite::SetAccelSound ( char *filename, int volume )
{
	if ( accel_sound_enabled )
	{
		accel_sound.UnloadSound () ;
		accel_sound_enabled = false ;
	}
	
	if ( filename != NULL )
	{
		accel_sound.LoadSound ( filename ) ;
		accel_sound.SetVolume ( volume ) ;
		accel_sound_enabled = TRUE ;
	}
}

void Sprite::SetBounceSound ( char *filename, int volume )
{
	if ( bump_sound_enabled )
	{
		bump_sound.UnloadSound () ;
		bump_sound_enabled = false ;
	}
	
	if ( filename != NULL )
	{
		bump_sound.LoadSound ( filename ) ;
		bump_sound.SetVolume ( volume ) ;
		bump_sound_enabled = TRUE ;
	}
}

void Sprite::SetPanning ( Audio_Sound sound_struct )
{
	KrImNode *parent = anim -> Parent () ;
	while ( parent -> Parent () )
		parent = parent -> Parent () ;
	
	int converted_pos = ConvertXPos ( parent, -GetWidth () / 2 ) * -1 ;
	int panning = (int) ( (double) 255 / (double) video . GetScreenX () *
							(double) converted_pos ) ;
	//fprintf ( stdout, "converted x pos = %d\n", converted_pos ) ;
	//fprintf ( stdout, "panning would be at %d\n", panning ) ;
	sound_struct.SetPanning ( panning ) ;
}

void Sprite::PlayMoveSound ( int channel, int times )
{
	if ( move_sound_enabled == TRUE )
	{
		move_sound.PlaySound ( channel == -1 ? move_sound_channel : channel, 
								times ) ;
		SetPanning ( move_sound ) ;
	}
}

void Sprite::PlayStopSound ( int channel, int times )
{
	if ( stop_sound_enabled == TRUE )
	{
		stop_sound.PlaySound ( channel == -1 ? stop_sound_channel : channel, 
								times ) ;
		SetPanning ( stop_sound ) ;
	}
}

void Sprite::PlayExtraSound ( int channel, int times )
{
	if ( extra_sound_enabled == TRUE )
	{
		extra_sound.PlaySound ( channel == -1 ? extra_sound_channel : channel, 
								times ) ;
		SetPanning ( extra_sound ) ;
	}
}

void Sprite::PlayExplodeSound ( int channel, int times )
{
	//if ( explosion_sound_enabled == TRUE )
	{
		explode_sound.PlaySound ( channel == -1 ? explosion_sound_channel : channel, 
								times ) ;
		SetPanning ( stop_sound ) ;
	}
}

void Sprite::PlayCreationSound ( int channel, int times )
{
	if ( creation_sound_enabled == TRUE )
	{
		creation_sound.PlaySound ( channel == -1 ? creation_sound_channel : channel, 
									times ) ;
		SetPanning ( creation_sound ) ;
	}
}

void Sprite::PlayAccelSound ( int channel, int times )
{
	if ( accel_sound_enabled == TRUE )
	{
		accel_sound.PlaySound ( channel == -1 ? accel_sound_channel : channel, 
									times ) ;
		SetPanning ( accel_sound ) ;
	}
}

void Sprite::PlayBounceSound ( int channel, int times )
{
	if ( bump_sound_enabled == TRUE )
	{
		bump_sound.PlaySound ( channel == -1 ? bump_sound_channel : channel, 
									times ) ;
		SetPanning ( bump_sound ) ;
	}
}

void Sprite::Glow ( bool activate )
{
	glow_activated = activate ;
}

void Sprite::Twinkling ( bool activate, int new_max_drops, int new_lifetime )
{
	if ( twinkling_activated == FALSE && activate == TRUE )
	{
		rain.Start () ;
	}
	if ( twinkling_activated == TRUE && activate == FALSE )
	{
		rain.Stop () ;
	}
	twinkling_activated = activate ;
	
	if ( activate == TRUE && new_max_drops != -1 )
		rain.SetMaxDrops ( new_max_drops ) ;
	
	if ( activate == TRUE && new_lifetime != -1 )
		rain.SetLifeTime ( new_lifetime ) ;
}

void Sprite::SetTwinklingSpeed ( int speed_x, int speed_y )
{
	rain.SetSpeed ( speed_x, speed_y ) ;
}

void Sprite::Jump ( int new_height )
{
	jump_height = new_height ;
	anim_z_speed = anim_xy_speed ; // = 5
	anim_z_accel_factor = anim_xy_accel_factor ;
	jump_factor = 0 ;
	anim_z_pos = 0 ;
	jump_pos = 0 ;
}

void Sprite::Explode ( unsigned int wait_time, unsigned int speed )
{
	if ( GetAlpha () < 1 )
	{
		pre_explosion_phase_done = TRUE ;
		exploding = TRUE ;
	}
		
	if ( pre_explosion_phase_done == TRUE || exploding == TRUE ) return ;
	
	if ( speed < 1 ) speed = ( rand () % 40 ) + 5 ;
	if ( speed > 255 ) speed = 255 ;
	
	exploding = TRUE ;
	
	pre_explosion_phase_done = FALSE ;
	pre_explosion_phase_pos = wait_time ;
	//glow_value = 0 ;
	glow_activated = FALSE ;
	if ( using_opengl )
		rain.SetMaxDrops ( 64 / 3 * level_of_detail ) ;
	else
		rain.SetMaxDrops ( 0/*8 / 3 * level_of_detail*/ ) ;
	Twinkling ( wait_time == 0 ) ;

	pre_explosion_speed = speed ;
}

void Sprite::ExplodePhaseTwo ()
{
	if ( quit_state == TRUE ) return ;
	
	KrEngine *engine = video . GetEngine () ;
	
	anim_cur_frame = 2 ;
	
	GLASSERT ( anim ) ;
	GLASSERT ( anim->Parent() ) ;
	KrImNode *node = anim->Parent() ;
	GLASSERT ( node ) ;
	
	engine->Tree()->DeleteNode ( anim ) ;
	
	animRes = base_vault->GetSpriteResource( explosion_resource ) ;
	GLASSERT ( animRes ) ;
	anim = new KrSprite ( animRes ) ;
	SetIPos ( anim_x_pos, anim_y_pos ) ;
	anim->SetFrame ( anim_cur_frame ) ;
	engine->Tree()->AddNode ( node, anim ) ;
	
	anim_frames = anim->NumFrames () ;
	
	anim->QueryBoundingBox ( &anim_rect ) ;
	anim_w = anim_rect.Width() ;
	anim_h = anim_rect.Height() ;
	
	SetTint ( explosion_tint_red, explosion_tint_green, explosion_tint_blue ) ;
	
	PlayExplodeSound () ;
}

void Sprite::PauseAnim ()
{
	anim_paused = TRUE ;
}

void Sprite::ContinueAnim ()
{
	anim_paused = FALSE ;
}

void Sprite::Draw ()
{
	if ( quit_state == TRUE ) return ;
	
	if ( quit == TRUE && exploding == FALSE && quit_state == FALSE )
	{
		int alpha = GetAlpha () ;
		
		if ( alpha <= 0 ) 
		{
			Twinkling ( FALSE ) ;
			anim->SetVisible ( false ) ;
			if ( rain.RainHasStopped () == TRUE )
			{
				//fprintf ( stdout, "1 - rain has stopped. quitting...\n" ) ;
				quit_state = TRUE ;
			}
		}
		
		SetAlpha ( alpha - 5 ) ;
	}
	else if ( current_alpha != dest_alpha ||
				current_brighten != dest_brighten ||
				current_darken != dest_darken )
	{
		if ( level_of_detail > 1 && quit_state == FALSE )
		{
			int speed = GetFadeSpeed () ;
			
			if ( current_alpha != dest_alpha )
			{
				KrColorTransform color = anim -> CTransform () ;
				
				if ( current_alpha < dest_alpha ) current_alpha += speed ;
				if ( current_alpha > dest_alpha ) current_alpha -= speed ;
				
				current_alpha = GlClamp ( current_alpha, 0, 255 ) ;
				dest_alpha    = GlClamp ( dest_alpha,    0, 255 ) ;
				
				color.SetAlpha ( current_alpha ) ;
				anim -> SetColor ( color ) ;
			}
			
			if ( current_brighten != dest_brighten )
			{
				KrColorTransform color = anim -> CTransform () ;
				
				if ( current_brighten < dest_brighten ) current_brighten += speed ;
				if ( current_brighten > dest_brighten ) current_brighten -= speed ;
				
				current_brighten = GlClamp ( current_brighten, 0, 255 ) ;
				dest_brighten    = GlClamp ( dest_brighten,    0, 255 ) ;
				
				color.Brighten ( current_brighten ) ;
				anim -> SetColor ( color ) ;
			}
			
			if ( current_darken != dest_darken )
			{
				KrColorTransform color = anim -> CTransform () ;
				
				if ( current_darken < dest_darken ) current_darken += speed ;
				if ( current_darken > dest_darken ) current_darken -= speed ;
				
				current_darken = GlClamp ( current_darken, 0, 255 ) ;
				dest_darken    = GlClamp ( dest_darken,    0, 255 ) ;
				
				color.Darken ( current_darken ) ;
				anim -> SetColor ( color ) ;
			}
		}
		
		if ( quit_state == false )
			if ( anim -> IsVisible () != ( current_alpha > 0 ) )
				anim -> SetVisible ( current_alpha > 0 ) ;
	}
	
	if ( exploding == TRUE && pre_explosion_phase_done == FALSE )
	{
		if ( pre_explosion_phase_pos < 255 )
		{
			unsigned int glow = max ( glow_value, pre_explosion_phase_pos ) ;
			if ( glow > 255 ) glow = 255 ;
			SetTint ( glow, glow, glow ) ;
			pre_explosion_phase_pos += pre_explosion_speed ;
		}
		else
		{
			pre_explosion_phase_done = TRUE ;
			ExplodePhaseTwo () ;
		}
	}
	
	if ( exploding == TRUE && pre_explosion_phase_done == TRUE )
	{
		anim->SetFrame ( anim_cur_frame ) ;
		
		if ( anim_cur_frame >= anim_frames && quit_state == FALSE )
		{
	
			Twinkling ( FALSE ) ;
			anim->SetVisible ( false ) ;
			//if ( rain.RainHasStopped () == TRUE )
			{
				//fprintf ( stdout, "2 - rain has stopped. quitting...\n" ) ;
				quit_state = TRUE ;
			}
			SetAlpha ( 0 ) ;
			exploding = FALSE ;
		}
		
		quit = TRUE ;
		
		anim_cur_frame++ ;
	}
	
	SetRainCreationArea () ;
	rain.Draw () ;
	
	if ( quit_state == TRUE ) return ;
		
	if ( already_drawn == FALSE )
	{
		if ( anim -> IsVisible () != ( current_alpha > 0 ) )
			anim->SetVisible ( current_alpha > 0 ) ;
		already_drawn = TRUE ;
	}
	
	// add some glow effect to our sprite
	if ( glow_counter >= 2 && ( glow_activated == TRUE || glow_value > 0 ) )
	{
		if ( glow_activated == TRUE )
		{
			if ( glow_value <= 0 && glow_direction < 0 ||
					glow_value >= 30 && glow_direction > 0 )
				glow_direction = -glow_direction ;
			
			glow_value += glow_direction * 2 ;
		}
		else
		{
			glow_value -= 2 ;
		}
		
		if ( glow_value > 255 ) glow_value = 255 ;
		if ( glow_value < 0 ) glow_value = 0 ;
		
		Brighten ( glow_value ) ;
		glow_counter = 0 ;
	}
	else
		glow_counter++ ;
	
	bool scaling_changed = FALSE ;
	// if our sprite wants to jump... :)
	if ( jump_height > 0 && jump_pos < jump_height )
	{
		if ( jump_pos < jump_height / 2 )
		{	// increase size
			jump_factor += 2 ;
		}
		else
		{	// decrease size
			jump_factor -= 2 ;
		}
		jump_pos += anim_xy_speed ;
		anim_z_pos = anim_z_speed * anim_z_accel_factor * jump_factor ;
		
		scaling_changed = TRUE ;
		
		//fprintf ( stdout, "jump_pos = %d ; anim_z_pos = %f ; jump_factor = %d\n",
		//			jump_pos, anim_z_pos, jump_factor ) ;
		
		anim_bounce_x_speed = anim_bounce_x_mod = 0 ;
		anim_bounce_y_speed = anim_bounce_y_mod = 0 ;
		SetXAccelFactor ( 1 ) ; SetYAccelFactor ( 1 ) ;
	}
	else
	{
		if ( anim_z_pos != 0 || anim_dest_z_pos != 0 )
		{
			anim_z_pos = 0 ;
			anim_dest_z_pos = 0 ;
			scaling_changed = TRUE ;
		}
	}
	
	int bounce_changed = FALSE ;
	// if our sprite has to bounce
	// FIXME: only tested with MoveUserGemToTop () !!!
	//        so x <> 0 and y > 0 are untested and unsupported :)
	if ( anim_bounce_x_speed != 0 || anim_bounce_y_speed != 0 )
	{
		if ( SlideFinishedY () )
		{
			if ( anim_bounce_y_mod < anim_bounce_y_speed && anim_bounce_y_speed < 0 )
				anim_bounce_y_speed = anim_bounce_y_speed / -2 ;
			
			if ( anim_bounce_y_mod > anim_bounce_y_speed && anim_bounce_y_speed > 0 )
				anim_bounce_y_speed = anim_bounce_y_speed / -2 ;
			
			anim_bounce_y_mod += anim_bounce_y_speed ;
			
			if ( anim_bounce_y_speed == 0 )
				anim_bounce_y_mod = 0 ;
			
			bounce_changed = TRUE ;
		}
		
		if ( bounce_changed && anim_bounce_list -> Count () )
		{
			int smooth_x = 0, smooth_y = 0 ;
			for ( unsigned int i = 0 ; i < anim_bounce_list -> Count () ; i++ )
			{
				Sprite *sprite = *(anim_bounce_list -> ItemPointer ( i )) ;
				if ( sprite -> SlideFinished () )
					sprite -> SetBounce ( anim_bounce_x_speed + smooth_x,
											anim_bounce_y_speed + smooth_y ) ;
				if ( anim_bounce_x_speed + smooth_x < 0 ) smooth_x++ ;
				if ( anim_bounce_x_speed + smooth_x > 0 ) smooth_x-- ;
				if ( anim_bounce_y_speed + smooth_y < 0 ) smooth_y++ ;
				if ( anim_bounce_y_speed + smooth_y > 0 ) smooth_y-- ;
			}
			
			anim_bounce_list -> Clear () ;
		}
	}
	
	// do we have to move?
	if ( anim_x_pos != anim_dest_x_pos || anim_y_pos != anim_dest_y_pos ||
			anim_z_pos != anim_dest_z_pos || scaling_changed || bounce_changed )
	{
		if ( anim_dest_x_pos < anim_x_pos )
		{
			anim_x_pos -= anim_x_speed * anim_x_accel_factor ;
			if ( anim_dest_x_pos > anim_x_pos )
				anim_x_pos += anim_dest_x_pos - anim_x_pos ;
		}
		if ( anim_dest_x_pos > anim_x_pos )
		{
			anim_x_pos += anim_x_speed * anim_x_accel_factor ;
			if ( anim_dest_x_pos < anim_x_pos )
				anim_x_pos -= anim_x_pos - anim_dest_x_pos ;
		}
		if ( anim_dest_y_pos < anim_y_pos )
		{
			anim_y_pos -= anim_y_speed * anim_y_accel_factor ;
			if ( anim_dest_y_pos > anim_y_pos )
				anim_y_pos += anim_dest_y_pos - anim_y_pos ;
		}
		if ( anim_dest_y_pos > anim_y_pos )
		{
			anim_y_pos += anim_y_speed * anim_y_accel_factor ;
			if ( anim_dest_y_pos < anim_y_pos )
				anim_y_pos -= anim_y_pos - anim_dest_y_pos ;
		}
		
		if ( ( anim_x_accel_factor > 1 && anim_dest_x_pos == anim_x_pos ) ||
				( anim_y_accel_factor > 1 && anim_dest_y_pos == anim_y_pos ) )
		{
			accel_sound.StopSound () ;
			PlayBounceSound () ;
			
			if ( anim_dest_x_pos == anim_x_pos )
				anim_x_accel_factor = 1 ;
			if ( anim_dest_y_pos == anim_y_pos )
				anim_y_accel_factor = 1 ;
		}
		
		double dscale_mod = 1 ; //( using_opengl ? 1.00 : 0.50 ) ;

		double dscale = (double) anim_z_pos * 0.01 * dscale_mod ;
		int sx = (int) ( ( (double) ( anim_w / 2 ) ) * dscale ) ;
		int sy = (int) ( ( (double) ( anim_h / 2 ) ) * dscale ) ;
		
		if ( level_of_detail < 3 ) { sx = 0 ; sy = 0 ; }

		SetIPos ( anim_x_pos + anim_x_pos_mod - sx + anim_bounce_x_mod, 
					anim_y_pos + anim_y_pos_mod - sy + anim_bounce_y_mod ) ;

		dscale = (double) 1 + ( (double) anim_z_pos * 0.01 * dscale_mod ) ;
		GlFixed scale = dscale ;
		if ( level_of_detail > 2 )
		{
			if ( use_cached_scaling &&
					!animRes->IsScaleCached ( scale, scale ) )
				animRes->CacheScale ( scale, scale ) ;
			anim->SetScale ( scale, scale ) ;
			//anim->SetZDepth ( anim_zdepth_pos + scale.v ) ;
			anim->SetZDepth ( (int) ( anim_zdepth_pos + ( anim_z_pos * dscale ) ) ) ;
		}
	}
	
	// sound to play when showing for the first time
	if ( creation_sound_enabled == true && creation_sound_started == false )
	{
		creation_sound_started = true ;
		PlayCreationSound () ;
	}
	
	// check if we have to slide...
	if ( anim_x_pos_mod != 0 || anim_y_pos_mod != 0 )
	{
		DoSlide () ;
	} else if ( SlideFinished() == true &&
				move_sound_play == true ) 
	{
		//move_sound.StopSound () ;
		move_sound_play = false ; move_sound_started = false ;
		stop_sound_play = true ; stop_sound_started = false ;
	}
	
	if ( stop_sound_play == true && stop_sound_started == false )
	{
		PlayStopSound () ;
		stop_sound_started = true ;
		stop_sound_play = false ;
	}

	if ( anim_paused ) return ;
	
	if ( anim_pause_interval_counter >= anim_pause_interval )
	{
		if ( anim_delay_counter >= anim_delay )
		{
			anim_delay_counter = 0 ;
			
			GLASSERT ( anim ) ;
			anim->SetFrame ( anim_cur_frame ) ;
			
			// extra sound?
			if ( extra_sound_enabled == true && anim_cur_frame == extra_sound_framenumber )
				PlayExtraSound () ;
			
			if ( anim_bounce == TRUE )
			{
				if ( anim_direction > 0 && anim_cur_frame >= anim_frames - 1 )
				{
					anim_direction = -anim_direction ;
				} else if ( anim_direction < 0 && anim_cur_frame <= 0 )
				{
					anim_direction = -anim_direction ;
					anim_pause_interval_counter = 0 ;
				}
			} else if ( anim_bounce == FALSE && anim_play_once == FALSE ) {
				if ( anim_direction > 0 && anim_cur_frame >= anim_frames - 1 )
				{
					anim_cur_frame = -1 ;
					anim_pause_interval_counter = 0 ;
				} else if ( anim_direction < 0 && anim_cur_frame <= -1 ) {
					anim_cur_frame = anim_frames ;
					anim_pause_interval_counter = 0 ;
				}
			} else if ( anim_bounce == FALSE && anim_play_once == TRUE ) {
				if ( anim_direction > 0 && anim_cur_frame >= anim_frames - 1 )
				{
					anim_cur_frame = anim_frames - 1 ;
					anim_pause_interval_counter = 0 ;
				} else if ( anim_direction < 0 && anim_cur_frame <= -1 ) {
					anim_cur_frame = 1 ;
					anim_pause_interval_counter = 0 ;
				}
			}
			
			anim_cur_frame += anim_direction ;
			
			if ( anim_pause_interval_counter == 0 &&
					anim_pause_interval_randomness == TRUE &&
					anim_play_once == FALSE )
			{
				anim_pause_interval = rand() % anim_max_pause_interval ;
			}
		} else anim_delay_counter++ ;
	} else anim_pause_interval_counter++ ;
}

bool Sprite::GetIsAlreadyDrawn ()
{
	return already_drawn ;
}

bool Sprite::GetIsQuitting ()
{
	return quit ;
}

bool Sprite::GetIsDone ()
{
	return quit_state ;
}

void Sprite::Quit ()
{
	quit = TRUE ;
}

void Sprite::Free ()
{
	if ( already_freed )
	{
		fprintf ( stdout, "sprite already freed!\n" ) ;
		return ;
	}
	KrEngine *engine = video . GetEngine () ;
	
	Twinkling ( false ) ;
	rain.Quit () ;
	anim->SetVisible ( false ) ;
	if ( move_sound_enabled == true ) move_sound.UnloadSound () ;
	if ( stop_sound_enabled == true ) stop_sound.UnloadSound () ;
	if ( extra_sound_enabled == true ) extra_sound.UnloadSound () ;
	if ( creation_sound_enabled == true ) creation_sound.UnloadSound () ;
	if ( accel_sound_enabled == true ) accel_sound.UnloadSound () ;
	if ( bump_sound_enabled == true ) bump_sound.UnloadSound () ;
	explode_sound.UnloadSound () ;
	engine->Tree()->DeleteNode ( anim ) ;
	anim = NULL ;
	anim_bounce_list -> Clear () ;
	delete anim_bounce_list ;
	
	quit_state = TRUE ;
	already_drawn = FALSE ;
	already_freed = TRUE ;
}
