#include "utils.h"

extern char rc_dir[], pax_dir[] ;

#include <stdio.h>
#include <sys/types.h>
#include <errno.h>

GlDynArray < std::string > BreakTextToMaxCols ( std::string text, uint max_cols )
{
	GlDynArray < std::string > textbuffer ;
	textbuffer . Clear () ;
	
	char line [ 256 ] ;
	char pre [ 5 ] ;
	
	sprintf ( line, "%s", (char*) text . c_str () ) ;
	sprintf ( pre, "%s", "" ) ;
	
	if ( strncmp ( line, "| ", 2 ) == 0 )
	{
		sprintf ( pre, "%s", "| " ) ;
		sprintf ( line, "%s", line + 2 ) ;
	}
	else if ( strncmp ( line, "|| ", 3 ) == 0 )
	{
		sprintf ( pre, "%s", "|| " ) ;
		sprintf ( line, "%s", line + 3 ) ;
	}
	
	max_cols -= strlen ( pre ) ;
	
	while ( strlen ( line ) > 0 )
	{
		const int t_line_size = max ( max_cols + strlen ( pre ), 256 ) ;
		char *t_line = new char [ t_line_size ] ;
		
		strncpy ( t_line, line, max_cols ) ;
		t_line [ max_cols ] = '\0' ;
		
		if ( strlen ( line ) > strlen ( t_line ) )
			sprintf ( line, "%s", line + strlen ( t_line ) ) ;
		else
			sprintf ( line, "%s", "" ) ;
		
		{
			char *temp = new char [ t_line_size ] ;
			sprintf ( temp, "%s", t_line ) ;
			sprintf ( t_line, "%s%s", pre, temp ) ;
			delete [] temp ;
		}
		
		std::string t_text = t_line ;
		
		textbuffer . PushBack ( t_text ) ;
		
		delete [] t_line ;
	}
	
	return textbuffer ;
}

std::string CreateKey ( Uint32 seed )
{
	char key [ 256 ] ;
	
	if ( seed != 0 )
		srand ( SDL_GetTicks () + seed ) ;
	
	int key_len = ( rand () % 27 ) + 5 ; // min 5 and max 32 chars
	
	for ( int i = 0 ; i < key_len ; i++ )
	{
		while ( ! ( ( key [ i ] >= '0' && key [ i ] <= '9' ) ||
				( key [ i ] >= 'a' && key [ i ] <= 'z' ) ||
				( key [ i ] >= 'A' && key [ i ] <= 'Z' ) ) )
			key [ i ] = (int) (255.0*rand()/(RAND_MAX+1.0)) ;
	}
	key [ key_len ] = '\0' ;
	
	std::string return_string = key ;
	
	return return_string ;
}
	
char *temp_file ( char *filename )
{
	int size = strlen ( filename ) ;

#ifndef __WIN32__
	mkdir ( CACHE_DIR, S_IRWXU ) ;
#else
	mkdir ( CACHE_DIR ) ;
#endif
	
	char orgname [ size ] ;
	sprintf ( orgname, "%s", filename ) ;
	
	char newname [ size ] ;
	int pos = 0 ;
	for ( ; pos < size ; pos++ )
	{
		if ( orgname [ pos ] == '/' )
			newname [ pos ] = '_' ;
		else
			newname [ pos ] = orgname [ pos ] ;
	}
	
	newname [ pos ] = '\0' ;
	
	char new_filename [ 2048 ] ;
	sprintf ( new_filename, "%s/%s", CACHE_DIR, newname ) ;
	
	//fprintf ( stdout, "tempfilename = '%s'\n", new_filename ) ;
	
	return strdup ( new_filename ) ;
}

void cleanup_temp_dir ( void )
{
	char cmd [ 2048 ] ;
	
#ifndef __WIN32__
	sprintf ( cmd, "rm -rf '%s'", CACHE_DIR ) ;
#else
	sprintf ( cmd, "deltree /Y '%s'", CACHE_DIR ) ;
#endif
	system ( cmd ) ;
}

int min ( int a, int b ) { if ( a < b ) return a ; else return b ; }
int max ( int a, int b ) { if ( a > b ) return a ; else return b ; }

int clamp ( int value, int min, int max )
{
	if ( value < min )
		value = min ;
	if ( value > max )
		value = max ;
	
	return value ;
}

void clamp ( int *value, int min, int max )
{
	*value = clamp ( *value, min, max ) ;
}

bool compare_start_of_string ( char *haystack, char *needle )
{
	if ( haystack == NULL || needle == NULL ) return false ;
	
	int haystack_len = strlen ( haystack ) ;
	int needle_len = strlen ( needle ) ;
	
	if ( needle_len > haystack_len ) return false ;
	
	return strncmp ( haystack, needle, needle_len ) == 0 ? true : false ;
}

int hex_string_to_int ( char *string, int start, int length )
{
	if ( length == -1 )
		length = strlen ( string ) ;

	char data [ strlen ( string ) + 2 ] ; // 2 chars as buffer :)
	sprintf ( data, "%s", string + start ) ;
	data [ length ] = '\0' ;
	
	int value = 0 ;

	// val
	// val * 2 ^ x
	// 2, 4, 8, 16, 32, 64, 128, 256
	// 2^1, 2^2, 2^3, 2^4, 2^5, 2^6, 2^7, 2^8
	
	for ( unsigned int pos = 0 ; pos < strlen ( data ) ; pos++ )
	{
		char c = data [ pos ] ;
		
		int val = c - '0' ;
		if ( c == 'A' || c == 'a' ) val = 10 ;
		if ( c == 'B' || c == 'b' ) val = 11 ;
		if ( c == 'C' || c == 'c' ) val = 12 ;
		if ( c == 'D' || c == 'd' ) val = 13 ;
		if ( c == 'E' || c == 'e' ) val = 14 ;
		if ( c == 'F' || c == 'f' ) val = 15 ;

		int x = strlen ( data ) - 1 - pos ;
		// pos = 0 ; x = 1 ;
		// pos = 1 ; x = 0 ;

		for ( int i = 0 ; i < x ; i++ )
			val = val * 16 ;
		
		value += val ;
	}

	return value ;
}

void create_config_dir ()
{
#ifndef __WIN32__
	struct stat buf ;
	
	char dir [ 256 ] ;
	
	if ( strncmp ( RC_DIR, "~", 1 ) == 0 )
		sprintf ( dir, "%s%s", getenv ( "HOME" ), RC_DIR + 1 ) ;
	else
		sprintf ( dir, "%s", RC_DIR ) ;
	
	sprintf ( rc_dir, "%s", dir ) ;
	
	if ( stat ( rc_dir, &buf ) != 0 )
	{
		fprintf ( stdout, "Creating config directory %s ...\n", rc_dir ) ;
		if ( mkdir ( rc_dir, S_IRWXU ) )
			fprintf ( stdout, "Sorry, couldn't create config dir %s !\n", rc_dir ) ;
	}
	
	sprintf ( pax_dir, "%s/pax", rc_dir ) ;
	
	if ( stat ( pax_dir, &buf ) != 0 )
	{
		fprintf ( stdout, "Creating pax directory %s ...\n", pax_dir ) ;
		if ( mkdir ( pax_dir, S_IRWXU ) )
			fprintf ( stdout, "Sorry, couldn't create pax dir %s !\n", pax_dir ) ;
	}
	
	fprintf ( stdout, "\n" ) ;
#endif
}

#ifdef __WIN32__
char *index ( char *s, int c )
{
	if ( s == NULL ) return NULL ;
	
	for ( int i = 0 ; i < strlen ( s ) ; i++ )
	{
		if ( s [ i ] == c ) return s + i ;
	}
	
	return NULL ;
}

char *rindex ( char *s, int c )
{
	if ( s == NULL ) return NULL ;
	
	for ( int i = strlen ( s ) - 1 ; i > 0 ; i-- )
	{
		if ( s [ i ] == c ) return s + i ;
	}
	
	return NULL ;
}
#endif

// myscandir - return number of items found in dir
int myscandir ( const char *dirname, GlDynArray < dirent > *namelist )
{
        DIR             *dip;
        struct dirent   *dit;

        int             i = 0;
		
		namelist -> Clear () ;

        /* DIR *opendir(const char *name);
         *
         * Open a directory stream to dirname and make sure
         * it's a readable and valid (directory) */
        if ((dip = opendir(dirname)) == NULL)
        {
                //perror("opendir");
                return 0;
        }

        //printf("Directory stream is now open\n");
		
		
        /*  struct dirent *readdir(DIR *dir);
         *
         * Read in the files from dirname and print */
        while ((dit = readdir(dip)) != NULL)
        {
                i++;
                //printf("\n%s", dit->d_name);
				
				namelist -> PushBack ( *dit ) ;
        }

        //printf("\n\nreaddir() found a total of %i files\n", i);

        /* int closedir(DIR *dir);
         *
         * Close the stream to dirname. And check for errors. */
        if (closedir(dip) == -1)
        {
                //perror("closedir");
                return 0;
        }

        //printf("\nDirectory stream is now closed\n");
        return i;
}

// returns whether if it's a directory or not
bool isdir ( char *filename )
{
        DIR *dir ;
        struct stat buf ;
		char *last_element = rindex ( filename, '/' ) ;

		if ( last_element == NULL )
			last_element = filename ;
		else
			last_element++ ;
		
        //fprintf ( stdout, "stat\n" ) ;
        stat ( filename, &buf ) ;

        if ( S_ISDIR ( buf.st_mode ) &&
				( dir = opendir ( filename ) ) &&
				strncmp ( last_element, ".", 1 ) != 0 )
        {
                closedir ( dir ) ;
                return TRUE ;
        }
        
        return FALSE ;
}

char *dirname ( char *filepath )
{
	char *copy = strdup ( filepath ) ;
	
	if ( rindex ( copy, '/' ) )
		*rindex ( copy, '/' ) = '\0' ;
	else
		sprintf ( copy, "%s", "." ) ;
	
	return copy ;
}

char *basename ( char *filepath )
{
	char *copy = strdup ( filepath ) ;
	
	if ( copy [ strlen ( copy ) - 1 ] == '/' )
		copy [ strlen ( copy ) - 1 ] = '\0' ;

	if ( rindex ( copy, '/' ) )
	{
		char *result = strdup ( rindex ( copy, '/' ) + 1 ) ;
		free ( copy ) ;
		return result ;
	}
	else
		return copy ;
}
