#include "video.h"

void Video::Init ()
{
	display_framerate = FALSE ;
	react_on_focus_lost = FALSE ;

	logic_frames = 0 ;
	rendered_frames = 0 ;
	
	// we prefer smooth rendering over speed
	if ( using_opengl )
		render_speed = RENDER_SPEED_NORMAL ;
	else
		render_speed = RENDER_SPEED_NORMAL ;
	
	fill_background = TRUE ;
	
	current_time = 0 ;
	
	c1_mode = FALSE ; // let's think that this is a normal machine at start...
	
	// RESIZE_MODE_NONE is the most direct rendering possible ( fastest )
	window_resize_mode = RESIZE_MODE_NONE ;
	
	// which color-depth do we prefer? kyra prefers 32bit !
	default_depth = 32 ;
	
	// this caches scaled sprites ( needs much memory!! )
	use_cached_scaling = FALSE ;
	
	// this will be discovered later.... shouldn't be changed!
	using_opengl = FALSE ;
	
	// this too...
	fullscreen_activated = FALSE ;
	
	// whether one can pause the game or not
	pause_possible = TRUE ;
	
	// should our mouse-cursor be visible?
	mouse_visible = false ;
	
	flags = 0 ;
	gl_flags = 0 ;
	force_opengl = FALSE ;
	force_software = ( window_resize_mode == RESIZE_MODE_OVERLAY ? TRUE : FALSE ) ;
	force_depth = FALSE ;
	
	// if wanted, these represent our zoom factors for the main window
	zoom_x = 1 ;
	zoom_y = 1 ;
	
	// when window is resizable we won't draw every frame to the screen...
	frame_skipper_max = 1 ;
	frame_skipper = 0 ;

	/// choose default overlay_format from these:
	//
	//    SDL_YV12_OVERLAY  0x32315659    Planar mode: Y + V + U  (3 planes)
	//    SDL_IYUV_OVERLAY  0x56555949    Planar mode: Y + U + V  (3 planes)
	//    SDL_YUY2_OVERLAY  0x32595559    Packed mode: Y0+U0+Y1+V0 (1 plane)
	//    SDL_UYVY_OVERLAY  0x59565955    Packed mode: U0+Y0+V0+Y1 (1 plane)
	//    SDL_YVYU_OVERLAY  0x55595659    Packed mode: Y0+V0+Y1+U0 (1 plane)
	//
	//SetOverlayMode ( SDL_YV12_OVERLAY ) ;
	SetOverlayMode ( SDL_YUY2_OVERLAY ) ;
	//
	///
	
	done = false ;
}

void Video::Setup ()
{
	SDL_InitSubSystem ( SDL_INIT_VIDEO ) ;
	
	const SDL_VideoInfo *info = SDL_GetVideoInfo () ;
	if ( !info )
	{
		fprintf ( stderr, "Video query failed: %s\n", SDL_GetError () ) ;
		exit ( -1 ) ;
	}
	
	char *videodriver_name = (char*) malloc ( 1024 * sizeof(char) ) ;
	SDL_VideoDriverName ( videodriver_name, 1024 ) ;
	if ( strcmp ( videodriver_name, "fbcon" ) == 0 )
	{
		fprintf ( stdout, "Detected framebuffer console.\n" ) ;
		fullscreen_activated = TRUE ;
	}

	{	// check for funny videomodes like 1024x480 ( vaio c1 )
		bool allow_override = FALSE ; // allow other modes than 640x480 and 1024x480
		int override_w = 0, override_h = 0 ;
		bool normal_mode_available = FALSE ;
		bool c1_mode_available = FALSE ;
		
		SDL_Rect **modes ;
		
		// Get available fullscreen/hardware modes
		modes = SDL_ListModes ( NULL, SDL_FULLSCREEN | SDL_HWSURFACE ) ;
		
		// Check is there are any modes available
		if ( modes == (SDL_Rect **)0 )
		{
			fprintf ( stdout, "No video-modes available!\n" ) ;
			exit ( -1 ) ;
		}
		
		// Check if our resolution is restricted
		if ( modes == (SDL_Rect **)-1 )
		{
			//fprintf ( stdout, "All resolutions available.\n" ) ;
			normal_mode_available = TRUE ;
		}
		else
		{
			int biggest_w = 640, biggest_h = 480 ;
			// Print valid modes
			//fprintf ( stdout, "Available Modes\n" ) ;
			for ( int i = 0 ; modes [ i ] ; ++i )
			{
				//fprintf ( stdout, "  %d x %d\n",
				//			modes [ i ] -> w, modes [ i ] -> h ) ;
				if ( modes [ i ] -> w == 640 && modes [ i ] -> h == 480 )
					normal_mode_available = TRUE ;
				if ( modes [ i ] -> w == 1024 && modes [ i ] -> h == 480 )
					c1_mode_available = TRUE ;
				if ( modes [ i ] -> w >= biggest_w &&
					modes [ i ] -> h >= biggest_h )
				{
					biggest_w = modes [ i ] -> w ;
					biggest_h = modes [ i ] -> h ;
				}
			}
			if ( ! normal_mode_available && ! c1_mode_available &&
				( biggest_w > 640 || biggest_h > 480 ) &&
				allow_override )
			{
				override_w = biggest_w ;
				override_h = biggest_h ;
			}
		}
		// now switch to c1-mode :) ( vaio c1 with 1024x480 pixels )
		if ( c1_mode_available )
		{
			c1_mode = TRUE ;
			fprintf ( stdout, "VAIO C1 display or similar detected!\n" ) ;
		}
		if ( override_w || override_h )
			fprintf ( stdout, "Using unusual mode %dx%d\n",
					override_w, override_h ) ;

		// general size of screen - for now these are FIXED vars!!!
		screenx = override_w ? override_w : ( c1_mode ? 1024 : 640 ) ;
		screeny = override_h ? override_h : 480 ;
	}
	
	/* This checks to see if surfaces can be stored in memory */
	//if ( info->hw_available ) flags |= SDL_HWSURFACE ;
	//else flags |= SDL_SWSURFACE ;

	/* This checks if hardware blits can be done */
	if ( info->blit_hw ) 
		flags |= SDL_HWACCEL ;

	flags |= SDL_RLEACCEL ; // try using colorkey acceleration
	
	gl_flags |= SDL_OPENGL | SDL_GL_DOUBLEBUFFER | SDL_HWPALETTE | SDL_SRCALPHA ;
	if ( window_resize_mode == RESIZE_MODE_AUTO )
		gl_flags |= SDL_OPENGLBLIT ;

	{	// is the abilitiy to load a rgba png sdl-version dependend?!
		//SDL_Surface *icon = IMG_Load ( "gfx/wmicon.bmp" ) ;
		SDL_Surface *icon = IMG_Load ( "gfx/icon.png" ) ;
		SDL_WM_SetIcon ( icon, NULL ) ;
		SDL_FreeSurface ( icon ) ;
	}
	
	SDL_WM_SetCaption ( "GemHunters - testing video...", 0 ) ;

	/* Create a display for testing */
	screen = SDL_SetVideoMode ( 1, 1, 0, flags | gl_flags | SDL_NOFRAME ) ;
	if ( screen == NULL ) // hmm... try the real thing...
		screen = SDL_SetVideoMode ( 1, 1, 0, flags | gl_flags ) ;
	if ( screen == NULL ) // and now without opengl?!...
		screen = SDL_SetVideoMode ( 1, 1, 0, flags ) ;
	if ( screen == NULL )
	{
		fprintf ( stdout, "couldn't create display for testing at line %d in %s\n",
				__LINE__ -4, __FILE__ ) ;
		exit ( 3 ) ;
	}

	engine = new KrEngine ( screen ) ;

	{
		char buf [ 512 ] ;
		std::string s ;
		engine->QueryRenderDesc ( screen, &s ) ;
		strcpy ( buf, s.c_str() ) ;
		fprintf ( stdout, "Available rendering device:\n%s\n", buf ) ;
		if ( ( strstr ( buf, "Software render" ) || force_software == TRUE ||
				( force_opengl == FALSE && strstr ( buf, "Mesa" ) ) ) &&
				force_opengl == FALSE )
		{
			if ( force_software == TRUE )
				fprintf ( stdout, "Software rendering forced!\n" ) ;
			else
			{
				fprintf ( stdout, "Using software renderer.\n" ) ;
				if ( strstr ( buf, "Mesa" ) && force_opengl == FALSE )
				{
					fprintf ( stdout, "If you're sure that OpenGL works for you\n" ) ;
					fprintf ( stdout, "try forcing it using -o as an argument.\n" ) ;
				}
			}
		}
		else
		{
			SDL_FreeSurface ( screen ) ;
			screen = SDL_SetVideoMode ( 1, 1, 0, flags | SDL_NOFRAME ) ;
			if ( screen == NULL )
			{
		                fprintf ( stdout, "couldn't create opengl display at line %d in %s\n",
                                __LINE__ -4, __FILE__ ) ;
				exit ( 3 ) ;
			}
			if ( force_opengl == TRUE )
				fprintf ( stdout, "OpenGL rendering forced!\n" ) ;
			else
				fprintf ( stdout, "Using OpenGL renderer.\n" ) ;
			flags |= gl_flags ;
			using_opengl = TRUE ;
		}
	}

	// now create our screen and initialize our engine
	delete engine ;
	SDL_FreeSurface ( screen ) ;
	flags |= ( fullscreen_activated ? SDL_FULLSCREEN : 0 ) ;
	flags |= ( window_resize_mode != RESIZE_MODE_NONE ? SDL_RESIZABLE : 0 ) ;
	depth = default_depth ;
	if ( force_depth == FALSE && window_resize_mode != RESIZE_MODE_NONE )
		depth = SDL_VideoModeOK ( screenx, screeny, default_depth, flags ) ;
	screen = SDL_SetVideoMode ( screenx, screeny, depth, flags ) ;
	if ( screen == NULL )
	{
                fprintf ( stdout, "couldn't create game display at line %d in %s\n",
                                __LINE__ -4, __FILE__ ) ;
		exit ( 3 ) ;
	}

	screen_width = screenx ;
	screen_height = screeny ;
	virtual_display = NULL ;
	if ( window_resize_mode == RESIZE_MODE_OVERLAY )
	{
		Uint32 rmask, gmask, bmask, amask;
		//fprintf ( stdout, "create display...\n" ) ;

		// SDL interprets each pixel as a 32-bit number, so our masks must depend
		// on the endianness (byte order) of the machine
		#if SDL_BYTEORDER == SDL_BIG_ENDIAN
			rmask = 0xff000000;
			gmask = 0x00ff0000;
			bmask = 0x0000ff00;
			amask = 0x000000ff;
		#else
			rmask = 0x000000ff;
			gmask = 0x0000ff00;
			bmask = 0x00ff0000;
			amask = 0xff000000;
		#endif
		
		virtual_display = SDL_CreateRGBSurface ( flags, screenx, screeny, 32, rmask, gmask, bmask, amask ) ;
		//fprintf ( stdout, "now generate engine...\n" ) ;
		KrRect dest_rect ;
		dest_rect.Set ( 0, 0, virtual_display->w, virtual_display->h ) ;
		engine = new KrEngine ( virtual_display, dest_rect, (KrRGBA*) NULL ) ;
		//fprintf ( stdout, "done.\n" ) ;
		
		video_overlay . Init () ;
	}
	else
		engine = new KrEngine ( screen ) ;

#ifndef __WIN32__
	SDL_WM_SetCaption ( "GemHunters - F1 toggles fullscreen - P pauses game - ESC quits", 0 ) ;
#else
	SDL_WM_SetCaption ( "GemHunters - P pauses game - ESC quits", 0 ) ;
#endif
	if ( fullscreen_activated && ! mouse_visible ) HideMouse () ;
	
	fprintf ( stdout, "Initialized video ( depth = %dbit ) and activated the following extensions:\n", depth ) ;
	if ( flags & SDL_HWSURFACE )
		fprintf ( stdout, "- SDL_HWSURFACE\n" ) ;
	else
		fprintf ( stdout, "- SDL_SWSURFACE\n" ) ;
	if ( flags & SDL_HWACCEL ) fprintf ( stdout, "- SDL_HWACCEL\n" ) ;
	if ( flags & SDL_OPENGL ) fprintf ( stdout, "- SDL_OPENGL\n" ) ;
	if ( flags & SDL_GL_DOUBLEBUFFER ) fprintf ( stdout, "- SDL_GL_DOUBLEBUFFER\n" ) ;
	if ( flags & SDL_HWPALETTE ) fprintf ( stdout, "- SDL_HWPALETTE\n" ) ;
	if ( virtual_display != NULL ) fprintf ( stdout, "- Internal video buffer ( depth = %dbit )\n", virtual_display->format->BitsPerPixel ) ;
	fprintf ( stdout, "\n" ) ;

	SetClearBackground ( fill_background ) ;
	
	c1_layer . Init () ;

	// FPS
	fps . Init () ;
	
	// OSD
	osd . Init () ;
	osd . DisplayMessage ( "loading..." ) ;

	// now display kyra splash-screen
	//if ( window_resize_mode == RESIZE_MODE_OVERLAY )
	//	ResizeVideo ( 1024, 768 ) ;

	engine->StartSplash ( 2 ) ;
	UpdateDisplayOnly () ;
	engine->UpdateSplash ( 200000 ) ;
	UpdateDisplayOnly () ;
	sleep ( 1 ) ;
	
	gamescreen . Init () ;
	
	engine->EndSplash () ;
}

KrEngine *Video::GetEngine () { return engine ; }

int Video::GetDefaultDepth () { return default_depth ; }
bool Video::GetForceSoftware () { return force_software ; }
bool Video::GetUsingOpenGL () { return using_opengl ; }
bool Video::GetUseCachedScaling () { return use_cached_scaling ; }
bool Video::GetDisplayFramerate () { return display_framerate ; }
bool Video::GetFramerateReached () { return fps . FpsReached () ; }
int  Video::GetFramesLeft () { return frames_left ; }
int  Video::GetLogicFrames () { return logic_frames ; }
int  Video::GetRenderedFrames () { return rendered_frames ; }

int Video::GetScreenX () { return c1_mode ? 640 : screenx ; }
int Video::GetScreenY () { return c1_mode ? 480 : screeny ; }
int Video::GetScreenWidth () { return screen_width ; }
int Video::GetScreenHeight () { return screen_height ; }
double Video::GetZoomX () { return zoom_x ; }
double Video::GetZoomY () { return zoom_y ; }
SDL_Surface *Video::GetScreenSurface () { return screen ; }
Uint32 Video::GetOverlayMode () { return video_overlay . GetOverlayMode () ; }
bool Video::GetReactOnFocusLost () { return react_on_focus_lost ; }
bool Video::GetC1ModeActive () { return c1_mode ; }


void Video::SetFullscreen ( bool value ) { fullscreen_activated = value ; }
void Video::SetForceOpenGL ( bool value ) { force_opengl = value ; }
void Video::SetForceSoftware ( bool value ) { force_software = value ; }
void Video::SetDefaultDepth ( int value ) { default_depth = value ; }
void Video::SetForceDepth ( bool value ) { force_depth = value ; }
void Video::SetWindowResizeMode ( int value ) { window_resize_mode = value ; }
void Video::SetUseCachedScaling ( int value ) { use_cached_scaling = value ; }
void Video::SetOverlayMode ( Uint32 value ) { video_overlay . SetOverlayMode ( value ) ; }
void Video::SetReactOnFocusLost ( bool value ) { react_on_focus_lost = value ; }
void Video::SetDisplayFramerate ( bool value ) { display_framerate = value ; }
void Video::SetPausePossible ( bool value ) { pause_possible = value ; }

void Video::SetClearBackground ( bool choice )
{
	if ( choice == fill_background ) return ;
	
	if ( choice == true )
	{
		KrRGBA color ;
		color . Set ( 0, 0, 0 ) ;
		engine -> FillBackground ( &color ) ;
		fill_background = true ;
	}
	else
	{
		engine -> FillBackground ( 0 ) ;
		fill_background = false ;
	}
}

void Video::SetDisplayC1Layer ( bool value )
{
	if ( value )
		c1_layer.Show () ;
	else
		c1_layer.Hide () ;
}

void Video::SetTime ( float time )
{
	current_time = time ;
}

float Video::GetTime ()
{
	return current_time ;
}

void Video::SetMouseVisible ( bool value )
{
	if ( fullscreen_activated && ! mouse_visible && value )
		ShowMouse () ;
	if ( fullscreen_activated && mouse_visible && ! value )
	{
		mouse_visible = value ;
		HideMouse () ;
	}
	mouse_visible = value ;
}

void Video::ShowMouse ()
{
	SDL_ShowCursor ( SDL_ENABLE ) ;
}

void Video::HideMouse ()
{
	if ( mouse_visible == false )
		SDL_ShowCursor ( SDL_DISABLE ) ;
}

void Video::Everytime ( SDL_Event event )
{
	fps.Update () ;
	
	gamescreen.ProcessEvents ( event ) ;
}

void Video::ResizeVideo ( int w, int h )
{
	bool update_message = false ;
	
	fprintf ( stdout, "resized main window to %dx%d - ", w, h ) ;
	
	if ( window_resize_mode == RESIZE_MODE_NONE )
	{
		fprintf ( stdout, "support for window resize is disabled.\n" ) ;
		fprintf ( stdout, "this means that your window manager is somehow broken\n" ) ;
		fprintf ( stdout, "and does not support the window manager API correctly.\n" ) ;
		return ;
	}
	
	zoom_x = (double) w / (double) screenx ;
	zoom_y = (double) h / (double) screeny ;
	fprintf ( stdout, "zoom_x = %f ; zoom_y = %f\n",
				zoom_x, zoom_y ) ;
	int new_screen_width = (int) ( (double) screenx * zoom_x ) ;
	int new_screen_height = (int) ( (double) screeny * zoom_y ) ;
	
	if ( new_screen_width == screen_width && new_screen_height == screen_height )
		return ;
	
	if ( new_screen_width < screenx )
	{
		new_screen_width = screenx ;
		zoom_x = 1 ;
		update_message = true ;
	}
	if ( new_screen_height < screeny )
	{
		new_screen_height = screeny ;
		zoom_y = 1 ;
		update_message = true ;
	}
	
	screen_width = new_screen_width ;
	screen_height = new_screen_height ;
	
	SDL_FreeSurface ( screen ) ;
	screen = SDL_SetVideoMode ( screen_width, screen_height, depth, flags ) ;

	if ( update_message )
		fprintf ( stdout, "resized main window to %dx%d - zoom_x = %f ; zoom_y = %f\n",
					screen_width, screen_height, zoom_x, zoom_y ) ;
}

void Video::ToggleFullscreen ()
{
	SDL_WM_ToggleFullScreen ( screen ) ;
	fullscreen_activated = ( fullscreen_activated ? FALSE : TRUE ) ;
	if ( fullscreen_activated )
		HideMouse () ;
	else
		ShowMouse () ;
}

void Video::DumpFPS ()
{
	fps . DumpStatus () ;
}

void Video::PrintFPSStatistics ()
{
	fps . PrintStatistics () ;
}

void Video::DisplayOSDMessage ( char *message )
{
	osd . DisplayMessage ( message ) ;
	
	bool dummy1, dummy2, dummy3 ;
	dummy1 = dummy2 = dummy3 = false ;
	Update ( dummy1, dummy2, dummy3, true ) ;
}

void Video::HoldOSDMessage ()
{
	osd . HoldMessage () ;
}

void Video::UnholdOSDMessage ()
{
	osd . UnholdMessage () ;
}

bool Video::OSDMessageIsHold ()
{
	return osd . MessageIsHold () ;
}

void Video::Update ( bool &pause, bool &toggle_pause_by_user,
						bool &focus_lost, int do_update,
						int frames_until_update )
{
	frames_left = frames_until_update ;
	
	if ( pause_possible == false && toggle_pause_by_user == true )
		toggle_pause_by_user = false ;
	
	//if ( fps . FpsReached () || do_update != false )
	{
		if ( do_update == AUTO ) // if we don't _have_to_ update our display...
		{
			// pause thingies...
			bool app_not_focused = 
					( ! ( SDL_GetAppState () & SDL_APPACTIVE ) ||
					! ( SDL_GetAppState () & SDL_APPINPUTFOCUS ) ) ?
					true : false ;
			if ( gamescreen . IsQuitting () ) app_not_focused = FALSE ;
			if ( pause == false && ( react_on_focus_lost == true && 
				  ( focus_lost == false && app_not_focused == true ) ||
				  toggle_pause_by_user == true ) )
			{	// now the user went away... :(
				if ( focus_lost == false && pause == false )
				{
					// pause game!
					gamescreen . Pause () ;
					pause = true ;
					if ( toggle_pause_by_user == false )
						focus_lost = true ;
					// if fullscreen was activated free / show the mouse-cursor
					if ( fullscreen_activated == true )
						ShowMouse () ;
					
					// throttle the framrate down a bit
					fps . SetSpeedToSlow () ;
					
					// display some osd-message
					osd . DisplayMessage ( "paused" ) ;
					osd . HoldMessage () ;
					
					// set our window title to represent our current state
					if ( focus_lost == true && toggle_pause_by_user == false )
						SDL_WM_SetCaption ( "GemHunters - paused - click me to continue playing!", 0 ) ;
					else
						SDL_WM_SetCaption ( "GemHunters - paused - press p to continue playing!", 0 ) ;
				}
			}
			else if ( pause == true && app_not_focused == false &&
					( focus_lost == true || toggle_pause_by_user == true ) )
			{	// now the user looks at us :)
				// unpause game!
				gamescreen . UnPause () ;
				pause = false ;
				focus_lost = false ;
				// if fullscreen was activated grab / hide the mouse-cursor again
				if ( fullscreen_activated == true )
					HideMouse () ;
				// set our window title to represent our current state
#ifndef __WIN32__
				SDL_WM_SetCaption ( "GemHunters - F1 toggles fullscreen - P pauses game - ESC quits", 0 ) ;
#else
				SDL_WM_SetCaption ( "GemHunters - P pauses game - ESC quits", 0 ) ;
#endif
	
				// 
				fps . SetSpeedToNormal () ;
				//
				osd . UnholdMessage () ;
				osd . DisplayMessage ( "unpaused" ) ;
			}
			toggle_pause_by_user = false ;
			//
		}
		
		// NEVER try to do this when do_update is TRUE !!!
		// otherwise you could try to render NULL pointers.....
		// !! do_update is also used when gamescreen is uninitialized.
		int limit = 1 ; // defaults to smooth rendering
		float framerate = fps . GetFramerate () ;
		if ( render_speed == RENDER_SPEED_FAST )
			limit = (int) ( framerate / 10.00 ) ;
		if ( render_speed == RENDER_SPEED_FASTEST )
			limit = (int) ( framerate / 5.00 ) ;
		limit = clamp ( limit, 1, min ( (int) framerate, FRAMERATE ) ) ;
		if ( do_update != true && frames_left < limit )
		{
			//if ( frames_left < 0 || frames_left > FRAMERATE )
			//	printf ( "frames_left = %d\n",
			//			frames_left ) ;
			//printf ( "limit = %5d\r", limit ) ;
			
			fps.Draw () ;
			osd.Draw () ;
			
			gamescreen.Draw () ;
			
			/*static int max_fl = 0 ;
			if ( frames_left > max_fl )
				max_fl = frames_left ;
			static int min_fl = 1000 ;
			if ( frames_left < min_fl )
				min_fl = frames_left ;
			printf ( "frames_left: min = %d ; max = %d ; cur = %d\r",
					min_fl, max_fl, frames_left ) ;*/
		}
		
		bool fps_reached = true ;
		if ( render_speed > RENDER_SPEED_SMOOTH )
			fps_reached = fps.FpsReached () ;
		
		// count logically rendered frames
		logic_frames++ ;
		
		if ( ( do_update == AUTO && fps_reached &&
			frames_left == 0 ) || do_update == true )
		{
			if ( virtual_display == NULL )
			{	// there's no virtual screen so render directly
				engine->Draw () ;
			}
			else
			{	// now that we have a virtual screen we have to render indirectly
				// this is YUV overlay resizing
				video_overlay . Draw ( virtual_display ) ;
			}
			// count visual updates
			rendered_frames++ ;
		}
		
		if ( do_update != true )
		{
			gamescreen . AfterDraw () ;
	
			// can we quit?
			if ( gamescreen . IsDone () ) 
				done = true ;
		}
	}
}

void Video::UpdateDisplayOnly ()
{
	bool dummy1, dummy2, dummy3 ;
	dummy1 = dummy2 = dummy3 = false ;

	Update ( dummy1, dummy2, dummy3, true ) ;
}

void Video::QuitGameScreen ()
{
	gamescreen . Quit () ;
}

bool Video::GetIsQuitting ()
{
	return gamescreen . IsQuitting () ;
}

bool Video::GetIsDone ()
{
	return done ;
}

void Video::Quit ()
{
	fps . Quit () ;
	osd . Quit () ;
	c1_layer . Quit () ;
	
	delete engine ;

	if ( virtual_display != NULL ) SDL_FreeSurface ( virtual_display ) ;
	video_overlay . Quit () ;

	SDL_QuitSubSystem ( SDL_INIT_VIDEO ) ;
}
