#include "video_overlay.h"

#include "video.h"
extern Video video ;

void Video_Overlay::Init ()
{
	overlay = NULL ;

	SDL_Surface *screen = video . GetScreenSurface () ;
	
	overlay = SDL_CreateYUVOverlay ( screen->w, screen->h,
						overlay_format,
						screen ) ;
	if ( overlay == NULL )
	{
		fprintf ( stdout, "couldn't create overlay: %s\n", SDL_GetError() ) ;
		exit ( 1 ) ;
	}
	printf ( "Created %dx%dx%d %s %s overlay\n", overlay->w, overlay->h,
				overlay->planes,
				overlay->hw_overlay ? "hardware" : "software",
				overlay_format == SDL_YV12_OVERLAY ? "YV12" :
				overlay_format == SDL_IYUV_OVERLAY ? "IYUV" :
				overlay_format == SDL_YUY2_OVERLAY ? "YUY2" :
				overlay_format == SDL_UYVY_OVERLAY ? "UYVY" :
				overlay_format == SDL_YVYU_OVERLAY ? "YVYU" :
				"Unknown" ) ;
	for ( int i = 0 ; i < overlay->planes ; i++ )
	{
		printf ( "  plane %d: pitch = %d\n", i, overlay->pitches [ i ] ) ;
	}
	
	if ( overlay_format == SDL_YV12_OVERLAY )
	{
		// Black initialization /
		memset ( overlay->pixels [ 0 ], 0, overlay->pitches [ 0 ] *overlay->h ) ;
		memset ( overlay->pixels [ 1 ], 128, overlay->pitches [ 1 ] * ( ( overlay->h + 1 ) / 2 ) ) ;
		memset ( overlay->pixels [ 2 ], 128, overlay->pitches [ 2 ] * ( ( overlay->h + 1 ) / 2 ) ) ;
	}
}

void Video_Overlay::Draw ( SDL_Surface *virtual_display )
{
	int monochrome = 0 ;
	int luminance = 100 ;
	SDL_Rect srect ;

	GlDynArray< KrRect >* rectangles = new GlDynArray< KrRect > ;
	video . GetEngine () -> Draw ( FALSE, rectangles ) ;
	//fprintf ( stdout, "drawing %d rectangles\n", rectangles -> Count () ) ;
	
	SDL_LockSurface ( virtual_display ) ;
	SDL_LockYUVOverlay ( overlay ) ;
	for ( unsigned int i = 0 ; i < rectangles -> Count () ; i++ )
	{
		KrRect *rectangle = rectangles -> ItemPointer ( i ) ;
		srect . x = rectangle -> xmin ;
		srect . y = rectangle -> ymin ;
		srect . w = rectangle -> Width () ;
		srect . h = rectangle -> Height () ;
		//fprintf ( stdout, "drawing %d,%d x %d,%d\n",
		//			srect.x, srect.y, srect.w, srect.h ) ;
		switch ( overlay_format )
		{
			case SDL_YV12_OVERLAY:
					ConvertRGBtoYV12 ( virtual_display, overlay->pixels, overlay->pitches, srect, monochrome, luminance ) ;
			break ;
			case SDL_YUY2_OVERLAY:
					ConvertRGBtoYUY2 ( virtual_display, overlay->pixels, overlay->pitches, srect, monochrome, luminance ) ;
			break ;
		}
	}
	SDL_UnlockYUVOverlay ( overlay ) ;
	SDL_UnlockSurface ( virtual_display ) ;
	rectangles -> Clear () ;
	delete rectangles ;
	//fprintf ( stdout, "draw...\n" ) ;
	SDL_Rect drect ;
	drect.x = 0 ; drect.y = 0 ;
	drect.w = video . GetScreenWidth () ; drect.h = video . GetScreenHeight () ;
	//fprintf ( stdout, "blitting overlay to %dx%d+%dx%d\n",
	//			drect.x, drect.y, drect.w, drect.h ) ;
	SDL_DisplayYUVOverlay ( overlay , &drect ) ;
}

void Video_Overlay::Quit ()
{
	if ( overlay != NULL ) SDL_FreeYUVOverlay ( overlay ) ;
}

void Video_Overlay::RGBtoYUV ( Uint8 *rgb, int *yuv, int monochrome, int luminance )
{
	int algo = 0 ; // algorhythm to use

	if ( monochrome )
	{	// these are the two formulas that I found on the FourCC site...
		if ( algo == 0 )
		{
			yuv [ 0 ] = (int) ( 0.299 * rgb [ 0 ] + 0.587 * rgb [ 1 ] + 0.114 * rgb [ 2 ] ) ;
			yuv [ 1 ] = 128 ;
			yuv [ 2 ] = 128 ;
		}
		else
		{
			yuv [ 0 ] = (int) ( ( 0.257 * rgb [ 0 ] ) + ( 0.504 * rgb [ 1 ] ) + ( 0.098 * rgb [ 2 ] ) + 16 ) ;
			yuv [ 1 ] = 128;
			yuv [ 2 ] = 128;
		}
	}
	else
	{	// these are the two formulas that I found on the FourCC site...
		if ( algo == 0 )
		{
			yuv [ 0 ] = (int) ( 0.299 * rgb [ 0 ] + 0.587 * rgb [ 1 ] + 0.114 * rgb [ 2 ] ) ;
			yuv [ 1 ] = (int) ( ( rgb [ 2 ] - yuv [ 0 ] ) * 0.565 + 128 ) ;
			yuv [ 2 ] = (int) ( ( rgb [ 0 ] - yuv [ 0 ] ) * 0.713 + 128 ) ;
		}
		else
		{
			yuv [ 0 ] = (int) ( ( 0.257 * rgb [ 0 ] ) + ( 0.504 * rgb [ 1 ] ) + ( 0.098 * rgb [ 2 ] ) + 16 ) ;
			yuv [ 1 ] = (int) ( 128 - ( 0.148 * rgb [ 0 ] ) - ( 0.291 * rgb [ 1 ] ) + ( 0.439 * rgb [ 2 ] ) ) ;
			yuv [ 2 ] = (int) ( 128 + ( 0.439 * rgb [ 0 ] ) - ( 0.368 * rgb [ 1 ] ) - ( 0.071 * rgb [ 2 ] ) ) ;
		}
	}

	if ( luminance != 100 )
	{
		yuv [ 0 ] = yuv [ 0 ] * luminance / 100 ;
		if ( yuv [ 0 ] > 255 )
			yuv [ 0 ] = 255 ;
	}

	/* clamp values...if you need to, we don't seem to have a need */
	/*
	for ( int i = 0 ; i < 3 ; i++ )
	{
		if ( yuv [ i ] < 0 )
			yuv [ i ] = 0 ;
		if ( yuv [ i ] > 255 )
			yuv [ i ] = 255 ;
	}
	*/
}

void __inline__ Video_Overlay::ConvertRGBtoYV12 ( SDL_Surface *s, Uint8 **o,
											Uint16 *pitches, SDL_Rect rect,
											int monochrome, int luminance )
{
	int x , y ;
	int yuv [ 3 ] ;
	Uint8 *p, *op [ 3 ] ;

	//fprintf ( stdout, "ConvertRGBtoYV12 %d,%d x %d,%d\n",
	//			rect.x, rect.y, rect.w, rect.h ) ;
	
	// Convert
	for ( y = rect.y ; y < s->h && y < rect.y + rect.h ; y++ )
	{
		int correction = 0 ;
		for ( x = 0 ; x < rect.x ; x++ )
			if ( x % 2 == 0 && y % 2 == 0 )
				correction ++ ;
		
		p = ( (Uint8 *) s->pixels ) + s->pitch * y + ( s->format->BytesPerPixel * rect.x ) ;
		op [ 0 ] = o [ 0 ] + pitches [ 0 ] * y + rect.x ;
		op [ 1 ] = o [ 1 ] + pitches [ 1 ] * ( y / 2 ) + correction ;
		op [ 2 ] = o [ 2 ] + pitches [ 2 ] * ( y / 2 ) + correction ;
		
		for ( x = rect.x ; x < s->w && x < rect.x + rect.w ; x++ )
		{
			//RGBtoYUV ( p, yuv, monochrome, luminance ) ;
			
			yuv [ 0 ] = ( 299 * p [ 0 ] + 587 * p [ 1 ] + 114 * p [ 2 ] ) / 1000 ;
			yuv [ 1 ] = ( ( p [ 2 ] - yuv [ 0 ] ) * 565 + 128000 ) / 1000 ;
			yuv [ 2 ] = ( ( p [ 0 ] - yuv [ 0 ] ) * 713 + 128000 ) / 1000 ;
			
			//fprintf ( stdout, "x = %d ; y = %d\n", x, y ) ;
			
			* ( op [ 0 ] ++ ) = yuv [ 0 ] ;
			if ( x % 2 == 0 && y % 2 == 0 )
			{
				* ( op [ 1 ] ++ ) = yuv [ 2 ] ;
				* ( op [ 2 ] ++ ) = yuv [ 1 ] ;
			}
			p += s->format->BytesPerPixel ;
		}
	}
}

void __inline__ Video_Overlay::ConvertRGBtoYUY2 ( SDL_Surface *s, Uint8 **o,
											Uint16 *pitches, SDL_Rect rect,
											int monochrome, int luminance )
{
	int x , y ;
	int yuv [ 3 ] ;
	Uint8 *p, *op ;

	//fprintf ( stdout, "ConvertRGBtoYUY2 %d,%d x %d,%d\n",
	//			rect.x, rect.y, rect.w, rect.h ) ;
	
	// Convert
	for ( y = rect.y ; y < s->h && y < rect.y + rect.h ; y++ )
	{
		p = ( (Uint8 *) s->pixels ) + s->pitch * y + ( s->format->BytesPerPixel * rect.x ) ;
		op = o [ 0 ] + pitches [ 0 ] * y + rect.x * 2 ;
		
		for ( x = rect.x ; x < s->w && x < rect.x + rect.w ; x++ )
		{
			//RGBtoYUV ( p, yuv, monochrome, luminance ) ;
			
			yuv [ 0 ] = ( 299 * p [ 0 ] + 587 * p [ 1 ] + 114 * p [ 2 ] ) / 1000 ;
			yuv [ 1 ] = ( ( p [ 2 ] - yuv [ 0 ] ) * 565 + 128000 ) / 1000 ;
			yuv [ 2 ] = ( ( p [ 0 ] - yuv [ 0 ] ) * 713 + 128000 ) / 1000 ;
			
			//fprintf ( stdout, "x = %d ; y = %d\n", x, y ) ;
			if ( x % 2 == 0 )
			{
				* ( op ++ ) = yuv [ 0 ] ;
				* ( op ++ ) = yuv [ 1 ] ;
				op [ 1 ] = yuv [ 2 ] ;
			}
			else
			{
				*op = yuv [ 0 ] ;
				op += 2 ;
			}
			
			p += s->format->BytesPerPixel ;
		}
	}
}
